
/**
 * Copyright © 2002 - 2006 Anton Altaparmakov
 * Copyright © 2004 - 2005 Richard Russon
 * Copyright © 2004 - 2008 Szabolcs Szakacsits
 * Copyright © 2011 - 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef BITMAP_H
#define BITMAP_H

/* fribee-ng */
#include "die.h"
#include "defs.h"
#include "haveinline.h"

/**
 * bitmap_set - set a bit in a field of bits
 *
 * @bitmap:	field of bits
 * @bit:	bit to set
 * @new_value:	value to set bit to (0 or 1)
 *
 * Set the bit @bit in the @bitmap to @new_value.
 */
static inline void bitmap_set(u8 * bitmap, const u64 bit, const u8 new_value)
{
	if (unlikely(!bitmap) || unlikely(new_value > 1))
		die_guru();

	if (!new_value)
		bitmap[bit >> 3] &= ~(1 << (bit & 7));
	else
		bitmap[bit >> 3] |= (1 << (bit & 7));
}

/**
 * bitmap_get - get value of a bit in a field of bits
 *
 * @bitmap:	field of bits
 * @bit:	bit to get
 *
 * Get and return the value of the bit @bit in @bitmap (0 or 1).
 */
static inline char bitmap_get(const u8 * bitmap, const u64 bit)
{
	if (unlikely(!bitmap))
		die_guru();

	return (bitmap[bit >> 3] >> (bit & 7)) & 1;
}

/**
 * bitmap_clear_bits - clear "count" bits, begin @ bit
 *
 * @bitmap:	field of bits
 * @bit:	first bit to zero out
 * @count:	number of bits to set to 0
 */
static inline void bitmap_clear_bits(u8 * bitmap, u64 bit, u64 count)
{
	register u64 i;

	for (i = 0; i < count; i++)
		bitmap_set(bitmap, bit + i, 0);
}

/**
 * bitmap_set_bits - set "count" bits, begin @ bit
 *
 * @bitmap:	field of bits
 * @bit:	first bit to be set
 * @count:	number of bits to set to 0
 */
static inline void bitmap_set_bits(u8 * bitmap, u64 bit, u64 count)
{
	register u64 i;

	for (i = 0; i < count; i++)
		bitmap_set(bitmap, bit + i, 1);
}

/**
 * bitmap_count_bits - return count of all set bits, begin @ bit
 *
 * @bitmap:	field of bits
 * @bit:	first bit to be test
 * @max:	number of bits to test
 */
static inline u64 bitmap_count_bits(u8 * bitmap, u64 bit, u64 max)
{
	register u64 i;
	u64 r = 0;

	for (i = 0; i < max; i++)
		if (bitmap_get(bitmap + bit, i))
			r++;

	return r;
}

#endif				/* BSWAP_H */
