
/**
 * Copyright © 2008 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 8
 */

/* libowfat */
#include "open.h"
#include "byte.h"
#include "fmt.h"
#include "errmsg.h"

/* frisbee-ng */
#include "errmsg.h"
#include "compr.h"
#include "debug.h"
#include "defs.h"
#include "die.h"
#include "fmtX.h"
#include "mem.h"
#include "image.h"
#include "xxhash.h"
#include "term.h"
#include "crypt.h"

#undef DEBUG

#ifdef DEBUG
#include "term.h"
#endif

void compr_init(struct fng_compr *c)
{
	/* cflags */
	c->id = 0;
	c->clen = 0;
	c->ulen = 0;
	c->crc = 0;
	c->seed = 0;		/* 4 byte seed */

	c->cbuf = 0;
	c->ubuf = 0;

	/* only used for encryption */
	c->salt = 0;		/* 64 byte salt */
	c->key = 0;		/* 32 byte k */
	c->nonce = 0;		/* 24 byte nonce */
	c->password = 0;

	c->writesalt = 0;

	return;
}

/**
 * compr_init_password
 * - get called once in the beginning, when writing encrypted data
 * - nonce gets also allocated and is set to zero
 * - salt is initialized with 64 byte random data
 * - key is derived from password + salt
 */
void compr_init_password(struct fng_compr *c, stralloc * password)
{
	const void *salt, *pass;
	void *p;

	/* allocate space for the key stuff */
	c->salt = mem_talloc(u8, 64);
	c->key = mem_tzalloc(u8, 32);
	c->nonce = mem_tzalloc(u8, 24);

	/* set up random values */
	p = &c->seed;
	if (fmt_urandom(p, 4) != 4)
		fatal("fmt_random()");
	if (fmt_urandom(c->salt, 64) != 64)
		fatal("fmt_random()");

	/* this fuction should be okay, even for weak passwords ;) */
	salt = (void *)c->salt;
	pass = (void *)password->s;
	crypto_scrypt(pass, password->len, salt, 64, 1 << 14, 8, 1, c->key, 32);

	/* salt must be written on new images ... */
	c->writesalt = 1;

	return;
}

/**
 * compr_init_seed
 * - get called once in the beginning, when writing plaintext data
 */
void compr_init_seed(struct fng_compr *c)
{
	void *p;

	/* set up new seed value */
	p = &c->seed;
	if (fmt_urandom(p, 4) != 4)
		fatal("fmt_random()");

	return;
}

/**
 * compr_set_password
 * - get called once in the beginning, when reading encrypted data
 * - next: salt has to be set via compr_set_salt()
 */
void compr_set_password(struct fng_compr *c, stralloc * password)
{
	c->password = password;

	return;
}

/**
 * compr_set_salt
 * - get called once in the beginning, when reading encrypted data
 * - will set 64 byte salt to c->salt
 * - next: nonce should be set when needed
 */
static void compr_set_salt(struct fng_compr *c, u8 * salt)
{
	const void *pass;

	if (!c->password)
		fatal("Password not set!");

	/* allocate space for the key stuff */
	c->salt = mem_talloc(u8, 64);
	c->key = mem_tzalloc(u8, 32);
	c->nonce = mem_tzalloc(u8, 24);

	/* set content */
	byte_copy(c->salt, 64, salt);

	/* this fuction should be okay, even for weak passwords ;) */
	pass = (void *)c->password->s;
	crypto_scrypt(pass, c->password->len, salt, 64, 1 << 14, 8, 1, c->key,
		      32);

	return;
}

/**
 * compr_set_nonce
 * - changes the 24 byte nonce @ c->nonce
 */
void compr_set_nonce(struct fng_compr *c, u64 n1, u64 n2, u64 n3)
{
	void *p;
	u64 *u;

	/* allocate space, if needed */
	if (!c->nonce)
		fatal("c->nonce not set");
	p = c->nonce;

	u = p;
	*u = cpu_to_le64(n1);
	u = p + sizeof(u64);
	*u = cpu_to_le64(n2);
	u = p + sizeof(u64) * 2;
	*u = cpu_to_le64(n3);

	return;
}

/**
 * setup compressed block header (16 bytes)
 */
void compr_set_fnghdr(struct fng_compr *c)
{
	void *cbuf = c->cbuf;
	u32 *u, crc;

	/* 1) id (4 bytes, little endian) */
	u = cbuf;
	*u = cpu_to_le32(c->id);

	/* 2) compressed length (4 bytes, little endian) */
	u = cbuf + sizeof(u32);
	*u = cpu_to_le32(c->clen);

	/* 3) uncompressed length (4 bytes, little endian) */
	u = cbuf + sizeof(u32) * 2;
	*u = cpu_to_le32(c->ulen);

	/* 4) set crc to zero */
	u = cbuf + sizeof(u32) * 3;
	*u = 0;

	/* 5) calculate crc */
	crc = XXH32(c->ubuf, c->ulen, c->seed);

	/* 6) set crc hash(id,clen,ulen,ubuf) */
	*u = cpu_to_le32(crc);

#ifdef DEBUG
	term_printflush("SET_fnghdr() id=%06X clen=%z ulen=%z crc=%0x\n", c->id,
			c->clen, c->ulen, crc);
#endif

	return;
}

/* set last compress method header bit (0x01) */
void compr_set_fnghdr_last(void *cbuf)
{
	u32 *u = cbuf, U;

	/* cbuf[0] |= 0x01 */
	U = le32_to_cpu(*u);
	set(U, 0x01);
	*u = cpu_to_le32(U);

	return;
}

/**
 * compr_free - free all allocated structures of c
 */
void compr_free(struct fng_compr *c)
{
	/* mem_free() checks if allocated */
	mem_free(c->cbuf);
	mem_free(c->ubuf);
	if (c->salt) {
		mem_free(c->salt);
		mem_free(c->key);
		mem_free(c->nonce);
	}
}

int bzip2_compress(struct fng_compr *c, int blockSize, int workFactor)
{
	char *cbuf, *ubuf = c->ubuf;
	unsigned int clen, ulen = c->ulen;
	int ret;

	clen = (ulen * 5) >> 2;
	cbuf = mem_alloc(clen);

	ret =
	    BZ2_bzBuffToBuffCompress(cbuf, &clen, ubuf, ulen, blockSize, 0,
				     workFactor);
	if (ret != BZ_OK) {
		die_compr("BZ2_bzBuffToBuffCompress() != BZ_OK: ", ret);
	}

	c->cbuf = (void *)cbuf;
	c->clen = (size_t) clen;

	return CFLAGS_BZIP2;
}

void bzip2_decompress(struct fng_compr *c)
{
	/* should be okay, needs testing... */
	bzip2_decompress_data(c);

	return;
}

/**
 * DECOMPRESS
 */

/**
 * decompress(chunk)
 *
 * struct fng_compr {
 * 	void *ubuf;   -> [-W] uncompressed data [pre allocated]
 * 	u32 ulen;     -> [R-] uncompressed data length
 * 	void *cbuf;   -> [R-] compressed data [pre allocated]
 * 	u32 clen;     -> [R-] compressed data length
 *
 *  pre allocated data for encrypting / decrypting:
 * 	u8 *key;      -> [R-] 32 byte key
 * 	u8 *salt;     -> [R-] 64 byte salt
 * 	u8 *nonce;    -> [R-] 24 byte nonce
 * };
 */

void lz4_decompress_data(struct fng_compr *c)
{
	const char *cbuf = c->cbuf;
	char *ubuf = c->ubuf;
	int ulen = c->ulen;
	int clen = c->clen;
	int ret;

	ret = LZ4_decompress_fast(cbuf, ubuf, ulen);
	debug(8, 5, "lz4_decompress()=%d clen=%d ulen=%d", ret, c->clen,
	      c->ulen);
	if (ret != clen) {
		die_compr("lz4_decompress() != clen: ", ret);
	}

	return;
}

void lzf_decompress_data(struct fng_compr *c)
{
	void *cbuf = c->cbuf;
	void *ubuf = c->ubuf;
	unsigned int clen = c->clen;
	unsigned int ulen = c->ulen;
	unsigned int ret;

	/* we have no errorchecking here :( */
	ret = lzf_decompress(cbuf, clen, ubuf, ulen);
	debug(8, 5, "lzf_decompress()=%d clen=%d ulen=%d", ret, c->clen,
	      c->ulen);

	return;
}

void lzjb_decompress_data(struct fng_compr *c)
{
	void *cbuf = c->cbuf;
	void *ubuf = c->ubuf;
	size_t ulen = c->ulen;
	int ret;

	/* we have no errorchecking here :( */
	ret = lzjb_decompress_it(cbuf, ubuf, ulen);
	debug(8, 5, "lzjb_decompress()=%d clen=%d ulen=%d", ret, c->clen,
	      c->ulen);
	if (ret != 0) {
		die_compr("lzjb_decompress() != ulen: ", ret);
	}

	return;
}

void gzip_decompress_data(struct fng_compr *c)
{
	Byte *cbuf = c->cbuf;
	Byte *ubuf = c->ubuf;
	uLong clen = c->clen;
	uLong ulen = c->ulen;
	int ret;

	ret = uncompress(ubuf, &ulen, cbuf, clen);
	debug(8, 5, "gzip_decompress()=%d clen=%d ulen=%d ulen2=%d", ret,
	      c->clen, c->ulen, ulen);
	if (ret != Z_OK) {
		die_compr("gzip_decompress() != Z_OK: ", ret);
	}

	return;
}

void lzma_decompress_data(struct fng_compr *c)
{
	unsigned char *cbuf = c->cbuf;
	unsigned char *ubuf = c->ubuf;
	size_t clen = c->clen;
	size_t ulen = c->ulen;
	int ret;

	ret = LzmaUncompress(ubuf, &ulen, cbuf, &clen);
	debug(8, 5, "lzma_decompress()=%d clen=%d ulen=%d", ret, c->clen,
	      c->ulen);
	if (ret != SZ_OK) {
		die_compr("lzma_decompress() != SZ_OK: ", ret);
	}

	return;
}

void bzip2_decompress_data(struct fng_compr *c)
{
	char *cbuf = c->cbuf;
	char *ubuf = c->ubuf;
	unsigned int clen = c->clen;
	unsigned int ulen = c->ulen;
	int ret;

	ret = BZ2_bzBuffToBuffDecompress(ubuf, &ulen, cbuf, clen, 0, 0);
	if (ret != BZ_OK) {
		die_compr("BZ2_bzBuffToBuffDecompress() != BZ_OK: ", ret);
	}
	debug(8, 5, "bzip2_decompress()=%d clen=%d ulen=%d", ret, c->clen,
	      c->ulen);

	return;
}

void delta_decompress_data(struct fng_compr *c)
{
	char *cbuf = c->cbuf;
	char *ubuf = c->ubuf;
	int clen = c->clen;
	int ulen = c->ulen;

	delta_decompress(ubuf, ulen, cbuf, &clen);

	return;
}

void fastlz_decompress_data(struct fng_compr *c)
{
	void *cbuf = c->cbuf;
	void *ubuf = c->ubuf;
	int clen = c->clen;
	int ulen = c->ulen;
	int ret;

	ret = fastlz_decompress_r(cbuf, clen, ubuf, ulen);
	if (ret == 0) {
		die_compr("fastlz_decompress() failed: ret=", ret);
	}
	debug(8, 5, "fastlz_decompress()=%d clen=%d ulen=%d", ret, c->clen,
	      c->ulen);

	return;
}

void zstd_decompress_data(struct fng_compr *c)
{
	void *cbuf = c->cbuf;
	void *ubuf = c->ubuf;
	size_t clen = c->clen;
	size_t ulen = c->ulen;
	size_t ret;

	ret = ZSTD_decompress(ubuf, ulen, cbuf, clen);
	if (ZSTD_isError(ret)) {
		die_compr("ZSTD_decompress() failed: ret=", ret);
	}
	debug(8, 5, "ZSTD_decompress()=%d clen=%d ulen=%d", ret, c->clen,
	      c->ulen);

	return;
}

void quicklz_decompress_data(struct fng_compr *c)
{
	void *cbuf = c->cbuf;
	void *ubuf = c->ubuf;
	int ulen = c->ulen;
	int ret = 0;
	u8 *level;

	/**
	 * first u8 of cbuf contains level
	 *
	 * 0x01 -> level 1
	 * 0x02 -> level 3
	 * 0x03 -> level 3
	 */

	level = c->cbuf;
	switch (*level) {
	case 1:
		ret = quicklz_decompress1(cbuf + 1, ubuf);
		break;
	case 2:
		ret = quicklz_decompress2(cbuf + 1, ubuf);
		break;
	case 3:
		ret = quicklz_decompress3(cbuf + 1, ubuf);
		break;
	default:
		die_compr("quicklz_decompress() failed: level=", *level);
	}

	if (ret == 0) {
		die_compr("quicklz_decompress() failed: ret=", ret);
	}

	if (ret != ulen) {
		die_compr("quicklz_decompress() failed: ulen=", ulen);
	}

	debug(8, 5, "quicklz_decompress(level %d)=%d clen=%d ulen=%d", level,
	      ret, ulen);

	return;
}

#define LZOHDR 2
void lzo_decompress_data(struct fng_compr *c)
{
	lzo_bytep src = c->cbuf + LZOHDR;
	lzo_bytep dst = c->ubuf;
	lzo_uint slen = c->clen;
	lzo_uint dlen = c->ulen;
	u16 *lzohdr = le16_to_cpu(c->cbuf);
	int r = !LZO_E_OK;

	if (r != LZO_E_OK)

		switch (*lzohdr) {
		case LZO1_MAGIC:
			r = lzo1_decompress(src, slen, dst, &dlen, 0);
			break;
		case LZO1A_MAGIC:
			r = lzo1a_decompress(src, slen, dst, &dlen, 0);
			break;
		case LZO1B_MAGIC:
			r = lzo1b_decompress(src, slen, dst, &dlen, 0);
			break;
		case LZO1C_MAGIC:
			r = lzo1c_decompress(src, slen, dst, &dlen, 0);
			break;
		case LZO1F_MAGIC:
			r = lzo1f_decompress(src, slen, dst, &dlen, 0);
			break;
		case LZO1X_MAGIC:
			r = lzo1x_decompress(src, slen, dst, &dlen, 0);
			break;
		case LZO1Y_MAGIC:
			r = lzo1y_decompress(src, slen, dst, &dlen, 0);
			break;
		case LZO1Z_MAGIC:
			r = lzo1z_decompress(src, slen, dst, &dlen, 0);
			break;
		case LZO2A_MAGIC:
			r = lzo2a_decompress(src, slen, dst, &dlen, 0);
			break;
		}

	if (r != LZO_E_OK)
		die_msg(111, "lzo decompression failed!");

	return;
}

void rep_decompress_data(struct fng_compr *c)
{
	int clen = c->clen;
	int ulen = c->ulen;

	rep_decompress(c->ubuf, ulen, c->cbuf, clen);

	return;
}

/**
 * COMPRESS
 */

void compress_lzf_data(struct fng_compr *c, void *options)
{
	void *cbuf;
	void *ubuf = c->ubuf;
	unsigned int clen, ulen = c->ulen;

	unused(options);

	clen = (ulen * 5) >> 2;
	cbuf = mem_alloc(clen);

	clen = lzf_compress(ubuf, ulen, cbuf + FNGHDR, clen);
	c->cbuf = (void *)cbuf;
	c->clen = (size_t) clen;
	c->id = CFLAGS_LZF;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_lzjb_data(struct fng_compr *c, void *options)
{
	void *cbuf;
	void *ubuf = c->ubuf;
	unsigned int clen, ulen = c->ulen;

	unused(options);

	clen = (ulen * 5) >> 2;
	cbuf = mem_alloc(clen);

	clen = lzjb_compress_it(ubuf, cbuf + FNGHDR, ulen, clen);
	c->cbuf = (void *)cbuf;
	c->clen = (size_t) clen;
	c->id = CFLAGS_LZJB;
	compr_set_fnghdr(c);

	/* success */
	return;
}

/* compression with level */
void compress_lz4_data(struct fng_compr *c, void *options)
{
	struct compr_method_with_level *o =
	    (struct compr_method_with_level *)options;

	c->clen = LZ4_compressBound(c->ulen);
	c->cbuf = mem_alloc(c->clen + FNGHDR);
	c->clen =
	    LZ4_compress_fast(c->ubuf, c->cbuf + FNGHDR, c->ulen, c->clen,
			      o->level);
	if (c->clen == 0)
		die_msg(111, "lz4 error");
	c->id = CFLAGS_LZ4;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_lz4hc_data(struct fng_compr *c, void *options)
{
	struct compr_method_with_level *o =
	    (struct compr_method_with_level *)options;

	c->clen = LZ4_compressBound(c->ulen);
	c->cbuf = mem_alloc(c->clen + FNGHDR);
	c->clen =
	    LZ4_compress_HC(c->ubuf, c->cbuf + FNGHDR, c->ulen,
			    c->clen - FNGHDR, o->level);
	if (c->clen == 0)
		die_msg(111, "lz4hc error");
	c->id = CFLAGS_LZ4;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_gzip_data(struct fng_compr *c, void *options)
{
	struct compr_method_with_level *o =
	    (struct compr_method_with_level *)options;
	Byte *cbuf, *ubuf = c->ubuf;
	uLong clen, ulen = c->ulen;
	int ret;

	clen = compressBound(ulen) + FNGHDR;
	cbuf = mem_alloc(clen);

	debug(8, 5, "clen=%d ulen=%d level=%d", clen, ulen, o->level);
	ret = compress2(cbuf + FNGHDR, &clen, ubuf, ulen, o->level);
	if (ret != Z_OK) {
		die_compr("compress2() != Z_OK: ", ret);
	}

	c->cbuf = cbuf;
	c->clen = clen;
	c->id = CFLAGS_GZIP;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_quicklz_data(struct fng_compr *c, void *options)
{
	struct compr_method_with_level *o =
	    (struct compr_method_with_level *)options;
	u8 *ldata;

	/* http://www.quicklz.com/manual.html */
	c->cbuf = mem_zalloc(c->ulen + 500);

	switch (o->level) {
	case 1:
		c->clen =
		    quicklz_compress1(c->ubuf, c->ulen, c->cbuf + FNGHDR + 1);
		break;
	case 2:
		c->clen =
		    quicklz_compress2(c->ubuf, c->ulen, c->cbuf + FNGHDR + 1);
		break;
	case 3:
		c->clen =
		    quicklz_compress3(c->ubuf, c->ulen, c->cbuf + FNGHDR + 1);
		break;
	}

	ldata = c->cbuf + FNGHDR;
	*ldata = (u8) o->level;
	c->id = CFLAGS_QUICKLZ;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_fastlz_data(struct fng_compr *c, void *options)
{
	struct compr_method_with_level *o =
	    (struct compr_method_with_level *)options;
	void *cbuf, *ubuf = c->ubuf;
	int clen, ulen = c->ulen;

	clen = (ulen * 5) >> 2;
	cbuf = mem_alloc(clen);

	clen = fastlz_compress_level(o->level, ubuf, ulen, cbuf + FNGHDR);

	c->cbuf = cbuf;
	c->clen = (size_t) clen;
	c->id = CFLAGS_FASTLZ;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_zstd_data(struct fng_compr *c, void *options)
{
	struct compr_method_with_level *o =
	    (struct compr_method_with_level *)options;
	void *cbuf, *ubuf = c->ubuf;
	size_t clen, ulen = c->ulen;
	size_t ret;

	clen = ZSTD_compressBound(ulen);
	cbuf = mem_alloc(clen + FNGHDR);

	debug(8, 5, "clen=%d ulen=%d level=%d", clen, ulen, o->level);
	ret = ZSTD_compress(cbuf + FNGHDR, clen, ubuf, ulen, o->level);
	if (ZSTD_isError(ret)) {
		die_compr("ZSTD_compress() != Err: ", ret);
	}

	c->cbuf = cbuf;
	c->clen = ret;
	c->id = CFLAGS_ZSTD;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_rep_data(struct fng_compr *c, void *options)
{
	struct compr_method_rep *o = (struct compr_method_rep *)options;
	void *cbuf, *ubuf = c->ubuf;
	int clen = c->clen;
	int ulen = c->ulen;

	clen = ulen + 300;
	cbuf = mem_alloc(clen);

	debug(8, 5, "clen=%d ulen=%d", clen, ulen);
	rep_compress(o->BlockSize, o->MinMatchLen, o->Barrier, o->SmallestLen,
		     o->HashBits, o->Amplifier, ubuf, ulen, cbuf + FNGHDR,
		     &clen);

	c->clen = clen;
	c->cbuf = cbuf;
	c->id = CFLAGS_REP;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_lzma_data(struct fng_compr *c, void *options)
{
	struct compr_method_lzma *o = (struct compr_method_lzma *)options;
	size_t clen, ulen;

	ulen = c->ulen;
	clen = (ulen * 5) >> 2;
	c->cbuf = mem_alloc(clen + FNGHDR);
	LzmaCompress(c->cbuf + FNGHDR, &clen, c->ubuf, ulen, o->level,
		     o->dictSize, o->lc, o->lp, o->pb, o->algo, o->fb,
		     o->btMode, o->HashBytes);
	c->clen = clen;
	c->ulen = ulen;
	c->id = CFLAGS_LZMA;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_lzo_data(struct fng_compr *c, void *options)
{
	struct compr_method_lzo *o = (struct compr_method_lzo *)options;
	lzo_voidp wrkmem;
	int r;
	u16 *lzohdr;
	lzo_uint clen, ulen;

	ulen = c->ulen;
	wrkmem = mem_alloc(o->memory);
	clen = ulen + ulen / 16 + 64 + 3;
	c->cbuf = mem_alloc(clen + FNGHDR);

	r = o->compress(c->ubuf, ulen, c->cbuf + FNGHDR + LZOHDR, &clen,
			wrkmem);
	if (r != LZO_E_OK)
		die_msg(111, "lzo compression failed!");
	mem_free(wrkmem);

	/* setup information for decompressing */
	lzohdr = c->cbuf + FNGHDR;
	*lzohdr = cpu_to_le16(o->magic);
	c->clen = clen;
	c->id = CFLAGS_LZO;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_bzip2_data(struct fng_compr *c, void *options)
{
	struct compr_method_bzip2 *o = (struct compr_method_bzip2 *)options;
	char *cbuf, *ubuf = c->ubuf;
	unsigned int clen, ulen = c->ulen;
	int ret;

	clen = ulen + ulen * 2 / 128 + 600;
	cbuf = mem_alloc(clen + FNGHDR);

	ret =
	    BZ2_bzBuffToBuffCompress(cbuf + FNGHDR, &clen, ubuf, ulen,
				     o->blockSize, 0, o->workFactor);
	if (ret != BZ_OK) {
		die_compr("BZ2_bzBuffToBuffCompress() != BZ_OK: ", ret);
	}

	c->cbuf = (void *)cbuf;
	c->clen = (size_t) clen;
	c->id = CFLAGS_BZIP2;
	compr_set_fnghdr(c);

	/* success */
	return;
}

void compress_delta_data(struct fng_compr *c, void *options)
{
	struct compr_method_delta *o = (struct compr_method_delta *)options;
	char *ubuf = c->ubuf, *cbuf;
	int ulen = c->ulen;
	int clen;

	clen = c->ulen * 2;
	cbuf = mem_zalloc(clen + FNGHDR);
	delta_compress(ubuf, ulen, cbuf + FNGHDR, &clen, o->blockSize);

	c->cbuf = cbuf;
	c->clen = clen;
	c->id = CFLAGS_DELTA;
	compr_set_fnghdr(c);

	/* success */
	return;
}

/**
 * c->ubuf will be encrypted and stored to c->cbuf
 */
#define SALTLEN 64
void compress_encrypt_data(struct fng_compr *c)
{
	const unsigned char *key, *nonce;
	unsigned char *cbuf;
	int r, extralen = 0;

	/* first enrypted block gets the 64 byte salt */
	if (c->writesalt) {
		extralen += SALTLEN;
		cbuf = mem_tzalloc(unsigned char, c->ulen + FNGHDR + extralen);

		byte_copy(cbuf + FNGHDR, SALTLEN, c->salt);
		c->writesalt = 0;
	} else {
		cbuf = mem_tzalloc(unsigned char, c->ulen + FNGHDR);
	}

	key = c->key;		/* 32 byte key, derived from password */
	nonce = c->nonce;	/* 24 byte nonce */
	r = crypto_stream_xsalsa208_xor(cbuf + FNGHDR + extralen, c->ubuf,
					c->ulen, nonce, key);
	if (r != 0)
		fatal("crypto_stream_xsalsa20_xor");

	c->cbuf = cbuf;
	c->clen = c->ulen + extralen;
	c->id = CFLAGS_ENCRYPT;
	compr_set_fnghdr(c);
}

/**
 * c->cbuf will be decrypted and stored @ c->ubuf
 */
void encrypt_decompress_data(struct fng_compr *c)
{
	const unsigned char *nonce, *key;
	unsigned char *ubuf = mem_tzalloc(unsigned char, c->ulen);
	int r, extralen = 0;

	/* first enrypted block has the 64 byte salt */
	if (!c->salt) {
		compr_set_salt(c, c->cbuf);
		extralen += 64;
	}

	key = c->key;		/* 32 byte key, derived from password */
	nonce = c->nonce;	/* 24 byte nonce */
	r = crypto_stream_xsalsa208_xor(ubuf, c->cbuf + extralen, c->ulen,
					nonce, key);
	if (r != 0)
		fatal("crypto_stream_xsalsa20_xor");

	c->ubuf = ubuf;
}

/**
 * decompress_data - decompress data ...
 */
int decompress_data(struct fng_compr *c)
{
	int lastone = 0;

	/* warning: clen is without fnghdr... */
	u32 hash, in_clen, *u;

 again:
	in_clen = c->clen - FNGHDR;

	/**
	 * this id should have one or two bits set
	 * - we may check for correctness here?
	 */
	u = c->cbuf;
	c->id = le32_to_cpu(*u);

	/* remove lastone bit, when set */
	if (is_set(c->id, 0x01)) {
		lastone = 1;
		unset(c->id, 0x01);
	}

	/* this should be the same as in_len */
	u = c->cbuf + sizeof(u32);
	c->clen = le32_to_cpu(*u);

	/* this should be used for allocating the buffer for uncompressed data */
	u = c->cbuf + sizeof(u32) * 2;
	c->ulen = le32_to_cpu(*u);

	/* for checking the input ... */
	u = c->cbuf + sizeof(u32) * 3;
	c->crc = le32_to_cpu(*u);

	if (in_clen != c->clen)
		die_msg(111, "Wrong Header (clen)!");

	/* skip fnghdr */
	c->cbuf += FNGHDR;
	c->ubuf = mem_alloc(c->ulen);

	switch (c->id) {
	case CFLAGS_LZO:
		lzo_decompress_data(c);
		break;
	case CFLAGS_LZF:
		lzf_decompress_data(c);
		break;
	case CFLAGS_LZ4:	/* lz4hc also */
		lz4_decompress_data(c);
		break;
	case CFLAGS_LZJB:
		lzjb_decompress_data(c);
		break;
	case CFLAGS_LZMA:
		lzma_decompress_data(c);
		break;
	case CFLAGS_GZIP:
		gzip_decompress_data(c);
		break;
	case CFLAGS_BZIP2:
		bzip2_decompress_data(c);
		break;
	case CFLAGS_FASTLZ:
		fastlz_decompress_data(c);
		break;
	case CFLAGS_QUICKLZ:
		quicklz_decompress_data(c);
		break;
	case CFLAGS_REP:
		rep_decompress_data(c);
		break;
	case CFLAGS_DELTA:
		delta_decompress_data(c);
		break;
	case CFLAGS_ZSTD:
		zstd_decompress_data(c);
		break;
	case CFLAGS_ENCRYPT:
		encrypt_decompress_data(c);
		break;
	default:
		die_msg(111, "This is no FNG Compressed Block :O");
	}

	/* set up next round... yes, we are working with goto's here! */
	c->cbuf -= FNGHDR;
	hash = XXH32(c->ubuf, c->ulen, c->seed);

#ifdef DEBUG
	term_printflush
	    ("fnghdr: id=%06X clen=%z ulen=%z crc=%0x | hash=%0x seed=%d\n",
	     c->id, c->clen, c->ulen, c->crc, hash, c->seed);
#endif
	if (hash != c->crc)
		die_msg(111, "CRC check failed!");

	mem_free(c->cbuf);

	if (!lastone) {
		c->cbuf = c->ubuf;
		c->clen = c->ulen;
		goto again;
	}

	return 0;
}
