
/**
 * Copyright © 2008 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef COMPR_H
#define COMPR_H

/* system */
#include <unistd.h>

/* libowfat */
#include "case.h"
#include "stralloc.h"

/* extra libs */
#include "bzlib/bzlib.h"
#include "quicklz/quicklz.h"
#include "lzf/lzf.h"
#include "zlib/zlib.h"
#include "lzma/lzma_lib.h"
#include "lz4/lz4.h"
#include "lz4/lz4hc.h"
#include "zstd/zstd.h"

#include "lzo/lzo1.h"
#include "lzo/lzo1a.h"
#include "lzo/lzo1b.h"
#include "lzo/lzo1c.h"
#include "lzo/lzo1f.h"
#include "lzo/lzo1x.h"
#include "lzo/lzo1y.h"
#include "lzo/lzo1z.h"
#include "lzo/lzo2a.h"

#include "haveinline.h"

#include "compr_fastlz.h"
#include "compr_lzjb.h"
#include "compr_delta.h"
#include "compr_rep.h"

#include "list.h"
#include "defs.h"

/**
 * compression flag (cflags) used in images or on network
 */
#define CFLAGS_DATA      1<<0	/*  1 | -c marker for last compr. method */
#define CFLAGS_REP       1<<1	/*  1 | -c rep (options) */
#define CFLAGS_DELTA     1<<2	/*  2 | -c delta */
#define CFLAGS_LZO       1<<3	/*  3 | -c lzo (options) */
#define CFLAGS_LZF       1<<4	/*  4 | -c lzf */
#define CFLAGS_LZ4       1<<5	/*  5 | -c lz4 1..19 */
#define CFLAGS_LZ4HC     1<<6	/*  6 | -c lz4hc 1..16 */
#define CFLAGS_LZJB      1<<7	/*  7 | -c lzjb */
#define CFLAGS_LZMA      1<<8	/*  8 | -c lzma (options) */
#define CFLAGS_GZIP      1<<9	/*  9 | -c gzip 1..9 */
#define CFLAGS_BZIP2     1<<10	/* 10 | -c bzip2 (options) */
#define CFLAGS_FASTLZ    1<<11	/* 11 | -c fastlz 1..2 */
#define CFLAGS_QUICKLZ   1<<12	/* 12 | -c quicklz 1..3 */
#define CFLAGS_ZSTD      1<<13	/* 13 | -c zstd 1..20 */
/* ... reserved ... */
#define CFLAGS_ENCRYPT   1<<29	/* 29 | -c encrypted */

/**
 * each compressed block has 4x4 bytes (little endian) @ start:
 *
 * 1) 1x u32 -> cflags id (algo)
 * 2) 1x u32 -> length of clen
 * 3) 1x u32 -> length of ulen
 * 4) 1x u32 -> crc(id,clen,ulen,cbuf) / xxhash32()
 */
#define FNGHDR 4 * 4

struct fng_compr {
	u32 id;
	u32 clen;		/* length of compressed data (without fnghdr) */
	u32 ulen;		/* length of uncompressed data */
	u32 crc;		/* we use xxhash - which should be very fast */
	u32 seed;		/* 4 byte seed for xxhash32 (in fnghdr) */

	void *ubuf;		/* uncompressed data block */
	void *cbuf;		/* compressed data block */

	u8 *salt;		/* 64 byte salt for xsalsa20 key derivation */
	u8 *key;		/* 32 byte secret key for xsalsa20 encryption */
	u8 *nonce;		/* 24 byte nonce */
	stralloc *password;	/* password, set by the user */
	u8 writesalt;
};

/**
 * set the first 16 bytes of cbuf to:
 * - id (id of used algo + 0x01 when last ccompressed block)
 * - clen (length of compressed data, without fnghdr!)
 * - ulen (data length of uncompressed data...)
 * - crc of complete ubuf
 *
 * - when excryption is used, first encrypted block gets 64byte salt
 */

/* init all needed variables to zero */
extern void compr_init(struct fng_compr *c);

/* init a new password/salt pair for new images */
extern void compr_init_password(struct fng_compr *c, stralloc * password);
extern void compr_init_seed(struct fng_compr *c);

/* set password/seed/salt for reading images */
extern void compr_set_password(struct fng_compr *c, stralloc * password);

/* set nonce for reading/writing images */
extern void compr_set_nonce(struct fng_compr *c, u64 n1, u64 n2, u64 n3);

/* set 16 byte compression header */
extern void compr_set_fnghdr(struct fng_compr *c);
extern void compr_set_fnghdr_last(void *cbuf);

/* free allocated data */
extern void compr_free(struct fng_compr *c);

/* method with level */
struct compr_method_with_level {
	int level;
};

/* rep -> "rep:l{N}:d{N}:s{N}:h{N}:a{N}" */
struct compr_method_rep {
	/* "bs{N} --  x MB */
	int BlockSize;
	/* "l{N} --  minimal match len (default: 512)" */
	int MinMatchLen;
	/* "s{N} --  minimal match len after barrier (0)" */
	int SmallestLen;
	/* "d{N} --  barrier for smaller matches (8MB)" */
	int Barrier;
	/* "h{N} --  hash bits (default: calculated from size)" */
	int HashBits;
	/* "a{N} --  coefficient of search \"amplification\"
	 * (default: 1)" */
	int Amplifier;
};

/* lzma */
struct compr_method_lzma {
	int level;		/*  0 <= level <= 9 */
	/* 1<<12 (4KB) .. 1<<27 (128M) */
	int dictSize;		/* (1 << 12) <= dictSize <= (1 << 27) for 32-bit version (1 << 12) <= dictSize <= (1 << 30) for 64-bit version default = (1 << 24) */
	int lc;			/* 0 <= lc <= 8, default = 3 */
	int lp;			/* 0 <= lp <= 4, default = 0 */
	int pb;			/* 0 <= pb <= 4, default = 2 */
	int algo;		/* 0 - fast, 1 - normal, default = 1 */
	int fb;			/* 5 <= fb <= 273, default = 32 */
	int btMode;		/* 0 - hashChain Mode, 1 - binTree mode - normal, default = 1 */
	int HashBytes;		/* 2, 3 or 4, default = 4 */
};

/* lzo */
struct compr_method_lzo {
	lzo_compress_t compress;
	int memory;

	/* 9 different decompression formats */
	u16 magic;
#define LZO1_MAGIC  0x0001
#define LZO1A_MAGIC 0x0002
#define LZO1B_MAGIC 0x0004
#define LZO1C_MAGIC 0x0008
#define LZO1F_MAGIC 0x0010
#define LZO1X_MAGIC 0x0020
#define LZO1Y_MAGIC 0x0040
#define LZO1Z_MAGIC 0x0080
#define LZO2A_MAGIC 0x0100
};

/* bzip2 */
struct compr_method_bzip2 {
	int blockSize;		/*  1 <= blockSize <= 9 */
	int workFactor;		/*  0 <= workfactor <= 250 (default: 30) */
};

/* delta */
struct compr_method_delta {
	int blockSize;		/*  blockSize in MB */
};

/* used for parsing multiple compresson modes */
struct compr_method {
	char *name;
	int id;

	union {
		/**
		 * no level compressions: none, lzf, lzjb, delta
		 * encryption also uses no special parameters
		 */

		/* compression with level */
		struct compr_method_with_level lz4;
		struct compr_method_with_level lz4hc;
		struct compr_method_with_level gzip;
		struct compr_method_with_level quicklz;
		struct compr_method_with_level fastlz;
		struct compr_method_with_level zstd;

		/* compression types with special options */
		struct compr_method_rep rep;
		struct compr_method_lzma lzma;
		struct compr_method_lzo lzo;
		struct compr_method_bzip2 bzip2;
		struct compr_method_delta delta;
	} u;

	/* pointer to compressing function */
	void (*compress) (struct fng_compr * compr, void *options);

	/* pointer to argument of compressing function */
	void *options;

	/* pointer to argument of -c option */
	const char *option;
};

extern int bzip2_compress(struct fng_compr *c, int blockSize, int workFactor);
extern void bzip2_decompress(struct fng_compr *c);

/* decompress fng compressed blocks */
extern void none_decompress_data(struct fng_compr *c);
extern void lz4_decompress_data(struct fng_compr *c);	/* also lz4hc */
extern void lzf_decompress_data(struct fng_compr *c);
extern void lzjb_decompress_data(struct fng_compr *c);
extern void gzip_decompress_data(struct fng_compr *c);
extern void lzma_decompress_data(struct fng_compr *c);
extern void bzip2_decompress_data(struct fng_compr *c);
extern void quicklz_decompress_data(struct fng_compr *c);
extern void fastlz_decompress_data(struct fng_compr *c);
extern void delta_decompress_data(struct fng_compr *c);
extern void lzo_decompress_data(struct fng_compr *c);
extern void rep_decompress_data(struct fng_compr *c);
extern void zstd_decompress_data(struct fng_compr *c);

/**
 * decompress fng encrypted blocks
 * - first encrypted block must contain 64byte salt!
 */
extern void encrypt_decompress_data(struct fng_compr *c);

/* id, name and both functions must be defined! */
struct compression_type {
	struct list_head list;

	/**
	 * id of lossless compression algorithm (CFLAGS_* defines)
	 */
	int id;

	/**
	 * name of lossless compression algorithm
	 */
	char *name;

	/**
	 * give usage information
	 */
	char *usage;

	/**
	 * used für chaining like "rep:l128+delta+gzip"
	 */
	struct compr_method *method;
	int method_count;	/* 1..N */
};

/* register all known compression types */
extern void register_compression_modes(void);

/**
 * list with all known compression types
 * - must be initialized by register_compression_modes()
 */
extern struct list_head compression_types;

/**
 * find_filesystem - parse compression option
 *
 * @option: string to search for (method:option...)
 *
 * returns number of chained compression types
 */
extern int find_compression(const char *option, struct compr_method **c);
extern void find_compression_help(void);
extern void find_compression_usage(int id);

/* compression without level */
void compress_lzf_data(struct fng_compr *c, void *options);
void compress_lzjb_data(struct fng_compr *c, void *options);

/* compression with level */
void compress_lz4_data(struct fng_compr *c, void *options);
void compress_lz4hc_data(struct fng_compr *c, void *options);
void compress_gzip_data(struct fng_compr *c, void *options);
void compress_quicklz_data(struct fng_compr *c, void *options);
void compress_fastlz_data(struct fng_compr *c, void *options);
void compress_zstd_data(struct fng_compr *c, void *options);

/* compression with special options */
void compress_rep_data(struct fng_compr *c, void *options);
void compress_lzma_data(struct fng_compr *c, void *options);
void compress_lzo_data(struct fng_compr *c, void *options);
void compress_bzip2_data(struct fng_compr *c, void *options);
void compress_delta_data(struct fng_compr *c, void *options);

void compress_encrypt_data(struct fng_compr *c);

/**
 * one function for all methods
 */
int decompress_data(struct fng_compr *c);

#endif
