
/**
 * Copyright © 2004 – 2013, Bulat Ziganshin
 * Copyright © 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *
 * Delta: binary tables preprocessor v1.51 (C Version)
 * - based on this code: http://freearc.org/download/research/delta151.zip
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* libowfat */
#include "byte.h"

/* frisbee-ng */
#include "compr_delta.h"
#include "die.h"
#include "mem.h"
#include "defs.h"
#include "bswap.h"

typedef uint8_t bool;

#define TRUE            1
#define FALSE           0

#define iterate_var(i, num) for(i=0; i<(num); i++)
#define MAX_ELEMENT_SIZE 30
#define SMALL_ELEMENT_SIZE 4
#define FAST_CYCLES 4
#define LINE 32
#define DELTA 8

struct Buffer {
	uint8_t *buf;
	uint8_t *p;
	uint8_t *end;
	uint8_t *bufend;
};

static void Buffer_init(struct Buffer *b)
{
	int size = 64 * 1024 * 1024;

	b->buf = b->p = b->end = (uint8_t *) mem_alloc(size);
	b->bufend = b->buf + size;
}

static int Buffer_len(struct Buffer *b)
{
	return max(b->p, b->end) - b->buf;
}

static void Buffer_empty(struct Buffer *b)
{
	b->p = b->end = b->buf;
}

static void Buffer_free(struct Buffer *b)
{
	mem_free(b->buf);
	b->buf = b->p = b->end = 0;
}

static void Buffer_put32(struct Buffer *b, uint32_t x)
{
	*(uint32_t *) b->p = cpu_to_le32(x);
	b->p += sizeof(uint32_t);
}

static void Buffer_reserve(struct Buffer *b, int size)
{
	if (b->p + size > b->bufend) {
		int newsize =
		    max(b->p + size - b->buf, (b->bufend - b->buf) * 2);
		uint8_t *newbuf = (uint8_t *) mem_realloc(b->buf, newsize);

		b->bufend = newbuf + newsize;
		b->p += newbuf - b->buf;
		b->end += newbuf - b->buf;
		b->buf = newbuf;
	}
}

static uint32_t Buffer_get32(struct Buffer *b)
{
	uint32_t x = *(uint32_t *) b->p;

	b->p += sizeof(uint32_t);
	return x;
}

static uint32_t encode_type(int N, bool immutable[])
{
	uint32_t type = 1 << N;
	int i;

	for (i = 0; i < N; i++) {
		type += immutable[i] << i;
	}
	return type;
}

static void decode_type(uint32_t type, int *N, bool doDiff[], bool immutable[])
{
	int i;

	for (i = 0; type > 1; i++, type >>= 1) {
		immutable[i] = type & 1;
		doDiff[i] = !immutable[i];
	}
	*N = i;
}

static void diff_table(int N, uint8_t * table_start, int table_len,
		       bool doDiff[])
{
	uint8_t *r;
	int i, carry;

	for (r = table_start + N * table_len; (r -= N) > table_start;)
		for (i = 0, carry = 0; i < N; i++)
			if (doDiff[i]) {
				int newcarry = r[i] < r[i - N] + carry;

				r[i] -= r[i - N] + carry;
				carry = newcarry;
			} else
				carry = 0;
}

static void undiff_table(int N, uint8_t * table_start, int table_len,
			 bool doDiff[])
{
	uint8_t *r;
	int i, carry;

	for (r = table_start + N; r < table_start + N * table_len; r += N)
		for (i = 0, carry = 0; i < N; i++)
			if (doDiff[i]) {
				int sum = r[i] + r[i - N] + carry;

				r[i] = sum;
				carry = sum / 256;
			} else
				carry = 0;
}

static void reorder_table(int N, uint8_t * table_start, int table_len,
			  bool immutable[], struct Buffer *tempbuf)
{
	Buffer_reserve(tempbuf, N * table_len);
	byte_copy(tempbuf->buf, N * table_len, table_start);

	uint8_t *p = table_start, *q = tempbuf->buf;
	int i, k;

	for (i = 0; i < table_len; i++)
		for (k = 0; k < N; k++, q++)
			if (immutable[k])
				*p++ = *q;

	q = tempbuf->buf;
	for (i = 0; i < table_len; i++)
		for (k = 0; k < N; k++, q++)
			if (!immutable[k])
				*p++ = *q;
}

static void unreorder_table(int N, uint8_t * table_start, int table_len,
			    bool immutable[], struct Buffer *tempbuf)
{
	int i, k, imm_columns = 0;

	iterate_var(i, N) imm_columns += immutable[i];
	if (imm_columns == 0 || imm_columns == N)
		return;

	Buffer_reserve(tempbuf, N * table_len);
	byte_copy(tempbuf->buf, N * table_len, table_start);

	uint8_t *p = table_start;
	uint8_t *q = tempbuf->buf;
	uint8_t *q1 = tempbuf->buf + imm_columns * table_len;

	for (i = 0; i < table_len; i++)
		for (k = 0; k < N; k++)
			*p++ = immutable[k] ? *q++ : *q1++;
}

static void analyze_table(int N, uint8_t * table_start, int table_len,
			  bool doDiff[], bool immutable[])
{
	int i, k;

	for (k = 0; k < N; k++) {
		uint8_t *p = table_start + k;
		int neq = 0;

		for (i = 1; i < table_len; i++, p += N) {
			neq += p[N] != p[0];
		}

		immutable[k] = neq * 4 < table_len && N != 2 && N != 4
		    && N != 8;
		doDiff[k] = !immutable[k];
	}
}

static uint8_t *search_for_table_boundary(int N, uint8_t * t,
					  uint8_t * bufstart, uint8_t * bufend,
					  int *_useless)
{
	int dir = (*(int16_t *) (t + N) - *(int16_t *) t < 0) ? -1 : 1, len =
	    0, omit = 0, useless = 0, bad = 0;
	uint8_t *lastpoint = t;
	int first_time = TRUE;

	for (t += N; bufstart <= t + N && t + N + sizeof(int16_t) <= bufend;
	     t += N) {
		int diff = *(int16_t *) t - *(int16_t *) (t - N);
		uint32_t abs_diff = abs(diff);
		uint32_t abs_item = abs(*(int16_t *) t);
		uint32_t cmp_item =
		    (abs_item > 4095 ? abs_item / 6 : abs_item / 3);

		if (diff == 0)
			useless++;
		else if (dir * diff > 0)
			abs_diff < cmp_item ? (len++, omit =
					       0) : (useless++, omit++);
		else {
			if (len >= 4 || first_time)
				bad = 0, lastpoint = t - N * omit, *_useless =
				    useless, first_time = FALSE;
			else if (++bad >= 2)
				break;
			dir =
			    (*(int16_t *) (t + N) - *(int16_t *) t <
			     0) ? -1 : 1;
			len = omit = 0;
			if (dir * diff > 0)
				t -= N;
		}
	}
	return lastpoint;
}

/* XXX: gcc __bultin_ stuff does not work on my box, why? */
#if 0
#ifndef __GNUC__
#include <math.h>
#define my_logb logb
#define my_sqrt sqrt
#else
#define my_logb(x) __builtin_logb((x))
#define my_sqrt(x) __builtin_sqrt((x))
#endif
#endif

/* no math.h please ... */
static double my_sqrt(const double x)
{
	const double xhalf = 0.5f * x;
	union {
		double x;
		int64_t i;
	} u;

	u.x = x;
	u.i = 6910469410427058089ULL - (u.i >> 1);
	u.x = u.x * (1.5f - xhalf * u.x * u.x);

	return x * u.x * (1.5f - xhalf * u.x * u.x);
}

/* no math.h please ... */
static double my_logb(double x)
{
#ifndef __GNUC__
#include <math.h>
	return logb(x);		/* hm, we need it :( */
#else
	register double value;
	register double junk;
	__asm __volatile__("fxtract\n\t":"=t"(junk), "=u"(value):"0"(x));

	return value;
#endif
}

static int slow_check_for_data_table(int N, uint8_t * p, uint32_t * type,
				     uint8_t ** table_start,
				     uint8_t ** table_end, uint8_t * bufstart,
				     uint8_t * bufend,
				     struct Buffer *ReorderingBuffer)
{
	int useless = 0;

	*table_start =
	    search_for_table_boundary(-N, p, bufstart, bufend, &useless);
	*table_end =
	    search_for_table_boundary(N, *table_start, bufstart, bufend,
				      &useless);

	int rows = (*table_end - *table_start) / N;
	int useful = rows - useless;
	double x = max(*table_start - bufstart, 1);
	double skipBits = my_logb(x);

	if (useful * my_sqrt((double)N) > 30 + 4 * skipBits) {
		bool doDiff[MAX_ELEMENT_SIZE], immutable[MAX_ELEMENT_SIZE];

		analyze_table(N, *table_start, rows, doDiff, immutable);
		diff_table(N, *table_start, rows, doDiff);
		reorder_table(N, *table_start, rows, immutable,
			      ReorderingBuffer);
		*type = encode_type(N, immutable);
		return TRUE;
	}

	return FALSE;
}

static inline int FAST_CHECK(int N, uint8_t * p)
{
	return ((uint32_t) (p[1] - p[N + 1] + DELTA) <= (2 * DELTA)
		&& (uint32_t) (p[N + 1] - p[2 * N + 1] + DELTA) <= (2 * DELTA)
		&& (uint32_t) (p[2 * N + 1] - p[3 * N + 1] + DELTA) <=
		(2 * DELTA)
		&& *(int16_t *) (p) + *(int16_t *) (p + N) !=
		*(int16_t *) (p + 2 * N) + *(int16_t *) (p + 3 * N));
}

#define encode_table(skip, type, rows)   \
{                                        \
    Buffer_put32(TSkip, skip);  \
    Buffer_put32(TType, type);  \
    Buffer_put32(TRows, rows);  \
}

static void delta_process(uint8_t * buf, int Size, struct Buffer *TSkip,
			  struct Buffer *TType, struct Buffer *TRows,
			  struct Buffer *ReorderingBuffer)
{
	int i;
	uint8_t *bufend = buf + Size;
	uint8_t *last_table_end = buf;
	uint8_t *hash[256], *hash1[256];

	iterate_var(i, 256) hash[i] = hash1[i] = buf - 1;
	uint8_t count[256], count1[256];

	memset(&count, 0, 256);
	memset(&count1, 0, 256);
	unsigned cycle = 0;

	uint8_t *ptr;

	for (ptr = buf + MAX_ELEMENT_SIZE; ptr + LINE < bufend; ptr += LINE) {
		if (*(int32_t *) ptr != *(int32_t *) (ptr + 3)) {

			iterate_var(i,
				    (cycle ==
				     0 ? MAX_ELEMENT_SIZE : SMALL_ELEMENT_SIZE))
			    count1[i + 1] = count[i + 1], count[i + 1] = 0;
			uint8_t *p;

			for (p = ptr; p < ptr + LINE; p++) {
				int n = p - hash[*p / 16];

				hash[*p / 16] = p;
				count[n & 255]++;
			}
			cycle = (cycle + 1) % FAST_CYCLES;

			iterate_var(i,
				    (cycle ==
				     0 ? MAX_ELEMENT_SIZE : SMALL_ELEMENT_SIZE))
			{
				if (count[i + 1] + count1[i + 1] >=
				    (i >= SMALL_ELEMENT_SIZE ? 24 : 12)) {
					int N = i + 1, j;
					uint8_t *p =
					    ptr - (i >=
						   SMALL_ELEMENT_SIZE
						   ? (FAST_CYCLES -
						      1) * LINE : 0);

					for (j = 0; j < N; j++, p++) {

						if (FAST_CHECK(N, p)) {
							uint8_t *table_start,
							    *table_end;
							uint32_t type;

							if (slow_check_for_data_table(N, p, &type, &table_start, &table_end, last_table_end, bufend, ReorderingBuffer)) {
								encode_table
								    (table_start
								     -
								     last_table_end,
								     type,
								     (table_end
								      -
								      table_start)
								     / N);
								last_table_end =
								    table_end;
								ptr =
								    last_table_end
								    - LINE;
								cycle = 0;
								goto found;
							}
						}
					}
				}
			}
		}
 found:	;
	}
}

/**
 * fdin  -> ubuf + ulen
 * fdout -> cbuf + clen
 */
int delta_compress(char *ubuf, int ulen, char *cbuf, int *clen, int BlockSize)
{
	struct Buffer TSkip, TType, TRows, ReorderingBuffer;
	int todo = ulen, done = 0, cbuf_len = *clen;
	void *dest;
	u32 *u;

	/* x MB */
	BlockSize *= 1024 * 1024;

	uint8_t *buf =
	    (uint8_t *) mem_alloc(BlockSize + MAX_ELEMENT_SIZE * 8 + 1);

	Buffer_init(&TSkip);
	Buffer_init(&TType);
	Buffer_init(&TRows);
	Buffer_init(&ReorderingBuffer);

	*clen = 0;
	dest = cbuf;
	while (todo > 0) {
		int size = min(BlockSize, todo);

		/* term_print("lese %d bytes\n", size); */
		byte_copy(buf, size, ubuf + done);
		delta_process(buf, size, &TSkip, &TType, &TRows,
			      &ReorderingBuffer);

		/* count compressed size */
		*clen += 8;
		*clen += Buffer_len(&TSkip);
		*clen += Buffer_len(&TType);
		*clen += Buffer_len(&TRows);
		*clen += size;
#if 0
#include "term.h"
		term_print("schreibe %d bytes\n", 8);
		term_print("schreibe %d bytes\n", Buffer_len(&TSkip));
		term_print("schreibe %d bytes\n", Buffer_len(&TType));
		term_print("schreibe %d bytes\n", Buffer_len(&TRows));
		term_print("schreibe %d bytes\n", size);
#endif

		if (*clen > cbuf_len)
			die_msg(111, "need more memory for delta");

		u = dest;
		*u = cpu_to_le32(size);
		dest += 4;

		u = dest;
		*u = cpu_to_le32(Buffer_len(&TType));
		dest += 4;

		byte_copy(dest, Buffer_len(&TSkip), TSkip.buf);
		dest += Buffer_len(&TSkip);

		byte_copy(dest, Buffer_len(&TType), TType.buf);
		dest += Buffer_len(&TType);

		byte_copy(dest, Buffer_len(&TRows), TRows.buf);
		dest += Buffer_len(&TRows);

		byte_copy(dest, size, buf);
		dest += size;

		Buffer_empty(&TSkip);
		Buffer_empty(&TType);
		Buffer_empty(&TRows);

		todo -= size;
		done += size;
	}

	Buffer_free(&TSkip);
	Buffer_free(&TType);
	Buffer_free(&TRows);
	Buffer_free(&ReorderingBuffer);
	mem_free(buf);

	return 0;
}

/**
 * fdin  -> ubuf + ulen
 * fdout -> cbuf + clen
 */
int delta_decompress(char *ubuf, int ulen, char *cbuf, int *clen)
{
	struct Buffer Data, TSkip, TType, TRows, ReorderingBuffer;
	void *source, *dest;
	int todo = ulen;
	u32 *u;

	unused(clen);

	Buffer_init(&Data);
	Buffer_init(&TSkip);
	Buffer_init(&TType);
	Buffer_init(&TRows);
	Buffer_init(&ReorderingBuffer);

	dest = ubuf;
	source = cbuf;
	for (;;) {
		uint32_t TableSize, DataSize;
		uint8_t *p;
		int i;

		/* read DataSize and TableSize ... */
		u = source;
		DataSize = le32_to_cpu(*u);
		source += 4;

		u = source;
		TableSize = le32_to_cpu(*u);
		source += 4;

		Buffer_reserve(&TSkip, TableSize);
		byte_copy(TSkip.buf, TableSize, source);
		source += TableSize;

		Buffer_reserve(&TType, TableSize);
		byte_copy(TType.buf, TableSize, source);
		source += TableSize;

		Buffer_reserve(&TRows, TableSize);
		byte_copy(TRows.buf, TableSize, source);
		source += TableSize;

		Buffer_reserve(&Data, DataSize);
		byte_copy(Data.buf, DataSize, source);
		source += DataSize;

		p = Data.buf;
		for (i = TableSize / sizeof(uint32_t); i; i--) {
			bool doDiff[MAX_ELEMENT_SIZE];
			bool immutable[MAX_ELEMENT_SIZE];
			int skip = Buffer_get32(&TSkip);
			int type = Buffer_get32(&TType);
			int rows = Buffer_get32(&TRows);
			int N = 0;

			decode_type(type, &N, doDiff, immutable);
			p += skip;
			unreorder_table(N, p, rows, immutable,
					&ReorderingBuffer);
			undiff_table(N, p, rows, doDiff);
			p += N * rows;
		}

		byte_copy(dest, DataSize, Data.buf);
		dest += DataSize;

		Buffer_empty(&Data);
		Buffer_empty(&TSkip);
		Buffer_empty(&TType);
		Buffer_empty(&TRows);

		todo -= DataSize;

		/* we are done ;) */
		if (todo == 0)
			break;
		if (todo == -1)
			die_msg(111,
				"Strange things happen @ delta decompress!");
	}

	Buffer_free(&Data);
	Buffer_free(&TSkip);
	Buffer_free(&TType);
	Buffer_free(&TRows);
	Buffer_free(&ReorderingBuffer);

	return 0;
}
