
/**
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/**
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/**
 * We keep our own copy of this algorithm for 2 main reasons:
 *	1. If we didn't, anyone modifying common/os/compress.c would
 *         directly break our on disk format
 *	2. Our version of lzjb does not have a number of checks that the
 *         common/os version needs and uses
 *	3. We initialize the lempel to ensure deterministic results,
 *	   so that identical blocks can always be deduplicated.
 * In particular, we are adding the "feature" that compress() can
 * take a destination buffer size and return -1 if the data will not
 * compress to d_len or less.
 *
 * debug section: 8
 */

#include <sys/types.h>

#include "defs.h"
#include "compr_lzjb.h"

#define	MATCH_BITS	6
#define	MATCH_MIN	3
#define	MATCH_MAX	((1 << MATCH_BITS) + (MATCH_MIN - 1))
#define	OFFSET_MASK	((1 << (16 - MATCH_BITS)) - 1)
#define	LEMPEL_SIZE	1024
#define NBBY		8

 /*ARGSUSED*/ size_t lzjb_compress_it(void *s_start, void *d_start,
				      size_t s_len, size_t d_len)
{
	unsigned char *src = s_start;
	unsigned char *dst = d_start;
	unsigned char *cpy, *copymap = 0;
	int copymask = 1 << (NBBY - 1);
	int mlen, offset, hash;
	u16 *hp;
	u16 lempel[LEMPEL_SIZE] = { 0 };

	while (src < (unsigned char *)s_start + s_len) {
		if ((copymask <<= 1) == (1 << NBBY)) {
			if (dst >=
			    (unsigned char *)d_start + d_len - 1 - 2 * NBBY)
				return (s_len);
			copymask = 1;
			copymap = dst;
			*dst++ = 0;
		}
		if (src > (unsigned char *)s_start + s_len - MATCH_MAX) {
			*dst++ = *src++;
			continue;
		}
		hash = (src[0] << 16) + (src[1] << 8) + src[2];
		hash += hash >> 9;
		hash += hash >> 5;
		hp = &lempel[hash & (LEMPEL_SIZE - 1)];
		offset = (intptr_t) (src - *hp) & OFFSET_MASK;
		*hp = (u16) (uintptr_t) src;
		cpy = src - offset;
		if (cpy >= (unsigned char *)s_start && cpy != src
		    && src[0] == cpy[0] && src[1] == cpy[1]
		    && src[2] == cpy[2]) {
			*copymap |= copymask;
			for (mlen = MATCH_MIN; mlen < MATCH_MAX; mlen++)
				if (src[mlen] != cpy[mlen])
					break;
			*dst++ =
			    ((mlen - MATCH_MIN) << (NBBY -
						    MATCH_BITS)) | (offset >>
								    NBBY);
			*dst++ = (unsigned char)offset;
			src += mlen;
		} else {
			*dst++ = *src++;
		}
	}
	return (dst - (unsigned char *)d_start);
}

 /*ARGSUSED*/ int lzjb_decompress_it(void *s_start, void *d_start, size_t d_len)
{
	unsigned char *src = s_start;
	unsigned char *dst = d_start;
	unsigned char *d_end = (unsigned char *)d_start + d_len;
	unsigned char *cpy, copymap = 0;
	int copymask = 1 << (NBBY - 1);

	while (dst < d_end) {
		if ((copymask <<= 1) == (1 << NBBY)) {
			copymask = 1;
			copymap = *src++;
		}
		if (copymap & copymask) {
			int mlen = (src[0] >> (NBBY - MATCH_BITS)) + MATCH_MIN;
			int offset = ((src[0] << NBBY) | src[1]) & OFFSET_MASK;

			src += 2;
			if ((cpy = dst - offset) < (unsigned char *)d_start)
				return (-1);
			while (--mlen >= 0 && dst < d_end)
				*dst++ = *cpy++;
		} else {
			*dst++ = *src++;
		}
	}

	return 0;
}
