
/**
 * © Copyright 2004 – 2010, Bulat Ziganshin
 * © Copyright 2014 - 2016, Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Based on this code: http://freearc.org/download/research/rep.zip
 */

/**
 * REP is an LZ77-family algorithm, i.e. it founds matches and outputs them as
 * (len,offset) pairs. It is oriented toward very fast compression and small
 * memory overhead (1/4 of buffer size), but limited to rather large values of
 * mimimal match length (say, 32), and don't search for optimum match. It's
 * intended to preprocess data before using full-fledged compressors. and in
 * this area it beats RZIP and, to some degree, LZP preprocessors. Small
 * memory overhead means that RZIP/LZP/REP are capable to find matches at very
 * long distances and this algorithm does it much better than RZIP and LZP.
 * The algorithm implemented in functions REPEncode() and REPDecode().
 *
 * Main differences comparing to RZIP:
 * 1) Sliding window which slides at 1/16 of buffer size each time
 * 2) Almost ideal hash function (see update_hash)
 * 3) Direct hashing without hash chains which 1.5x cuts memory requirements
 * 4) Tags are not saved in hashtable, which again halves memory requirements.
 *      Instead, a few lower bits of hash table entry are used to save a few
 *      bits of tag (see chksum)
 * 5) Hash size is proportional to buffer size (which is equal to the maximum
 *      search distance) and by default limited to 1/4 of buffer size
 * 6) In order to find strings of length >=MinLen, blocks of length L=MinLen/2
 *      are indexed via hash. Of all those possible blocks, only 1/sqrt(L) are
 *      indexed and only 1/sqrt(L) are searched. It is alternative to solution
 *      described in RZIP paper where 1/L of blocks are indexed and each block
 *      searched. This means that logb(sqrt(L)) lower bits of hash entry are
 *      zeroes which allows to use trick 4.
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/* libowfat */
#include "byte.h"

#include "defs.h"
#include "compr_rep.h"
#include "bswap.h"
#include "mem.h"

typedef unsigned char byte;
typedef int int32;

static unsigned power(unsigned base, unsigned n)
{
	int result;

	for (result = 1; n != 0; result *= base, n--) ;
	return result;
}

static int roundup_to_power_of(unsigned n, unsigned base)
{
	int result;

	for (result = base, n--; (n /= base) != 0; result *= base) ;
	return result;
}

static unsigned sqrtb(unsigned n, unsigned base)
{
	int result;

	for (result = 1; (n /= base * base) != 0; result *= base) ;
	return result;
}

static inline byte *find_match_start(byte * p, byte * q, byte * start)
{
	while (q > start)
		if (*--p != *--q)
			return q + 1;
	return q;
}

static inline byte *find_match_end(byte * p, byte * q, byte * end)
{
	while (q < end && *p == *q)
		p++, q++;
	return q;
}

struct Buffer {
	byte *buf;
	byte *p;
	byte *end;
};

static void Buffer_init(struct Buffer *b, int size)
{
	b->buf = b->p = b->end = (byte *) mem_alloc(size);
}

static int Buffer_len(struct Buffer *b)
{
	return max(b->p, b->end) - b->buf;
}

static void Buffer_free(struct Buffer *b)
{
	mem_free(b->buf);
	b->buf = b->p = b->end = 0;
}

static void Buffer_put32(struct Buffer *b, int x)
{
	*(int32 *) b->p = x;
	b->p += sizeof(int32);
}

static void Buffer_rewind(struct Buffer *b)
{
	b->end = max(b->p, b->end);
	b->p = b->buf;
}

static int Buffer_get32(struct Buffer *b)
{
	int x = *(int32 *) b->p;

	b->p += sizeof(int32);
	return x;
}

static int Buffer_eof(struct Buffer *b)
{
	return b->p >= b->end;
}

/* or any other prime number */
#define PRIME          153191
#define POWER_PRIME_L  power(PRIME,L)

#define update_hash(sub,add)                          \
{                                                     \
    hash = hash * PRIME + add - sub * cPOWER_PRIME_L; \
}

#define chksum   ((hash>>28)&k1)

/* I leave In() and Out() for better understanding */
static int In(byte * source, byte * buffer, int count)
{
	byte_copy(buffer, count, source);

	return count;
}

static int Out(byte * dest, byte * buffer, int count)
{
	byte_copy(dest, count, buffer);

	return count;
}

#define Put32(dest, x)                 \
{                                                     \
    struct Buffer header;                             \
    Buffer_init(&header, sizeof(int32_t));            \
    Buffer_put32(&header, x);                         \
    dest += Out(dest, header.buf, Buffer_len(&header));  \
    Buffer_free(&header);                             \
}

int rep_compress(int BlockSize, int MinMatchLen, int Barrier, int SmallestLen,
		 int HashBits, int Amplifier, uint8_t * ubuf, int ulen,
		 uint8_t * cbuf, int *clen)
{
	int L = roundup_to_power_of(SmallestLen / 2, 2);
	int k = sqrtb(L * 2, 2), k1 = k - 1;
	int test = min(k * Amplifier, L), cPOWER_PRIME_L = POWER_PRIME_L;
	int HashSize, HashMask = 0, *hasharr = NULL, hash = 0;
	int Base = 0, last_i = 0, last_match = 0;
	int FirstTime = 1;

	uint8_t *buf;
	uint8_t *source = ubuf;
	void *dest = cbuf;

	BlockSize *= 1024 * 1024;
	Barrier *= 1024 * 1024;

	buf = (uint8_t *) mem_alloc(BlockSize);

	for (;;) {
		int Size = min(BlockSize - Base,
			       FirstTime ? BlockSize : BlockSize / 16);
		int done = source - ubuf;
		int i, literals, bsize;
		struct Buffer lens, offsets, datalens, dataOffsets;

		if (done + Size > ulen)
			Size = ulen - done;

		Size = In(source, buf + Base, Size);
		source += Size;

		if (Size == 0)
			break;

		if (Size < 0) {
			goto exit;
		}

		if (FirstTime) {
			buf = (uint8_t *) mem_realloc(buf, Size);
			HashSize =
			    HashBits >
			    0 ? (1 << HashBits) : roundup_to_power_of(Size,
								      2) /
			    max(k, 16);
			HashMask = HashSize - 1;
			hasharr = (int *)mem_calloc(HashSize, sizeof(int));
			if (HashSize && hasharr == NULL) {
				goto exit;
			}
			Put32(dest, Size);
			FirstTime = 0;
		}

		if (Base == 0) {
			hash = 0;
			for (i = 0; i < min(L, Size); i++)
				update_hash(0, buf[i]);
		}

		literals = 0;
		bsize = (Size / SmallestLen + 1) * sizeof(int32_t);

		Buffer_init(&lens, bsize);
		Buffer_init(&offsets, bsize);
		Buffer_init(&datalens, bsize);
		Buffer_init(&dataOffsets, bsize);

		for (i = last_i; i + L * 2 < Base + Size; last_i = i) {
			int j;

			for (j = 0; j < test; j++, i++) {
				if (i >= last_match) {
					int match = hasharr[hash & HashMask];

					if (match && chksum == (match & k1)) {
						match &= ~k1;
						if (match >= i
						    && match < Base + Size)
							goto no_match;

						int LowBound =
						    match <
						    i ? i - match : match -
						    (Base + Size) >
						    i ? 0 : i - (match -
								 (Base + Size));
						int HighBound =
						    BlockSize - match + i;

						int start =
						    find_match_start(buf +
								     match,
								     buf + i,
								     buf +
								     max
								     (last_match,
								      LowBound))
						    - buf;
						int end =
						    find_match_end(buf + match,
								   buf + i,
								   buf +
								   min(Base +
								       Size,
								       HighBound))
						    - buf;

						if (end - start >=
						    (i - match <
						     Barrier ? MinMatchLen :
						     SmallestLen)) {
							int offset = i - match;

							if (offset < 0)
								offset +=
								    BlockSize;

							Buffer_put32
							    (&dataOffsets,
							     last_match);
							Buffer_put32(&datalens,
								     start -
								     last_match);
							Buffer_put32(&offsets,
								     offset);
							Buffer_put32(&lens,
								     end -
								     start);

							literals +=
							    start - last_match;
							last_match = end;
						}
					}
				}
 no_match:
				if ((i & k1) == 0)
					hasharr[hash & HashMask] = i + chksum;
				update_hash(buf[i], buf[i + L]);
			}

			while ((i & (L - 1)) != 0) {
				hasharr[hash & HashMask] = i + chksum;
				int j;

				for (j = 0; j < k; j++, i++)
					update_hash(buf[i], buf[i + L]);
			}
		}

		Base += Size;
		if (Base == BlockSize) {
			Buffer_put32(&dataOffsets, last_match);
			Buffer_put32(&datalens, Base - last_match);
			literals += Base - last_match;
			Base = last_match = last_i = 0;
		} else {
			if (Buffer_len(&lens) == 0)
				goto skip_output;
			Buffer_put32(&datalens, 0);
		}

		{
			struct Buffer header;

			Buffer_init(&header, 2 * sizeof(int32_t));
			Buffer_put32(&header,
				     sizeof(int32_t) + Buffer_len(&lens) +
				     Buffer_len(&offsets) +
				     Buffer_len(&datalens) + literals);
			Buffer_put32(&header,
				     Buffer_len(&lens) / sizeof(int32_t));

			dest += Out(dest, header.buf, Buffer_len(&header));
			dest += Out(dest, lens.buf, Buffer_len(&lens));
			dest += Out(dest, offsets.buf, Buffer_len(&offsets));
			dest += Out(dest, datalens.buf, Buffer_len(&datalens));

			Buffer_rewind(&dataOffsets);
			Buffer_rewind(&datalens);
			while (!Buffer_eof(&dataOffsets)) {
				dest +=
				    Out(dest, buf + Buffer_get32(&dataOffsets),
					Buffer_get32(&datalens));
			}
			Buffer_free(&header);

		}
 skip_output:	Buffer_free(&lens);
		Buffer_free(&offsets);
		Buffer_free(&datalens);
		Buffer_free(&dataOffsets);
	}

	{
		int datalen = Base - last_match;

		Put32(dest, sizeof(int32_t) * 2 + datalen);
		Put32(dest, 0);
		Put32(dest, datalen);
		dest += Out(dest, buf + last_match, datalen);
		Put32(dest, 0);
	}
 exit:
	mem_free(hasharr);
	mem_free(buf);

	buf = dest;
	*clen = buf - cbuf;

	return 0;
}

int rep_decompress(uint8_t * ubuf, int ulen, uint8_t * cbuf, int clen)
{
	uint8_t *data, *data0;
	uint8_t *last_data;

	void *source = cbuf;
	void *dest = ubuf;
	u32 *u;
	int BlockSize;

	unused(ulen);
	unused(clen);

	/* read 4 bytes -> blocksize */
	u = source;
	BlockSize = *u;
	source += sizeof(int32_t);

	data = mem_alloc(BlockSize);
	data0 = data;

	for (last_data = data;; last_data = data) {
		void *buf, *buf0;
		int ComprSize, num, i;
		int32_t *lens, *offsets, *datalens;

		u = source;
		ComprSize = le32_to_cpu(*u);
		source += sizeof(int32_t);

		if (ComprSize == 0)
			break;

		buf = mem_alloc(ComprSize);
		buf0 = buf;

		source += In(source, buf, ComprSize);

		u = (u32 *) buf;
		num = le32_to_cpu(*u);
		buf += sizeof(int32_t);

		lens = buf;
		buf += num * sizeof(int32_t);

		offsets = buf;
		buf += num * sizeof(int32_t);

		datalens = buf;
		buf += (num + 1) * sizeof(int32_t);

		for (i = 0; i < num; i++) {
			int offset;

			byte_copy(data, datalens[i], buf);
			buf += datalens[i];
			data += datalens[i];
			offset =
			    offsets[i] <=
			    data - data0 ? offsets[i] : offsets[i] - BlockSize;
			byte_copy(data, lens[i], data - offset);
			data += lens[i];
		}
		byte_copy(data, datalens[num], buf);

		buf += datalens[num];
		data += datalens[num];

		dest += Out(dest, last_data, data - last_data);

		if (data == data0 + BlockSize)
			data = data0;
		mem_free(buf0);
	}

	mem_free(data0);
	return 0;
}
