
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Derived from public domain C code in NACL, see http://nacl.cr.yp.to/
 * and https://131002.net/siphash/
 */

#include "crypto_auth.h"
#include "crypto_hashblocks.h"
#include "crypto_verify.h"

#define blocks crypto_hashblocks_sha512

static const unsigned char iv[64] = {
	0x6a, 0x09, 0xe6, 0x67, 0xf3, 0xbc, 0xc9, 0x08,
	0xbb, 0x67, 0xae, 0x85, 0x84, 0xca, 0xa7, 0x3b,
	0x3c, 0x6e, 0xf3, 0x72, 0xfe, 0x94, 0xf8, 0x2b,
	0xa5, 0x4f, 0xf5, 0x3a, 0x5f, 0x1d, 0x36, 0xf1,
	0x51, 0x0e, 0x52, 0x7f, 0xad, 0xe6, 0x82, 0xd1,
	0x9b, 0x05, 0x68, 0x8c, 0x2b, 0x3e, 0x6c, 0x1f,
	0x1f, 0x83, 0xd9, 0xab, 0xfb, 0x41, 0xbd, 0x6b,
	0x5b, 0xe0, 0xcd, 0x19, 0x13, 0x7e, 0x21, 0x79
};

/* by Daniel J. Bernstein */
int crypto_auth_hmacsha512256(unsigned char *out, const unsigned char *in,
			      unsigned long long inlen, const unsigned char *k)
{
	unsigned char h[64];
	unsigned char padded[256];
	unsigned int i;
	unsigned long long bytes = 128 + inlen;

	for (i = 0; i < 64; ++i)
		h[i] = iv[i];

	for (i = 0; i < 32; ++i)
		padded[i] = k[i] ^ 0x36;
	for (i = 32; i < 128; ++i)
		padded[i] = 0x36;

	blocks(h, padded, 128);
	blocks(h, in, inlen);
	in += inlen;
	inlen &= 127;
	in -= inlen;

	for (i = 0; i < inlen; ++i)
		padded[i] = in[i];
	padded[inlen] = 0x80;

	if (inlen < 112) {
		for (i = inlen + 1; i < 119; ++i)
			padded[i] = 0;
		padded[119] = bytes >> 61;
		padded[120] = bytes >> 53;
		padded[121] = bytes >> 45;
		padded[122] = bytes >> 37;
		padded[123] = bytes >> 29;
		padded[124] = bytes >> 21;
		padded[125] = bytes >> 13;
		padded[126] = bytes >> 5;
		padded[127] = bytes << 3;
		blocks(h, padded, 128);
	} else {
		for (i = inlen + 1; i < 247; ++i)
			padded[i] = 0;
		padded[247] = bytes >> 61;
		padded[248] = bytes >> 53;
		padded[249] = bytes >> 45;
		padded[250] = bytes >> 37;
		padded[251] = bytes >> 29;
		padded[252] = bytes >> 21;
		padded[253] = bytes >> 13;
		padded[254] = bytes >> 5;
		padded[255] = bytes << 3;
		blocks(h, padded, 256);
	}

	for (i = 0; i < 32; ++i)
		padded[i] = k[i] ^ 0x5c;
	for (i = 32; i < 128; ++i)
		padded[i] = 0x5c;

	for (i = 0; i < 64; ++i)
		padded[128 + i] = h[i];
	for (i = 0; i < 64; ++i)
		h[i] = iv[i];

	for (i = 64; i < 128; ++i)
		padded[128 + i] = 0;
	padded[128 + 64] = 0x80;
	padded[128 + 126] = 6;

	blocks(h, padded, 256);
	for (i = 0; i < 32; ++i)
		out[i] = h[i];

	return 0;
}

int crypto_auth_hmacsha512256_verify(const unsigned char *h,
				     const unsigned char *in,
				     unsigned long long inlen,
				     const unsigned char *k)
{
	unsigned char correct[32];

	crypto_auth_hmacsha512256(correct, in, inlen, k);
	return crypto_verify_32(h, correct);
}

#include <inttypes.h>
typedef uint64_t u64;
typedef uint32_t u32;
typedef uint8_t u8;

#define ROTL(x,b) (u64)( ((x) << (b)) | ( (x) >> (64 - (b))) )

#define U32TO8_LE(p, v)       \
    (p)[0] = (u8)((v)      ); \
    (p)[1] = (u8)((v) >>  8); \
    (p)[2] = (u8)((v) >> 16); \
    (p)[3] = (u8)((v) >> 24);

#define U64TO8_LE(p, v)         \
  U32TO8_LE((p),     (u32)((v)      ));   \
  U32TO8_LE((p) + 4, (u32)((v) >> 32));

#define U8TO64_LE(p) \
  (((u64)((p)[0])      ) | \
   ((u64)((p)[1]) <<  8) | \
   ((u64)((p)[2]) << 16) | \
   ((u64)((p)[3]) << 24) | \
   ((u64)((p)[4]) << 32) | \
   ((u64)((p)[5]) << 40) | \
   ((u64)((p)[6]) << 48) | \
   ((u64)((p)[7]) << 56))

#define SIPROUND            \
  do {              \
    v0 += v1; v1=ROTL(v1,13); v1 ^= v0; v0=ROTL(v0,32); \
    v2 += v3; v3=ROTL(v3,16); v3 ^= v2;     \
    v0 += v3; v3=ROTL(v3,21); v3 ^= v0;     \
    v2 += v1; v1=ROTL(v1,17); v1 ^= v2; v2=ROTL(v2,32); \
  } while(0)

/* SipHash-2-4 by Jean-Philippe Aumasson */
int crypto_auth_siphash24(unsigned char *out, const unsigned char *in,
			  unsigned long long inlen, const unsigned char *k)
{
	/* "somepseudorandomlygeneratedbytes" */
	u64 v0 = 0x736f6d6570736575ULL;
	u64 v1 = 0x646f72616e646f6dULL;
	u64 v2 = 0x6c7967656e657261ULL;
	u64 v3 = 0x7465646279746573ULL;
	u64 b;
	u64 k0 = U8TO64_LE(k);
	u64 k1 = U8TO64_LE(k + 8);
	u64 m;
	const u8 *end = in + inlen - (inlen % sizeof(u64));
	const int left = inlen & 7;

	b = ((u64) inlen) << 56;
	v3 ^= k1;
	v2 ^= k0;
	v1 ^= k1;
	v0 ^= k0;

	for (; in != end; in += 8) {
		m = U8TO64_LE(in);
		v3 ^= m;
		SIPROUND;
		SIPROUND;
		v0 ^= m;
	}

	switch (left) {
	case 7:
		b |= ((u64) in[6]) << 48;

	case 6:
		b |= ((u64) in[5]) << 40;

	case 5:
		b |= ((u64) in[4]) << 32;

	case 4:
		b |= ((u64) in[3]) << 24;

	case 3:
		b |= ((u64) in[2]) << 16;

	case 2:
		b |= ((u64) in[1]) << 8;

	case 1:
		b |= ((u64) in[0]);
		break;

	case 0:
		break;
	}

	v3 ^= b;
	SIPROUND;
	SIPROUND;
	v0 ^= b;
	v2 ^= 0xff;
	SIPROUND;
	SIPROUND;
	SIPROUND;
	SIPROUND;
	b = v0 ^ v1 ^ v2 ^ v3;
	U64TO8_LE(out, b);

	return 0;
}

int crypto_auth_siphash24_verify(const unsigned char *h,
				 const unsigned char *in,
				 unsigned long long inlen,
				 const unsigned char *k)
{
	unsigned char correct[8];

	crypto_auth_siphash24(correct, in, inlen, k);
	return crypto_verify_8(h, correct);
}
