
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Derived from public domain C code in NACL, see http://nacl.cr.yp.to/
 */

#ifndef crypto_box_H
#define crypto_box_H

#define crypto_box_PUBLICKEYBYTES 32
#define crypto_box_SECRETKEYBYTES 32
#define crypto_box_BEFORENMBYTES  32
#define crypto_box_NONCEBYTES     24
#define crypto_box_ZEROBYTES      32
#define crypto_box_BOXZEROBYTES   16

#define crypto_box crypto_box_curve25519xsalsa20poly1305
#define crypto_box_open crypto_box_curve25519xsalsa20poly1305_open
#define crypto_box_keypair crypto_box_curve25519xsalsa20poly1305_keypair
#define crypto_box_beforenm crypto_box_curve25519xsalsa20poly1305_beforenm
#define crypto_box_afternm crypto_box_curve25519xsalsa20poly1305_afternm
#define crypto_box_open_afternm crypto_box_curve25519xsalsa20poly1305_open_afternm

/* The crypto_box function encrypts and authenticates a message m[0], ...,
 * m[mlen-1] using the sender's secret key sk[0], sk[1], ...,
 * sk[crypto_box_SECRETKEYBYTES-1], the receiver's public key pk[0], pk[1],
 * ..., pk[crypto_box_PUBLICKEYBYTES-1], and a nonce n[0], n[1], ...,
 * n[crypto_box_NONCEBYTES-1]. The crypto_box function puts the ciphertext
 * into c[0], c[1], ..., c[mlen-1]. It then returns 0.
 */
int crypto_box_curve25519xsalsa20poly1305(unsigned char *c,	/* ciphertext */
					  const unsigned char *m,	/* message */
					  unsigned long long mlen,	/* message length */
					  const unsigned char *n,	/* nonce - 24 byte */
					  const unsigned char *pk,	/* public key, 32 byte */
					  const unsigned char *sk);	/* secret key, 32 byte */

/* The crypto_box_keypair function randomly generates a secret key and a
 * corresponding public key. It puts the secret key into sk[0], sk[1], ...,
 * sk[crypto_box_SECRETKEYBYTES-1] and puts the public key into pk[0], pk[1],
 * ..., pk[crypto_box_PUBLICKEYBYTES-1]. It then returns 0.
 */
int crypto_box_curve25519xsalsa20poly1305_keypair(unsigned char *pk,
						  unsigned char *sk);

/* The crypto_box_open function verifies and decrypts a ciphertext c[0], ...,
 * c[clen-1] using the receiver's secret key sk[0], sk[1], ...,
 * sk[crypto_box_SECRETKEYBYTES-1], the sender's public key pk[0], pk[1],
 * ..., pk[crypto_box_PUBLICKEYBYTES-1], and a nonce n[0], ...,
 * n[crypto_box_NONCEBYTES-1]. The crypto_box_open function puts the
 * plaintext into m[0], m[1], ..., m[clen-1]. It then returns 0.
 *
 * If the ciphertext fails verification, crypto_box_open instead returns -1,
 * possibly after modifying m[0], m[1], etc.
 *
 * The caller must ensure, before calling the crypto_box_open function, that the
 * first crypto_box_BOXZEROBYTES bytes of the ciphertext c are all 0. The
 * crypto_box_open function ensures (in case of success) that the first
 * crypto_box_ZEROBYTES bytes of the plaintext m are all 0. 
 */
int crypto_box_curve25519xsalsa20poly1305_open(unsigned char *m,
					       const unsigned char *c,
					       unsigned long long clen,
					       const unsigned char *n,
					       const unsigned char *pk,
					       const unsigned char *sk);

/* Applications that send several messages to the same receiver can gain speed
 * by splitting crypto_box into two steps, crypto_box_beforenm and
 * crypto_box_afternm. Similarly, applications that receive several messages
 * from the same sender can gain speed by splitting crypto_box_open into two
 * steps, crypto_box_beforenm and crypto_box_open_afternm.
 *
 * The intermediate data computed by crypto_box_beforenm is suitable for both
 * crypto_box_afternm and crypto_box_open_afternm, and can be reused for any
 * number of messages.
 */
int crypto_box_curve25519xsalsa20poly1305_beforenm(unsigned char *k,
						   const unsigned char *pk,
						   const unsigned char *sk);

int crypto_box_curve25519xsalsa20poly1305_afternm(unsigned char *c,
						  const unsigned char *m,
						  unsigned long long mlen,
						  const unsigned char *n,
						  const unsigned char *k);

int crypto_box_curve25519xsalsa20poly1305_open_afternm(unsigned char *m, const unsigned char
						       *c,
						       unsigned long long clen,
						       const unsigned char
						       *n, const unsigned char
						       *k);

#endif
