
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Derived from public domain C code in NACL, see http://nacl.cr.yp.to/
 */

#ifndef crypto_secretbox_H
#define crypto_secretbox_H

#define crypto_secretbox_KEYBYTES     32
#define crypto_secretbox_NONCEBYTES   24
#define crypto_secretbox_ZEROBYTES    32
#define crypto_secretbox_BOXZEROBYTES 16

#define crypto_secretbox      crypto_secretbox_xsalsa20poly1305
#define crypto_secretbox_open crypto_secretbox_xsalsa20poly1305_open

/* The crypto_secretbox function encrypts and authenticates a message m[0],
 * m[1], ..., m[mlen-1] using a secret key k[0], ...,
 * k[crypto_secretbox_KEYBYTES-1] and a nonce n[0], n[1], ...,
 * n[crypto_secretbox_NONCEBYTES-1]. The crypto_secretbox function puts the
 * ciphertext into c[0], c[1], ..., c[mlen-1]. It then returns 0.
 *
 * WARNING: Messages in the C NaCl API are 0-padded versions of messages in the
 * C++ NaCl API. Specifically: The caller must ensure, before calling the C NaCl
 * crypto_secretbox function, that the first crypto_secretbox_ZEROBYTES bytes of
 * the message m are all 0. Typical higher-level applications will work with the
 * remaining bytes of the message; note, however, that mlen counts all of the
 * bytes, including the bytes required to be 0.
 * 
 * Similarly, ciphertexts in the C NaCl API are 0-padded versions of messages in
 * the C++ NaCl API. Specifically: The crypto_secretbox function ensures that
 * the first crypto_secretbox_BOXZEROBYTES bytes of the ciphertext c are all 0.
 */
int crypto_secretbox_xsalsa20poly1305(unsigned char *c, const unsigned char *m,
				      unsigned long long mlen,
				      const unsigned char *n,
				      const unsigned char *k);

/* The crypto_secretbox_open function verifies and decrypts a ciphertext
 * c[0], c[1], ..., c[clen-1] using a secret key k[0], k[1], ...,
 * k[crypto_secretbox_KEYBYTES-1] and a nonce n[0], ...,
 * n[crypto_secretbox_NONCEBYTES-1]. The crypto_secretbox_open function puts
 * the plaintext into m[0], m[1], ..., m[clen-1]. It then returns 0.
 * 
 * If the ciphertext fails verification, crypto_secretbox_open instead returns
 * -1, possibly after modifying m[0], m[1], etc.
 * 
 * The caller must ensure, before calling the crypto_secretbox_open function,
 * that the first crypto_secretbox_BOXZEROBYTES bytes of the ciphertext c are
 * all 0. The crypto_secretbox_open function ensures (in case of success) that
 * the first crypto_secretbox_ZEROBYTES bytes of the plaintext m are all 0.
 */
int crypto_secretbox_xsalsa20poly1305_open(unsigned char *m,
					   const unsigned char *c,
					   unsigned long long clen,
					   const unsigned char *n,
					   const unsigned char *k);

#endif
