
/**
 * Copyright © 2012 - 2013 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Derived from public domain C code in NACL and supercop, see
 * http://nacl.cr.yp.to/ and http://bench.cr.yp.to/supercop.html
 */

#include "crypto_core.h"
#include "crypto_stream.h"

static const unsigned char sigma[16] = "expand 32-byte k";

int crypto_stream_xsalsa20(unsigned char *c, unsigned long long clen,
			   const unsigned char *n, const unsigned char *k)
{
	unsigned char subkey[32];

	crypto_core_hsalsa20(subkey, n, k, sigma);
	return crypto_stream_salsa20(c, clen, n + 16, subkey);
}

int crypto_stream_xsalsa20_xor(unsigned char *c, const unsigned char *m,
			       unsigned long long mlen, const unsigned char *n,
			       const unsigned char *k)
{
	unsigned char subkey[32];

	crypto_core_hsalsa20(subkey, n, k, sigma);
	return crypto_stream_salsa20_xor(c, m, mlen, n + 16, subkey);
}

int crypto_stream_xsalsa208(unsigned char *c, unsigned long long clen,
			    const unsigned char *n, const unsigned char *k)
{
	unsigned char subkey[32];

	crypto_core_hsalsa208(subkey, n, k, sigma);
	return crypto_stream_salsa208(c, clen, n + 16, subkey);
}

int crypto_stream_xsalsa208_xor(unsigned char *c, const unsigned char *m,
				unsigned long long mlen, const unsigned char *n,
				const unsigned char *k)
{
	unsigned char subkey[32];

	crypto_core_hsalsa208(subkey, n, k, sigma);
	return crypto_stream_salsa208_xor(c, m, mlen, n + 16, subkey);
}

int crypto_stream_salsa20(unsigned char *c, unsigned long long clen,
			  const unsigned char *n, const unsigned char *k)
{
	unsigned char in[16];
	unsigned char block[64];
	unsigned int i;
	unsigned int u;

	if (!clen)
		return 0;

	for (i = 0; i < 8; ++i)
		in[i] = n[i];
	for (i = 8; i < 16; ++i)
		in[i] = 0;

	while (clen >= 64) {
		crypto_core_salsa20(c, in, k, sigma);

		u = 1;
		for (i = 8; i < 16; ++i) {
			u += (unsigned int)in[i];
			in[i] = u;
			u >>= 8;
		}

		clen -= 64;
		c += 64;
	}

	if (clen) {
		crypto_core_salsa20(block, in, k, sigma);
		for (i = 0; i < clen; ++i)
			c[i] = block[i];
	}
	return 0;
}

int crypto_stream_salsa20_xor(unsigned char *c, const unsigned char *m,
			      unsigned long long mlen, const unsigned char *n,
			      const unsigned char *k)
{
	unsigned char in[16];
	unsigned char block[64];
	unsigned int i;
	unsigned int u;

	if (!mlen)
		return 0;

	for (i = 0; i < 8; ++i)
		in[i] = n[i];
	for (i = 8; i < 16; ++i)
		in[i] = 0;

	while (mlen >= 64) {
		crypto_core_salsa20(block, in, k, sigma);
		for (i = 0; i < 64; ++i)
			c[i] = m[i] ^ block[i];

		u = 1;
		for (i = 8; i < 16; ++i) {
			u += (unsigned int)in[i];
			in[i] = u;
			u >>= 8;
		}

		mlen -= 64;
		c += 64;
		m += 64;
	}

	if (mlen) {
		crypto_core_salsa20(block, in, k, sigma);
		for (i = 0; i < mlen; ++i)
			c[i] = m[i] ^ block[i];
	}
	return 0;
}

int crypto_stream_salsa208(unsigned char *c, unsigned long long clen,
			   const unsigned char *n, const unsigned char *k)
{
	unsigned char in[16];
	unsigned char block[64];
	unsigned int i;
	unsigned int u;

	if (!clen)
		return 0;

	for (i = 0; i < 8; ++i)
		in[i] = n[i];
	for (i = 8; i < 16; ++i)
		in[i] = 0;

	while (clen >= 64) {
		crypto_core_salsa208(c, in, k, sigma);

		u = 1;
		for (i = 8; i < 16; ++i) {
			u += (unsigned int)in[i];
			in[i] = u;
			u >>= 8;
		}

		clen -= 64;
		c += 64;
	}

	if (clen) {
		crypto_core_salsa208(block, in, k, sigma);
		for (i = 0; i < clen; ++i)
			c[i] = block[i];
	}
	return 0;
}

int crypto_stream_salsa208_xor(unsigned char *c, const unsigned char *m,
			       unsigned long long mlen, const unsigned char *n,
			       const unsigned char *k)
{
	unsigned char in[16];
	unsigned char block[64];
	unsigned int i;
	unsigned int u;

	if (!mlen)
		return 0;

	for (i = 0; i < 8; ++i)
		in[i] = n[i];
	for (i = 8; i < 16; ++i)
		in[i] = 0;

	while (mlen >= 64) {
		crypto_core_salsa208(block, in, k, sigma);
		for (i = 0; i < 64; ++i)
			c[i] = m[i] ^ block[i];

		u = 1;
		for (i = 8; i < 16; ++i) {
			u += (unsigned int)in[i];
			in[i] = u;
			u >>= 8;
		}

		mlen -= 64;
		c += 64;
		m += 64;
	}

	if (mlen) {
		crypto_core_salsa208(block, in, k, sigma);
		for (i = 0; i < mlen; ++i)
			c[i] = m[i] ^ block[i];
	}
	return 0;
}
