
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef __FreeBSD__

/* system */
#include <sys/stat.h>
#include <sys/disk.h>
#include <sys/cdio.h>

/* libowfat */
#include "stralloc.h"

/* frisbee-ng */
#include "strallocX.h"
#include "term.h"
#include "defs.h"
#include "dev.h"
#include "die.h"
#include "mem.h"

static int testdisk_and_add(const char *path);

static int testdisk_and_add(const char *path)
{
	struct stat st;
	int r;

	r = lstat(path, &st);
	if (r == -1)
		return 0;	/* not found or no permission */

	/* no duplication of the same disk */
	if (!S_ISLNK(st.st_mode))
		return 0;

	return 1;
}

/**
 * get all harddisk devices we can access and return them in one stralloc
 *
 * @return: "/dev/ad0\0/dev/da0\0 ..."

a = root
b = swap
c = whole disk
d = obselete whole disk
e-h = everything else

 */
stralloc *dev_get_disks(void)
{
	stralloc *rv, sa;
	int i, j;

	stralloc_init(&sa);
	if (!stralloc_copys(&sa, ""))
		die_nomem();

	rv = mem_talloc(stralloc, 1);	/* return value */
	stralloc_init(rv);
	if (!stralloc_copys(rv, ""))
		die_nomem();

	for (i = 0; i < 32; i++) {

		/* IDE disk devices */
		if (!stralloc_copys(&sa, "/dev/ad"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* IDE disk devices (before 4.0-RELEASE) */
		if (!stralloc_copys(&sa, "/dev/wd"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* SCSI disk devices */
		if (!stralloc_copys(&sa, "/dev/da"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* SCSI disk devices (before 3.0-RELEASE) */
		if (!stralloc_copys(&sa, "/dev/sd"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* SATA disk devices */
		if (!stralloc_copys(&sa, "/dev/ada"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

	}
	stralloc_free(&sa);

	return rv;
}

/**
 * dev_getdisksize - return disksize in bytes
 *
 * #define DIOCGMEDIASIZE  _IOR('d', 129, off_t)
 * Get the size of the entire device in bytes. This should be a
 * multiple of the sectorsize.
 */
u64 dev_getdisksize(int fd)
{
	u64 bytes;
	int r;

	r = ioctl(fd, DIOCGMEDIASIZE, &bytes);
	if (r == 0)
		return bytes;

	/* may be a file? */
	if (r == -1) {
		off_t o = lseek(fd, 0, SEEK_END);

		if (o != (off_t) - 1) {
			bytes = o;
			return bytes;
		}
	}

	die_msg(111, "ioctl(DIOCGMEDIASIZE) and lseek(fd) failed!");
	return 0;		/* not reached */
}

/**
 * dev_getsectorsize - return logical sector size of disk
 *
 * #define DIOCGSECTORSIZE _IOR('d', 128, u_int)
 * Get the sectorsize of the device in bytes.  The sectorsize is the
 * smallest unit of data which can be transfered from this device.
 * Usually this is a power of two but it may not be. (ie: CDROM audio)
 */
int dev_getsectorsize(int fd)
{
	int lbs;

	if (ioctl(fd, DIOCGSECTORSIZE, &lbs) == -1)
		return 512;

	return lbs;
}

/**
 * dev_reread_ptable - re-read the partition table of disk
 */
void dev_reread_ptable(int fd)
{
	return;			/* is there sth. we can do? */
}

#endif
