
/**
 * Copyright © 2008 - 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef __linux__

/* system */
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "stralloc.h"
#include "str.h"
#include "fmt.h"
#include "byte.h"
#include "io.h"

/* frisbee-ng */
#include "strallocX.h"
#include "term.h"
#include "defs.h"
#include "dev.h"
#include "die.h"
#include "mem.h"
#include "wrapper.h"

static int testdisk_and_add(const char *path);
static int is_blockdev(int fd);

static int testdisk_and_add(const char *path)
{
	struct stat st;
	int r;

	r = lstat(path, &st);
	if (r == -1)
		return 0;	/* not found or no permission */
	if (!S_ISBLK(st.st_mode))
		return 0;	/* not an block device */

	return 1;
}

static int is_blockdev(int fd)
{
	struct stat st;
	int r;

	r = fstat(fd, &st);
	if (r == -1)
		fatal("fstat() failed");

	if (S_ISBLK(st.st_mode))
		return 1;

	return 0;
}

/**
 * get all harddisk devices we can access and return them in one stralloc
 * - not to much intelligence here... try and fail is okay!
 * - no loop devices (use shell scripts!)
 * - no device mapper stuff (use shell scripts!)
 *
 * @return: "/dev/sda\0/dev/sdb\0.../dev/hdz\0"
 */
stralloc *dev_get_disks(void)
{
	stralloc *rv, sa;
	int i, j;

	stralloc_init(&sa);
	if (!stralloc_copys(&sa, ""))
		die_nomem();

	rv = mem_talloc(stralloc, 1);	/* return value */
	stralloc_init(rv);
	if (!stralloc_copys(rv, ""))
		die_nomem();

	for (i = 0; i < 26; i++) {	/* 'a'..'z' */
		char x = 'a' + i;

		/* SCSI, SAS and SATA disk devices */
		if (!stralloc_copys(&sa, "/dev/sd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* MFM, RLL and IDE hard disk */
		if (!stralloc_copys(&sa, "/dev/hd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* ACSI disk (68k/Atari) */
		if (!stralloc_copys(&sa, "/dev/ad"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Flash Translation Layer (FTL) filesystems */
		if (!stralloc_copys(&sa, "/dev/ftl"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Parallel port IDE disk devices */
		if (!stralloc_copys(&sa, "/dev/pd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* I2O hard disk */
		if (!stralloc_copys(&sa, "/dev/i2o/hd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* NAND Flash Translation Layer filesystem */
		if (!stralloc_copys(&sa, "/dev/nftl"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* IBM S/390 DASD block storage */
		if (!stralloc_copys(&sa, "/dev/dasd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Inverse NAND Flash Translation Layer */
		if (!stralloc_copys(&sa, "/dev/inftl"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* User-mode virtual block device */
		if (!stralloc_copys(&sa, "/dev/ubd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Compressed block device */
		if (!stralloc_copys(&sa, "/dev/cbd/"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* IBM iSeries virtual disk */
		if (!stralloc_copys(&sa, "/dev/iseries/vd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* USB block devices */
		if (!stralloc_copys(&sa, "/dev/ub"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Xen Virtual Block Device */
		if (!stralloc_copys(&sa, "/dev/xvd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Resident Flash Disk Flash Translation Layer */
		if (!stralloc_copys(&sa, "/dev/rfd"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* SSFDC Flash Translation Layer filesystem */
		if (!stralloc_copys(&sa, "/dev/ssfdc"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}
	}

	for (i = 0; i < 16; i++) {	/* 0..9 (controller) */
		for (j = 0; j < 32; j++) {	/* 0..32 (disk) */

			/* Mylex DAC960 PCI RAID controller */
			if (!stralloc_copys(&sa, "/dev/rd/c"))
				die_nomem();
			if (!stralloc_cat_uint(&sa, i))
				die_nomem();
			if (!stralloc_cats(&sa, "d"))
				die_nomem();
			if (!stralloc_cat_uint(&sa, j))
				die_nomem();
			if (!stralloc_0(&sa))
				die_nomem();
			if (testdisk_and_add(sa.s)) {
				if (!stralloc_cat(rv, &sa))
					die_nomem();
			}

			/* Compaq Intelligent Drive Array */
			if (!stralloc_copys(&sa, "/dev/ida/c"))
				die_nomem();
			if (!stralloc_cat_uint(&sa, i))
				die_nomem();
			if (!stralloc_cats(&sa, "d"))
				die_nomem();
			if (!stralloc_cat_uint(&sa, j))
				die_nomem();
			if (!stralloc_0(&sa))
				die_nomem();
			if (testdisk_and_add(sa.s)) {
				if (!stralloc_cat(rv, &sa))
					die_nomem();
			}

			/* Compaq Next Generation Drive Array */
			if (!stralloc_copys(&sa, "/dev/cciss/c"))
				die_nomem();
			if (!stralloc_cat_uint(&sa, i))
				die_nomem();
			if (!stralloc_cats(&sa, "d"))
				die_nomem();
			if (!stralloc_cat_uint(&sa, j))
				die_nomem();
			if (!stralloc_0(&sa))
				die_nomem();
			if (testdisk_and_add(sa.s)) {
				if (!stralloc_cat(rv, &sa))
					die_nomem();
			}
		}
	}

	for (i = 0; i < 10; i++) {	/* '0'..'9' */
		char x = '0' + i;

		/* Metadisk (RAID) devices (hm?) */
		if (!stralloc_copys(&sa, "/dev/md"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* MMC block devices */
		if (!stralloc_copys(&sa, "/dev/mmcblk"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Parallel port ATAPI disk devices */
		if (!stralloc_copys(&sa, "/dev/pf"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* AMI HyperDisk RAID controller */
		if (!stralloc_copys(&sa, "/dev/amiraid/ar"))
			die_nomem();
		if (!stralloc_catb(&sa, &x, 1))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}
	}

	for (i = 0; i < 16; i++) {	/* 0..15 */

		/* PPDD encrypted disk driver */
		if (!stralloc_copys(&sa, "/dev/ppdd"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* IDE BIOS powered software RAID interfaces */
		if (!stralloc_copys(&sa, "/dev/ataraid/d"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Distributed Replicated Block Device (DRBD) */
		if (!stralloc_copys(&sa, "/dev/drbd"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* Enhanced Metadisk RAID (EMD) storage units */
		if (!stralloc_copys(&sa, "/dev/emd/"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}
	}

	for (i = 0; i < 32; i++) {	/* 0..31 */

		/*  Carmel 8-port SATA Disks on Second Controller */
		if (!stralloc_copys(&sa, "/dev/carmel/"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}
	}

	for (i = 0; i < 255; i++) {	/* 0..254 */

		/* Scramdisk/DriveCrypt encrypted devices */
		if (!stralloc_copys(&sa, "/dev/scramdisk/"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* NetWare (NWFS) Devices (0-255) */
		if (!stralloc_copys(&sa, "/dev/nwfs/v"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}

		/* EtherDrive Block Devices */
		if (!stralloc_copys(&sa, "/dev/etherd/"))
			die_nomem();
		if (!stralloc_cat_uint(&sa, i))
			die_nomem();
		if (!stralloc_0(&sa))
			die_nomem();
		if (testdisk_and_add(sa.s)) {
			if (!stralloc_cat(rv, &sa))
				die_nomem();
		}
	}

	stralloc_free(&sa);

	return rv;
}

/* we define it self... */
#define BLKGETSIZE64	_IOR(0x12,114,size_t)
#define BLKSSZGET	_IO(0x12,104)
#define BLKRRPART	_IO(0x12,95)

/**
 * dev_getdisksize - return disksize in bytes
 */
u64 dev_getdisksize(int fd)
{
	u64 bytes;
	int r;

	if (is_blockdev(fd)) {
		r = ioctl(fd, BLKGETSIZE64, &bytes);
		if (r == 0)
			return bytes;
		fatal("ioctl(BLKGETSIZE64) failed");
	} else {
		/* may be a file? */
		off_t o = lseek(fd, 0, SEEK_END);

		if (o != (off_t) - 1) {
			bytes = (o << 9) >> 9;
			return bytes;
		}
		fatal("lseek(SEEK_END) failed");
	}

	return 0;		/* not reached */
}

/**
 * dev_getsectorsize - return logical sector size of disk
 */
int dev_getsectorsize(int fd)
{
	int r;

	if (is_blockdev(fd))
		if (ioctl(fd, BLKSSZGET, &r) != -1)
			return r;

	return 512;
}

/**
 * re-read partition table
 */
void dev_reread_ptable(int fd)
{
	ioctl(fd, BLKRRPART);
}

/**
 * dev_create_loop - create loop device, and return the fd
 *
 * 1) create loop dev under /tmp
 * 2) test if loop is unused
 * 3) loop it and return the fd
 */

#define LOOP_SET_FD		0x4C00
#define LOOP_CLR_FD		0x4C01
#define LOOP_SET_STATUS64	0x4C04
#define LOOP_GET_STATUS64	0x4C05
#define LOOP_CTL_GET_FREE	0x4C82

struct loop_info64 {
	uint64_t lo_device;
	uint64_t lo_inode;
	uint64_t lo_rdevice;
	uint64_t lo_offset;
	uint64_t lo_sizelimit;	/* bytes, 0 == max available */
	uint32_t lo_number;
	uint32_t lo_encrypt_type;
	uint32_t lo_encrypt_key_size;
	uint32_t lo_flags;
	char lo_file_name[64];
	uint8_t lo_crypt_name[64];
	uint8_t lo_encrypt_key[32];
	uint64_t lo_init[2];
};

int dev_loop_create(int ffd, char *name, u64 offset, u64 size)
{
	char loop_device[] = "/dev/loop#";
	struct loop_info64 li;
	int i, r, lfd, cfd;

	if (geteuid() != 0)
		fatal("Need root privileges for creating loop devices!");

	cfd = open("/dev/loop-control", O_RDWR);
	if (cfd >= 0) {
		/* LOOP_CTL_GET_FREE seems never to fail.. */
		lfd = ioctl(cfd, LOOP_CTL_GET_FREE);
		close(cfd);
		loop_device[9] = '0' + lfd;
		/* .. but if it fails, lfs will be -1 in the end */
		lfd = open(loop_device, O_RDONLY);
	} else {
		/* searching the old way */
		for (i = 0; i < 10; i++) {
			loop_device[9] = '0' + i;

			lfd = open(loop_device, O_RDONLY);
			if (lfd == -1)
				continue;

			r = ioctl(lfd, LOOP_GET_STATUS64, &li);
			/* warning, dietlibc has some problems here,
			 * strace says: ENXIO
			 * but errno = 0 ?!
			 */
			if (r == -1 && errno == ENXIO)
				break;

			close(lfd);
			lfd = -1;
		}
	}

	if (lfd == -1) {
		fatal("dev_loop_create: no free loop devices?");
	}

	/* we write max 63, so the last one is zero */
	byte_zero(&li, sizeof(struct loop_info64));
	fmt_strn(li.lo_file_name, name, 63);
	li.lo_offset = offset;
	li.lo_sizelimit = size;
	li.lo_flags = 1;	/* read-only */

	r = ioctl(lfd, LOOP_SET_FD, ffd);
	if (r == -1) {
		close(lfd);
		fatal("dev_loop_create: LOOP_SET_FD failed?");
	}

	r = ioctl(lfd, LOOP_SET_STATUS64, &li);
	if (r == -1) {
		close(lfd);
		fatal("dev_loop_create: LOOP_SET_STATUS64 failed?");
	}

	io_fd2(lfd);
	return lfd;
}

void dev_loop_close(int lfd)
{
	(void)ioctl(lfd, LOOP_CLR_FD);
	(void)io_close(lfd);

	return;
}

#endif
