
/**
 * Copyright © 2008 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "buffer.h"
#include "open.h"
#include "scan.h"
#include "fmt.h"
#include "str.h"
#include "errmsg.h"

/* fribee-ng */
#include "list.h"
#include "term.h"
#include "compr.h"
#include "mem.h"
#include "main.h"
#include "bswap.h"

static int find_option(const char *option, struct compr_method *c);

static int find_option_none(const char *option);

static int find_option_level(const char *option,
			     struct compr_method_with_level *o, int min,
			     int max, int std);

static int find_option_lzma(const char *option, struct compr_method_lzma *o);
static int find_option_rep(const char *option, struct compr_method_rep *o);
static int find_option_lzo(const char *option, struct compr_method_lzo *o);
static int find_option_bzip2(const char *option, struct compr_method_bzip2 *o);
static int find_option_delta(const char *option, struct compr_method_delta *o);

static int find_option_none(const char *option)
{
	const char *x = option;

	/* the only thing should be '+' */
	if (*x == '+')
		return 1;

	if (*x)
		return -1;

	/* we have no options, so what should we read here... */
	return 0;
}

static int find_option_level(const char *option,
			     struct compr_method_with_level *o, int min,
			     int max, int std)
{

	const char *x = option;
	int len = 0;

	/* dafault */
	o->level = std;

	while (*x) {
		/* more options */
		if (*x == ':') {
			x++;
			continue;
		}

		/* last option */
		if (*x == '+') {
			x++;
			break;
		}

		/*  min <= level <= max */
		if (str_start(x, "l")) {
			len = scan_int(x + 1, &o->level);
			if (len < 1)
				return -1;
			if (o->level > max)
				return -1;
			if (o->level < min)
				return -1;
			x += 1 + len;
			continue;
		}

		/* invalid option */
		return -1;
	}

	/* return ... and maybe parse more compression options */
	return x - option;
}

static int find_option_bzip2(const char *option, struct compr_method_bzip2 *o)
{

	const char *x = option;
	int len = 0;

	/* set up default values */
	o->blockSize = 3;
	o->workFactor = 30;

	while (*x) {
		/* more options */
		if (*x == ':') {
			x++;
			continue;
		}

		/* last option */
		if (*x == '+') {
			x++;
			break;
		}

		/*  1 <= blockSize <= 9 */
		if (str_start(x, "l")) {
			len = scan_int(x + 1, &o->blockSize);
			if (len < 1)
				return -1;
			if (o->blockSize > 9)
				return -1;
			if (o->blockSize < 1)
				return -1;
			x += 1 + len;
			continue;
		}

		/*  1 <= workFactor <= 250 */
		if (str_start(x, "wf")) {
			len = scan_int(x + 2, &o->workFactor);
			if (len < 1)
				return -1;
			if (o->workFactor > 9)
				return -1;
			if (o->workFactor < 1)
				return -1;
			x += 2 + len;
			continue;
		}

		/* invalid option */
		return -1;
	}

	/* return ... and maybe parse more compression options */
	return x - option;
}

static int find_option_delta(const char *option, struct compr_method_delta *o)
{

	const char *x = option;
	int len = 0;

	/* set up default value */
	o->blockSize = 100;

	while (*x) {
		/* more options */
		if (*x == ':') {
			x++;
			continue;
		}

		/* last option */
		if (*x == '+') {
			x++;
			break;
		}

		/*  1 <= blockSize <= 9 */
		if (str_start(x, "bs")) {
			len = scan_int(x + 2, &o->blockSize);
			if (len < 1)
				return -1;
			if (o->blockSize > 2000)
				return -1;
			if (o->blockSize < 2)
				return -1;
			x += 2 + len;
			continue;
		}

		/* invalid option */
		return -1;
	}

	/* return ... and maybe parse more compression options */
	return x - option;
}

static int find_option_lzma(const char *option, struct compr_method_lzma *o)
{
	const char *x = option;
	int len = 0;

	while (*x) {
		/* more options */
		if (*x == ':') {
			x++;
			continue;
		}

		/* last option */
		if (*x == '+') {
			x++;
			break;
		}

		/* 0 <= lc <= 8, default = 3 */
		if (str_start(x, "lc")) {
			len = scan_int(x + 2, &o->lc);
			if (len != 1)
				return -1;
			if (o->lc > 8)
				return -1;
			if (o->lc < 0)
				return -1;
			x += 2 + len;
			continue;
		}

		/* 0 <= lp <= 4, default = 0 */
		if (str_start(x, "lp")) {
			len = scan_int(x + 2, &o->lc);
			if (len != 1)
				return -1;
			if (o->lc > 4)
				return -1;
			if (o->lc < 0)
				return -1;
			x += 2 + len;
			continue;
		}

		/* 0 <= pb <= 4, default = 2 */
		if (str_start(x, "pb")) {
			len = scan_int(x + 2, &o->lc);
			if (len != 1)
				return -1;
			if (o->lc > 4)
				return -1;
			if (o->lc < 0)
				return -1;
			x += 2 + len;
			continue;
		}

		/* 5 <= fb <= 273, default = 32 */
		if (str_start(x, "fb")) {
			len = scan_int(x + 2, &o->fb);
			if (len < 0)
				return -1;
			if (o->fb > 273)
				return -1;
			if (o->fb < 5)
				return -1;
			x += 2 + len;
			continue;
		}

		/* 0 - hashChain Mode, 1 - binTree mode (default) */
		if (str_start(x, "m")) {
			len = scan_int(x + 1, &o->btMode);
			if (len != 1)
				return -1;
			if (o->btMode > 1)
				return -1;
			if (o->btMode < 0)
				return -1;
			x += 1 + len;
			continue;
		}

		/* 2, 3 or 4, default = 4 */
		if (str_start(x, "hb")) {
			len = scan_int(x + 2, &o->HashBytes);
			if (len != 1)
				return -1;
			if (o->HashBytes > 4)
				return -1;
			if (o->HashBytes < 2)
				return -1;
			x += 2 + len;
			continue;
		}

		/* 0 - fast, 1 - normal, default = 1 */
		if (str_start(x, "a")) {
			len = scan_int(x + 1, &o->algo);
			if (len != 1)
				return -1;
			if (o->algo > 1)
				return -1;
			if (o->algo < 0)
				return -1;
			x += 1 + len;
			continue;
		}

		/* 1<<12 (4KB) .. 1<<27 (128M), XXX: maybe KB? */
		if (str_start(x, "d")) {
			len = scan_int(x + 1, &o->dictSize);
			if (len != 2)
				return -1;
			if (o->dictSize > 27)
				return -1;
			if (o->dictSize < 12)
				return -1;
			x += 1 + len;
			continue;
		}

		/*  0 <= level <= 9 */
		if (str_start(x, "l")) {
			len = scan_int(x + 1, &o->level);
			if (len < 1)
				return -1;
			if (o->level > 9)
				return -1;
			if (o->level < 0)
				return -1;
			x += 1 + len;
			continue;
		}

		/* invalid option */
		return -1;
	}

	/* setup default, if no option was given */
	if (len == 0)
		o->level = 3;

	/* return ... and maybe parse more compression options */
	return x - option;
}

static int find_option_rep(const char *option, struct compr_method_rep *o)
{
	const char *x = option;
	int len = 0;

	/* set up default values */
	o->BlockSize = 32;	/* -b -> in MiB */
	o->MinMatchLen = 256;	/* -l */
	o->Barrier = 8;		/* -d -> in MiB */
	o->SmallestLen = 0;	/* -s */
	o->HashBits = 0;	/* -h */
	o->Amplifier = 1;	/* -a */

	while (*x) {
		/* more options */
		if (*x == ':') {
			x++;
			continue;
		}

		/* last option */
		if (*x == '+') {
			x++;
			break;
		}

		/* "b{N} - blocksize in mb (default: 64)" */
		if (str_start(x, "b")) {
			len = scan_int(x + 1, &o->BlockSize);
			if (len < 1)
				return -1;
			if (o->BlockSize > 4096)
				return -1;
			if (o->BlockSize < 1)
				return -1;
			x += 1 + len;
			continue;
		}

		/* "l{N} - minimal match len (default: 8192)" */
		if (str_start(x, "l")) {
			len = scan_int(x + 1, &o->MinMatchLen);
			if (len < 2)
				return -1;
			if (o->MinMatchLen > 32 * 1024)
				return -1;
			if (o->MinMatchLen < 10)
				return -1;
			x += 1 + len;
			continue;
		}

		/* "d{N} - barrier for smaller matches (8MB)" */
		if (str_start(x, "d")) {
			len = scan_int(x + 1, &o->Barrier);
			if (len < 1)
				return -1;
			if (o->Barrier > 1024)
				return -1;
			if (o->Barrier < 0)
				return -1;
			x += 1 + len;
			continue;
		}

		/* "s{N} - minimal match len after barrier (0)" */
		if (str_start(x, "s")) {
			len = scan_int(x + 1, &o->SmallestLen);
			if (len < 1)
				return -1;
			if (o->SmallestLen > 32 * 1024)
				return -1;
			if (o->SmallestLen < 10)
				return -1;
			x += 1 + len;
			continue;
		}

		/* "h{N} - hash bits (default: calculated from size)" */
		if (str_start(x, "h")) {
			len = scan_int(x + 1, &o->HashBits);
			if (len < 1)
				return -1;
			if (o->HashBits > 32)
				return -1;
			if (o->HashBits < 0)
				return -1;
			x += 1 + len;
			continue;
		}

		/* "a{N} - coefficient of search \"amplification\" (default: 1)" */
		if (str_start(x, "a")) {
			len = scan_int(x + 1, &o->Amplifier);
			if (len != 1)
				return -1;
			if (o->Amplifier > 9)
				return -1;
			if (o->Amplifier < 0)
				return -1;
			x += 1 + len;
			continue;
		}

		/* invalid option */
		return -1;
	}

	if (o->SmallestLen == 0)
		o->SmallestLen = o->MinMatchLen;

	if (o->SmallestLen > o->MinMatchLen)
		o->SmallestLen = o->MinMatchLen;

	/* return ... and maybe parse more compression options */
	return x - option;
}

static int find_option_lzo(const char *option, struct compr_method_lzo *o)
{
	const char *x = option;

	o->compress = lzo1c_1_compress;
	o->memory = LZO1C_MEM_COMPRESS;
	o->magic = LZO1C_MAGIC;

	while (*x) {
		/* more options */
		if (*x == ':') {
			x++;
			continue;
		}

		/* last option */
		if (*x == '+') {
			x++;
			break;
		}

		/* -c lzo:1-1 */
		if (str_equal(x, "1-1")) {
			o->compress = lzo1_compress;
			o->memory = LZO1_MEM_COMPRESS;
			o->magic = LZO1_MAGIC;
			x += 3;
			continue;
		}

		/* -c lzo:1-99 */
		if (str_equal(x, "1-99")) {
			o->compress = lzo1_99_compress;
			o->memory = LZO1_99_MEM_COMPRESS;
			o->magic = LZO1_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1a-1 */
		if (str_equal(x, "1a-1")) {
			o->compress = lzo1a_compress;
			o->memory = LZO1A_MEM_COMPRESS;
			o->magic = LZO1A_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1a-99 */
		if (str_equal(x, "1a-99")) {
			o->compress = lzo1a_99_compress;
			o->memory = LZO1A_99_MEM_COMPRESS;
			o->magic = LZO1A_MAGIC;
			x += 5;
			continue;
		}

		/* -c lzo:1b-1 */
		if (str_equal(x, "1b-1")) {
			o->compress = lzo1b_1_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		if (str_equal(x, "1b-2")) {
			o->compress = lzo1b_2_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		if (str_equal(x, "1b-3")) {
			o->compress = lzo1b_3_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		if (str_equal(x, "1b-4")) {
			o->compress = lzo1b_4_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		if (str_equal(x, "1b-5")) {
			o->compress = lzo1b_5_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		if (str_equal(x, "1b-6")) {
			o->compress = lzo1b_6_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		if (str_equal(x, "1b-7")) {
			o->compress = lzo1b_7_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		if (str_equal(x, "1b-8")) {
			o->compress = lzo1b_8_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1b-9 */
		if (str_equal(x, "1b-9")) {
			o->compress = lzo1b_9_compress;
			o->memory = LZO1B_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1b-99 */
		if (str_equal(x, "1b-99")) {
			o->compress = lzo1b_99_compress;
			o->memory = LZO1B_99_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 5;
			continue;
		}

		/* -c lzo:1b-999 */
		if (str_equal(x, "1b-999")) {
			o->compress = lzo1b_999_compress;
			o->memory = LZO1B_999_MEM_COMPRESS;
			o->magic = LZO1B_MAGIC;
			x += 6;
			continue;
		}

		/* -c lzo:1c-1 */
		if (str_equal(x, "1c-1")) {
			o->compress = lzo1c_1_compress;
			o->memory = LZO1C_MEM_COMPRESS;
			o->magic = LZO1C_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1c-9 */
		if (str_equal(x, "1c-9")) {
			o->compress = lzo1c_1_compress;
			o->memory = LZO1C_MEM_COMPRESS;
			o->magic = LZO1C_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1c-99 */
		if (str_equal(x, "1c-99")) {
			o->compress = lzo1c_99_compress;
			o->memory = LZO1C_99_MEM_COMPRESS;
			o->magic = LZO1C_MAGIC;
			x += 5;
			continue;
		}

		/* -c lzo:1c-999 */
		if (str_equal(x, "1c-999")) {
			o->compress = lzo1c_999_compress;
			o->memory = LZO1C_999_MEM_COMPRESS;
			o->magic = LZO1C_MAGIC;
			x += 6;
			continue;
		}

		/* -c lzo:1f-1 */
		if (str_equal(x, "1f-1")) {
			o->compress = lzo1f_1_compress;
			o->memory = LZO1F_MEM_COMPRESS;
			o->magic = LZO1F_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1f-999 */
		if (str_equal(x, "1f-999")) {
			o->compress = lzo1f_999_compress;
			o->memory = LZO1F_999_MEM_COMPRESS;
			o->magic = LZO1F_MAGIC;
			x += 6;
			continue;
		}

		/* -c lzo:1x-1 */
		if (str_equal(x, "1x-1")) {
			o->compress = lzo1x_1_compress;
			o->memory = LZO1X_MEM_COMPRESS;
			o->magic = LZO1X_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1x-999 */
		if (str_equal(x, "1x-999")) {
			o->compress = lzo1x_999_compress;
			o->memory = LZO1X_999_MEM_COMPRESS;
			o->magic = LZO1X_MAGIC;
			x += 6;
			continue;
		}

		/* -c lzo:1y-1 */
		if (str_equal(x, "1y-1")) {
			o->compress = lzo1y_1_compress;
			o->memory = LZO1Y_MEM_COMPRESS;
			o->magic = LZO1Y_MAGIC;
			x += 4;
			continue;
		}

		/* -c lzo:1y-999 */
		if (str_equal(x, "1y-999")) {
			o->compress = lzo1y_999_compress;
			o->memory = LZO1Y_999_MEM_COMPRESS;
			o->magic = LZO1Y_MAGIC;
			x += 6;
			continue;
		}

		/* -c lzo:1z-999 */
		if (str_equal(x, "1z-999")) {
			o->compress = lzo1z_999_compress;
			o->memory = LZO1Z_999_MEM_COMPRESS;
			o->magic = LZO1Z_MAGIC;
			x += 6;
			continue;
		}

		/* -c lzo:2a-999 */
		if (str_equal(x, "2a-999")) {
			o->compress = lzo2a_999_compress;
			o->memory = LZO2A_999_MEM_COMPRESS;
			o->magic = LZO2A_MAGIC;
			x += 6;
			continue;
		}

		/* invalid option */
		return -1;
	}

	/* return ... and maybe parse more compression options */
	return x - option;
}

/* returns count of parsed bytes */
static int find_option(const char *option, struct compr_method *c)
{
	int r, len = str_len(c->name);
	const char *x = option + len;

	/* (id 1) rep -> "" */
	if (c->id == CFLAGS_REP) {
		c->options = &c->u.rep;
		c->compress = compress_rep_data;
		r = find_option_rep(x, &c->u.rep);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 2) delta -> "no options" */
	if (c->id == CFLAGS_DELTA) {
		c->options = &c->u.delta;
		c->compress = compress_delta_data;
		r = find_option_delta(x, &c->u.delta);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 3) lzo -> "lzo:1x-999" */
	if (c->id == CFLAGS_LZO) {
		c->options = &c->u.lzo;
		c->compress = compress_lzo_data;
		r = find_option_lzo(x, &c->u.lzo);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 4) lzf */
	if (c->id == CFLAGS_LZF) {
		c->options = 0;
		c->compress = compress_lzf_data;
		r = find_option_none(x);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 5) lz4 -> "lz4:l{N}" */
	if (c->id == CFLAGS_LZ4) {
		c->options = &c->u.lz4;
		c->compress = compress_lz4_data;
		r = find_option_level(x, &c->u.lz4, 1, 200, 20);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 6) lz4hc -> "lz4hc:l{N}" */
	if (c->id == CFLAGS_LZ4HC) {
		c->options = &c->u.lz4hc;
		c->compress = compress_lz4hc_data;
		r = find_option_level(x, &c->u.lz4hc, 0, 16, 0);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 7) lzjb */
	if (c->id == CFLAGS_LZJB) {
		c->options = 0;
		c->compress = compress_lzjb_data;
		r = find_option_none(x);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 8) lzma -> "options" */
	if (c->id == CFLAGS_LZMA) {
		c->options = &c->u.lzma;
		c->compress = compress_lzma_data;
		r = find_option_lzma(x, &c->u.lzma);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 9) gzip -> "gzip:l{N}" */
	if (c->id == CFLAGS_GZIP) {
		c->options = &c->u.gzip;
		c->compress = compress_gzip_data;
		r = find_option_level(x, &c->u.gzip, 1, 9, 5);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 10) bzip2 -> "bzip2:l{N}:" */
	if (c->id == CFLAGS_BZIP2) {
		c->options = &c->u.bzip2;
		c->compress = compress_bzip2_data;
		r = find_option_bzip2(x, &c->u.bzip2);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 11) fastlz -> "fastlz:l{N}" */
	if (c->id == CFLAGS_FASTLZ) {
		c->options = &c->u.fastlz;
		c->compress = compress_fastlz_data;
		r = find_option_level(x, &c->u.fastlz, 1, 2, 2);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 12) quicklz -> "quicklz:l{N}" */
	if (c->id == CFLAGS_QUICKLZ) {
		c->options = &c->u.quicklz;
		c->compress = compress_quicklz_data;
		r = find_option_level(x, &c->u.quicklz, 1, 3, 3);
		if (r == -1)
			goto help;
		return len + r;
	}

	/* (id 13) zstd -> "zstd:l{N}" */
	if (c->id == CFLAGS_ZSTD) {
		c->options = &c->u.zstd;
		c->compress = compress_zstd_data;
		r = find_option_level(x, &c->u.zstd, 1, 21, 5);
		if (r == -1)
			goto help;
		return len + r;
	}

 help:
	find_compression_usage(c->id);
	fng_exit(0);

	/* make gcc happy */
	return 0;
}

/**
 * find_compression - parse -c commandline
 *
 * @name: name/options to parse
 * sample: "rzip+delta+gzip:l3"
 *
 * returns the number of found methods within the option string
 */
int find_compression(const char *option, struct compr_method **cm)
{
	struct list_head *l;
	int i, methods, found = 0;
	const char *x = option;
	struct compr_method *c;

	/* count '+' with the compression option */
	for (i = 0, methods = 1; i < (int)str_len(option); i++)
		if (option[i] == '+')
			methods++;

	/* allocate space for the options */
	c = mem_tzalloc(struct compr_method, methods);

	/* sth. like: "rzip:l1+delta+gzip:l1" -> means 3 methods */
	for (i = 0; i < methods; i++) {
		list_for_each(l, &compression_types) {
			struct compression_type *ct;
			ct = list_entry(l, struct compression_type, list);

			if (str_start(x, ct->name)) {
				c[i].id = ct->id;
				c[i].name = ct->name;
				c[i].option = x;
				x += find_option(x, &c[i]);
				found++;
				break;
			}
		}
	}			/* for all + */

	if (methods != found)
		find_compression_help();

	*cm = c;
	return methods;
}

/**
 * find_compression_usage - give usage information for all methods
 */
void find_compression_help()
{

	term_print("\nPre-Compression methods:\n");
	find_compression_usage(CFLAGS_REP);
	find_compression_usage(CFLAGS_DELTA);

	term_print("\nCompressing Methods:\n");
	find_compression_usage(CFLAGS_LZO);
	find_compression_usage(CFLAGS_LZF);
	find_compression_usage(CFLAGS_LZ4);
	find_compression_usage(CFLAGS_LZ4HC);
	find_compression_usage(CFLAGS_LZJB);
	find_compression_usage(CFLAGS_FASTLZ);
	find_compression_usage(CFLAGS_QUICKLZ);
	find_compression_usage(CFLAGS_LZMA);
	find_compression_usage(CFLAGS_GZIP);
	find_compression_usage(CFLAGS_BZIP2);
	find_compression_usage(CFLAGS_ZSTD);

	fng_exit(0);
}

/**
 * find_compression_usage - give usage info for one method
 */
void find_compression_usage(int id)
{
	struct list_head *l;

	list_for_each(l, &compression_types) {
		struct compression_type *ct;
		ct = list_entry(l, struct compression_type, list);

		if (id != ct->id)
			continue;
		term_hline(78, "-");
		term_print("\n");
		term_print(ct->name);
		term_print(" ");
		term_print(ct->usage);
		term_print("\n");
	}
}
