
/**
 * Copyright © 2012 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

/* libowfat */
#include "socket.h"
#include "buffer.h"
#include "scan.h"
#include "open.h"
#include "io.h"
#include "str.h"
#include "ip4.h"
#include "ip6.h"

/* frisbee-ng */
#include "strallocX.h"
#include "term.h"
#include "crypt.h"
#include "compr.h"
#include "main.h"
#include "tty.h"
#include "dev.h"
#include "die.h"
#include "getkey.h"
#include "cdb.h"
#include "log.h"
#include "wrapper.h"
#include "list.h"
#include "bitmap.h"
#include "os.h"
#include "sc.h"
#include "fs.h"
#include "pt_reader.h"
#include "sysstat.h"
#include "openX.h"
#include "progress.h"
#include "mem.h"
#include "read_skipfile.h"
#include "packet.h"

#include "fng-adhoc.h"

/* internal functions: */
static void usage(void);
static void AdhocServerDisk(struct adhoc_config *cfg);
static void AdhocServerPartition(struct adhoc_config *cfg);
static void AdhocClient(struct adhoc_config *cfg);

static void usage(void)
{
	term_print("\nUsage:\n");
	term_print(" fng-adhoc -s groupname [options] -d /dev/sda\n");
	term_print(" fng-adhoc -s groupname [options] -d /dev/sda,1,2\n");
	term_print(" fng-adhoc -g groupname [options] -d /dev/sda\n");
	term_print(" fng-adhoc [options] -d /dev/disk\n");
	term_print("\nGeneral Options:\n");
	term_print(" %16s%s", "-h or -?", "show help\n");
	term_print(" %16s%s", "-v", "show version\n");
	term_print("\nServer Mode Options:\n");
	term_print(" %16s%s", "-s gname",
		   "start in server mode with group named <gname>\n");
	term_print(" %16s%s", "-b blocks", "number of blocks per chunk\n");
	term_print(" %16s%s", "-d device", "select <device> as source\n");
	term_print(" %16s%s", "-m bytes",
		   "tune the max size of network packets (default: 1472)\n");
	term_print(" %16s%s", "-q len",
		   "max chunk count in the outgoing queue\n");
	term_print("\nClient Mode Options: (Default Mode)\n");
	term_print(" %16s%s", "-g <name>",
		   "start in client mode and join the group <name>\n");
	term_print(" %16s%s", "-d device",
		   "destination device, you may specify -d multiple times\n");
	term_print("\nApplication Options:\n");
	term_print(" %16s%s", "-S skipfile",
		   "do not clone files, which are given in Skipfile\n");
	term_print(" %16s%s", "-c <type>",
		   "use this compression type (default \"rep+delta+zstd\")\n");
	term_print(" %16s%s", "-l <level>",
		   "use this compression level (default 3)\n");
	term_print(" %16s%s", "-i <interface>",
		   "use this network <interface> (ipv6)\n");
	term_print(" %16s%s", "-M <ip>",
		   "set multicast ip address to <ip> (ipv4 or ipv6)\n");
	fng_exit(0);
}

/**
 * create a loop device to read the bitmap of some filesystem
 * - close it in the end and just return
 * - if sth. fails, create panic
 */
static void AdhocServerReadFilesystem(struct sstate *s, int i, int j)
{
	struct adhoc_config *cfg = s->cfg;
	struct todo *todo = &s->todo[j];
	struct fsinfo *fsi = &s->todo[j].fsi;
	struct list_head *l;
	ssize_t r = -1;
	u64 c, chunk_index;
	int cur_blks_chunk, isfirst, fd;

	if (geteuid() != 0)
		fatal("Need root privileges for creating loop devices!");

	/* should call fatal, if no loopdev can be used! */
	fd = dev_loop_create(s->fd, s->device, s->sl[i].start * 512,
			     s->sl[i].size * 512);

	io_setcookie(fd, fsi);
	fsi->fd = fd;
	fsi->id = FSID_AUTO;
	fsi->path = s->device;
	fsi->psize = s->sl[i].size * 512;

	// term_printflush("nr=%d index=%d sl=%d\n", j, i, s->sl[i].number);
	s->todo[j].index = j;
	s->todo[j].number = s->sl[i].number;
	s->todo[j].start = s->sl[i].start;	/* sectors, each 512 */
	s->todo[j].size = s->sl[i].size;	/* sectors, each 512 */

	/* try all known filesystems, use raw in the end */
	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->fs_getbitmap)
			r = fs->fs_getbitmap(fd);
		if (r == 0) {
			break;
		}
	}

	if (r != 0)
		fatal("Can not read filesystem information :/");

	/* loop can be closed now... */
	dev_loop_close(fd);

	/* calculate needed chunks */
	todo->clen = fsi->bused + cfg->blks_chunk - 1;
	todo->clen /= cfg->blks_chunk;

	/* build chunk structure for this partition */
	todo->c = mem_talloc(struct chunk, todo->clen);

	/* bitmab will be used all the time, we don't remember all the offsets,
	   just the bitmap, for saving memory on very big partitions */
	cur_blks_chunk = 0;
	chunk_index = 0;
	isfirst = 1;		/* next block is startblock for current chunk */
	for (c = 0; c < fsi->blocks; c++) {
		if (!bitmap_get(fsi->bitmap, c))
			continue;

		/* get first used block of chunk, store it */
		if (isfirst) {
			todo->c[chunk_index].idx_chunk = chunk_index;
			todo->c[chunk_index].idx_block = c;
			isfirst = 0;
		}

		/* check, if chunk is full */
		cur_blks_chunk += 1;
		todo->c[chunk_index].st_blocks = cur_blks_chunk;
		if (cur_blks_chunk == cfg->blks_chunk) {
			cur_blks_chunk = 0;	/* reset counter */
			chunk_index += 1;
			isfirst = 1;
		}
	}			/* for(all blocks) */

	term_printflush("chunk_index=%d clen=%d\n", chunk_index, todo->clen);
	for (c = 0; c < todo->clen; c++) {
		term_printflush("chunk[%d] st_blocks=%d\n",
				todo->c[c].idx_chunk, todo->c[c].st_blocks);
	}

	return;
}

static void AdhocServerPartition(struct adhoc_config *cfg)
{
	unused(cfg);
	fatal("later");
}

/**
 * 1) checken welche disk / partitionen geclont werden sollen
 * 2) checken, wie groß muss zielplatte mind. sein
 * 3) errechnen, wie viele packete usw. benötigt werden

 * 4) check via broadcast, ob anderer server schon da ist...
 * 5) entsprechende sockets auf machen
 * 6) auf clients warten
 * 7) abarbeiten der geforderten arbeit
 */
static void AdhocServerDisk(struct adhoc_config *cfg)
{
	struct sstate *s = mem_talloc(struct sstate, 1);
	u32 i, j, copyall = 0;
	size_t len;

	/* init the server sstate structure */
	s->cfg = cfg;
	s->device = cfg->devices.s;
	s->group = cfg->group.s;

	/* replace first ',' with \0 -> /dev/sda,2,4,5 -> part 2,5,6 to copy */
	len = str_chr(s->device, ',');
	if (len != str_len(s->device)) {
		/* else: open_device_ro() would fail */
		s->device[len] = 0;
	} else
		copyall = 1;
	s->fd = open_device_ro(s->device);
	io_setcookie(s->fd, s);

	/* check and read partitions */
	s->ptype = pt_getlist(s->fd, &s->sl, &s->slen, &s->type);
	if (s->ptype == 0) {
		term_printflush("no partition found on %s\n\n", s->device);
		return;
	}

	if (copyall) {
		/* the esay way, clone all slices... */
		for (i = 0; i < s->slen; i++)
			s->sl[i].todo = 1;
		s->tlen = s->slen;
	} else {
		/* check, what needs to be cloned */
		char *x = s->device + len + 1;

		if (scan_uint(x, &j) == 0)
			goto kaputt;

		for (i = 0; i < s->slen; i++)
			if (s->sl[i].number == j) {
				s->sl[i].todo = 1;
				s->tlen = 1;	/* first given slice */
			}

		while (*x) {
			if (*x++ != ',')
				continue;
			if (scan_uint(x, &j) == 0)
				goto kaputt;
			for (i = 0; i < s->slen; i++)
				if (s->sl[i].number == j) {
					s->sl[i].todo = 1;
					s->tlen += 1;
				}
		}
	}

 kaputt:
	if (s->tlen == 0) {
		term_printflush("Don't know what todo!\n");
		return;
	}

	/* 2) checken, wie groß muss zielplatte mind. sein */
	s->min_devsize = 0;
	for (i = 0; i < s->slen; i++)
		if (s->sl[i].todo == 1) {
			u64 sz = s->sl[i].start + s->sl[i].size - 1;

			if (sz > s->min_devsize)
				s->min_devsize = sz;
		}

	s->todo = mem_talloc(struct todo, s->tlen);

	term_print("device:%s group:%s partitions:%d todo=%d minsize=%U\n",
		   s->device, s->group, s->slen, s->tlen, s->min_devsize);
	term_flush();

	/* 3) errechnen, wie viele packete usw. benötigt werden */
	for (i = 0, j = 0; i < s->slen; i++)
		if (s->sl[i].todo == 1)
			AdhocServerReadFilesystem(s, i, j++);

	for (i = 0; i < s->tlen; i++) {
		term_print
		    ("device:%s group:%s partitions:%d todo=%d minsize=%U\n",
		     s->device, s->group, s->slen, s->tlen, s->min_devsize);
		term_flush();
	}

	len = scan_ip6(cfg->mip, cfg->fng_mcip);
	if (len != str_len(cfg->mip)) {
		fatal("ip");
	}

	/* XXX: ttl, port, ip usw. als parameter?! */
	s->fd_mc = sc_multicast(cfg->fng_mcip, cfg->fng_port, 10, 1);
	s->fd_bc = sc_broadcast();
	io_setcookie(s->fd_mc, s);
	io_setcookie(s->fd_bc, s);

	/* this would be done anyway... */
	io_wantread(s->fd_mc);
	io_wantread(s->fd_bc);
	term_printflush("fd_mc=%d fd_bc=%d\n", s->fd_mc, s->fd_bc);

	/*
	 * packet details server:
	 * 1) immer allocated -> antwort zum client hello
	 * 2) immer allocated -> antwort zum client get-configuration
	 * 3) immer allocated -> antwort zum client get-bootrecord
	 * 4) je nach anfrage allocated -> antwort zum client get-chunk
	 */

	term_s(T_CLEAR);
	while (1) {
		term_printflush(".");

		ClientHello(s->fd_bc);
		sleep(1);
	}

	//term_s(T_CLEAR);
	term_flush();

	for (;;) {
		int64 fd;

		io_wait();
		while ((fd = io_canread()) != -1) {
			if (fd == s->fd_mc) {
				/* receive multicast packet */
				handle_adhoc_server_read(s->fd_mc);
			} else {
				/* handle user input */
				die_msg(111,
					"don't know which input comes in here!");
			}
		}

#if 0
		while ((fd = io_canwrite()) != -1) {
			if (fd == s->fd_mc) {
				handle_adhoc_server_read(s->fd_mc);
			} else {
				die_msg(111,
					"server, kann fd ohne cookie schreiben?!");
				continue;
			}
			// io_dontwantwrite(i);
		}
#endif

	}

	return;
}

/*
 * 1) client hello (broadcast check for servers)
 * 2) client get-configuration (via udp, maxmtu, chunksize, min_devsize)
 * 3) client get-bootrecord (via udp, mbr bzw gpt... packet 1/10 2/10 ... je nach
 *    maxmtu, je partition die werte für belegt... und max chunkcount)
 * 4) client get-chunk 0..chunkcount
 * 5) client get-chunkpart (0..chunkcount) part 0..N
 */
static void AdhocClient(struct adhoc_config *cfg)
{
	struct cstate *s = mem_talloc(struct cstate, cfg->devcount);
	u32 i, j;
	size_t len;
	int64 fd_bc;		/* socket for broadcast */
	int64 fd_mc;		/* socket for multicast */

	for (i = 0, j = 0; i < cfg->devcount; i++) {
		s[i].path = cfg->devices.s + j;
		s->fd = open_device_rw(s[i].path);
		s[i].devsize = dev_getdisksize(s->fd);
		j += str_len(cfg->devices.s + j) + 1;
		// can be done, when we know, how much data is to be written!
		// progress_init(&d->p, total, current);
	}

	for (i = 0; i < cfg->devcount; i++) {
		term_print("dest[%d]: path=%s size=%U\n", i, s[i].path,
			   s[i].devsize);
	}

	len = scan_ip6(cfg->mip, cfg->fng_mcip);
	if (len != str_len(cfg->mip)) {
		fatal("ip");
	}

	/* XXX: ttl, port, ip usw. als parameter?! */
	fd_mc = sc_multicast(cfg->fng_mcip, cfg->fng_port, 10, 1);
	fd_bc = sc_broadcast();
	io_setcookie(fd_mc, s);
	io_setcookie(fd_bc, s);

	/* this would be done anyway... */
	io_wantread(fd_mc);
	io_wantread(fd_bc);
	//io_wantwrite(fd_bc);
	term_printflush("fd_mip=%d fd_bcip=%d\n", fd_mc, fd_bc);

	/*
	 * packet details client:
	 * 1) immer allocated -> antwort zum client hello
	 * 2) immer allocated -> antwort zum client get-configuration
	 * 3) immer allocated -> antwort zum client get-bootrecord
	 * 4) je nach anfrage allocated -> antwort zum client get-chunk
	 */

	s->state = STATE_HELLO;

	for (;;) {
		int64 fd;

		io_wait();
		while ((fd = io_canread()) != -1) {
			if (fd == fd_mc) {
				handle_adhoc_client_read(fd);
				/* receive multicast packet */
			} else if (fd == 0) {
				term_printflush("fd=0 lesen...\n");
				/* handle user input */
			} else {
				/* handle user input */
				die_msg(111,
					"don't know which input comes in here!");
			}
		}
	}

	return;
}

int fng_adhoc(int argc, char **argv)
{
	struct adhoc_config *cfg = mem_talloc(struct adhoc_config, 1);
	int mode = FNG_ADHOC_CLIENT;
	int opt;

	/* default values, user may change them... */
	cfg->mtu = 1472;	/* 1500 - 20 - 8 */
	cfg->queuelen = 3;
	cfg->blks_chunk = 256;

	cfg->mip = "224.102.110.103";
	cfg->fng_port = 7983;
	cfg->fng_bcip[0] = cfg->fng_bcip[1] = 255;
	cfg->fng_bcip[2] = cfg->fng_bcip[3] = 255;

	/* set up our default group name */
	stralloc_init(&cfg->group);
	if (!stralloc_copys(&cfg->group, "adhoc@"))
		die_nomem();
	if (!stralloc_cats(&cfg->group, os_gethostname()))
		die_nomem();

	stralloc_init(&cfg->devices);
	cfg->devcount = 0;

	/* compression to use */
	stralloc_init(&cfg->compr);
	cfg->compr_count = 0;

	while ((opt = getopt(argc, argv, "h?vM:s:g:i:b:m:q:d:p:c:S:")) != -1) {
		switch (opt) {
		case 'h':
		case '?':
			usage();
		case 'v':
			version();
		case 'M':
			cfg->mip = optarg;
		case 's':
			/* -s groupname -> create this group  */
			mode = FNG_ADHOC_SERVER_DISK;
			if (!stralloc_copys(&cfg->group, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->group))
				die_nomem();
			break;
		case 'g':
			/* -g groupname -> join this group */
			mode = FNG_ADHOC_CLIENT;
			if (!stralloc_copys(&cfg->group, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->group))
				die_nomem();
			break;
		case 'i':
			/* -i interface */
			sc_scope_id = socket_getifidx(optarg);
			break;
		case 'b':
			/* -b blocks (blocks per chunk) */
			if (scan_int(optarg, &cfg->blks_chunk) == 0)
				usage();
			break;
		case 'm':
			/* -m bytes (sth. like mtu for the network) */
			if (scan_int(optarg, &cfg->mtu) == 0)
				usage();
			break;
		case 'q':
			/* -q length (global variable) */
			if (scan_int(optarg, &cfg->queuelen) == 0)
				usage();
			break;
		case 'd':
			/* -d device / -d device,1,2,3 */
			if (!stralloc_cats(&cfg->devices, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->devices))
				die_nomem();
			cfg->devcount++;
			break;
		case 'p':
			/* -p partition */
			mode = FNG_ADHOC_SERVER_PART;
			if (!stralloc_copys(&cfg->devices, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->devices))
				die_nomem();
			/* -1 means one partiton only, no mbr and so on... */
			cfg->devcount = -1;
			break;
		case 'c':
			if (!optarg)
				usage();

			if (case_equals(optarg, "help"))
				find_compression_usage(0);

			/* -c compression */
			if (!stralloc_cats(&cfg->compr, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->compr))
				die_nomem();
			cfg->compr_count++;
			break;
		case 'S':
			/* -S skipfile */
			read_skipfile(optarg);
			break;
		default:
			usage();
		}
	}

	/* we need some device ;) */
	if (cfg->devcount == 0)
		usage();

	/* device_count can be -1 or 0 @ server mode! */
	if (mode == FNG_ADHOC_SERVER_DISK)
		if (cfg->devcount > 1)
			die_msg(111, "Only one device needed @ server mode!");

	/* start client or server mode ... */
	if (mode == FNG_ADHOC_SERVER_DISK)
		AdhocServerDisk(cfg);
	else if (mode == FNG_ADHOC_SERVER_PART)
		AdhocServerPartition(cfg);
	else
		AdhocClient(cfg);

	return fng_exit(0);
}
