
/**
 * Copyright © 2010 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 1
 */

/* system */
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

/* libowfat */
#include "byte.h"
#include "open.h"
#include "openreadclose.h"
#include "scan.h"
#include "fmt.h"
#include "buffer.h"
#include "io.h"
#include "str.h"
#include "case.h"
#include "term.h"
#include "errmsg.h"

/* frisbee-ng */
#include "read_skipfile.h"
#include "main.h"
#include "tty.h"
#include "fs.h"
#include "crypt.h"
#include "die.h"
#include "log.h"
#include "openX.h"
#include "pt_reader.h"
#include "mem.h"
#include "progress.h"
#include "dev.h"
#include "transport.h"

/**
 * fng-clone - clone some partition to one or more destinations
 */

#define FNG_CLONE_UNKNOWN 0
#define FNG_CLONE_DISK    1
#define FNG_CLONE_PART    2

/* options of fng-clone */
struct config {
	int transportmode;
	int readtest;
	int threads;

	/* source device */
	stralloc source;

	/* destination devices */
	stralloc devices;
	u32 devcount;
};

/* generic handling structure for cloning */
struct state {
	struct config *cfg;	/* for easy access of the configuration */

	int64 fd;		/* opened source device in read only mode */
	char *source;		/* path to source device or file */
	u64 min_devsize;	/* needed disksize in bytes */

	/* all slices found on source device */
	struct slice *sl;	/* slice information of device */
	u32 slice_count;	/* number of slices in *sl */
	char *type;		/* name of partition type (gpt, bsd, mbr...) */
	int ptype;		/* id of partition type, see pt_reader.h  */

	struct partition *todo;	/* each slice gets it's own thing here */
	u32 tlen;		/* number of todo's */

	/* destination(s), cfg->devcount */
	stralloc *dest;
	char **dst_path;	/* ptr to dest stralloc */
	int64 *dst_fd;		/* destination fds */
	u64 *dst_size;		/* devsize of destination fds */

	/* overall progress status */
	progress_t *progress;
};

static void usage(void);
static void CloneDisk(struct config *cfg);
static void ClonePartition(struct config *cfg);

static void usage(void)
{
	term_print("\nUsage:\n");
	term_print(" fng-clone [options] -d <harddisk> <dest1> [ .. destN ]\n");
	term_print
	    (" fng-clone [options] -p <partition> <dest1> [ .. destN ]\n");
	term_print("\nGeneral Options:\n");
	term_print(" %16s%s", "-h or -?", "show help\n");
	term_print(" %16s%s", "-v", "show version\n");
	term_print("\nApplication Commands:\n");
	term_print(" %16s%s", "-d <harddisk>",
		   "disk cloning, argument is source harddisk\n");
	term_print(" %16s%s", "-p <partition>",
		   "partiton cloning, argument is source partition\n");
	term_print("\nApplication Options:\n");
	term_print(" %16s%s", "-t threadcount",
		   "number of threads per disk (default: nr of cpu cores)\n");
	term_print(" %16s%s", "-S skipfile",
		   "do not clone files, which are given in Skipfile\n");
	term_print(" %16s%s", "-r",
		   "read the written data and check integrity (good harddisk test)\n");
	term_print("\nTransport Options:\n");
	term_print(" %16s%s", "-m readwrite",
		   "read(2) write(2) mode (default)\n");
	term_print(" %16s%s", "-m rdwr", "alias for readwrite\n");
#ifdef HAVE_SPLICE
	term_print(" %16s%s", "-m splice", "use splice(2)\n");
	term_print(" %16s%s", "-m splice-tee",
		   "use splice(2)-tee(2) combination\n");
#endif
	term_print(" %16s%s", "-m sendfile", "use sendfile(2)\n");
	term_print("\nExamples:\n");
	term_print(" fng-clone [options] -p /dev/sda1 /dev/sdb1 /dev/sdc1\n");
	term_print
	    (" -> copy partition /dev/sda1 to /dev/sdb1 and /dev/sdc1 (they have to exist)\n\n");
	term_print(" fng-clone [options] -d /dev/sda,1,2 /dev/sdd\n");
	term_print(" -> copy partition table and partiton 1+2 to /dev/sdd\n");
	fng_exit(0);
}

/**
 * create a loop device to read the bitmap of some filesystem
 * - close it in the end and just return
 * - if sth. fails, create panic
 */
static void CloneDisk_ReadFS(struct state *s, int i, int j)
{
	struct fsinfo *fsi = &s->todo[j].fsi;
	struct list_head *l;
	ssize_t r = -1;

	/* will call fatal, if no loopdev can be used! */
	fsi->fd =
	    dev_loop_create(s->fd, s->source, s->sl[i].start * 512,
			    s->sl[i].size * 512);

	io_setcookie(fsi->fd, fsi);
	fsi->id = FSID_AUTO;
	fsi->path = s->source;
	fsi->psize = s->sl[i].size * 512;

	/* term_printflush("nr=%d index=%d sl=%d\n", j, i, s->sl[i].number); */
	s->todo[j].index = j;
	s->todo[j].number = s->sl[i].number;
	s->todo[j].start = s->sl[i].start * 512;
	s->todo[j].size = s->sl[i].size * 512;

	/* try all known filesystems, use raw in the end */
	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->fs_getbitmap)
			r = fs->fs_getbitmap(fsi->fd);
		if (r == 0) {
			/* must be set for postclone() */
			fsi->fs = fs;
			break;
		}
	}

	/* partition progress */
	fsi->progress = mem_talloc(progress_t, 1);
	progress_init(fsi->progress, fsi->bused * fsi->bsize);
	debug(1, 1, "progress[P]: %U", fsi->bused * fsi->bsize);

	if (r != 0)
		fatal("Can not read filesystem information :/");

	/* loop can be closed now... */
	dev_loop_close(fsi->fd);

	return;
}

/**
 * 1) checken welche disk / partitionen geclont werden sollen
 * 2) checken, wie groß muss zielplatte mind. sein
 * 3) prüfen, ob zielplatten groß genug sind
 * 4) suchen der zu kopierenden blöcke
 * 5) klonen
 */
static void CloneDisk(struct config *cfg)
{
	struct state *s = mem_tzalloc(struct state, 1);
	u32 i, j, copyall = 0;
	size_t len;

	/* init the state structure */
	s->cfg = cfg;
	s->source = cfg->source.s;

	/* setup destination devices */
	s->dest = &cfg->devices;
	s->dst_path = mem_talloc(char *, cfg->devcount);

	s->dst_fd = mem_talloc(int64, cfg->devcount);
	s->dst_size = mem_talloc(u64, cfg->devcount);
	for (i = 0, j = 0; i < cfg->devcount; i++) {
		s->dst_path[i] = cfg->devices.s + j;
		s->dst_fd[i] = open_device_rw(s->dst_path[i]);
		s->dst_size[i] = dev_getdisksize(s->dst_fd[i]);
		j += str_len(cfg->devices.s + j) + 1;
	}

	/* replace first ',' with \0 -> /dev/sda,1,2,3 -> part 1,2,3 to copy */
	len = str_chr(s->source, ',');
	if (len != str_len(s->source)) {
		/* else: open_device_ro() would fail */
		s->source[len] = 0;
	} else
		copyall = 1;
	s->fd = open_device_ro(s->source);

	/* check and read partitions */
	s->ptype = pt_getlist(s->fd, &s->sl, &s->slice_count, &s->type);
	if (s->ptype == 0) {
		term_printflush("no partition found on %s\n\n", s->source);
		return;
	}

	if (copyall) {
		/* the esay way, clone all slices... */
		s->tlen = s->slice_count;
		for (i = 0; i < s->slice_count; i++) {
			if (is_extended(s->sl[i].container)) {
				s->sl[i].todo = 0;
				s->tlen -= 1;
				continue;
			}
			s->sl[i].todo = 1;
		}
	} else {
		/* check, what needs to be cloned */
		char *x = s->source + len + 1;

		if (scan_uint(x, &j) == 0)
			goto kaputt;

		for (i = 0; i < s->slice_count; i++)
			if (s->sl[i].number == j) {
				s->sl[i].todo = 1;
				s->tlen = 1;	/* first given slice */
			}

		while (*x) {
			if (*x++ != ',')
				continue;
			if (scan_uint(x, &j) == 0)
				goto kaputt;
			for (i = 0; i < s->slice_count; i++)
				if (s->sl[i].number == j) {
					s->sl[i].todo = 1;
					s->tlen += 1;
				}
		}
	}

 kaputt:
	if (s->tlen == 0) {
		term_printflush("Don't know what todo!\n");
		return;
	}

	/* 2) checken, wie groß muss zielplatte sein muss */
	s->min_devsize = 0;
	for (i = 0; i < s->slice_count; i++)
		if (s->sl[i].todo == 1) {
			u64 sz = s->sl[i].start + s->sl[i].size - 1;

			if (sz > s->min_devsize)
				s->min_devsize = sz;
		}

	/* 3) prüfen, ob zielplatten groß genug sind */
	for (i = 0, j = 0; i < cfg->devcount; i++) {
		if (s->dst_size[i] < s->min_devsize * 512) {
			term_printflush
			    ("Disk %d (%s) is to small, size=%U needed=%U\n",
			     i + 1, s->dst_path[i], s->dst_size[i],
			     s->min_devsize * 512);
			fng_exit(2);
		}
	}

	/* 4) suchen der zu kopierenden blöcke */
	s->todo = mem_talloc(struct partition, s->tlen);

	for (i = 0, j = 0; i < s->slice_count; i++)
		if (s->sl[i].todo == 1)
			CloneDisk_ReadFS(s, i, j++);

	/* 5) clone partiton tables */
	pt_clone(s->fd, s->ptype, s->sl, s->slice_count, cfg->devcount,
		 s->dst_fd, s->dst_size);

	/* 6a) gesamt progress zeugs berechnen  */
	{
		u64 used = 0;

		for (i = 0; i < s->tlen; i++) {
			struct fsinfo *fsi = &s->todo[i].fsi;

			used += fsi->bused * fsi->bsize;
		}

		/* progress for whole disk */
		s->progress = mem_tzalloc(struct progress, 1);

		progress_init(s->progress, used);
		debug(1, 1, "progress[A]: %U", used);
	}

	/* 6b) info what is done... */
	if (!stralloc_copys(&term_left_status_top, cfg->source.s))
		die_nomem();
	if (!stralloc_cats(&term_left_status_top, " -> "))
		die_nomem();
	for (i = 0; i < cfg->devcount; i++) {
		if (!stralloc_cats(&term_left_status_top, s->dst_path[i]))
			die_nomem();
		if (!stralloc_cats(&term_left_status_top, " "))
			die_nomem();
	}

	/* 7) klonen der partitionen */
	for (i = 0; i < s->tlen; i++) {
		struct fsinfo *fsi = &s->todo[i].fsi;

		/* for postclone(), s->fd must held fsi structure! */
		io_setcookie(s->fd, fsi);
		transport_doit(s->fd, &s->todo[i], cfg->devcount, s->dst_fd,
			       cfg->transportmode, cfg->readtest, s->progress,
			       i + 1, s->tlen + 1, cfg->threads);
	}

	/* 8) free resources... not needed, cause we end here! */

	return;
}

static void ClonePartition(struct config *cfg)
{
	struct state *s = mem_talloc(struct state, 1);
	struct fsinfo *fsi;
	struct list_head *l;
	ssize_t r = -1;
	u32 i, j;

	/* init the state structure */
	s->cfg = cfg;
	s->source = cfg->source.s;
	s->fd = open_device_ro(s->source);
	s->tlen = 1;		/* we have just ONE partition ;) */
	s->todo = mem_talloc(struct partition, s->tlen);

	/* unused things */
	s->type = 0;
	s->ptype = 0;
	s->sl = 0;
	s->slice_count = 0;

	/* setup destination devices */
	s->dest = &cfg->devices;
	s->dst_path = mem_talloc(char *, cfg->devcount);

	s->dst_fd = mem_talloc(int64, cfg->devcount);
	s->dst_size = mem_talloc(u64, cfg->devcount);
	for (i = 0, j = 0; i < cfg->devcount; i++) {
		s->dst_path[i] = cfg->devices.s + j;
		s->dst_fd[i] = open_device_rw(s->dst_path[i]);
		s->dst_size[i] = dev_getdisksize(s->dst_fd[i]);
		j += str_len(cfg->devices.s + j) + 1;
	}

	/* checken, wie groß muss zielplatte sein muss */
	s->min_devsize = dev_getdisksize(s->fd);

	/* prüfen, ob ziele groß genug sind */
	for (i = 0, j = 0; i < cfg->devcount; i++) {
		if (s->dst_size[i] < s->min_devsize) {
			term_printflush
			    ("Partition %d (%s) is to small, size=%U needed=%U\n",
			     i + 1, s->dst_path[i], s->dst_size[i],
			     s->min_devsize);
			fng_exit(2);
		}
	}

	/* suchen der zu kopierenden blöcke */
	s->todo[0].index = 0;
	s->todo[0].number = 0;
	s->todo[0].start = 0;
	s->todo[0].size = s->min_devsize;

	fsi = &s->todo[0].fsi;
	fsi->fd = s->fd;
	fsi->id = FSID_AUTO;
	fsi->path = s->source;
	fsi->psize = s->min_devsize;
	io_setcookie(s->fd, fsi);

	/* try all known filesystems, use raw in the end */
	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->fs_getbitmap)
			r = fs->fs_getbitmap(fsi->fd);
		if (r == 0) {
			/* must be set for postclone() */
			fsi->fs = fs;
			break;
		}
	}

	if (r != 0)
		fatal("Can not read filesystem information :/");

	/* progress for the partition */
	fsi->progress = mem_tzalloc(struct progress, 1);

	progress_init(fsi->progress, fsi->bused * fsi->bsize);

	/* overall progress is the same as the one partition */
	s->progress = mem_tzalloc(struct progress, 1);

	progress_init(s->progress, fsi->bused * fsi->bsize);

	/* info what is done... */
	if (!stralloc_copys(&term_left_status_top, cfg->source.s))
		die_nomem();
	if (!stralloc_cats(&term_left_status_top, " -> "))
		die_nomem();
	for (i = 0; i < cfg->devcount; i++) {
		if (!stralloc_cats(&term_left_status_top, s->dst_path[i]))
			die_nomem();
		if (!stralloc_cats(&term_left_status_top, " "))
			die_nomem();
	}

	/* klonen der partition */
	transport_doit(s->fd, &s->todo[0], cfg->devcount, s->dst_fd,
		       cfg->transportmode, cfg->readtest, s->progress, 1, 1,
		       cfg->threads);

	return;
}

int fng_clone(int argc, char **argv)
{
	struct config *cfg = mem_talloc(struct config, 1);
	int mode = FNG_CLONE_UNKNOWN;
	int opt;

	cfg->readtest = 0;
	cfg->transportmode = FNG_TMODE_DEFAULT;
	stralloc_init(&cfg->source);
	stralloc_init(&cfg->devices);
	cfg->devcount = 0;

	/* default worker count = nr of cpu cores */
	cfg->threads = fng_cpucount;

	while ((opt = getopt(argc, argv, "h?vd:p:t:S:m:r")) != -1) {

		switch (opt) {
		case 'h':
		case '?':
			usage();	/* never returns */

		case 'v':
			version();	/* never returns */

		case 'd':
			/* -d device / -d device,1,2,3 */
			if (!stralloc_copys(&cfg->source, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->source))
				die_nomem();
			mode = FNG_CLONE_DISK;
			break;

		case 'p':
			/* -p partition */
			if (!stralloc_copys(&cfg->source, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->source))
				die_nomem();
			mode = FNG_CLONE_PART;
			break;

		case 't':
			/* -t N -> number of worker threads */
			if (scan_int(optarg, &cfg->threads) == 0)
				die_msg(2, "Wrong thread count!");
			if (cfg->threads > fng_cpucount)
				die_msg(2, "To many threads!");
			if (cfg->threads < 1)
				die_msg(2,
					"Zero worker, who should make the work?");
			break;

		case 'S':
			/* -S skipfile */
			read_skipfile(optarg);
			break;

		case 'm':
			/* -m transport modes and other ones... */
			cfg->transportmode = find_transportmode(optarg);
			break;

		case 'r':
			/* -r => re-read written data */
			cfg->readtest = 1;
			break;
		default:
			usage();	/* never returns */
		}
	}

	/* -p or -d must be given! */
	if (mode == FNG_CLONE_UNKNOWN)
		usage();

	if (argc == optind)
		usage();

	for (;;) {
		if (!stralloc_cats(&cfg->devices, argv[optind++]))
			die_nomem();
		if (!stralloc_0(&cfg->devices))
			die_nomem();
		cfg->devcount++;
		if (argc == optind)
			break;
	}

	/* init terminal stuff */
	term_resize();

	if (mode == FNG_CLONE_DISK)
		CloneDisk(cfg);
	else
		ClonePartition(cfg);

	sync();
	sync();
	sync();

	return fng_exit(0);
}
