
/**
 * Copyright © 2014 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

/* libowfat */
#include "byte.h"
#include "open.h"
#include "openreadclose.h"
#include "scan.h"
#include "fmt.h"
#include "str.h"
#include "sig.h"
#include "buffer.h"
#include "io.h"
#include "case.h"
#include "term.h"
#include "errmsg.h"

/* frisbee */
#include "main.h"
#include "tty.h"
#include "io.h"
#include "mem.h"
#include "fs.h"
#include "crypt.h"
#include "pt_reader.h"
#include "die.h"
#include "log.h"
#include "wrapper.h"
#include "openX.h"
#include "compr.h"
#include "progress.h"
#include "image.h"
#include "read_skipfile.h"
#include "fmtX.h"
#include "dev.h"
#include "get_cycles.h"

/**
 * testing compression modes and options
 * - this code is ugly to read, but for some testing... it's ok :)
 */

struct stats {
	char *name;		/* name of method */
	u64 cycles;		/* cpu cycles */
};

struct config {

	/* running encode or decode? */
	int mode;
#define FNG_UNKNOWN  0
#define FNG_ENCODE   1
#define FNG_DECODE   2

	char *infile;
	int fd_in;

	char *outfile;
	int fd_out;

	/**
	 * compression type and options to use
	 * - only used by encoding
	 */
	stralloc compr;
	int compr_count;

	/* for benchmarking */
	struct stats *stats;

	/**
	 * save the best compression, when multiple -c options are there
	 */
	void *best_cbuf;
	u32 best_clen;

	int benchmark;
	int syncfile;

	/* encryption */
	stralloc password;
};

/* internal functions */
static void usage(int mode);
static void DecodeFile(struct config *cfg);
static void EncodeFile(struct config *cfg);

static void usage(int mode)
{
	term_print("\nUsage:\n");

	if (mode == FNG_DECODE)
		term_print(" fng-decode [options] in-file out-file\n");
	else
		term_print(" fng-encode [options] in-file out-file\n");

	term_print("\nGeneral Options:\n");
	term_print(" %16s%s", "-h or -?", "show help\n");
	term_print(" %16s%s", "-v", "show version\n");
	term_print("\nApplication Commands:\n");
	term_print(" %16s%s", "-n N", "run benchmark N times\n");

	if (mode == FNG_ENCODE) {
		term_print(" %16s%s", "-s", "fsync fresh written file\n");
		term_print("\nCompressing Options:\n");
		term_print(" %16s%s",
			   "-c <method1:option1+method2:option2+...>",
			   "use this compression type\n");
		term_print("\nEncrypting Options:\n");
		term_print(" %16s%s", "-P <password>",
			   "encrypt the image with xsalsa20\n");
	} else {
		term_print("\nDecrypting Options:\n");
		term_print(" %16s%s", "-P <password>",
			   "decrypt with password\n");
	}

	fng_exit(0);
}

/* do some decoding */
static void DecodeFile(struct config *cfg)
{
	struct fng_compr *c = mem_talloc(struct fng_compr, 1);
	struct stats *s = mem_tzalloc(struct stats, 1);
	struct tms tms;
	int b, slen;

	/* how much data do we have, get it once in beginning */
	slen = lseek(cfg->fd_in, 0, SEEK_END);

	/* some basic input checking... */
	if (slen < FNGHDR) {
		die_msg(11, "Input to Small");
	}

	compr_init(c);
	if (cfg->password.len)
		compr_set_password(c, &cfg->password);

	/* cbuf -> ubuf */
	for (b = 0; b < cfg->benchmark; b++) {
		cycles_t ts, te;

		/* read the full source file */
		c->clen = slen;
		c->cbuf = mem_alloc(c->clen);
		pread_loop2(cfg->fd_in, c->cbuf, c->clen, 0);

		ts = get_cycles();
		decompress_data(c);
		te = get_cycles();

		/* for calculating the average later */
		s->cycles += te - ts;

		if (b == 0) {
			/* write it */
			write_loop2(cfg->fd_out, c->ubuf, c->ulen);
		}
		mem_free(c->ubuf);
	}

	times(&tms);
	term_printflush
	    ("uncompress %z -> %z (cycles=%4H) utime=%3U stime=%2U\n", slen,
	     c->ulen, s->cycles / cfg->benchmark, tms.tms_utime, tms.tms_stime);

	if (cfg->syncfile)
		fsync(cfg->fd_out);
	close(cfg->fd_out);

	mem_free(c);
	fng_exit(0);
}

/* do some encoding */
static void EncodeFile(struct config *cfg)
{
	struct fng_compr *c = mem_talloc(struct fng_compr, 1);
	struct compr_method *cm;
	struct tms tms;
	int i, methods, option, b, ulen;
	void *ubuf;
	char *x;

	/* we need some basic input... */
	if (cfg->compr_count == 0)
		die_msg(11, "Hey, I need some compression ;)");

	/* read the full source file in the beginning... */
	compr_init(c);
	c->ulen = lseek(cfg->fd_in, 0, SEEK_END);
	c->ubuf = mem_alloc(c->ulen);
	pread_loop2(cfg->fd_in, c->ubuf, c->ulen, 0);
	close(cfg->fd_in);

	/* save these buffers... */
	ulen = c->ulen;
	ubuf = c->ubuf;

	/* how much -c options should be tested... */
	cfg->stats = mem_tzalloc(struct stats, cfg->compr_count);

	cfg->best_clen = 0;
	cfg->best_cbuf = 0;

	x = cfg->compr.s;
	for (option = 0; option < cfg->compr_count; option++) {
		cfg->stats[option].name = x;
		x += str_len(x) + 1;
	}

	for (option = 0; option < cfg->compr_count; option++) {
		struct stats *s = &cfg->stats[option];
		cycles_t ts, te;	/* start, end */

		/* get options/methods of current -c option */
		methods = find_compression(s->name, &cm);
		c->clen = 0;
		c->cbuf = 0;

		/* loop, cfg->benchmark times */
		for (b = 0; b < cfg->benchmark; b++) {

			/* next round */
			c->ulen = ulen;
			c->ubuf = ubuf;

			/* when in benchmark mode, c->cbuf != zero */
			mem_free(c->cbuf);

			/* the whole compr. chain... */
			ts = get_cycles();
			for (i = 0; i < methods; i++) {

				/* term_printflush("[%4d] %20s compress()\n", i,
				   cm[i].name); */

				/* will allocate one new cbuf */
				cm[i].compress(c, cm[i].options);

				/* first method has to set the data id 0x01 */
				if (i == 0)
					compr_set_fnghdr_last(c->cbuf);

				/* free old compressed data, but not the first ubuf */
				if (c->ubuf != ubuf)
					mem_free(c->ubuf);

				/**
				 * next uncompressed buffer is the currently new
				 * compressed one... clen only represents the
				 * compressed data part, not the hdr... */
				c->ulen = c->clen + FNGHDR;
				c->ubuf = c->cbuf;
			}
			te = get_cycles();

			/* for calculating the average later */
			s->cycles += te - ts;
		}

		if (cfg->best_cbuf == 0) {
			/* first result is the best ;) */
			cfg->best_cbuf = c->cbuf;
			cfg->best_clen = c->clen;
		} else if (cfg->best_clen > c->clen) {
			/* save better result and free old best one */
			cfg->best_cbuf = c->cbuf;
			cfg->best_clen = c->clen;
		}

		/* free unused compressed data */
		if (c->cbuf != cfg->best_cbuf)
			mem_free(c->cbuf);

		times(&tms);
		term_printflush
		    ("[%4d] %20s utime=%3U stime=%2U | %z -> %z (cycles=%4H)\n",
		     option, s->name, tms.tms_utime, tms.tms_stime, ulen,
		     cfg->best_clen, s->cycles / cfg->benchmark);

		/* next option will re-allocate it */
		mem_free(cm);
	}			/* for all -c options */

	if (cfg->best_cbuf == 0)
		die_msg(111, "Compression not working ?!");

	/* last step, encryption if wanted ... */
	if (cfg->password.len) {
		compr_init_password(c, &cfg->password);
		c->seed = 0;	/* we use 0 only here */
		c->ubuf = cfg->best_cbuf;
		c->ulen = cfg->best_clen + FNGHDR;
		compress_encrypt_data(c);
		cfg->best_cbuf = c->cbuf;
		cfg->best_clen = c->clen;
	}

	/* write + cleanup */
	write_loop2(cfg->fd_out, cfg->best_cbuf, cfg->best_clen + FNGHDR);

	if (cfg->syncfile)
		fsync(cfg->fd_out);
	close(cfg->fd_out);

	/* eflags stuff is not freed */
	mem_free(cfg->best_cbuf);
	mem_free(c);
	fng_exit(0);
}

int fng_encode(int argc, char **argv, int mode)
{
	struct config *cfg = mem_talloc(struct config, 1);
	int opt;

	/* default values, user may change them... */
	cfg->mode = mode;	/* 1 / 2 */
	cfg->benchmark = 1;	/* default: one time */

	/* compression to use */
	stralloc_init(&cfg->password);
	stralloc_init(&cfg->compr);
	cfg->compr_count = 0;
	cfg->syncfile = 0;

	while ((opt = getopt(argc, argv, "h?vn:c:sP:")) != -1) {
		switch (opt) {
		case 'h':
		case '?':
			/* never returns */
			usage(mode);

		case 'v':
			/* never returns */
			version();

		case 'n':
			if (scan_int(optarg, &cfg->benchmark) == 0)
				die_msg(111, "benchmark needs some count");
			break;

		case 'c':
			if (!optarg)
				usage(mode);

			if (case_equals(optarg, "help"))
				find_compression_help();

			/* -c compression */
			if (!stralloc_cats(&cfg->compr, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->compr))
				die_nomem();
			cfg->compr_count++;
			break;

		case 's':
			cfg->syncfile = 1;
			break;

		case 'P':
			/* -P password */
			if (!stralloc_copys(&cfg->password, optarg))
				die_nomem();
			break;

		default:
			usage(mode);
		}
	}

	/**
	 * prog -c method:options src dst
	 */
	if (argc != optind + 2)
		usage(mode);

	cfg->infile = argv[optind];
	cfg->fd_in = open_device_ro(cfg->infile);

	cfg->outfile = argv[optind + 1];
	cfg->fd_out = open_trunc(cfg->outfile);

	if (mode == FNG_ENCODE)
		EncodeFile(cfg);
	else if (mode == FNG_DECODE)
		DecodeFile(cfg);
	else
		usage(mode);

	return fng_exit(0);
}
