
/**
 * Copyright © 2010 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 1
 */

/* system */
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

/* libowfat */
#include "byte.h"
#include "open.h"
#include "openreadclose.h"
#include "scan.h"
#include "fmt.h"
#include "str.h"
#include "sig.h"
#include "buffer.h"
#include "io.h"
#include "case.h"
#include "term.h"
#include "errmsg.h"

/* frisbee */
#include "main.h"
#include "term.h"
#include "lshw.h"
#include "tty.h"
#include "io.h"
#include "mem.h"
#include "fs.h"
#include "crypt.h"
#include "pt_reader.h"
#include "die.h"
#include "log.h"
#include "wrapper.h"
#include "openX.h"
#include "compr.h"
#include "progress.h"
#include "image.h"
#include "read_skipfile.h"
#include "get_cycles.h"
#include "fmtX.h"
#include "dev.h"
#include "splice.h"

#define FNG_IMAGE_UNKNOWN 0
#define FNG_IMAGE_DISK    1
#define FNG_IMAGE_PART    2

/* internal functions */
static void usage(void);
static void ImageDisk(struct image_config *cfg);
static void ImagePartition(struct image_config *cfg);

static void OpenImageFiles(struct img_image *s);
static void OpenSourceDevice(struct img_image *s, char *source, int diskid);
static void ReadFS(struct img_disk *s, int i, int j);

static void usage(void)
{
	term_print("\nUsage:\n");
	term_print(" fng-image [options] -d /dev/sda,1,2 image.fng\n");
	term_print(" fng-image [options] -d /dev/sda -d /dev/sdb image.fng\n");
	term_print(" fng-image [options] -p /dev/sda1 image.fng\n");
	term_print
	    (" fng-image [options] -d /dev/sdb fng://user:pass@server:sdb-neu.fng\n");
	term_print("\nGeneral Options:\n");
	term_print(" %16s%s", "-h or -?", "show help\n");
	term_print(" %16s%s", "-v", "show version\n");
	term_print("\nDisk and Partition Options:\n");
	term_print(" %16s%s", "-d <disk>", "full disk should be imaged\n");
	term_print(" %16s%s", "-d <disk,1>",
		   "partition table and first partition should be imaged\n");
	term_print(" %16s%s", "-p <partition>",
		   "partition, which should be imaged\n");
	term_print(" %16s%s", "-p / -d",
		   "can be specified more then once...\n");
	term_print("\nApplication Options:\n");
	term_print(" %16s%s", "-b chunksize",
		   "chunksize in KiB (default: 512)\n");
	term_print(" %16s%s", "-t threadcount",
		   "number of threads per disk (default: nr of cpu cores)\n");
	term_print(" %16s%s", "-S <skipfile>",
		   "can be used for skipping unwanted files on some filesystems\n");
	term_print(" %16s%s", "-C <comment>",
		   "adding some comment for image description\n");
	term_print("\nCompressing Options:\n");
	term_print(" %s\n%16s%s",
		   "-c <method1+method2:option+methodX:optionX:optionX2>", "",
		   "use this compression type (default: \"rep+delta+zstd\")\n");
	term_print("\nEncrypting Options:\n");
	term_print(" %16s%s", "-P <password>",
		   "encrypt the image with xsalsa20\n");
	fng_exit(0);
}

/**
 * open destination image files
 */
static void OpenImageFiles(struct img_image *s)
{
	char *x;
	int i;

	/**
	 * image filename ptr, may be needed later ...
	 */
	s->image_fds = mem_talloc(int64, s->cfg->image_count);
	s->image_files = mem_talloc(char *, s->cfg->image_count);

	x = s->cfg->images.s;
	for (i = 0; i < (int)s->cfg->image_count; i++) {
		s->image_files[i] = x;
		s->image_fds[i] = open_device_trunc(x);
		x += str_len(x) + 1;
	}

	return;
}

/**
 * open source disk(s)
 */
static void OpenSourceDevice(struct img_image *s, char *source, int diskid)
{
	u32 i, j, copyall = 0;
	size_t len;
	struct img_disk *disk = &s->disk[diskid];

	/* copy the source string, we modify it */
	disk->device = mem_tzalloc(char, str_len(source) + 1);

	fmt_str(disk->device, source);

	/* replace first ',' with \0 -> /dev/sda \0 1,2,3 */
	len = str_chr(disk->device, ',');
	if (len != str_len(disk->device)) {
		/* else: open_device_ro() would fail */
		disk->device[len] = 0;
	} else
		copyall = 1;

	/* open source disk */
	disk->fd = open_device_ro(disk->device);
	disk->devsize = dev_getdisksize(disk->fd);
	disk->img = s;		/* access to the big state structure */

	/* check and read partitions */
	disk->ptype =
	    pt_getlist(disk->fd, &disk->sl, &disk->slice_count, &disk->type);
	if (disk->ptype == 0) {
		term_printflush("no partition found on %s\n\n", disk->device);
		return;
	}

	if (copyall) {
		/* the esay way, image all slices... */
		disk->tlen = disk->slice_count;
		for (i = 0; i < disk->slice_count; i++) {
			if (is_extended(disk->sl[i].container)) {
				disk->sl[i].todo = 0;
				disk->tlen -= 1;
				continue;
			}
			disk->sl[i].todo = 1;
		}
	} else {
		/* check, what needs to be imaged */
		char *x = disk->device + len + 1;

		if (scan_uint(x, &j) == 0)
			goto kaputt;

		for (i = 0; i < disk->slice_count; i++)
			if (disk->sl[i].number == j) {
				disk->sl[i].todo = 1;
				disk->tlen = 1;	/* first given slice */
			}

		while (*x) {
			if (*x++ != ',')
				continue;
			if (scan_uint(x, &j) == 0)
				goto kaputt;
			for (i = 0; i < disk->slice_count; i++)
				if (disk->sl[i].number == j) {
					disk->sl[i].todo = 1;
					disk->tlen += 1;
				}
		}
	}

 kaputt:
	if (disk->tlen == 0) {
		term_printflush("Don't know what todo!\n");
		return;
	}

	/* 2) suchen der zu kopierenden blöcke */
	disk->todo = mem_talloc(struct partition, disk->tlen);
	for (i = 0, j = 0; i < disk->slice_count; i++)
		if (disk->sl[i].todo == 1)
			ReadFS(disk, i, j++);

}

/**
 * create a loop device to read the bitmap of some filesystem
 * - close it in the end and return
 * - if sth. fails, create panic
 */
static void ReadFS(struct img_disk *s, int i, int j)
{
	struct fsinfo *fsi = &s->todo[j].fsi;
	struct list_head *l;
	ssize_t r = -1;

	/* will call fatal, if no loopdev can be used! */
	fsi->fd =
	    dev_loop_create(s->fd, s->device, s->sl[i].start * 512,
			    s->sl[i].size * 512);

	io_setcookie(fsi->fd, fsi);
	fsi->id = FSID_AUTO;
	fsi->path = s->device;
	fsi->psize = s->sl[i].size * 512;

	/* term_printflush("nr=%d index=%d sl=%d\n", j, i, s->sl[i].number); */
	s->todo[j].index = j;
	s->todo[j].number = s->sl[i].number;
	s->todo[j].start = s->sl[i].start * 512;
	s->todo[j].size = s->sl[i].size * 512;

	/* try all known filesystems, use raw in the end */
	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->fs_getbitmap)
			r = fs->fs_getbitmap(fsi->fd);
		if (r == 0) {
			/* must be set for postclone() */
			fsi->fs = fs;
			break;
		}
	}

	/* partition progress */
	fsi->progress = mem_talloc(progress_t, 1);
	progress_init(fsi->progress, fsi->bused * fsi->bsize);
	progress_initadd(s->img->progress, fsi->bused * fsi->bsize);

	if (r != 0)
		fatal("Can not read filesystem information :/");

	/* loop can be closed now... */
	dev_loop_close(fsi->fd);

	return;
}

/**
 * 1) checken welche disk / partitionen in das image sollen
 * 2) suchen der zu kopierenden blöcke
 * 3) image erstellen
 */
static void ImageDisk(struct image_config *cfg)
{
	struct img_image *s = mem_tzalloc(struct img_image, 1);
	char *source;
	int i;

	/* init the img_image structure */
	s->cfg = cfg;

	/* this will initialize also an private key, if password was set */
	s->c = mem_tzalloc(struct fng_compr, 1);

	compr_init(s->c);
	if (s->cfg->password.len)
		compr_init_password(s->c, &s->cfg->password);
	else
		compr_init_seed(s->c);

	s->disks = cfg->source_count;
	s->disk = mem_tzalloc(struct img_disk, s->disks);

	s->partitions = 0;

	/* overall progress */
	s->progress = mem_talloc(progress_t, 1);
	progress_init(s->progress, 0);

	/* create destination files and open them with rw */
	OpenImageFiles(s);

	/* open source disks */
	source = s->cfg->source.s;
	for (i = 0; i < (int)s->cfg->source_count; i++) {
		OpenSourceDevice(s, source, i);
		source += str_len(source) + 1;
		s->partitions += s->disk[i].tlen;
	}

	/* create image(s) */
	image_create(s);
	return;
}

static void ImagePartition(struct image_config *cfg)
{
	term_print("later");
	return;
}

int fng_image(int argc, char **argv)
{
	struct image_config *cfg = mem_talloc(struct image_config, 1);
	int mode = FNG_IMAGE_UNKNOWN;
	int opt;

	/* setup default values */
	stralloc_init(&cfg->lshw);
	stralloc_init(&cfg->comment);
	stralloc_init(&cfg->password);

	/* no encryption per default */
	cfg->chunksize = 512;

	/* compression to use */
	stralloc_init(&cfg->compr);
	if (!stralloc_copys(&cfg->compr, "rep+delta+zstd"))
		die_nomem();
	if (!stralloc_0(&cfg->compr))
		die_nomem();

	/* source device(s) */
	stralloc_init(&cfg->source);
	cfg->source_count = 0;

	/* destination files(s) */
	stralloc_init(&cfg->images);
	cfg->image_count = 0;

	/* default worker count = nr of cpu cores */
	cfg->threads = fng_cpucount;

	while ((opt = getopt(argc, argv, "h?vb:d:p:c:l:t:C:S:P:")) != -1) {
		switch (opt) {
		case 'h':
		case '?':
			/* never returns */
			usage();

		case 'v':
			/* never returns */
			version();

		case 'b':
			/* -b 1 -> 1MiB uncompressed data per chunk */
			if (scan_uint(optarg, &cfg->chunksize) == 0) {
				die_msg(2, "reading cluster per chunk count");
			}
			break;

		case 'd':
			/* -d /dev/sda,1,2,3 */
			if (!stralloc_cats(&cfg->source, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->source))
				die_nomem();
			cfg->source_count++;

			/* image with one or more disks */
			set(mode, FNG_IMAGE_DISK);
			break;

		case 'p':
			/* -p /dev/sda1 */
			if (!stralloc_cats(&cfg->source, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->source))
				die_nomem();
			cfg->source_count++;

			/* image with just one partition */
			set(mode, FNG_IMAGE_PART);
			break;

		case 'c':
			if (!optarg)
				usage();

			if (case_equals(optarg, "help"))
				find_compression_usage(0);

			/* -c compression */
			if (!stralloc_copys(&cfg->compr, optarg))
				die_nomem();
			if (!stralloc_0(&cfg->compr))
				die_nomem();
			break;

		case 't':
			/* -t N -> number of worker threads */
			if (scan_int(optarg, &cfg->threads) == 0)
				die_msg(2, "Wrong thread count!");
			if (cfg->threads > fng_cpucount)
				die_msg(2, "To many threads!");
			if (cfg->threads < 1)
				die_msg(2,
					"Zero worker, who should make the work?");
			break;

		case 'C':
			/* -C comment */
			if (!stralloc_copys(&cfg->comment, optarg))
				die_nomem();
			break;

		case 'S':
			/* -S skipfile */
			read_skipfile(optarg);
			break;

		case 'P':
			/* -P password */
			if (!stralloc_copys(&cfg->password, optarg))
				die_nomem();
			break;

		default:
			usage();
		}
	}

	/* -p or -d must be given! */
	if (mode == FNG_IMAGE_UNKNOWN)
		usage();

	/* -p and -d specified, this is not okay! */
	if (mode == FNG_IMAGE_DISK + FNG_IMAGE_PART)
		usage();

	if (argc == optind)
		usage();

	/* user can specify more than one destination image */
	for (;;) {
		/* file1.fng file2.fng */
		if (!stralloc_cats(&cfg->images, argv[optind++]))
			die_nomem();
		if (!stralloc_0(&cfg->images))
			die_nomem();
		cfg->image_count++;

		if (argc == optind)
			break;
	}

	/* get some hardware info if lshw is installed */
	lshw(&cfg->lshw);

	/* chunksize is in KiB */
	cfg->chunksize *= 1024;

	/* init terminal stuff */
	term_resize();

	/**
	 * image partition(s)
	 */
	if (mode == FNG_IMAGE_PART) {
		ImagePartition(cfg);
		fng_exit(0);
	}

	/**
	 * image whole disk(s)
	 */
	ImageDisk(cfg);

	return fng_exit(0);
}
