
/**
 * Copyright © 2010 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 1
 */

/* system */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/statvfs.h>
#include <unistd.h>

/* libowfat */
#include "buffer.h"
#include "open.h"
#include "io.h"
#include "str.h"

/* frisbee-ng */
#include "strallocX.h"
#include "term.h"
#include "crypt.h"
#include "main.h"
#include "tty.h"
#include "main.h"
#include "dev.h"
#include "die.h"
#include "cdb.h"
#include "log.h"
#include "wrapper.h"
#include "list.h"
#include "bitmap.h"
#include "fs.h"
#include "pt_reader.h"
#include "read_skipfile.h"
#include "openX.h"
#include "mem.h"
#include "image.h"

/* internal functions: */
static void usage(void);

#define FNG_INFO_INFO        1
#define FNG_INFO_DEBUG       2
#define FNG_INFO_BITMAP      3
#define FNG_INFO_SBITMAP     4
#define FNG_INFO_BBITMAP     5
#define FNG_INFO_STATFS      6
#define FNG_INFO_PARTITIONS  7

static void doit_info(int fd);	/* -i partition */
static void doit_debug(int fd);	/* -x partition */
static void doit_bitmap(int fd);	/* -d partition */
static void doit_sbitmap(int fd);	/* -D partition */
static void doit_bbitmap(int fd);	/* -B partition */
static void doit_statvfs(int fd);	/* -s mountpoint */
static void doit_partitions(int fd);	/* -l harddisk */
static void doit_harddisks(void);	/* -L */
static void doit_filesystems(void);	/* -a */

static void usage(void)
{
	term_print("\nUsage:\n");
	term_print(" fng-info [options] [command [<device>|<file> ...]\n");
	term_print("\nGeneral Options:\n");
	term_print(" %16s%s", "-h or -?", "show help\n");
	term_print(" %16s%s", "-v", "show version\n");
	term_print("\nApplication Options:\n");
	term_print(" %16s%s", "-S <skipfile>",
		   "can be used together with -d (see manpage)\n");
	term_print(" %16s%s", "-t <type>",
		   "set fstype to <type> - used together with -d\n");
	term_print("\nApplication Commands:\n");
	term_print(" %16s%s", "-a", "show known filesystems\n");
	term_print(" %16s%s", "-L", "show available harddisks\n");
	term_print(" %16s%s", "-X", "show available harddisks (long format)\n");
	term_print(" %16s%s", "-l <harddisk>", "list partitions of harddisk\n");
	term_print(" %16s%s", "-d <partition>",
		   "dump bitmap of partition to stdout\n");
	term_print(" %16s%s", "-D <partition>",
		   "give summary of bitmap to stdout\n");
	term_print(" %16s%s", "-B <partition>",
		   "dump binary bitmap of partition to file\n");
	term_print(" %16s%s", "-i <partition>",
		   "show information of partition in a parseable format\n");
	term_print(" %16s%s", "-x <partition>",
		   "show all known information of filesystem at partition\n");
	term_print(" %16s%s", "-P <password>",
		   "use this password for opening the frisbee-ng image\n");
	term_print(" %16s%s", "-s <path>",
		   "show statvfs(2) information of mounted path\n");
	fng_exit(0);
}

static void doit_info(int fd)
{
	struct list_head *l;
	struct fsinfo *fsi = io_getcookie(fd);
	int r = -1;

	fsi = io_getcookie(fd);
	fsi->psize = dev_getdisksize(fd);

	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->id == fsi->id || fsi->id == FSID_AUTO) {
			if (fs->fs_info)
				r = fs->fs_info(fd);
			if (r == 0) {
				break;
			}
		}
	}

	if (r == -1) {
		term_printflush("Sorry, no information found @ %s :/\n",
				fsi->path);
		return;
	}

	/* wir hatten erfolg, also raus mit den infos: */
	term_print("Information for filesystem @ %s\n\n", fsi->path);
	term_print("filesystem type:   %s\n", fsi->name);
	term_print("partiton size:     %U bytes\n", (u64) fsi->psize);
	term_print("partiton size:     %HiB\n", (u64) fsi->psize);
	term_print("partiton size:     %hB\n\n", (u64) fsi->psize);
	term_print("filesystem size:   %U bytes\n", (u64) fsi->fssize);
	term_print("filesystem size:   %HiB\n", (u64) fsi->fssize);
	term_print("filesystem size:   %hB\n\n", (u64) fsi->fssize);
	term_print("fs block size:     %u\n", (u32) fsi->bsize);
	term_print("fs bitmap length:  %U\n", (u64) fsi->bitmaplen);
	term_print("fs # of blocks:    %U\n", (u64) fsi->blocks);
	term_print("fs # used blocks:  %U\n", (u64) fsi->bused);
	term_print("fs # free blocks:  %U\n\n", (u64) fsi->bfree);
	term_print("fs # total space:  %HiB\n", (u64) fsi->blocks * fsi->bsize);
	term_print("fs # used space:   %HiB\n", (u64) fsi->bused * fsi->bsize);
	term_print("fs # free space:   %HiB\n\n",
		   (u64) fsi->bfree * fsi->bsize);
	term_printflush("\n");

	return;
}

static void doit_debug(int fd)
{
	struct list_head *l;
	struct fsinfo *fsi;
	int r = -1;

	fsi = io_getcookie(fd);
	term_print("Extended information for \"%s\" ...", fsi->path);
	fsi->psize = dev_getdisksize(fd);

	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->id == fsi->id || fsi->id == FSID_AUTO) {
			if (fs->fs_debug)
				r = fs->fs_debug(fd);
			if (r == 0) {
				break;
			}
		}
	}

	if (r == -1) {
		term_printflush(" found noting :/\n");
		return;
	}

	return;
}

static void doit_bitmap(int fd)
{
	struct list_head *l;
	struct fsinfo *fsi;
	int r = -1;
	u64 i, nl;

	fsi = io_getcookie(fd);

	fsi->psize = dev_getdisksize(fd);

	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->id == fsi->id || fsi->id == FSID_AUTO) {
			if (fs->fs_getbitmap)
				r = fs->fs_getbitmap(fd);
			if (r == 0) {
				break;
			}
		}
	}

	if (r == -1) {
		term_printflush("Sorry, no bitmap found for %s :/\n",
				fsi->path);
		return;
	}

	/* FF FF FF FF │ 00 00 00 C0 */

	/* 1111:1111 1111:1111 1111:1111 1111:1111 | 1111:1111 1111:1111 1111:1111 1111:1111 */
	/* 1100:0000 0000:0000 0000:0000 0000:0000 | 0000:0000 0000:0000 0000:0000 0000:0000 */

	nl = 0;
	for (i = 0; i < fsi->blocks; i++, nl++) {
		if (nl % 640 == 0)
			term_flush();
		if (nl && nl % 64 == 0) {
			/* current offset */
			term_print(" %8U", i);
			term_s("\n");
		}
		if (nl % 8 == 0) {
			term_s(" ");
		} else if (nl % 4 == 0) {
			term_s(":");
		}
		if (bitmap_get(fsi->bitmap, i)) {
			term_s("1");
		} else {
			term_s("0");
		}
	}
	term_s("\n");
	term_flush();

	return;
}

static void doit_sbitmap(int fd)
{
	struct list_head *l;
	struct fsinfo *fsi;
	int r = -1;
	u64 i, used, free;

	fsi = io_getcookie(fd);
	fsi->psize = dev_getdisksize(fd);

	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->id == fsi->id || fsi->id == FSID_AUTO) {
			if (fs->fs_getbitmap)
				r = fs->fs_getbitmap(fd);
			if (r == 0) {
				break;
			}
		}
	}

	if (r == -1) {
		term_printflush("Sorry, no bitmap found for %s :/\n",
				fsi->path);
		return;
	}

	free = used = 0;
	/* count the clusters @ bitmap ... should be smaller now */
	for (i = 0; i < fsi->blocks; i++) {
		if (bitmap_get(fsi->bitmap, i)) {
			used++;
		} else {
			free++;
		}
	}

	term_print("fsname: %s\n", fsi->name);
	term_print("bitmap: free: %U used: %U maplen: %U\n", free, used,
		   fsi->bitmaplen * 8);
	term_print("fsinfo: free: %U used: %U total:  %U\n\n", fsi->bfree,
		   fsi->bused, fsi->blocks);
	term_flush();

	return;
}

static void doit_bbitmap(int fd)
{
	struct list_head *l;
	struct fsinfo *fsi;
	ssize_t r = -1;

	fsi = io_getcookie(fd);
	fsi->psize = dev_getdisksize(fd);

	list_for_each(l, &filesystems) {
		struct file_system *fs;

		fs = list_entry(l, struct file_system, list);

		if (fs->id == fsi->id || fsi->id == FSID_AUTO) {
			if (fs->fs_getbitmap)
				r = fs->fs_getbitmap(fd);
			if (r == 0) {
				break;
			}
		}
	}

	if (r == -1) {
		term_printflush("Sorry, no bitmap found for %s :/\n",
				fsi->path);
		return;
	}

	if (is_tty(1)) {
		fatal("will not write binary to terminal!");
		return;
	}

	write_loop2(1, fsi->bitmap, fsi->bitmaplen);

	return;
}

static void doit_statvfs(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	struct statvfs buf;

	fstatvfs(fd, &buf);	/* POSIX.1-2001 -> works also under bsd? sunos? */
	term_print("statvfs() information for filesystem @ %s\n\n", fsi->path);
	term_print("filesystem size:   %U bytes\n", buf.f_bsize * buf.f_blocks);
	term_print("filesystem size:   %HiB\n", buf.f_bsize * buf.f_blocks);
	term_print("filesystem size:   %hB\n\n", buf.f_bsize * buf.f_blocks);
	term_print("fs block size:     %U\n", buf.f_bsize);	/* file system block size */
	term_print("fragment size:     %U\n", buf.f_frsize);	/* fragment size */
	term_print("mount flags:       %U\n", buf.f_flag);	/* mount flags */
	term_print("filename max:      %U\n\n", buf.f_namemax);	/* maximum filename length */
	term_print("fs # of blocks:    %U\n", buf.f_blocks);	/* size of fs in f_frsize units */
	term_print("fs # used blocks:  %U\n", buf.f_blocks - buf.f_bfree);
	term_print("fs # free blocks:  %U\n\n", buf.f_bfree);	/* # free blocks */
	term_print("fs # of inodes:    %U\n", buf.f_files);	/* # inodes */
	/* some fs take 0 as total number :-/ */
	term_print("fs # used inodes:  %U\n", buf.f_files - buf.f_ffree);
	term_print("fs # free inodes:  %U\n", buf.f_ffree);	/* # free inodes */
	term_printflush("\n");

	return;
}

static void doit_partitions(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	struct slice *sl;
	char *type;
	u32 r, slen, i;

	r = pt_getlist(fd, &sl, &slen, &type);
	if (r == 0) {
		term_printflush("no partition found on %s\n\n", fsi->path);
		return;
	}

	fsi->psize = dev_getdisksize(fd);
	term_print("partition information for %s: type=%s, count=%d\n",
		   fsi->path, type, slen);
	term_print("nr start end size\n");
	for (i = 0; i < slen; i++)
		if (!is_extended(sl[i].container))
			term_print("%d %U %U %U (%HiB)\n", sl[i].number,
				   sl[i].start, sl[i].start + sl[i].size - 1,
				   sl[i].size * 512, sl[i].size * 512);
	term_flush();

	return;
}

static void doit_harddisks(void)
{
	stralloc *disks = dev_get_disks();
	unsigned int i;
	char vendor[9];
	char model[17];
	char rev[5];
	char serial[255];
	u64 disksize = 0;

	if (disks->len == 0) {
		die_msg(0, "No harddisks found ?!");
	}

	term_print("%11s %8s %16s %-4s %-20s\n", "device", "vendor", "model",
		   "rev", "serial");
	for (i = 0; i < disks->len; i += str_len(disks->s + i) + 1) {
		char *dsk = disks->s + i;
		int fd = open_read(dsk);

		/* take the next one... */
		if (fd == -1)
			continue;
		/* 1) device  size information */
		disksize = dev_getdisksize(fd);
		if (disksize == 0) {
			close(fd);
			continue;
		}

		/* 2) inquiry information */
		cdb_inquiry(fd, vendor, model, rev);
		/* 3) serial number */
		cdb_serial(fd, serial);
		term_print("%11s %8s %16s %-4s %-20s %5hB %5HiB\n", dsk, vendor,
			   model, rev, serial, disksize, disksize);
		close(fd);
	}
	stralloc_free(disks);

	fng_exit(0);
}

static void doit_harddisks_long(void)
{
	stralloc *disks = dev_get_disks();
	unsigned int i;
	char vendor[9];
	char model[17];
	char rev[5];
	char serial[255];
	int lbs = 0;
	u64 disksize = 0;

	if (disks->len == 0) {
		die_msg(0, "No harddisks found ?!");
	}

	term_print("%11s %8s %16s %-4s %-5s %-16s %-20s\n", "device", "vendor",
		   "model", "rev", "lbs", "devsize", "serial");
	for (i = 0; i < disks->len; i += str_len(disks->s + i) + 1) {
		char *dsk = disks->s + i;
		int fd = open_read(dsk);

		/* take the next one... */
		if (fd == -1)
			continue;
		/* 1) device  size information */
		disksize = dev_getdisksize(fd);
		if (disksize == 0)
			continue;
		/* 2) inquiry information */
		cdb_inquiry(fd, vendor, model, rev);
		/* 3) sector size information */
		lbs = dev_getsectorsize(fd);
		/* 4) serial number */
		cdb_serial(fd, serial);
		term_print("%11s %8s %16s %-4s %5u %16U %-20s %5hB %5HiB\n",
			   dsk, vendor, model, rev, lbs, disksize, serial,
			   disksize, disksize);
		close(fd);
	}
	stralloc_free(disks);

	fng_exit(0);
}

static void doit_filesystems(void)
{
	struct list_head *l;

	term_print("These Filesystems are supported:\n\n");
	list_for_each(l, &filesystems) {
		struct file_system *fs;
		fs = list_entry(l, struct file_system, list);

		term_print("%s:", fs->module);
		if (fs->fs_test)
			term_print(" test");
		if (fs->fs_info)
			term_print(" info");
		if (fs->fs_debug)
			term_print(" debug");
		if (fs->fs_getbitmap)
			term_print(" getbitmap");
		if (fs->fs_postclone)
			term_print(" postclone");
		term_print("\n");
	}
	term_print
	    ("\nPlease help me with adding more filesystems... write me an email ;)\n");

	fng_exit(0);
}

int fng_info(int argc, char **argv)
{
	/* show some information about filesystem by default */
	int mode = FNG_INFO_INFO;

	/* do a try and fail, with known filesystems */
	int opt_fsid = FSID_AUTO;

	/* for getopt() */
	int opt;

	while ((opt = getopt(argc, argv, "h?vaLXlIdDBS:isxt:P:")) != -1) {

		switch (opt) {
		case 'h':
		case '?':
			usage();
		case 'v':
			version();
		case 'a':
			/* list supported fs @ stdout */
			doit_filesystems();	/* no return */
		case 'L':
			/* dump harddisk information to stdout */
			doit_harddisks();	/* no return */
		case 'X':
			/* dump extended harddisk information to stdout */
			doit_harddisks_long();	/* no return */
		case 'l':
			/* list partitions of some disk */
			mode = FNG_INFO_PARTITIONS;
			break;
		case 'd':
			/* dump fs bitmap to stdout (no tty!) */
			mode = FNG_INFO_BITMAP;
			break;
		case 'D':
			/* give some bitmap summary to stdout */
			mode = FNG_INFO_SBITMAP;
			break;
		case 'B':
			/* dump fs bitmap to stdout (no tty!) */
			mode = FNG_INFO_BBITMAP;
			break;
		case 'S':
			/* -S <skipfile> */
			read_skipfile(optarg);
			break;
		case 'i':
			/* show fs info, frisbee would use for imaging */
			mode = FNG_INFO_INFO;
			break;
		case 's':
			/* show fs info from statvfs syscall */
			mode = FNG_INFO_STATFS;
			break;
		case 'x':
			/* show fs debuginfo */
			mode = FNG_INFO_DEBUG;
			break;
		case 't':
			/* set explicit some filesystem */
			opt_fsid = find_filesystem(optarg);
			break;
		default:
			usage();
		}
	}

	if (argc == optind)
		usage();

	for (;;) {
		struct fsinfo src;

		/* get next partition or file */
		src.path = argv[optind++];
		if (!src.path)
			break;
		src.fd = open_device_ro(src.path);
		src.id = opt_fsid;
		io_setcookie(src.fd, &src);

		/* get some info for this fs or file */
		switch (mode) {
		case FNG_INFO_PARTITIONS:
			doit_partitions(src.fd);
			break;
		case FNG_INFO_BITMAP:
			doit_bitmap(src.fd);
			break;
		case FNG_INFO_BBITMAP:
			doit_bbitmap(src.fd);
			break;
		case FNG_INFO_SBITMAP:
			doit_sbitmap(src.fd);
			break;
		case FNG_INFO_STATFS:
			doit_statvfs(src.fd);
			break;
		case FNG_INFO_DEBUG:
			doit_debug(src.fd);
			break;
		case FNG_INFO_INFO:
			doit_info(src.fd);
			break;
		}

		/* clean up */
		io_close(src.fd);
		term_flush();
	}			/* for all gives devices ... */

	return fng_exit(0);
}
