
/**
 * Copyright © 2010 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 1
 */

/* system */
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

/* libowfat */
#include "byte.h"
#include "open.h"
#include "scan.h"
#include "fmt.h"
#include "buffer.h"
#include "io.h"
#include "errmsg.h"

/* frisbee */
#include "fs.h"
#include "main.h"
#include "die.h"
#include "log.h"
#include "crypt.h"
#include "term.h"
#include "mem.h"
#include "image.h"
#include "openX.h"

#define FNG_RESTORE_UNKNOWN 0
#define FNG_RESTORE_DISK    1
#define FNG_RESTORE_PART    2
#define FNG_RESTORE_INFO    4
#define FNG_RESTORE_DEBUG   8

static void usage(void)
{
	term_print("\nUsage:\n");
	term_print(" fng-restore -i image.fng .. imageN.fng\n");
	term_print(" fng-restore -D image.fng .. imageN.fng\n");
	term_print(" fng-restore image.fng /dev/sda\n");
	term_print(" fng-restore -d 1 image.fng /dev/sda /dev/sdc\n");
	term_print(" fng-restore -d 1 -p 1 image.fng /dev/sda1\n");
	term_print("\nGeneral Options:\n");
	term_print(" %16s%s", "-h or -?", "show help\n");
	term_print(" %16s%s", "-v", "show version\n");
	term_print("\nDisk and Partition Selection:\n");
	term_print(" %16s%s", "-d N",
		   "for selecting the disk number in image (default: 1)\n");
	term_print(" %16s%s", "-p N",
		   "for selecting an partition, which should be restored (default: all of disk)\n");
	term_print("\nApplication Options:\n");
	term_print(" %16s%s", "-i", "Print information about image file(s).\n");
	term_print(" %16s%s", "-D",
		   "Print debug information about image file(s).\n");
	term_print("\nEncrypting Options:\n");
	term_print(" %16s%s", "-P <password>",
		   "decrypt image by using this password\n");
	fng_exit(0);
}

int fng_restore(int argc, char **argv)
{
	struct restore_config *cfg = mem_talloc(struct restore_config, 1);
	int mode = FNG_RESTORE_UNKNOWN;
	int opt;

	/* setup default values */
	stralloc_init(&cfg->password);

	/* image file to use */
	stralloc_init(&cfg->image);

	/* destination disks or partitions */
	stralloc_init(&cfg->dest);
	cfg->dest_count = 0;

	/* default: all partitions of first disk in image: */
	cfg->disk = 0;		/* 0=first 1=second */
	cfg->partition = 0;	/* 0=all >0 partition number X */

	/* default worker count = nr of cpu cores */
	cfg->threads = fng_cpucount;

	while ((opt = getopt(argc, argv, "h?vd:p:t:P:iD")) != -1) {
		switch (opt) {
		case 'h':
		case '?':
			/* never returns */
			usage();

		case 'v':
			/* never returns */
			version();

		case 'd':
			/* -d 1 */
			if (scan_uint(optarg, &cfg->disk) == 0)
				die_msg(2, "reading disk number failed");
			if (is_set(mode, FNG_RESTORE_DISK))
				die_msg(2,
					"You can not select multiple disks via -d on for restore!");

			set(mode, FNG_RESTORE_DISK);
			break;

		case 'p':
			/* -p 2 */
			if (scan_uint(optarg, &cfg->partition) == 0)
				die_msg(2, "reading disk number failed");
			if (is_set(mode, FNG_RESTORE_PART))
				die_msg(2,
					"You can not select multiple partitions via -p on for restore!");
			if (cfg->partition)
				set(mode, FNG_RESTORE_PART);
			break;

		case 't':
			/* -t N -> number of worker threads */
			if (scan_int(optarg, &cfg->threads) == 0)
				die_msg(2, "Wrong thread count!");
			if (cfg->threads > fng_cpucount)
				die_msg(2, "To many threads!");
			if (cfg->threads < 1)
				die_msg(2,
					"Zero worker, who should make the work?");
			break;

		case 'P':
			/* -P password */
			if (!stralloc_copys(&cfg->password, optarg))
				die_nomem();
			break;

		case 'i':
			/* -i image.fng ... */
			if (mode != FNG_RESTORE_UNKNOWN)
				die_msg(2,
					"You can not select -i together with restore options!");

			set(mode, FNG_RESTORE_INFO);
			break;

		case 'D':
			/* -D image.fng ... */
			if (mode != FNG_RESTORE_UNKNOWN)
				die_msg(2,
					"You can not select -D together with restore options!");

			set(mode, FNG_RESTORE_DEBUG);
			break;

		default:
			usage();
		}
	}

	/* default: restore all partitions of first disk in image */
	if (mode == FNG_RESTORE_UNKNOWN)
		mode = FNG_RESTORE_DISK;

	/* -d and -p specified, this is okay */
	if (is_set(mode, FNG_RESTORE_PART)) {
		if (cfg->partition) {
			mode = FNG_RESTORE_PART;
		} else {
			mode = FNG_RESTORE_DISK;
		}
	}

	if (argc == optind)
		usage();

	if (mode == FNG_RESTORE_INFO) {

		/**
		 * fng-restore -i image.fng image2.fng image3.fng
		 * -> info
		 */

		for (;;) {
			if (!stralloc_copys(&cfg->image, argv[optind++]))
				die_nomem();
			if (!stralloc_0(&cfg->image))
				die_nomem();

			cfg->fd = open_device_ro(cfg->image.s);
			image_info(cfg);
			close(cfg->fd);

			if (argc == optind)
				break;
		}
	} else if (mode == FNG_RESTORE_DEBUG) {

		/**
		 * fng-restore -D image.fng image2.fng image3.fng
		 * -> debug info
		 */

		for (;;) {
			if (!stralloc_copys(&cfg->image, argv[optind++]))
				die_nomem();
			if (!stralloc_0(&cfg->image))
				die_nomem();
			cfg->fd = open_device_ro(cfg->image.s);
			image_debug(cfg);
			close(cfg->fd);

			if (argc == optind)
				break;
		}
	} else if (mode == FNG_RESTORE_PART) {

		/**
		 * fng-restore -d2 -p 1 image.fng /dev/sda
		 * - von disk2 im image die erste partition
		 *
		 * fng-restore -p 1 image.fng /dev/sda
		 * - stellt aus einem image mit mehreren platten nur die platte 1 wieder her
		 * - ziel platte ist /dev/sda
		 */

		/* image file */
		if (!stralloc_copys(&cfg->image, argv[optind++]))
			die_nomem();
		if (!stralloc_0(&cfg->image))
			die_nomem();
		if (argc == optind)
			usage();

		cfg->fd = open_device_ro(cfg->image.s);
		for (;;) {
			if (!stralloc_cats(&cfg->dest, argv[optind++]))
				die_nomem();
			if (!stralloc_0(&cfg->dest))
				die_nomem();
			cfg->dest_count++;

			if (argc == optind)
				break;
		}

		/* init terminal stuff */
		term_resize();
		image_restore_partition(cfg);

	} else if (mode == FNG_RESTORE_DISK) {

		/*
		 * fng-restore image.fng /dev/sda
		 * -> darf nur genutzt werden, wenn nur eine disk
		 *
		 * fng-restore -d 1 image.fng /dev/sda
		 * - stellt aus einem image mit mehreren platten nur die platte 1 wieder her
		 * - ziel platte ist /dev/sda
		 *
		 * fng-restore -d 1 image.fng /dev/sda /dev/sdb
		 * - stellt aus einem image mit mehreren platten nur die erste wieder her
		 * - ziel platten sind: /dev/sda /dev/sdb
		 *
		 * fng-restore image.fng -d 2 -p1 /dev/sdc /dev/sdd
		 * - restore partition 1 of disk 2 in image
		 */

		/* image file */
		if (!stralloc_copys(&cfg->image, argv[optind++]))
			die_nomem();
		if (!stralloc_0(&cfg->image))
			die_nomem();
		if (argc == optind)
			usage();

		cfg->fd = open_device_ro(cfg->image.s);
		for (;;) {
			if (!stralloc_cats(&cfg->dest, argv[optind++]))
				die_nomem();
			if (!stralloc_0(&cfg->dest))
				die_nomem();
			cfg->dest_count++;

			if (argc == optind)
				break;
		}

		/* init terminal stuff */
		term_resize();
		image_restore_disk(cfg);
	}

	fng_exit(0);

	return 0;
}
