
/**
 * Copyright © 2010 - 2013 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>

/* libowfat */
#include "buffer.h"
#include "byte.h"
#include "scan.h"
#include "open.h"
#include "openreadclose.h"
#include "str.h"
#include "fmt.h"
#include "io.h"

/* frisbee-ng */
#include "strallocX.h"
#include "progress.h"
#include "wrapper.h"
#include "term.h"
#include "main.h"
#include "tty.h"
#include "dev.h"
#include "die.h"
#include "getkey.h"
#include "sig.h"
#include "openX.h"
#include "log.h"
#include "mem.h"
#include "crypt.h"
#include "fmtX.h"
#include "tty.h"
#include "main.h"

#define DEF_BLOCKSIZE (1024 * 1024) * 1	/* 1MB */

static pthread_mutex_t term_mutex = PTHREAD_MUTEX_INITIALIZER;

/* internal functions: */
static void usage(void);
static void *shredder(void *data);
static void *shredder_per_disk(void *data);

static void usage(void)
{
	term_print("\nUsage:\n");
	term_print(" fng-shredder [options] <device1>|<file1> <device2> ...\n");
	term_print("\nGeneral Options:\n");
	term_print(" %16s%s", "-h or -?", "show help\n");
	term_print(" %16s%s", "-v", "show version\n");
	term_print("\nApplication Options:\n");
	term_print(" %16s%s", "-b bytes",
		   "blocksize in bytes (default: 1MiB)\n");
	term_print(" %16s%s", "-n count",
		   "the data will be overwritten <count> times (default: 3)\n");
	term_print(" %16s%s", "-s blocks",
		   "skip N blocks (will not be overwritten)\n");
	term_print(" %16s%s", "-t threadcount",
		   "number of threads per disk (default: nr of cpu cores)\n");
	term_print(" %16s%s", "-y",
		   "don't ask before overwriting the file/disk - dangerous!\n");
	term_print(" %16s%s", "-r",
		   "read the written data and check integrity (harddisk testing)\n");
	term_print(" %16s%s", "-z ", "zero out with binary '0'\n");
	fng_exit(0);
}

/* data blob per disk */
struct disk_blob {
	int disk_id;		/* disk id */
	char *device;
	int fd;
	progress_t *pd;		/* progress of disk */
	u64 diskbytes;
	int flag_threads;
	u64 flag_skip;

	/* these flags go directly to the worker_blob */
	int flag_blocksize;
	int flag_ntimes;
	int flag_zero;
	int flag_rwtest;
};

/* data blob for the real worker */
struct worker_blob {
	int wid;		/* worker id */
	u64 start;
	u64 bytes;

	/* caller */
	struct disk_blob *db;
};

static void *shredder(void *data)
{
	struct worker_blob *wb = (struct worker_blob *)data;
	struct disk_blob *db = wb->db;
	progress_t *pd = db->pd;
	int blocksize = db->flag_blocksize;
	unsigned char *random, *random2 = 0;
	u64 c, x;

	random = mem_alloc(db->flag_ntimes * db->flag_blocksize);

	if (db->flag_rwtest)
		random2 = mem_alloc(db->flag_ntimes * db->flag_blocksize);

	for (c = 0; c != wb->bytes; c += blocksize) {
		int n;

		/* last round */
		if (blocksize + c >= wb->bytes)
			blocksize = wb->bytes - c;

		x = blocksize * db->flag_ntimes;

		if (db->flag_zero)
			fmt_uzero(random, x);
		else
			fmt_urandom(random, x);

		for (n = 0; n < db->flag_ntimes; n++) {
			x = blocksize * n;

			pwrite_loop2(db->fd, random + x, blocksize,
				     wb->start + c);

			/* sync() each written block! */
			fsync(db->fd);

			if (db->flag_rwtest) {
				pread_loop2(db->fd, random2 + x, blocksize,
					    wb->start + c);

				if (!byte_equal
				    (random + x, blocksize, random2 + x)) {
					term_printflush
					    ("read write test failed @ %2U .. %2U\n",
					     wb->start + c,
					     wb->start + c + blocksize);
				}
			}
		}

		progress_add(pd, blocksize);

		if (progress(pd) && db->disk_id == 0) {
			pthread_mutex_lock(&term_mutex);
			term_gotoxy(1, db->disk_id + 1);
			term_print("time: %02u:%02u:%02u", pd->ctt_h, pd->ctt_m,
				   pd->ctt_s);
			term_flush();
			pthread_mutex_unlock(&term_mutex);
		}

		if (progress(pd)) {
			pthread_mutex_lock(&term_mutex);
			term_gotoxy(1, db->disk_id + 3);
			term_print("eta: %02u:%02u:%02u %hB/%hB %d%%",
				   pd->eta_h, pd->eta_m, pd->eta_s, pd->current,
				   pd->total, pd->percent);
			term_flush();
			pthread_mutex_unlock(&term_mutex);
		}

	}

	return 0;
}

static void *shredder_per_disk(void *data)
{
	struct disk_blob *db = (struct disk_blob *)data;
	u64 work4each, todo;
	u64 skip = db->flag_skip * db->flag_blocksize;
	int r, i = 0;
	struct worker_blob *wb;
	pthread_t *wt;		/* worker threads per disk */

	/**
	 * falls zu viele blöcke angegeben wurden...
	 */
	if (skip > db->diskbytes) {
		term_printflush("\nNot that many blocks to skip are here!");
		return 0;
	}

	wb = mem_talloc(struct worker_blob, db->flag_threads);
	wt = mem_talloc(pthread_t, db->flag_threads);

	/**
	 * todo = bytes_all - bytes_skipped
	 * work4each = todo / nr_cpus
	 */
	todo = db->diskbytes - skip;
	work4each = todo / db->flag_threads;

	for (i = 0; i < db->flag_threads; i++) {

		wb[i].wid = i;
		wb[i].start = i * work4each + skip;
		wb[i].db = db;

		/* last thread gets the last not 'done' bytes... */
		todo -= work4each;
		if (todo < work4each)
			work4each += todo;
		wb[i].bytes = work4each;

		/* running new threads is cheap @ linux */
		r = pthread_create(&wt[i], 0, shredder, &wb[i]);
		if (r != 0)
			die_pthread(r);
	}

	/* wait for the workers */
	for (i = 0; i < db->flag_threads; i++) {
		r = pthread_join(wt[i], 0);
		if (r != 0)
			die_pthread(r);
	}

	return 0;
}

int fng_shredder(int argc, char **argv)
{
	unsigned long flag_skip = 0;
	int opt, disks;
	int r, i, x, y;
	struct disk_blob *db;
	pthread_t *tpd;		/* thread per disk */

	int flag_force = 0;
	int flag_threads = fng_cpucount;
	int flag_blocksize = DEF_BLOCKSIZE;
	int flag_ntimes = 3;
	int flag_zero = 0;
	int flag_rwtest = 0;

	/* ☠ ☢ ☣ ☯ ✄ */
	while ((opt = getopt(argc, argv, "h?vb:n:yrs:zt:")) != -1) {
		switch (opt) {
		case 'h':
		case '?':
			usage();
		case 'v':
			version();
		case 'b':
			/* -b blocksize in kilobytes */
			if (optarg)
				if (scan_int(optarg, &flag_blocksize) < 1) {
					die_msg(2, "reading blocksize failed");
				}
			break;
		case 'n':
			/* -n == overwrite it n times */
			if (optarg)
				if (scan_int(optarg, &flag_ntimes) < 1) {
					die_msg(2, "wrong count...");
				}
			break;
		case 'y':	/* force flag */
			flag_force = 1;
			break;
		case 'r':	/* read write test */
			flag_rwtest = 1;
			break;
		case 's':	/* skip n blocks */
			if (optarg)
				if (scan_ulong(optarg, &flag_skip) < 1) {
					die_msg(2, "reading blocks to skip :(");
				}
			break;
		case 'z':	/* just zeros */
			flag_zero = 1;
			flag_ntimes = 1;
			break;
		case 't':
			if (scan_int(optarg, &flag_threads) < 1) {
				die_msg(2, "reading number of threads to use");
			}
			if (flag_threads < 1) {
				die_msg(2, "wrong count...");
			}
			break;
		default:
			usage();
		}
	}

	/* how many disks should be shreddered? */
	disks = argc - optind;
	if (disks == 0)
		usage();

	/* how many disk fit to screen? */
	tty_getsize(&x, &y);
	if (disks + 5 > y)
		die_msg(111, "Terminal to small for so many disks!");

	/* each disk has some specific stuff we need to store */
	db = mem_talloc(struct disk_blob, disks);

	tpd = mem_talloc(pthread_t, disks);

	/* first check of all files/disks */
	for (i = 0; i < disks; i++) {
		db[i].disk_id = i;
		db[i].device = argv[i + optind];
		//term_print("check disk%d: %s\n", i, db[i].device);
		r = access(db[i].device, R_OK | W_OK);
		if (r != 0)
			die_msg(2, "Can't access device/file!");
	}

	/* should we really do it? */
	tty_init();
	if (!flag_force) {
		keyset_t k[] = { {0, "y"}
		, {1, "n"}
		, {-1, 0}
		};

		term_print("Do you really want to delete all contents of");
		for (i = 0; i < disks; i++) {
			term_print(" %s", db[i].device);
		}
		term_print("?\n");
		term_printflush("Type 'y' to continue or 'n' to abort.");

		r = getkey(k);
		if (r != 0) {
			term_printflush(".. aborting.\n");
			fng_exit(0);
		}
	}

	term_s(term_str_clear);
	term_print("starting ...");
	term_flush();

	/* begin with the real work for each disk */
	for (i = 0; i < disks; i++) {

		/* one manager thread per disk */
		progress_t *pd = mem_talloc(progress_t, 1);

		db[i].fd = open_device_rw(db[i].device);
		if (db[i].fd == -1)
			die_msg(2, "Can't access device/file!");

		/* get bytes of disks */
		db[i].diskbytes = dev_getdisksize(db[i].fd);
		db[i].pd = pd;	/* one progress per disk */

		db[i].flag_skip = flag_skip;
		db[i].flag_threads = flag_threads;
		db[i].flag_blocksize = flag_blocksize;
		db[i].flag_ntimes = flag_ntimes;
		db[i].flag_zero = flag_zero;
		db[i].flag_rwtest = flag_rwtest;

		/* setup progress calculator */
		progress_init(db[i].pd, db[i].diskbytes);

		/* running new threads is cheap @ linux */
		r = pthread_create(&tpd[i], 0, shredder_per_disk, &db[i]);
		if (r != 0)
			die_pthread(r);
	}

	/* wait for each disk thread */
	for (i = 0; i < disks; i++) {
		r = pthread_join(tpd[i], 0);
		if (r != 0)
			die_pthread(r);
	}

	term_gotoxy(1, disks + 3);
	term_print("\nFinished with erasing, syncing now ...\n\n");

	sync();
	sync();
	sync();

	return fng_exit(0);
}
