
/**
 * Copyright © 2010 - 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef FNG_FS_H
#define FNG_FS_H

/* libowfat */
#include "stralloc.h"

/* frisbee-ng */
#include "defs.h"
#include "bswap.h"
#include "list.h"
#include "case.h"
#include "haveinline.h"
#include "progress.h"

/**
 * this struct is used for source filesystems (partitions)
 * - fs_clonefs() -> läuft!
 */
struct fsinfo {
	int fd;			/* opened fd for input filesystem */
	char *path;		/* path to device or file (/dev/sda1 or ./lala.ntfs) */
	char *name;		/* name of filesystem (only ptr) */
	int id;			/* id of filesystem */

	/* pointer to filesystem */
	struct file_system *fs;

	/* dev_getsize_bytes() on real devices */
	u64 psize;		/* size in bytes of partition or file */

	/* returned by fs_info() / fs_getbitmap() */
	u32 bsize;		/* blocksize of filesystem */
	u64 fssize;		/* really used space in bytes of partition or file */

	u64 blocks;		/* total data/meta blocks used by filesystem */
	u64 bused;		/* used blocks in filesystem */
	u64 bfree;		/* free blocks in filesystem */

	/* returned by fs_getbitmap() */
	u8 *bitmap;		/* bitmap of used clusters */
	u64 bitmaplen;		/* legth of bitmap */

	/* used by cloning, progress of partition */
	progress_t *progress;
};

/* each partition gets it's own structure on most programs */
struct partition {
	int index;		/* index number in some array */
	int number;		/* slice number to show, 1,2,5,6 */
	u64 start;		/* offset @ disk */
	u64 size;		/* size on disk */
	struct fsinfo fsi;	/* filesystem info */
};

/**
 * each supported filesystem exports some or all of these functions
 */
struct file_system {
	struct list_head list;

	/**
	 * name of filesystem module
	 */
	char *module;

	/**
	 * frisbee's filesystem id (you can add, but not remove or change!)
	 */
	int id;
#define FSID_DISK	 -1	/* this is not a partition, it's a DISK!  */

#define FSID_AUTO	 0	/* search */
#define FSID_RAW	 1	/* full backup */
#define FSID_FRISBEE	 2	/* reserved... */
#define FSID_NTFS	 3	/* working */
#define FSID_EXT2	 4	/* working */
#define FSID_JFS	 5	/* working */
#define FSID_REISER	 6	/* working */
#define FSID_XFS	 7	/* working */
#define FSID_EXFAT	 8	/* working */
#define FSID_MINIX	 9	/* working */
#define FSID_HPFS	10	/* working */
#define FSID_BEFS	11	/* working */
#define FSID_AFS	12	/* working */
#define FSID_HFS	13	/* working */
#define FSID_HFSP	14	/* working */
#define FSID_FAT	15	/* NOT working */
#define FSID_QNX4	16	/* not working */
#define FSID_QNX6	17	/* not working */
#define FSID_UFS	18	/* NOT working */
#define FSID_F2FS	19	/* not working */

	/**
	 * fs_test - test for some specific filesystem
	 *
	 * @fd: points to opened(ro) filesystem file/device
	 *
	 * returns 1 if filesystem is correct, 0 on error
	 */
	int (*fs_test) (int fd);

	/**
	 * fs_info - get information for some specific filesystem
	 *
	 * @fd: points to opened(ro) filesystem file/device
	 *
	 * returns 0 on success, -1 on error
	 */
	int (*fs_info) (int fd);

	/**
	 * fs_getbitmap - get allocation bitmap for some filesystem
	 *
	 * @fd: points to opened(ro) filesystem file/device
	 *
	 * returns 0 on success, -1 on error
	 */
	int (*fs_getbitmap) (int fd);

	/**
	 * fs_postclone - do sth. after cloning this fs
	 *
	 * @fd: points to opened(ro) filesystem file/device
	 * -> fs_getbitmap() must be called before this function!
	 *
	 * returns nothing
	 */
	void (*fs_postclone) (int fd_in, int fd_out, off_t offset);

	/**
	 * fs_debug - return all information, which can be read by us
	 *
	 * @fd: points to opened(ro) filesystem file/device
	 *
	 * returns 0 on success, -1 on error
	 */
	int (*fs_debug) (int fd);
};

/**
 * register filesystems known to frisbee
 * - must be called once @ startup
 */
extern void register_filesystems(void);

/**
 * list with all known filesystems
 * - this variable will be initialized by register_filesystems()
 */
extern struct list_head filesystems;

/**
 * find_filesystem - search filesystem named *name, and return it's id
 *
 * @name: name to search for
 *
 * returns id of some filesystem, defaults to FSID_RAW
 */
static inline int find_filesystem(const char *name)
{
	struct list_head *l;

	list_for_each(l, &filesystems) {
		struct file_system *fs;
		fs = list_entry(l, struct file_system, list);

		if (case_equals(fs->module, name))
			if (fs->fs_getbitmap)
				return fs->id;
	}

	/* alias names for filesystems: */
	if (case_equals(name, "ext"))
		return FSID_EXT2;
	if (case_equals(name, "ext2"))
		return FSID_EXT2;
	if (case_equals(name, "ext3"))
		return FSID_EXT2;
	if (case_equals(name, "ext4"))
		return FSID_EXT2;

	if (case_equals(name, "reiser3"))
		return FSID_REISER;
	if (case_equals(name, "reiserfs"))
		return FSID_REISER;

	if (case_equals(name, "vfat"))
		return FSID_FAT;
	if (case_equals(name, "msdos"))
		return FSID_FAT;
	if (case_equals(name, "fat12"))
		return FSID_FAT;
	if (case_equals(name, "fat16"))
		return FSID_FAT;
	if (case_equals(name, "fat32"))
		return FSID_FAT;

	if (case_equals(name, "minix"))
		return FSID_MINIX;

	if (case_equals(name, "hfsx"))
		return FSID_HFSP;

	if (case_equals(name, "ntfs-3g"))
		return FSID_NTFS;

	return FSID_RAW;
}

#endif
