
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"

/**
 * AtheOS file system (AFS) -> shares a lot with BeFS
 *
 * systems: AtheOS, Syllable
 * endian:  big or little
 */

static char fs_name[] = "afs";

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

typedef struct {
	u32 ag;
	u16 start;
	u16 len;
} addr_t;

struct afs_sb {
	u8 name[32];

	u32 magic1;
	u32 byte_order;
	u32 block_size;
	u32 block_shift;

	u64 num_blocks;
	u64 used_blocks;
	u32 inode_size;
	u32 magic2;

	u32 blocks_per_ag;
	u32 ag_shift;
	u32 num_ags;
	u32 flags;

	addr_t log_blocks;
	u64 log_start;
	u32 valid_logblocks;
	u32 log_size;

	u32 magic3;
	addr_t root_dir;
	addr_t deleted_files;
	addr_t index_dir;
	u32 bootloader_size;
	u32 as_anPad[7];
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

#define AFS_SB_MAGIC1  0x41465331	/* AFS1 */
#define AFS_SB_MAGIC2  0xdd121031
#define AFS_SB_MAGIC3  0x15b6830e

/* global afs variables */
static struct afs_sb sb;
static int fs_endian;

/**
 * fs_test - check for afs filesystem
 *
 * @return 0 - no afs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	ssize_t r;

	r = pread_loop(fd, &sb, sizeof(struct afs_sb), 1024);
	if (r == -1) {
		return 0;
	}

	/* check first magic with LE and BE endianness */
	fs_endian = BYTESEX_LE;
	if (fs32_to_cpu(fs_endian, sb.magic1) != AFS_SB_MAGIC1) {
		fs_endian = BYTESEX_BE;
		if (fs32_to_cpu(fs_endian, sb.magic1) != AFS_SB_MAGIC1) {
			/* no afs @ all */
			return 0;
		}
	}

	/* magic1 was fount, endian is set ... try some more things */
	if (fs32_to_cpu(fs_endian, sb.magic2) != AFS_SB_MAGIC2)
		return 0;
	if (fs32_to_cpu(fs_endian, sb.magic3) != AFS_SB_MAGIC3)
		return 0;

	/* check blocksize */
	if (fs32_to_cpu(fs_endian, sb.block_size) != 1024
	    && fs32_to_cpu(fs_endian, sb.block_size) != 2048
	    && fs32_to_cpu(fs_endian, sb.block_size) != 4096
	    && fs32_to_cpu(fs_endian, sb.block_size) != 8192) {
		/* not valid */
		return 0;
	}

	/* no afs found */
	return 1;
}

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	fsi->name = fs_name;
	fsi->bsize = fs32_to_cpu(fs_endian, sb.block_size);
	fsi->blocks = fs64_to_cpu(fs_endian, sb.num_blocks);
	fsi->bused = fs64_to_cpu(fs_endian, sb.used_blocks);
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bfree = fsi->blocks - fsi->bused;

	/* calculate bitmap for later */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s%s\n", "volume name:", sb.name);
	term_print("%20s%u\n", "block size:",
		   fs32_to_cpu(fs_endian, sb.block_size));
	term_print("%20s%u\n", "block shift:",
		   fs32_to_cpu(fs_endian, sb.block_shift));
	term_print("%20s%U\n", "# fs blocks:",
		   fs64_to_cpu(fs_endian, sb.num_blocks));
	term_print("%20s%U\n", "# used blocks:",
		   fs64_to_cpu(fs_endian, sb.used_blocks));
	term_print("%20s%u\n", "inode_size:", sb.inode_size);

	term_print("%20s0x%x\n", "magic1:", fs32_to_cpu(fs_endian, sb.magic1));
	term_print("%20s0x%x\n", "magic2:", fs32_to_cpu(fs_endian, sb.magic2));
	term_print("%20s0x%x\n", "magic3:", fs32_to_cpu(fs_endian, sb.magic3));
	term_print("%20s%u\n", "bootloader size:",
		   fs32_to_cpu(fs_endian, sb.bootloader_size));
	term_printflush("\n");

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	off_t offset;

	if (fs_info(fd) == -1) {
		return -1;
	}

	/* allocate space for the map */
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + 8);

	/**
	 * +--------------------------+
	 * |       Boot block         |
	 * |         0-1023           |
	 * +--------------------------+
	 * |       Super block        |
	 * |       1024 - 2047        |
	 * +~~~~~~~~~~~~~~~~~~~~~~~~~~+
	 * |    Boot loader area      |
	 * +~~~~~~~~~~~~~~~~~~~~~~~~~~+
	 * |      Bitmap blocks       |
	 * +--------------------------+
	 * |     Journal blocks       |
	 * +--------------------------+
	 */
	offset = fs32_to_cpu(fs_endian, sb.bootloader_size) * fsi->bsize;
	offset += 2048;
	pread_loop2(fd, fsi->bitmap, fsi->blocks / 8, offset);

	return 0;
};

struct file_system afs_fs = {
	.id = FSID_AFS,
	.module = "afs",
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
