
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"

/**
 * BeOS file system (BeFS)
 *
 * systems: BeOS, ZETA, Haiku, SkyOS, Syllable
 * endian:  big or little
 */

static char fs_name[] = "befs";

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

typedef struct {
	u32 ag;
	u16 start;
	u16 len;
} addr_t;

struct befs_sb {
	u8 name[32];

	u32 magic1;
	u32 byte_order;
	u32 block_size;
	u32 block_shift;

	u64 num_blocks;
	u64 used_blocks;
	u32 inode_size;
	u32 magic2;

	/* Allocation group information */
	u32 blocks_per_ag;
	u32 ag_shift;
	u32 num_ags;
	u32 flags;

	/* jornal log entry */
	addr_t log_blocks;
	u64 log_start;
	u64 log_end;

	u32 magic3;
	addr_t root_dir;
	addr_t indices;
	u8 pad[8];
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

#define BEFS_SB_MAGIC1  0x42465331	/* BFS1 */
#define BEFS_SB_MAGIC2  0xdd121031
#define BEFS_SB_MAGIC3  0x15b6830e

#define BEFS_BYTESEX_BE 0x45474942	/* EGIB */
#define BEFS_BYTESEX_LE 0x42494745	/* BIGE */

#define BEFS_CLEAN      0x434c454e	/* CLEN */
#define BEFS_DIRTY      0x44495254	/* DIRT */

/* global befs variables */
static struct befs_sb sb;

/* my test images were with bb */
static int befs_offset;
static int fs_endian;

/**
 * fs_test_at - check for befs filesystem at offset
 *
 * @return 0 - no befs
 * @return 1 - yes
 */
static int fs_test_at(int fd, int offset)
{
	ssize_t r;

	r = pread_loop(fd, &sb, sizeof(struct befs_sb), offset);
	if (r == -1) {
		return 0;
	}

	/* get byte order */
	if (sb.byte_order == BEFS_BYTESEX_LE) {
		/* little endian */
		//term_printflush("[LE]");
		fs_endian = BYTESEX_LE;
	} else if (sb.byte_order == BEFS_BYTESEX_BE) {
		/* big endian */
		//term_printflush("[BE]");
		fs_endian = BYTESEX_BE;
	} else
		return 0;

	/* check magic */
	if (fs32_to_cpu(fs_endian, sb.magic1) != BEFS_SB_MAGIC1)
		return 0;
	if (fs32_to_cpu(fs_endian, sb.magic2) != BEFS_SB_MAGIC2)
		return 0;
	if (fs32_to_cpu(fs_endian, sb.magic3) != BEFS_SB_MAGIC3)
		return 0;

	/* when we got until here, it is befs ;) */
	return 1;
}

/**
 * fs_test - check for befs filesystem
 *
 * @return 0 - no befs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	befs_offset = 0;
	if (fs_test_at(fd, befs_offset) != 0) {
		return 1;	/* befs found */
	}

	befs_offset = 512;
	if (fs_test_at(fd, befs_offset) != 0) {
		return 1;	/* befs found */
	}

	/* check blocksize */
	if (fs32_to_cpu(fs_endian, sb.block_size) != 1024
	    && fs32_to_cpu(fs_endian, sb.block_size) != 2048
	    && fs32_to_cpu(fs_endian, sb.block_size) != 4096
	    && fs32_to_cpu(fs_endian, sb.block_size) != 8192) {
		/* not valid */
		return 0;
	}

	/* no befs found */
	return 0;
}

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	fsi->name = fs_name;
	fsi->bsize = fs32_to_cpu(fs_endian, sb.block_size);
	fsi->blocks = fs64_to_cpu(fs_endian, sb.num_blocks);
	fsi->bused = fs64_to_cpu(fs_endian, sb.used_blocks);
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bfree = fsi->blocks - fsi->bused;

	/* calculate bitmap for later */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s%s\n", "volume name:", sb.name);
	term_print("%20s0x%x\n", "byte order:",
		   fs32_to_cpu(fs_endian, sb.byte_order));
	term_print("%20s%u\n", "block size:",
		   fs32_to_cpu(fs_endian, sb.block_size));
	term_print("%20s%u\n", "block shift:",
		   fs32_to_cpu(fs_endian, sb.block_shift));
	term_print("%20s%U\n", "# fs blocks:",
		   fs64_to_cpu(fs_endian, sb.num_blocks));
	term_print("%20s%U\n", "# used blocks:",
		   fs64_to_cpu(fs_endian, sb.used_blocks));
	term_print("%20s%u\n", "inode_size:", sb.inode_size);

	term_print("%20s0x%x\n", "magic1:", fs32_to_cpu(fs_endian, sb.magic1));
	term_print("%20s0x%x\n", "magic2:", fs32_to_cpu(fs_endian, sb.magic2));
	term_print("%20s0x%x\n", "magic3:", fs32_to_cpu(fs_endian, sb.magic3));
	term_print("%20s%u\n", "ag shift:",
		   fs32_to_cpu(fs_endian, sb.ag_shift));
	term_print("%20s%u\n", "num ags:", fs32_to_cpu(fs_endian, sb.num_ags));
	term_print("%20s%s\n", "flags:",
		   fs32_to_cpu(fs_endian,
			       sb.flags) == BEFS_CLEAN ? "clean" : "dirty");
	term_printflush("\n");

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	/* allocate space for the map */
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + 8);

	/* bitmap is on the first map */
	pread_loop2(fd, fsi->bitmap, fsi->blocks / 8, fsi->bsize);
	/* endian? */

	return 0;
};

struct file_system befs_fs = {
	.id = FSID_BEFS,
	.module = "befs",
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
