
/**
 * Copyright © 2009, 2010 Andrew Nayenko
 * Copyright © 2012, 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"

/**
 * Extended File Allocation Table (exFAT)
 *
 * systems: Windows CE 6, Windows Vista SP1, Windows XP SP2, Linux,
 *          Mac OS X 10.6.5, Memory-Stick-XC and SD-XC
 * endian:  little
 */

static char fs_name[] = "exFAT";

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

struct exfat_sb {
	u8 jump[3];		/* 0x00 jmp and nop instructions */
	const char oem_name[8];	/* 0x03 "EXFAT   " */
	u8 __unused1[53];	/* 0x0B always 0 */
	u64 sector_start;	/* 0x40 partition first sector */
	u64 sector_count;	/* 0x48 partition sectors count */
	u32 fat_sector_start;	/* 0x50 FAT first sector */
	u32 fat_sector_count;	/* 0x54 FAT sectors count */
	u32 cluster_sector_start;	/* 0x58 first cluster sector */
	u32 cluster_count;	/* 0x5C total clusters count */
	u32 rootdir_cluster;	/* 0x60 first cluster of the root dir */
	u32 volume_serial;	/* 0x64 volume serial number */
	struct {		/* 0x68 FS version */
		u8 minor;
		u8 major;
	} version;
	u16 volume_state;	/* 0x6A volume state flags */
	u8 sector_bits;		/* 0x6C sector size as (1 << n) */
	u8 spc_bits;		/* 0x6D sectors per cluster as (1 << n) */
	u8 fat_count;		/* 0x6E always 1 */
	u8 drive_no;		/* 0x6F always 0x80 */
	u8 allocated_percent;	/* 0x70 percentage of allocated space */
	u8 __unused2[397];	/* 0x71 always 0 */
	u16 boot_signature;	/* the value of 0xAA55 */
} __FNG_ATTR_PACKED;		/* 512 byte */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global ext2fs variables */
static struct exfat_sb sb;

/**
 * fs_test - check if exfat filesystem
 *
 * @return 0 - no exfat
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	ssize_t r;

	r = pread_loop(fd, &sb, sizeof(struct exfat_sb), 0);
	if (r == -1) {
		return 0;
	}

	if (str_start(sb.oem_name, "EXFAT   ")) {
		return 1;
	}

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	off_t off;
	u8 *map;
	u64 c;

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* sector_size * sec_per_cluster => 512 Byte .. 32MiB */
	fsi->bsize = (1 << sb.sector_bits) * (1 << sb.spc_bits);
	fsi->blocks = le64_to_cpu(sb.sector_count) >> sb.spc_bits;

	/* calculate bitmap length */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;

	/* allocate real map and tmp map */
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + 1);
	map = mem_tzalloc(u8, fsi->bitmaplen);

	/**
	 * load the bitmap into the temporary map
	 * - the bitmap starts with cluster 2 = bit 0 ?!?!
	 * - all the bits before cluster_sector_start must be allocated per
	 *   hand in the bitmap (cluster heap)
	 */
	off = le32_to_cpu(sb.cluster_sector_start) * (1 << sb.sector_bits);
	pread_loop2(fd, map, fsi->bitmaplen, off);

	/* used blocks in begin of filesystem */
	off = le32_to_cpu(sb.cluster_sector_start) / (1 << sb.spc_bits) + 2;

	fsi->bfree = fsi->bused = 0;
	for (c = 0; c < (u64) off; c++) {
		bitmap_set(fsi->bitmap, c, 1);
		fsi->bused++;
	}

	for (c = 0; c < fsi->blocks; c++) {
		if (bitmap_get(map, c)) {
			bitmap_set(fsi->bitmap, c + off, 1);
			fsi->bused++;
		}
	}

	fsi->bfree = fsi->blocks - fsi->bused;
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->name = fs_name;

	mem_free(map);

	return 0;
};

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_getbitmap(fd) == -1) {
		return -1;
	}

	/* not needed @ fs_info() */
	free(fsi->bitmap);
	fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	if (fs_info(fd) == -1) {
		return -1;
	}

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fs_name);
	term_print("%20s0x%08x\n", "volume serial:",
		   le32_to_cpu(sb.volume_serial));
	term_print("%20s%u.%u\n", "fs version:", (u32) sb.version.major,
		   (u32) sb.version.minor);
	term_print("%20s%u\n", "sector size:", (u32) 1 << sb.sector_bits);
	term_print("%20s%u\n", "cluster size:",
		   (u32) (1 << sb.sector_bits) * (1 << sb.spc_bits));
	term_print("%20s%U\n", "sector start:", le64_to_cpu(sb.sector_start));
	term_print("%20s%U\n", "sector count:", le64_to_cpu(sb.sector_count));
	term_print("%20s%u\n", "fat sector start:",
		   le32_to_cpu(sb.fat_sector_start));
	term_print("%20s%u\n", "fat sector count:",
		   le32_to_cpu(sb.fat_sector_count));
	term_print("%20s%u\n", "first cluster:",
		   le32_to_cpu(sb.cluster_sector_start));
	term_print("%20s%u\n", "cluster count:", le32_to_cpu(sb.cluster_count));
	term_print("%20s%u\n", "rootdir cluster:",
		   le32_to_cpu(sb.rootdir_cluster));
	term_print("%20s%u\n", "volume state:", (u32) sb.volume_state);
	term_print("%20s%u\n", "fat_count:", (u32) sb.fat_count);
	term_print("%20s%u%%\n", "allocated percent:",
		   (u32) sb.allocated_percent);

	term_printflush("\n");

	return 0;
}

struct file_system exfat_fs = {
	.id = FSID_EXFAT,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
