
/**
 * Copyright © 1991 - 1992 Linus Torvalds
 * Copyright © 1992 - 1995 Remy Card
 * Copyright © 1993 - 1996 Theodore Ts'o.
 * Copyright © 2012 - 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/debug.h"
#include "src/bitmap.h"

/**
 * Second / Third / Fourth Extended File System
 *
 * systems: Linux, BSD, Mac OS X, Windows via IFS
 * endian:  little
 */

#define SB_MAGIC_EXT2         0xEF53
#define EXT2_MIN_BLOCK_SIZE   1024	/* log 0 */
#define EXT2_MAX_BLOCK_SIZE   65536	/* log 6 */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

/*
 * Structure of a blocks group descriptor
 */
struct ext2_group_desc {
	u32 bg_block_bitmap;	/* Blocks bitmap block */
	u32 bg_inode_bitmap;	/* Inodes bitmap block */
	u32 bg_inode_table;	/* Inodes table block */
	u16 bg_free_blocks_count;	/* Free blocks count */
	u16 bg_free_inodes_count;	/* Free inodes count */
	u16 bg_used_dirs_count;	/* Directories count */
	u16 bg_flags;		/* EXT4_BG_flags (INODE_UNINIT, etc) */

	u32 bg_exclude_bitmap_lo;	/* Exclude bitmap for snapshots */
	u16 bg_block_bitmap_csum_lo;	/* crc32c(s_uuid+grp_num+bitmap) LSB */
	u16 bg_inode_bitmap_csum_lo;	/* crc32c(s_uuid+grp_num+bitmap) LSB */
	u16 bg_itable_unused;	/* Unused inodes count */
	u16 bg_checksum;	/* crc16(sb_uuid+group+desc) */
} __FNG_ATTR_PACKED;		/* 32 byte */

/*
 * Structure of a blocks group descriptor for 64 Bit
 */
struct ext4_group_desc {
	u32 bg_block_bitmap;	/* Blocks bitmap block */
	u32 bg_inode_bitmap;	/* Inodes bitmap block */
	u32 bg_inode_table;	/* Inodes table block */
	u16 bg_free_blocks_count;	/* Free blocks count */
	u16 bg_free_inodes_count;	/* Free inodes count */
	u16 bg_used_dirs_count;	/* Directories count */
	u16 bg_flags;		/* EXT4_BG_flags (INODE_UNINIT, etc) */

	u32 bg_exclude_bitmap_lo;	/* Exclude bitmap for snapshots */
	u16 bg_block_bitmap_csum_lo;	/* crc32c(s_uuid+grp_num+bitmap) LSB */
	u16 bg_inode_bitmap_csum_lo;	/* crc32c(s_uuid+grp_num+bitmap) LSB */
	u16 bg_itable_unused;	/* Unused inodes count */
	u16 bg_checksum;	/* crc16(sb_uuid+group+desc) */

	/* EXT4_FEATURE_INCOMPAT_64BIT */
	u32 bg_block_bitmap_hi;	/* Blocks bitmap block MSB */
	u32 bg_inode_bitmap_hi;	/* Inodes bitmap block MSB */
	u32 bg_inode_table_hi;	/* Inodes table block MSB */
	u16 bg_free_blocks_count_hi;	/* Free blocks count MSB */
	u16 bg_free_inodes_count_hi;	/* Free inodes count MSB */
	u16 bg_used_dirs_count_hi;	/* Directories count MSB */
	u16 bg_itable_unused_hi;	/* Unused inodes count MSB */
	u32 bg_exclude_bitmap_hi;	/* Exclude bitmap block MSB */
	u16 bg_block_bitmap_csum_hi;	/* crc32c(s_uuid+grp_num+bitmap) MSB */
	u16 bg_inode_bitmap_csum_hi;	/* crc32c(s_uuid+grp_num+bitmap) MSB */
	u32 bg_reserved;
} __FNG_ATTR_PACKED;		/* 64 byte */

/*
 * bg_flags @ ext4
 */
#define EXT2_BG_INODE_UNINIT	0x01	/* Inode table/bitmap not initialized */
#define EXT2_BG_BLOCK_UNINIT	0x02	/* Block bitmap not initialized */
#define EXT2_BG_INODE_ZEROED	0x04	/* On-disk itable initialized to zero */

/**
 * Structure of the super block
 */
struct ext2_sb {
	u32 s_inodes_count;	/* Inodes count */
	u32 s_blocks_count;	/* Blocks count */
	u32 s_r_blocks_count;	/* Reserved blocks count */
	u32 s_free_blocks_count;	/* Free blocks count */
	u32 s_free_inodes_count;	/* Free inodes count */
	u32 s_first_data_block;	/* First Data Block */
	u32 s_log_block_size;	/* Block size */
	s32 s_log_frag_size;	/* Fragment size */
	u32 s_blocks_per_group;	/* # Blocks per group */
	u32 s_frags_per_group;	/* # Fragments per group */
	u32 s_inodes_per_group;	/* # Inodes per group */
	u32 s_mtime;		/* Mount time */
	u32 s_wtime;		/* Write time */
	u16 s_mnt_count;	/* Mount count */
	s16 s_max_mnt_count;	/* Maximal mount count */
	u16 s_magic;		/* Magic signature */
	u16 s_state;		/* File system state */
	u16 s_errors;		/* Behaviour when detecting errors */
	u16 s_minor_rev_level;	/* minor revision level */
	u32 s_lastcheck;	/* time of last check */
	u32 s_checkinterval;	/* max. time between checks */

	u32 s_creator_os;	/* OS */
#define EXT2_OS_LINUX		0
#define EXT2_OS_HURD		1
#define EXT2_OS_MASIX		2
#define EXT2_OS_FREEBSD		3
#define EXT2_OS_LITES		4

	u32 s_rev_level;	/* Revision level */
#define EXT2_GOOD_OLD_REV	0	/* The good old (original) format */
#define EXT2_DYNAMIC_REV	1	/* V2 format w/ dynamic inode sizes */

	u16 s_def_resuid;	/* Default uid for reserved blocks */
	u16 s_def_resgid;	/* Default gid for reserved blocks */
	/*
	 * These fields are for EXT2_DYNAMIC_REV superblocks only.
	 *
	 * Note: the difference between the compatible feature set and
	 * the incompatible feature set is that if there is a bit set
	 * in the incompatible feature set that the kernel doesn't
	 * know about, it should refuse to mount the filesystem.
	 *
	 * e2fsck's requirements are more strict; if it doesn't know
	 * about a feature in either the compatible or incompatible
	 * feature set, it must abort and not try to meddle with
	 * things it doesn't understand...
	 */
	u32 s_first_ino;	/* First non-reserved inode */
	u16 s_inode_size;	/* size of inode structure */
	u16 s_block_group_nr;	/* block group # of this superblock */

	u32 s_feature_compat;	/* compatible feature set */
#define EXT2_FEATURE_COMPAT_DIR_PREALLOC	0x0001
#define EXT2_FEATURE_COMPAT_IMAGIC_INODES	0x0002
#define EXT3_FEATURE_COMPAT_HAS_JOURNAL		0x0004
#define EXT2_FEATURE_COMPAT_EXT_ATTR		0x0008
#define EXT2_FEATURE_COMPAT_RESIZE_INO		0x0010
#define EXT2_FEATURE_COMPAT_DIR_INDEX		0x0020
#define EXT2_FEATURE_COMPAT_LAZY_BG		0x0040
#define EXT2_FEATURE_COMPAT_EXCLUDE_INODE	0x0080

	u32 s_feature_incompat;	/* incompatible feature set */
#define EXT2_FEATURE_INCOMPAT_COMPRESSION	0x0001
#define EXT2_FEATURE_INCOMPAT_FILETYPE		0x0002
#define EXT3_FEATURE_INCOMPAT_RECOVER		0x0004	/* Needs recovery */
#define EXT3_FEATURE_INCOMPAT_JOURNAL_DEV	0x0008	/* Journal device */
#define EXT2_FEATURE_INCOMPAT_META_BG		0x0010
#define EXT4_FEATURE_INCOMPAT_EXTENTS		0x0040
#define EXT4_FEATURE_INCOMPAT_64BIT		0x0080
#define EXT4_FEATURE_INCOMPAT_MMP		0x0100
#define EXT4_FEATURE_INCOMPAT_FLEX_BG		0x0200
#define EXT4_FEATURE_INCOMPAT_EA_INODE		0x0400
#define EXT4_FEATURE_INCOMPAT_DIRDATA		0x1000

	u32 s_feature_ro_compat;	/* readonly-compatible feature set */
#define EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER	0x0001
#define EXT2_FEATURE_RO_COMPAT_LARGE_FILE	0x0002
#define EXT2_FEATURE_RO_COMPAT_BTREE_DIR	0x0004	/* not used... */
#define EXT4_FEATURE_RO_COMPAT_HUGE_FILE	0x0008
#define EXT4_FEATURE_RO_COMPAT_GDT_CSUM		0x0010	/* uninit_bg */
#define EXT4_FEATURE_RO_COMPAT_DIR_NLINK	0x0020
#define EXT4_FEATURE_RO_COMPAT_EXTRA_ISIZE	0x0040
#define EXT4_FEATURE_RO_COMPAT_HAS_SNAPSHOT	0x0080

	u8 s_uuid[16];		/* 128-bit uuid for volume */
	char s_volume_name[16];	/* volume name */
	char s_last_mounted[64];	/* directory where last mounted */
	u32 s_algorithm_usage_bitmap;	/* For compression */
	/*
	 * Performance hints.  Directory preallocation should only
	 * happen if the EXT2_FEATURE_COMPAT_DIR_PREALLOC flag is on.
	 */
	u8 s_prealloc_blocks;	/* Nr of blocks to try to preallocate */
	u8 s_prealloc_dir_blocks;	/* Nr to preallocate for dirs */
	u16 s_reserved_gdt_blocks;	/* Per group table for online growth */
	/*
	 * Journaling support valid if EXT3_FEATURE_COMPAT_HAS_JOURNAL set.
	 */
	u8 s_journal_uuid[16];	/* uuid of journal superblock */
	u32 s_journal_inum;	/* inode number of journal file */
	u32 s_journal_dev;	/* device number of journal file */
	u32 s_last_orphan;	/* start of list of inodes to delete */
	u32 s_hash_seed[4];	/* HTREE hash seed */
	u8 s_def_hash_version;	/* Default hash version to use */
	u8 s_jnl_backup_type;	/* Default type of journal backup */
	u16 s_desc_size;	/* Group desc. size: INCOMPAT_64BIT */
	u32 s_default_mount_opts;
	u32 s_first_meta_bg;	/* First metablock group */
	u32 s_mkfs_time;	/* When the filesystem was created */
	u32 s_jnl_blocks[17];	/* Backup of the journal inode */
	u32 s_blocks_count_hi;	/* Blocks count high 32bits */
	u32 s_r_blocks_count_hi;	/* Reserved blocks count high 32 bits */
	u32 s_free_blocks_hi;	/* Free blocks count */
	u16 s_min_extra_isize;	/* All inodes have at least # bytes */
	u16 s_want_extra_isize;	/* New inodes should reserve # bytes */
	u32 s_flags;		/* Miscellaneous flags */
	u16 s_raid_stride;	/* RAID stride */
	u16 s_mmp_interval;	/* # seconds to wait in MMP checking */
	u64 s_mmp_block;	/* Block for multi-mount protection */
	u32 s_raid_stripe_width;	/* blocks on all data disks (N*stride) */
	u8 s_log_groups_per_flex;	/* FLEX_BG group size */
	u8 s_reserved_char_pad;
	u16 s_reserved_pad;	/* Padding to next 32bits */
	u64 s_kbytes_written;	/* nr of lifetime kilobytes written */

	u32 s_snapshot_inum;	/* Inode number of active snapshot */
	u32 s_snapshot_id;	/* sequential ID of active snapshot */
	u64 s_snapshot_r_blocks_count;	/* reserved blocks for active
					   snapshot's future use */
	u32 s_snapshot_list;	/* inode number of the head of the on-disk snapshot list */
	u32 s_error_count;	/* number of fs errors */
	u32 s_first_error_time;	/* first time an error happened */
	u32 s_first_error_ino;	/* inode involved in first error */
	u64 s_first_error_block;	/* block involved of first error */
	u8 s_first_error_func[32];	/* function where the error happened */
	u32 s_first_error_line;	/* line number where error happened */
	u32 s_last_error_time;	/* most recent time of an error */
	u32 s_last_error_ino;	/* inode involved in last error */
	u32 s_last_error_line;	/* line number where error happened */
	u64 s_last_error_block;	/* block involved of last error */
	u8 s_last_error_func[32];	/* function where the error happened */
	u8 s_mount_opts[64];
	u32 s_usr_quota_inum;	/* inode number of user quota file */
	u32 s_grp_quota_inum;	/* inode number of group quota file */
	u32 s_overhead_blocks;	/* overhead blocks/clusters in fs */
	u32 s_backup_bgs[2];	/* If sparse_super2 enabled */
	u32 s_reserved[106];	/* Padding to the end of the block */
	u32 s_checksum;		/* crc32c(superblock) */
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global ext2fs variables */
static struct ext2_sb sb;

/**
 * fs_test - check if ext2 filesystem
 *
 * @return 0 - no ext2fs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	ssize_t r;

	r = pread_loop(fd, &sb, sizeof(struct ext2_sb), 1024);
	if (r == -1) {
		return 0;
	}

	if (SB_MAGIC_EXT2 == le16_to_cpu(sb.s_magic)) {
		return 1;
	}

	return 0;
}

/**
 * fs_info -  gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	u32 u;

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* simple tests for ext3 and ext4 fs */
	fsi->name = "ext2";	/* default */
	u = le32_to_cpu(sb.s_feature_compat);
	if (is_set(u, EXT3_FEATURE_COMPAT_HAS_JOURNAL))
		fsi->name = "ext3";
	u = le32_to_cpu(sb.s_feature_ro_compat);
	if (is_set(u, EXT4_FEATURE_RO_COMPAT_HUGE_FILE))
		fsi->name = "ext4";

	fsi->bsize = EXT2_MIN_BLOCK_SIZE << le32_to_cpu(sb.s_log_block_size);

	/* 64bit block count */
	u = le32_to_cpu(sb.s_feature_incompat);
	if (is_set(u, EXT4_FEATURE_INCOMPAT_64BIT)) {
		fsi->blocks = le32_to_cpu(sb.s_blocks_count) | (u64)
		    le32_to_cpu(sb.s_blocks_count_hi)
		    << 32;
		fsi->bfree = le32_to_cpu(sb.s_free_blocks_count) | (u64)
		    le32_to_cpu(sb.s_free_blocks_hi)
		    << 32;
	} else {
		fsi->blocks = le32_to_cpu(sb.s_blocks_count);
		fsi->bfree = le32_to_cpu(sb.s_free_blocks_count);
	}

	fsi->bused = fsi->blocks - fsi->bfree;
	fsi->fssize = fsi->blocks * fsi->bsize;

	/* calculate bitmap for later */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;

	return 0;
}

/**
 * ext2_showflags - show some value, or not
 */
static void ext2_showflags(stralloc * sa, u32 flags, u32 val, char *str)
{
	if (is_set(flags, val)) {
		if (!stralloc_cats(sa, str))
			die_nomem();
		if (!stralloc_cats(sa, " "))
			die_nomem();
	}
}

static const char *m1 =
    "\n\nSorry, your ext4 filesystem has not initialized all needed Bitmaps!\n";
static const char *m2 =
    "Please run these two commands for getting frisbee-ng to work:\n\n";
static const char *m3 = "# tune2fs -O ^uninit_bg <partition>\n";
static const char *m4 = "# fsck.ext4 -y <partition>\n";

static void die_uninit(void)
{
	die_msg(111, m1, m2, m3, m4);

	return;
}

static int test_root(int bg, int n)
{
	while (1) {
		if (bg < n)
			return 0;
		if (bg == n)
			return 1;
		if ((bg % n) != 0)
			return 0;
		bg = bg / n;
	}
}

static int bg_has_sb(int bg)
{
	if (bg == 0)
		return 1;

	if (bg == 1)
		return 1;

	if (test_root(bg, 3))
		return 1;

	if (test_root(bg, 5))
		return 1;

	if (test_root(bg, 7))
		return 1;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	stralloc sa;
	u32 bg;			/* block groups */
	off_t off;
	u32 i, u;

	if (fs_info(fd) == -1) {
		return -1;
	}

	stralloc_init(&sa);
	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s%s\n", "volume name:", sb.s_volume_name);
	u = le32_to_cpu(sb.s_blocks_per_group);
	term_print("%20s%u\n", "blocks per group:", u);
	u = le32_to_cpu(sb.s_log_block_size);
	term_print("%20s%u\n", "log block size:", u);
	u = EXT2_MIN_BLOCK_SIZE << le32_to_cpu(sb.s_log_block_size);
	term_print("%20s%u\n", "fs block size:", u);
	u = EXT2_MIN_BLOCK_SIZE << le32_to_cpu(sb.s_log_frag_size);
	term_print("%20s%u\n\n", "fragment size:", u);
	term_print("%20s%u\n", "fs # of blocks:", fsi->blocks);
	term_print("%20s%u\n", "fs # used blocks:", fsi->bused);
	term_print("%20s%u\n\n", "fs # free blocks:", fsi->bfree);
	u = le32_to_cpu(sb.s_inodes_count);
	term_print("%20s%u\n", "fs # of inodes:", u);
	u = le32_to_cpu(sb.s_inodes_count) -
	    le32_to_cpu(sb.s_free_inodes_count);
	term_print("%20s%u\n", "fs # used inodes:", u);
	u = le32_to_cpu(sb.s_free_inodes_count);
	term_print("%20s%u\n\n", "fs # free inodes:", u); {
		static const char *os_type[] = {
			"Linux", "Hurd",
			"Masix", "FreeBSD", "Lites"
		};
		u = le32_to_cpu(sb.s_creator_os);
		term_print("%20s%s\n", "creator os:", os_type[u]);
	}

	u = le32_to_cpu(sb.s_feature_compat);
	if (!stralloc_copys(&sa, ""))
		die_nomem();
	ext2_showflags(&sa, u, EXT2_FEATURE_COMPAT_DIR_PREALLOC,
		       "dir_prealloc");
	ext2_showflags(&sa, u, EXT2_FEATURE_COMPAT_IMAGIC_INODES,
		       "imagic_inodes");
	ext2_showflags(&sa, u, EXT3_FEATURE_COMPAT_HAS_JOURNAL, "has_journal");
	ext2_showflags(&sa, u, EXT2_FEATURE_COMPAT_EXT_ATTR, "ext_attr");
	ext2_showflags(&sa, u, EXT2_FEATURE_COMPAT_RESIZE_INO, "resize_ino");
	ext2_showflags(&sa, u, EXT2_FEATURE_COMPAT_DIR_INDEX, "dir_index");
	ext2_showflags(&sa, u, EXT2_FEATURE_COMPAT_LAZY_BG, "lazy_bg");
	ext2_showflags(&sa, u, EXT2_FEATURE_COMPAT_EXCLUDE_INODE,
		       "exclude_inode");
	term_print("%20s%S\n", "feature compat:", &sa);
	u = le32_to_cpu(sb.s_feature_incompat);
	if (!stralloc_copys(&sa, ""))
		die_nomem();
	ext2_showflags(&sa, u, EXT2_FEATURE_INCOMPAT_COMPRESSION,
		       "compression");
	ext2_showflags(&sa, u, EXT2_FEATURE_INCOMPAT_FILETYPE, "filetype");
	ext2_showflags(&sa, u, EXT3_FEATURE_INCOMPAT_RECOVER, "recover");
	ext2_showflags(&sa, u, EXT3_FEATURE_INCOMPAT_JOURNAL_DEV,
		       "journal_dev");
	ext2_showflags(&sa, u, EXT2_FEATURE_INCOMPAT_META_BG, "meta_bg");
	ext2_showflags(&sa, u, EXT4_FEATURE_INCOMPAT_EXTENTS, "extends");
	ext2_showflags(&sa, u, EXT4_FEATURE_INCOMPAT_64BIT, "64bit");
	ext2_showflags(&sa, u, EXT4_FEATURE_INCOMPAT_MMP, "mmp");
	ext2_showflags(&sa, u, EXT4_FEATURE_INCOMPAT_FLEX_BG, "flex_bg");
	ext2_showflags(&sa, u, EXT4_FEATURE_INCOMPAT_EA_INODE, "ea_inode");
	ext2_showflags(&sa, u, EXT4_FEATURE_INCOMPAT_DIRDATA, "dirdata");
	term_print("%20s%S\n", "feature incompat:", &sa);
	u = le32_to_cpu(sb.s_feature_ro_compat);
	if (!stralloc_copys(&sa, ""))
		die_nomem();
	ext2_showflags(&sa, u, EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER,
		       "sparse_super");
	ext2_showflags(&sa, u, EXT2_FEATURE_RO_COMPAT_LARGE_FILE, "large_file");
	ext2_showflags(&sa, u, EXT2_FEATURE_RO_COMPAT_BTREE_DIR, "btree_super");
	ext2_showflags(&sa, u, EXT4_FEATURE_RO_COMPAT_HUGE_FILE, "huge_file");
	ext2_showflags(&sa, u, EXT4_FEATURE_RO_COMPAT_GDT_CSUM, "gdt_csum");
	ext2_showflags(&sa, u, EXT4_FEATURE_RO_COMPAT_DIR_NLINK, "dir_nlink");
	ext2_showflags(&sa, u, EXT4_FEATURE_RO_COMPAT_EXTRA_ISIZE,
		       "extra_isize");
	ext2_showflags(&sa, u, EXT4_FEATURE_RO_COMPAT_HAS_SNAPSHOT,
		       "has_snapshot");
	term_print("%20s%S\n", "feature ro compat:", &sa);
	stralloc_free(&sa);
	term_print("\nmore info of superblock:\n");
	u = le32_to_cpu(sb.s_inodes_count);
	term_print("%28s%u\n", "s_inodes_count:", u);
	u = le32_to_cpu(sb.s_blocks_count);
	term_print("%28s%u\n", "s_blocks_count:", u);
	u = le32_to_cpu(sb.s_r_blocks_count);
	term_print("%28s%u\n", "s_r_blocks_count:", u);
	u = le32_to_cpu(sb.s_free_blocks_count);
	term_print("%28s%u\n", "s_free_blocks_count:", u);
	u = le32_to_cpu(sb.s_free_inodes_count);
	term_print("%28s%u\n", "s_free_inodes_count:", u);
	u = le32_to_cpu(sb.s_first_data_block);
	term_print("%28s%u\n", "s_first_data_block:", u);
	u = le32_to_cpu(sb.s_log_block_size);
	term_print("%28s%u\n", "s_log_block_size:", u);
	u = le32_to_cpu(sb.s_log_frag_size);
	term_print("%28s%d\n", "s_log_frag_size:", u);
	u = le32_to_cpu(sb.s_blocks_per_group);
	term_print("%28s%u\n", "s_blocks_per_group:", u);
	u = le32_to_cpu(sb.s_frags_per_group);
	term_print("%28s%u\n", "s_frags_per_group:", u);
	u = le32_to_cpu(sb.s_inodes_per_group);
	term_print("%28s%u\n", "s_inodes_per_group:", u);
	u = le32_to_cpu(sb.s_mtime);
	term_print("%28s%u\n", "s_mtime:", u);
	u = le32_to_cpu(sb.s_wtime);
	term_print("%28s%u\n", "s_wtime:", u);
	u = le32_to_cpu(sb.s_lastcheck);
	term_print("%28s%u\n", "s_lastcheck:", u);
	u = le32_to_cpu(sb.s_checkinterval);
	term_print("%28s%u\n", "s_checkinterval:", u);
	u = le32_to_cpu(sb.s_rev_level);
	term_print("%28s%u\n", "s_rev_level:", u);
	u = le32_to_cpu(sb.s_first_ino);
	term_print("%28s%u\n", "s_first_ino:", u);
	u = le32_to_cpu(sb.s_algorithm_usage_bitmap);
	term_print("%28s%u\n", "s_algorithm_usage_bitmap:", u);
	u = le32_to_cpu(sb.s_journal_inum);
	term_print("%28s%u\n", "s_journal_inum:", u);
	u = le32_to_cpu(sb.s_journal_dev);
	term_print("%28s%u\n", "s_journal_dev:", u);
	u = le32_to_cpu(sb.s_last_orphan);
	term_print("%28s%u\n", "s_last_orphan:", u);
	u = le32_to_cpu(sb.s_default_mount_opts);
	term_print("%28s%u\n", "s_default_mount_opts:", u);
	u = le32_to_cpu(sb.s_first_meta_bg);
	term_print("%28s%u\n", "s_first_meta_bg:", u);
	u = le32_to_cpu(sb.s_mkfs_time);
	term_print("%28s%u\n", "s_mkfs_time:", u);
	u = (u32) le16_to_cpu(sb.s_reserved_gdt_blocks);
	term_print("%28s%u\n", "s_reserved_gdt_blocks:", u);
	u = (u32) (1 << sb.s_log_groups_per_flex);
	term_print("%28s%u\n", "s_log_groups_per_flex:", u);
	u = (u32) (1 << le32_to_cpu(sb.s_first_meta_bg));
	term_print("%28s%u\n", "s_first_meta_bg:", u);
	term_printflush("\n");

	bg = le32_to_cpu(sb.s_blocks_count);
	bg -= le32_to_cpu(sb.s_first_data_block);
	bg += le32_to_cpu(sb.s_blocks_per_group) - 1;
	bg /= le32_to_cpu(sb.s_blocks_per_group);
	term_print("%28s%u\n", "block groups:", bg);

	/*
	 * s_reserved_gdt_blocks: 1017
	 * -> 1x superblock + X block desc + 1017 GDT's + block map + inode map + itable + data
	 * Gruppendeskriptoren:
	 * sparse_super: 0 or a power of 3, 5, or 7
	 * 0, 1, 3, 5, 7, 9, 25, 27, 49, 81, 125, 243, 343, 625, 729
	 */
	off = fsi->bsize == 1024 ? 2048 : fsi->bsize;
	{
		struct ext2_group_desc *gd0, *gd;
		int l = bg * sizeof(struct ext2_group_desc);

		gd = gd0 = mem_alloc(l);
		pread_loop2(fd, gd, l, off);
		for (i = 0; i < bg; i++) {
			term_print("\n%28s%u", "block group:", i);
			u = le32_to_cpu(gd->bg_block_bitmap);
			if (!u) {
				term_print(" -> empty!!!\n");
				continue;
			}
			term_print("\n");
			u = le32_to_cpu(gd->bg_block_bitmap);
			term_print("%28s%u\n", "bg_block_bitmap:", u);
			u = le32_to_cpu(gd->bg_inode_bitmap);
			term_print("%28s%u\n", "bg_inode_bitmap:", u);
			u = le32_to_cpu(gd->bg_inode_table);
			term_print("%28s%u\n", "bg_inode_table:", u);
			u = le16_to_cpu(gd->bg_free_blocks_count);
			term_print("%28s%u\n", "bg_free_blocks_count:", u);
			u = le16_to_cpu(gd->bg_free_inodes_count);
			term_print("%28s%u\n", "bg_free_inodes_count:", u);
			u = le16_to_cpu(gd->bg_used_dirs_count);
			term_print("%28s%u\n", "bg_used_dirs_count:", u);
			u = le16_to_cpu(gd->bg_flags);
			term_print("%28s%u\n", "bg_flags:", u);
			if (is_set(gd->bg_flags, EXT2_BG_BLOCK_UNINIT)) {
				term_print("-> group has no block map! sb=%d\n",
					   bg_has_sb(i));
			}
			if (is_set(gd->bg_flags, EXT2_BG_INODE_UNINIT)) {
				term_print("-> group has no inode map! sb=%d\n",
					   bg_has_sb(i));
			}

			gd++;
		}
		mem_free(gd0);
	}

	term_printflush("\n");
	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	u64 *bgl;		/* bitmap locations */
	u32 bg;			/* block groups */
	off_t off;
	u32 i, u;

	if (fs_info(fd) == -1) {
		return -1;
	}

	if (le32_to_cpu(sb.s_blocks_per_group) != fsi->bsize * 8) {
		die_msg(111, "ext2fs: blocksize passt nicht!");
	}

	bg = le32_to_cpu(sb.s_blocks_count);
	bg -= le32_to_cpu(sb.s_first_data_block);
	bg += le32_to_cpu(sb.s_blocks_per_group) - 1;
	bg /= le32_to_cpu(sb.s_blocks_per_group);
	bgl = mem_talloc(u64, bg);

	/**
	 * locate offset of the first block group header
	 * - 1024 is a special case
	 * - @all other sizes: the beginning of the next data block
	 */
	off = fsi->bsize == 1024 ? 2048 : fsi->bsize;

	/* get the block group descriptors */
	u = le32_to_cpu(sb.s_feature_incompat);
	if (is_set(u, EXT4_FEATURE_INCOMPAT_64BIT)) {
		struct ext4_group_desc *gd0, *gd;
		int l = bg * sizeof(struct ext4_group_desc);

		gd = gd0 = mem_alloc(l);
		pread_loop2(fd, gd, l, off);
		for (i = 0; i < bg; i++) {
			bgl[i] = le32_to_cpu(gd->bg_block_bitmap) | (u64)
			    le32_to_cpu(gd->bg_block_bitmap_hi) << 32;
			if (is_set(gd->bg_flags, EXT2_BG_BLOCK_UNINIT))
				die_uninit();
			if (is_set(gd->bg_flags, EXT2_BG_INODE_UNINIT))
				die_uninit();
			gd++;
		}
		mem_free(gd0);
	} else {
		struct ext2_group_desc *gd0, *gd;
		int l = bg * sizeof(struct ext2_group_desc);

		gd = gd0 = mem_alloc(l);
		pread_loop2(fd, gd, l, off);
		for (i = 0; i < bg; i++) {
			bgl[i] = le32_to_cpu(gd->bg_block_bitmap);
			if (is_set(gd->bg_flags, EXT2_BG_BLOCK_UNINIT))
				die_uninit();
			if (is_set(gd->bg_flags, EXT2_BG_INODE_UNINIT))
				die_uninit();
			gd++;
		}
		mem_free(gd0);
	}

	/* fsi->bitmaplen is set in fs_info() */
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + 1);
	/* read the maps */
	u = le32_to_cpu(sb.s_blocks_count);
	for (i = 0; i < bg; i++) {
		size_t todo = fsi->bsize * 8 < u ? fsi->bsize : (u + 7) / 8;

		//term_print("pread_loop2(fd=%u, len=%u, block=%u)\n", fd, todo, bgl[i]);
		pread_loop2(fd, fsi->bitmap + i * fsi->bsize, todo,
			    fsi->bsize * bgl[i]);
		u -= le32_to_cpu(sb.s_blocks_per_group);
	}
	mem_free(bgl);

	/**
	 * the bitmap is done now... BUT:
	 * if EXT2_FEATURE_COMPAT_RESIZE_INO is set, we have also to
	 * copy this special resize-inode, which is mostly not set in
	 * the block bitmaps... this seems not okay?!
	 *
	 * possible work around #1:
	 * - disable EXT2_FEATURE_COMPAT_RESIZE_INO feature
	 * - set sb.s_reserved_gdt_blocks to zero
	 *
	 * possible work around #2:
	 * - ignore this, cause fsck.ext2 -fy device will fix it ;)
	 *
	 * -> currently I choosed workaround #2 /TR 2014-07-20
	 */
	term_flush();
	return 0;
};

struct file_system ext2_fs = {
	.id = FSID_EXT2,
	.module = "ext2",
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
