
/**
 * Copyright © 2012 Samsung Electronics Co., Ltd. http://www.samsung.com/
 * Copyright © 2013 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"

/**
 * F2FS, Flash Friendly File System by Samsung
 *
 * systems: Linux, maybe others in future
 * endian:  little
 */

#define F2FS_SUPER_MAGIC	0xF2F52010
#define F2FS_SUPER_OFFSET	1024	/* byte-size offset */
#define F2FS_BLKSIZE		4096	/* support only 4KB block */
#define F2FS_MAX_EXTENSION	64	/* # of extension entries */

#define MAX_ACTIVE_LOGS	16
#define MAX_ACTIVE_NODE_LOGS	8
#define MAX_ACTIVE_DATA_LOGS	8

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

/*
 * f2fs on disk superblock
 */
struct f2fs_sb {
	u32 magic;		/* Magic Number */
	u16 major_ver;		/* Major Version */
	u16 minor_ver;		/* Minor Version */
	u32 log_sectorsize;	/* log2 sector size in bytes */
	u32 log_sectors_per_block;	/* log2 # of sectors per block */
	u32 log_blocksize;	/* log2 block size in bytes */
	u32 log_blocks_per_seg;	/* log2 # of blocks per segment */
	u32 segs_per_sec;	/* # of segments per section */
	u32 secs_per_zone;	/* # of sections per zone */
	u32 checksum_offset;	/* checksum offset inside super block */
	u64 block_count;	/* total # of user blocks */
	u32 section_count;	/* total # of sections */
	u32 segment_count;	/* total # of segments */
	u32 segment_count_ckpt;	/* # of segments for checkpoint */
	u32 segment_count_sit;	/* # of segments for SIT */
	u32 segment_count_nat;	/* # of segments for NAT */
	u32 segment_count_ssa;	/* # of segments for SSA */
	u32 segment_count_main;	/* # of segments for main area */
	u32 segment0_blkaddr;	/* start block address of segment 0 */
	u32 cp_blkaddr;		/* start block address of checkpoint */
	u32 sit_blkaddr;	/* start block address of SIT */
	u32 nat_blkaddr;	/* start block address of NAT */
	u32 ssa_blkaddr;	/* start block address of SSA */
	u32 main_blkaddr;	/* start block address of main area */
	u32 root_ino;		/* root inode number */
	u32 node_ino;		/* node inode number */
	u32 meta_ino;		/* meta inode number */
	u8 uuid[16];		/* 128-bit uuid for volume */
	u16 volume_name[512];	/* volume name */
	u32 extension_count;	/* # of extensions below */
	u8 extension_list[F2FS_MAX_EXTENSION][8];	/* extension array */
} __FNG_ATTR_PACKED;

struct f2fs_checkpoint {
	u64 checkpoint_ver;	/* checkpoint block version number */
	u64 user_block_count;	/* # of user blocks */
	u64 valid_block_count;	/* # of valid blocks in main area */
	u32 rsvd_segment_count;	/* # of reserved segments for gc */
	u32 overprov_segment_count;	/* # of overprovision segments */
	u32 free_segment_count;	/* # of free segments in main area */

	/* information of current node segments */
	u32 cur_node_segno[MAX_ACTIVE_NODE_LOGS];
	u16 cur_node_blkoff[MAX_ACTIVE_NODE_LOGS];
	/* information of current data segments */
	u32 cur_data_segno[MAX_ACTIVE_DATA_LOGS];
	u16 cur_data_blkoff[MAX_ACTIVE_DATA_LOGS];
	u32 ckpt_flags;		/* Flags : umount and journal_present */
	u32 cp_pack_total_block_count;	/* total # of one cp pack */
	u32 cp_pack_start_sum;	/* start block number of data summary */
	u32 valid_node_count;	/* Total number of valid nodes */
	u32 valid_inode_count;	/* Total number of valid inodes */
	u32 next_free_nid;	/* Next free node number */
	u32 sit_ver_bitmap_bytesize;	/* Default value 64 */
	u32 nat_ver_bitmap_bytesize;	/* Default value 256 */
	u32 checksum_offset;	/* checksum offset inside cp block */
	u64 elapsed_time;	/* mounted time */
	/* allocation type of current segment */
	unsigned char alloc_type[MAX_ACTIVE_LOGS];

	/* SIT and NAT version bitmap */
	unsigned char sit_nat_version_bitmap[1];
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global variables */
static int init_done = 0;
static struct f2fs_sb *sb;
static struct f2fs_checkpoint *cp;

/**
 * fs_test - check for f2fs filesystem
 *
 * @return 0 - no f2fs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	ssize_t r;

	if (!init_done) {
		sb = mem_talloc(struct f2fs_sb, 1);
		cp = mem_talloc(struct f2fs_checkpoint, 1);

		init_done = 1;
	}

	r = pread_loop(fd, sb, sizeof(struct f2fs_sb), F2FS_SUPER_OFFSET);
	if (r != sizeof(struct f2fs_sb))
		return 0;

	if (le32_to_cpu(sb->magic) != F2FS_SUPER_MAGIC)
		return 0;

	/* 4096 bytes */
	if (le32_to_cpu(sb->log_blocksize) != 12)
		return 0;

	/* 512 bytes */
	if (le32_to_cpu(sb->log_sectorsize) != 9)
		return 0;

	/* F2FS_LOG_SECTORS_PER_BLOCK, 4KB */
	if (le32_to_cpu(sb->log_sectors_per_block) != 3)
		return 0;

	return 1;
}

/**
 * fs_info -  gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	fsi->name = "f2fs";
	fsi->bsize = 1 << le32_to_cpu(sb->log_blocksize);
	fsi->blocks = le64_to_cpu(sb->block_count);
	fsi->bused = 20;
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bfree = fsi->blocks - fsi->bused;

	/* calculate bitmap for later */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	int r;

	if (fs_info(fd) == -1) {
		return -1;
	}

	term_print("\n\n");

	term_print("sizeof sb=%u\n", sizeof(struct f2fs_sb));
	term_print("sizeof f2fs_checkpoint=%u\n",
		   sizeof(struct f2fs_checkpoint));

	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s%08x\n", "magic:", le32_to_cpu(sb->magic));
	term_print("%20s%u\n", "major_ver:", le16_to_cpu(sb->major_ver));
	term_print("%20s%u\n", "minor_ver:", le16_to_cpu(sb->minor_ver));

	term_print("%20s%u\n", "sectorsize:",
		   1 << le32_to_cpu(sb->log_sectorsize));
	term_print("%20s%u\n", "sectors_per_block:",
		   1 << le32_to_cpu(sb->log_sectors_per_block));
	term_print("%20s%u\n", "blocksize:",
		   1 << le32_to_cpu(sb->log_blocksize));
	term_print("%20s%u\n", "blocks_per_seg:",
		   1 << le32_to_cpu(sb->log_blocks_per_seg));
	term_print("%20s%u\n", "segs_per_sec:", le32_to_cpu(sb->segs_per_sec));
	term_print("%20s%u\n", "secs_per_zone:",
		   le32_to_cpu(sb->secs_per_zone));
	term_print("%20s%u\n", "checksum_offset:",
		   le32_to_cpu(sb->checksum_offset));

	term_print("%20s%U\n", "block_count:", le64_to_cpu(sb->block_count));
	term_print("%20s%u\n", "section_count:",
		   le32_to_cpu(sb->section_count));
	term_print("%20s%u\n", "segment_count:",
		   le32_to_cpu(sb->segment_count));
	term_print("%20s%u\n", "segment_count_ckpt:",
		   le32_to_cpu(sb->segment_count_ckpt));
	term_print("%20s%u\n", "segment_count_sit:",
		   le32_to_cpu(sb->segment_count_sit));
	term_print("%20s%u\n", "segment_count_nat:",
		   le32_to_cpu(sb->segment_count_nat));
	term_print("%20s%u\n", "segment_count_ssa:",
		   le32_to_cpu(sb->segment_count_ssa));
	term_print("%20s%u\n", "segment_count_main:",
		   le32_to_cpu(sb->segment_count_main));
	term_print("%20s%u\n", "segment0_blkaddr:",
		   le32_to_cpu(sb->segment0_blkaddr));
	term_print("%20s%u (%u)\n", "cp_blkaddr:", le32_to_cpu(sb->cp_blkaddr),
		   le32_to_cpu(sb->cp_blkaddr) * fsi->bsize);
	term_print("%20s%u (%u)\n", "sit_blkaddr:",
		   le32_to_cpu(sb->sit_blkaddr),
		   le32_to_cpu(sb->sit_blkaddr) * fsi->bsize);
	term_print("%20s%u (%u)\n", "nat_blkaddr:",
		   le32_to_cpu(sb->nat_blkaddr),
		   le32_to_cpu(sb->nat_blkaddr) * fsi->bsize);
	term_print("%20s%u (%u)\n", "ssa_blkaddr:",
		   le32_to_cpu(sb->ssa_blkaddr),
		   le32_to_cpu(sb->ssa_blkaddr) * fsi->bsize);
	term_print("%20s%u (%u)\n", "main_blkaddr:",
		   le32_to_cpu(sb->main_blkaddr),
		   le32_to_cpu(sb->main_blkaddr) * fsi->bsize);
	term_print("%20s%u\n", "root_ino:", le32_to_cpu(sb->root_ino));
	term_print("%20s%u\n", "node_ino:", le32_to_cpu(sb->node_ino));
	term_print("%20s%u\n", "meta_ino:", le32_to_cpu(sb->meta_ino));
	term_print("%20s%u\n", "extension_count:",
		   le32_to_cpu(sb->extension_count));

	r = pread_loop(fd, cp, sizeof(struct f2fs_checkpoint),
		       le32_to_cpu(sb->cp_blkaddr) * fsi->bsize);

	if (r != sizeof(struct f2fs_checkpoint))
		return 0;

	term_print("\n1. cp\n");
	term_print("%20s%U\n", "checkpoint_ver:",
		   le64_to_cpu(cp->checkpoint_ver));
	term_print("%20s%U\n", "user_block_count:",
		   le64_to_cpu(cp->user_block_count));
	term_print("%20s%U\n", "valid_block_count:",
		   le64_to_cpu(cp->valid_block_count));

	term_print("%20s%u\n", "rsvd_segment_count:",
		   le32_to_cpu(cp->rsvd_segment_count));
	term_print("%20s%u\n", "overprov_segment_count:",
		   le32_to_cpu(cp->overprov_segment_count));
	term_print("%20s%u\n", "free_segment_count:",
		   le32_to_cpu(cp->free_segment_count));

	term_print("%20s%u\n", "ckpt_flags:", le32_to_cpu(cp->ckpt_flags));
	term_print("%20s%u\n", "cp_pack_total_block_count:",
		   le32_to_cpu(cp->cp_pack_total_block_count));
	term_print("%20s%u\n", "cp_pack_start_sum:",
		   le32_to_cpu(cp->cp_pack_start_sum));
	term_print("%20s%u\n", "valid_node_count:",
		   le32_to_cpu(cp->valid_node_count));
	term_print("%20s%u\n", "valid_inode_count:",
		   le32_to_cpu(cp->valid_inode_count));
	term_print("%20s%u\n", "next_free_nid:",
		   le32_to_cpu(cp->next_free_nid));
	term_print("%20s%u\n", "sit_ver_bitmap_bytesize:",
		   le32_to_cpu(cp->sit_ver_bitmap_bytesize));
	term_print("%20s%u\n", "nat_ver_bitmap_bytesize:",
		   le32_to_cpu(cp->nat_ver_bitmap_bytesize));
	term_print("%20s%u\n", "checksum_offset:",
		   le32_to_cpu(cp->checksum_offset));
	term_print("%20s%U\n", "elapsed_time:", le64_to_cpu(cp->elapsed_time));

	r = pread_loop(fd, cp, sizeof(struct f2fs_checkpoint),
		       le32_to_cpu(sb->cp_blkaddr) * fsi->bsize * 2);

	if (r != sizeof(struct f2fs_checkpoint))
		return 0;

	term_print("\n2. cp\n");
	term_print("%20s%U\n", "checkpoint_ver:",
		   le64_to_cpu(cp->checkpoint_ver));
	term_print("%20s%U\n", "user_block_count:",
		   le64_to_cpu(cp->user_block_count));
	term_print("%20s%U\n", "valid_block_count:",
		   le64_to_cpu(cp->valid_block_count));

	term_print("%20s%u\n", "rsvd_segment_count:",
		   le32_to_cpu(cp->rsvd_segment_count));
	term_print("%20s%u\n", "overprov_segment_count:",
		   le32_to_cpu(cp->overprov_segment_count));
	term_print("%20s%u\n", "free_segment_count:",
		   le32_to_cpu(cp->free_segment_count));

	term_print("%20s%u\n", "ckpt_flags:", le32_to_cpu(cp->ckpt_flags));
	term_print("%20s%u\n", "cp_pack_total_block_count:",
		   le32_to_cpu(cp->cp_pack_total_block_count));
	term_print("%20s%u\n", "cp_pack_start_sum:",
		   le32_to_cpu(cp->cp_pack_start_sum));
	term_print("%20s%u\n", "valid_node_count:",
		   le32_to_cpu(cp->valid_node_count));
	term_print("%20s%u\n", "valid_inode_count:",
		   le32_to_cpu(cp->valid_inode_count));
	term_print("%20s%u\n", "next_free_nid:",
		   le32_to_cpu(cp->next_free_nid));
	term_print("%20s%u\n", "sit_ver_bitmap_bytesize:",
		   le32_to_cpu(cp->sit_ver_bitmap_bytesize));
	term_print("%20s%u\n", "nat_ver_bitmap_bytesize:",
		   le32_to_cpu(cp->nat_ver_bitmap_bytesize));
	term_print("%20s%u\n", "checksum_offset:",
		   le32_to_cpu(cp->checksum_offset));
	term_print("%20s%U\n", "elapsed_time:", le64_to_cpu(cp->elapsed_time));

#if 0
	u64 checkpoint_ver;	/* checkpoint block version number */
	u64 user_block_count;	/* # of user blocks */
	u64 valid_block_count;	/* # of valid blocks in main area */
	u32 rsvd_segment_count;	/* # of reserved segments for gc */
	u32 overprov_segment_count;	/* # of overprovision segments */
	u32 free_segment_count;	/* # of free segments in main area */

	/* information of current node segments */
	u32 cur_node_segno[MAX_ACTIVE_NODE_LOGS];
	u16 cur_node_blkoff[MAX_ACTIVE_NODE_LOGS];

	/* information of current data segments */
	u32 cur_data_segno[MAX_ACTIVE_DATA_LOGS];
	u16 cur_data_blkoff[MAX_ACTIVE_DATA_LOGS];

	u32 ckpt_flags;		/* Flags : umount and journal_present */
	u32 cp_pack_total_block_count;	/* total # of one cp pack */
	u32 cp_pack_start_sum;	/* start block number of data summary */
	u32 valid_node_count;	/* Total number of valid nodes */
	u32 valid_inode_count;	/* Total number of valid inodes */
	u32 next_free_nid;	/* Next free node number */
	u32 sit_ver_bitmap_bytesize;	/* Default value 64 */
	u32 nat_ver_bitmap_bytesize;	/* Default value 256 */
	u32 checksum_offset;	/* checksum offset inside cp block */
	u64 elapsed_time;	/* mounted time */

	/* allocation type of current segment */
	unsigned char alloc_type[MAX_ACTIVE_LOGS];

	struct super_block *sb = dentry->d_sb;
	struct f2fs_sb_info *sbi = F2FS_SB(sb);
	u64 id = huge_encode_dev(sb->s_bdev->bd_dev);
	block_t total_count, user_block_count, start_count, ovp_count;

	total_count = le64_to_cpu(sbi->raw_super->block_count);
	user_block_count = sbi->user_block_count;
	start_count = le32_to_cpu(sbi->raw_super->segment0_blkaddr);
	ovp_count = SM_I(sbi)->ovp_segments << sbi->log_blocks_per_seg;
	buf->f_type = F2FS_SUPER_MAGIC;
	buf->f_bsize = sbi->blocksize;

	buf->f_blocks = total_count - start_count;
	buf->f_bfree = buf->f_blocks - valid_user_blocks(sbi) - ovp_count;
	buf->f_bavail = user_block_count - valid_user_blocks(sbi);

	buf->f_files = sbi->total_node_count;
	buf->f_ffree = sbi->total_node_count - valid_inode_count(sbi);

	buf->f_namelen = F2FS_MAX_NAME_LEN;
	buf->f_fsid.val[0] = (u32) id;
	buf->f_fsid.val[1] = (u32) (id >> 32);
#endif

	term_printflush("\n");

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	if (fs_info(fd) == -1) {
		return -1;
	}

	return -1;
};

struct file_system f2fs_fs = {
	.id = FSID_F2FS,
	.module = "f2fs",
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
