
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"

/* XXX: not ready! */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

/**
 * fat12 / fat16 / fat32 store little endian on disk
 * - all info was taken from fatgen103.doc, see this site for more info:
 *   http://msdn.microsoft.com/de-de/windows/hardware/gg463084
 */
struct fat_sb {			/* offset: description */
	u8 BS_jmpBoot[3];	/* 00: Boot strap short or near jump */
	u8 BS_OEMName[8];	/* 03: system name */
	u16 BPB_BytsPerSec;	/* 11: bytes per logical sector */
	u8 BPB_SecPerClus;	/* 13: sectors/cluster */
	u16 BPB_RsvdSecCnt;	/* 14: reserved sectors */
	u8 BPB_NumFATs;		/* 16: number of FATs */
	u16 BPB_RootEntCnt;	/* 17: number of root directory entries */
	u16 BPB_TotSec16;	/* 19: if 0, BPB_TotSec32 supersedes */
	u8 BPB_Media;		/* 21: media code */
	u16 BPB_FATSz16;	/* 22: sectors/FAT for FAT12/16 */
	u16 BPB_SecPerTrk;	/* 24: sectors per track */
	u16 BPB_NumHeads;	/* 26: number of heads */
	u32 BPB_HiddSec;	/* 28: hidden sectors (partition start) */
	u32 BPB_TotSec32;	/* 32: no. of sectors (if sectors == 0) */

	union {
		/* fat12 / fat16 */
		struct {
			u8 BS_DrvNum;	/* 36: Int 0x13 drive number */
			u8 BS_Reserved1;	/* 37: reserved, used by Windows NT */
			u8 BS_BootSig;	/* 38: Extended boot signature (0x29) */
			u32 BS_VolID;	/* 39: Volume serial number */
			u8 BS_Label[11];	/* 43: Volume label */

			/* this field should not be used for type indentication!! */
			u8 BS_SysType[8];	/* 54: "FAT12   ", "FAT16   " ... */
			u8 boot_code[448];	/* Boot code (or message) */
		} fat16;

		/* fat32 */
		struct {
			u32 BPB_FATSz32;	/* 36: count of sectors in ONE FAT */
			u16 BPB_ExtFlags;	/* 40: special field, must be checked */
			u16 BPB_FSVer;	/* 42: High=major; Low byte=minor */
			u32 BPB_RootClus;	/* 44: cluster of rootdir */
			u16 BPB_FSInfo;	/* 48: sector number of fsinfo structure */
			u16 BPB_BkBootSec;	/* 50: ... */
			u8 BPB_Reserved[12];	/* 52: reserved, */
			u8 BS_DrvNum;	/* 64: ... */
			u8 BS_Reserved1;	/* 65: all should be '\0' */
			u8 BS_BootSig;	/* 66: Extended boot signature (0x29) */
			u32 BS_VolID;	/* 67: Volume serial number */
			u8 BS_VolLab[11];	/* 71: Volume label */
			u8 BS_FilSysType[8];	/* 82: should be "FAT32   " */
			u8 boot_code[420];	/* Boot code (or message) */
		} fat32;
	} u;
	u16 boot_sign;		/* 510: always 0xAA55 */
} __FNG_ATTR_PACKED;		/* 512 byte */

#define FAT32_FSINFO_MAGIC  0x41615252
#define FAT32_FSINFO_SIG    0x61417272
struct FatFsInfo {
	u32 magic;		/* 0x41615252L */
	u8 rev[480];
	u32 signature;		/* 0x61417272L */
	u32 free_count;		/* free clusters or -1 */
	u32 next_free;		/* recently allocated cluster */
	u32 reserved2[3];
	u32 trail;
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* maximum number of clusters */
#define MAX_FAT12   0xff4
#define MAX_FAT16   0xfff4
#define MAX_FAT32   0x0ffffff6

/* bad cluster mark */
#define BAD_FAT12   0xff7
#define BAD_FAT16   0xfff7
#define BAD_FAT32   0x0ffffff7

/* standard EOF */
#define EOF_FAT12   0xfff
#define EOF_FAT16   0xffff
#define EOF_FAT32   0x0fffffff

/* global fat variables */
static struct fat_sb sb;

static void die_fat(char *s)
{
	die_msg(111, s);
}

#define FAT12     1
#define FAT16     2
#define FAT32     3
#define FAT32PLUS 4

/**
 * fs_test - check if exfat filesystem
 *
 * @return 0 - no exfat
 * @return >0 - it is fat
 */
static int fs_test(int fd)
{
	ssize_t r;

	r = pread_loop(fd, &sb, sizeof(struct fat_sb), 0);
	if (r == -1) {
		return 0;
	}

	/* a first fast check */
	if (le16_to_cpu(sb.boot_sign) != 0xaa55)
		return 0;

	/**
	 * check for FAT32
	 */
	if (sb.u.fat32.BS_BootSig == 0x29)
		if (le16_to_cpu(sb.BPB_RootEntCnt) == 0)
			if (le16_to_cpu(sb.BPB_TotSec16) == 0) {
				/* * 0.0 = FAT32 */
				if (sb.u.fat32.BPB_FSVer == 0) {
					return FAT32;
				}

				/* 0.1 = FAT32+ (filesize up to 256GB) */
				return FAT32PLUS;

				/* there will be no 0.2 I hope! */
			}

	/**
	 * check for fat12 / fat16
	 */
	if (sb.u.fat16.BS_BootSig == 0x29) {
		/* FAT12 will be tested @ fs_info() */
		return FAT16;
	}

	/* XXX: check other things here? */
	term_print("fat.c sb.u.fat16.BS_BootSig=%u\n", sb.u.fat16.BS_BootSig);
	term_flush();

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	u32 RootDirSectors, TotSec, FATSz, DataSec;
	int fat_type = 0;

	fat_type = fs_test(fd);
	if (fat_type == 0) {
		return -1;
	}

	/* like described @ fatgen103.pdf */
	RootDirSectors =
	    le16_to_cpu(sb.BPB_RootEntCnt) * 32 +
	    le16_to_cpu(sb.BPB_BytsPerSec) - 1;

	/* calculate total number of sectors per fat */
	if (le16_to_cpu(sb.BPB_FATSz16) != 0) {
		FATSz = le16_to_cpu(sb.BPB_FATSz16);
	} else {
		FATSz = le16_to_cpu(sb.u.fat32.BPB_FATSz32);
	}

	/* calculate total number of sectors */
	if (le16_to_cpu(sb.BPB_TotSec16) != 0) {
		TotSec = le16_to_cpu(sb.BPB_TotSec16);
	} else {
		TotSec = le16_to_cpu(sb.BPB_TotSec32);
	}

	DataSec =
	    TotSec + le16_to_cpu(sb.BPB_RsvdSecCnt) + (sb.BPB_NumFATs * FATSz) +
	    RootDirSectors;
	fsi->blocks = DataSec / sb.BPB_SecPerClus;

	/* check for fat16 / fat12 */
	if (fat_type == FAT16) {
		if (fsi->blocks <= MAX_FAT12) {
			fat_type = FAT12;
		} else if (fsi->blocks <= MAX_FAT16) {
			fat_type = FAT16;
		} else
			die_fat("fat16 type, but to much clusters for it!");
	}

	/* set fs name */
	switch (fat_type) {
	case FAT12:
		fsi->name = "FAT12";
		break;
	case FAT16:
		fsi->name = "FAT16";
		break;
	case FAT32:
		fsi->name = "FAT32";
		break;
	case FAT32PLUS:
		fsi->name = "FAT32+";
		break;
	}

	/* now, we have to read the fat ... */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = mem_talloc(u8, fsi->bitmaplen);

	fsi->fssize = DataSec + 2;	/* 2 reserved cluster */
	fsi->bsize = sb.BPB_SecPerClus * 1024;
	fsi->bused = fsi->blocks;
	fsi->bfree = 0;

	return 0;
};

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	if (fs_getbitmap(fd) == -1) {
		return -1;
	}

	/* not needed @ fs_info() */
	//free(fsi->bitmap);
	//fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	if (fs_getbitmap(fd) == -1) {
		return -1;
	}

	term_print("\n\ntodo");
	// term_print("%20s%s\n", "filesystem type:", fs_name);
	// term_print("%20s%u%%\n", "allocated percent:",
	//      (u32)sb.allocated_percent);

	term_printflush("\n");

	return 0;
}

struct file_system fat_fs = {
	.id = FSID_FAT,
	.module = "fat",
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
