
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * based on Technical Note TN1150 / HFS Plus Volume Format 2004-03-05
 * and source code of xnu
 */

/* system */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/bswap.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/term.h"
#include "src/mem.h"

/**
 * Apple Hierarchical File System (HFS)
 *
 * systems: Mac OS, Mac OS X, BeOS, Linux
 * endian:  big endian
 */

static char fs_name[] = "hfs";

#define HFS_SIG_HFS   0x4244	/* "BD": HFS MDB (super block) */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

struct hfs_extent {
	u16 block;
	u16 count;
};

typedef struct hfs_extent hfs_ext_rec[3];

struct hfs_sb {
	u16 drSigWord;		/* Signature word indicating fs type */
	u32 drCrDate;		/* fs creation date/time */
	u32 drLsMod;		/* fs modification date/time */
	u16 drAtrb;		/* fs attributes */
	u16 drNmFls;		/* number of files in root directory */
	u16 drVBMSt;		/* location (in 512-byte blocks)
				   of the volume bitmap */
	u16 drAllocPtr;		/* location (in allocation blocks)
				   to begin next allocation search */
	u16 drNmAlBlks;		/* number of allocation blocks */
	u32 drAlBlkSiz;		/* bytes in an allocation block */
	u32 drClpSiz;		/* clumpsize, the number of bytes to
				   allocate when extending a file */
	u16 drAlBlSt;		/* location (in 512-byte blocks)
				   of the first allocation block */
	u32 drNxtCNID;		/* CNID to assign to the next
				   file or directory created */
	u16 drFreeBks;		/* number of free allocation blocks */
	u8 drVN[28];		/* the volume label */
	u32 drVolBkUp;		/* fs backup date/time */
	u16 drVSeqNum;		/* backup sequence number */
	u32 drWrCnt;		/* fs write count */
	u32 drXTClpSiz;		/* clumpsize for the extents B-tree */
	u32 drCTClpSiz;		/* clumpsize for the catalog B-tree */
	u16 drNmRtDirs;		/* # of directories in the root directory */
	u32 drFilCnt;		/* # of files in the fs */
	u32 drDirCnt;		/* # of directories in the fs */
	u8 drFndrInfo[32];	/* data used by the Finder */
	u16 drEmbedSigWord;	/* embedded volume signature */
	u32 drEmbedExtent;	/* starting block number (xdrStABN)
				   and number of allocation blocks
				   (xdrNumABlks) occupied by volume */
	u32 drXTFlSize;		/* bytes in the extents B-tree */
	hfs_ext_rec drXTExtRec;	/* extents B-tree's first 3 extents */
	u32 drCTFlSize;		/* bytes in the catalog B-tree */
	hfs_ext_rec drCTExtRec;	/* catalog B-tree's first 3 extents */
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* gloabl */
struct hfs_sb sb;

/**
 * fs_test - check if it is a hfs filesysem
 *
 * @return 0 - no hfs fs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	ssize_t r;

	r = pread_loop(fd, &sb, sizeof(struct hfs_sb), 1024);
	if (r == -1) {
		return 0;
	}

	/* check for one of the fs magic numbers */
	if (be16_to_cpu(sb.drSigWord) != HFS_SIG_HFS) {
		return 0;
	}

	/* set the name */
	fsi->name = fs_name;
	return 1;
}

/**
 * fs_info
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* set block stuff */
	fsi->bsize = be32_to_cpu(sb.drAlBlkSiz);
	fsi->blocks = be16_to_cpu(sb.drNmAlBlks);
	fsi->bfree = be16_to_cpu(sb.drFreeBks);
	fsi->bused = fsi->blocks - fsi->bfree;

	/* real used space on disk */
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s%s\n", "fs label:", sb.drVN);
	term_print("%20s0x%4x\n", "fs signature:", be16_to_cpu(sb.drSigWord));
	term_print("%20s%u\n", "fs ctime", be32_to_cpu(sb.drCrDate));
	term_print("%20s%u\n", "fs mtime", be32_to_cpu(sb.drLsMod));
	term_print("%20s%u\n", "fs backup time", be32_to_cpu(sb.drVolBkUp));
	term_print("%20s%u\n", "fs write count", be32_to_cpu(sb.drWrCnt));
	term_print("%20s%u\n", "# of files", be32_to_cpu(sb.drFilCnt));
	term_print("%20s%u\n", "# of directories", be32_to_cpu(sb.drDirCnt));
	term_print("%20s0x%04x\n", "fs attributes:", be16_to_cpu(sb.drAtrb));
	term_print("%20s%u\n", "files @ root:", be16_to_cpu(sb.drNmFls));
	term_print("%20s%u\n", "bitmap @ block:", be16_to_cpu(sb.drVBMSt));

	term_print("%20s%u\n", "# of blocks:", be16_to_cpu(sb.drNmAlBlks));
	term_print("%20s%u\n", "# free blocks:", be16_to_cpu(sb.drFreeBks));
	term_print("%20s%u\n", "# used blocks:", fsi->bused);
	term_print("%20s%u\n", "bytes per block:", be32_to_cpu(sb.drAlBlkSiz));
	term_print("%20s%u\n", "first data block:", be16_to_cpu(sb.drAlBlSt));
	term_flush();

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen);
	pread_loop2(fd, fsi->bitmap, fsi->bitmaplen,
		    be16_to_cpu(sb.drVBMSt) * 512);

	return 0;
}

/**
 * fs_postclone
 * - copy 3rd sector to the second-to-last sector
 */
static void fs_postclone(int fd, int fd_out, off_t offset)
{
	struct fsinfo *fsi = io_getcookie(fd);
	char *x = mem_alloc(512);
	u64 dest;

	if (!fsi)
		fatal("hfs fs_postclone() fsi not defined!");

	/* read 3rd sector of restored data (hfs_sb) */
	pread_loop2(fd_out, x, 512, 2 * 512 + offset);

	/* write it to it's special place */
	dest = (fsi->psize + 511) / 512;
	dest = (dest * 512) - 3 * 512;
	pwrite_loop2(fd_out, x, 512, dest + offset);
	mem_free(x);

	return;
}

struct file_system hfs_fs = {
	.id = FSID_HFS,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_postclone = fs_postclone,
	.fs_debug = fs_debug
};
