
/**
 * Copyright © 2012, 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * based on Technical Note TN1150 / HFS Plus Volume Format 2004-03-05
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/bswap.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/mem.h"
#include "src/fs.h"
#include "src/term.h"
#include "src/defs.h"

/**
 * Apple Hierarchical File System plus (HFS+)
 *
 * systems: Mac OS 8, Mac OS 9, Mac OS X, Linux
 * endian:  big endian
 */

static char fs_nameP[] = "hfs+";
static char fs_nameX[] = "hfsx";

#define HFS_SIG_HFSP  0x482B	/* 'H+' HFS Plus  (Verison 4) */
#define HFS_SIG_HFSX  0x4858	/* 'HX' HFSX (Verison 5) */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

struct hfs_extent {
	u32 start_block;
	u32 block_count;
};

struct hfs_fork {
	u64 logical_size;
	u32 clump_size;
	u32 total_blocks;
	struct hfs_extent ext[8];
};

struct hfs_sb {
	u16 signature;		/* 'H+' => HFS+ / 'HX' => HFSX */
	u16 version;		/* 4 => HFS+  5 => HFSX */
	u32 attributes;
	u32 last_mounting_version;
	u32 journal_info_block;

	u32 date_created;
	u32 date_modified;
	u32 date_backedup;
	u32 date_checked;

	u32 file_count;
	u32 folder_count;

	u32 block_size;
	u32 total_blocks;
	u32 free_blocks;

	u32 next_alloc_block;
	u32 rsrc_clump_size;
	u32 data_clump_size;
	u32 next_cnid;

	u32 write_count;
	u64 encodings;

	u32 finder_info[8];

	struct hfs_fork allocation_file;
	struct hfs_fork extents_file;
	struct hfs_fork catalog_file;
	struct hfs_fork attributes_file;
	struct hfs_fork startup_file;
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

static struct hfs_sb sb;

/**
 * fs_test - check for hfs+
 *
 * @return 0 - no hfs+ / hfsx
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	ssize_t r;

	r = pread_loop(fd, &sb, sizeof(struct hfs_sb), 1024);
	if (r == -1) {
		return 0;
	}

	if (be16_to_cpu(sb.signature) == HFS_SIG_HFSP) {
		fsi->name = fs_nameP;
		return 1;
	}

	if (be16_to_cpu(sb.signature) == HFS_SIG_HFSX) {
		fsi->name = fs_nameX;
		return 1;
	}

	return 0;
}

/**
 * fs_info
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* set block stuff */
	fsi->bsize = be32_to_cpu(sb.block_size);
	fsi->blocks = be32_to_cpu(sb.total_blocks);
	fsi->bfree = be32_to_cpu(sb.free_blocks);
	fsi->bused = fsi->blocks - fsi->bfree;

	/* real used space on disk */
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	term_print("\n\n");

	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s0x%4x\n", "fs signature:", be16_to_cpu(sb.signature));
	term_print("%20s%u\n", "fs version", be32_to_cpu(sb.version));
	term_print("%20s%u\n", "fs last mounted on",
		   be32_to_cpu(sb.last_mounting_version));
	term_print("%20s%u\n", "journal info block",
		   be32_to_cpu(sb.journal_info_block));

	term_print("%20s%u\n", "date created", be32_to_cpu(sb.date_created));
	term_print("%20s%u\n", "date modified", be32_to_cpu(sb.date_modified));
	term_print("%20s%u\n", "date backedup", be32_to_cpu(sb.date_backedup));
	term_print("%20s%u\n", "date checked:", be32_to_cpu(sb.date_checked));

	term_print("%20s%u\n", "file count:", be32_to_cpu(sb.file_count));
	term_print("%20s%u\n", "folder count:", be32_to_cpu(sb.folder_count));

	term_print("%20s%u\n", "block size:", be32_to_cpu(sb.block_size));
	term_print("%20s%u\n", "total blocks:", be32_to_cpu(sb.total_blocks));
	term_print("%20s%u\n", "free_blocks:", be32_to_cpu(sb.free_blocks));
	term_print("%20s%u\n", "used blocks:", fsi->bused);

	term_print("%20s%u\n", "next_alloc_block:",
		   be32_to_cpu(sb.next_alloc_block));
	term_print("%20s%u\n", "rsrc_clump_size:",
		   be32_to_cpu(sb.rsrc_clump_size));
	term_print("%20s%u\n", "data_clump_size:",
		   be32_to_cpu(sb.data_clump_size));
	term_print("%20s%u\n", "next_cnid:", be32_to_cpu(sb.next_cnid));
	term_print("%20s%u\n", "write_count:", be32_to_cpu(sb.write_count));
	term_flush();

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	u32 mapstart;

	if (fs_info(fd) == -1) {
		return -1;
	}

	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen);

	/**
	 * bs =  512 => start = 3 * 512
	 * bs = 1024 => start = 2 * 1024
	 * bs = 2048 => start = bs
	 * bs = 4096 => start = bs
	*/
	if (fsi->bsize == 512) {
		mapstart = 3 * 512;
	} else if (fsi->bsize == 1024) {
		mapstart = 2048;
	} else
		mapstart = fsi->bsize;
	pread_loop2(fd, fsi->bitmap, fsi->bitmaplen, mapstart);

	return 0;
}

/**
 * fs_postclone
 * - copy 3rd sector to the second-to-last sector
 */
static void fs_postclone(int fd, int fd_out, off_t offset)
{
	struct fsinfo *fsi = io_getcookie(fd);
	char *x = mem_alloc(512);
	u64 dest;

	/* read 3rd sector (hfs_sb) */
	pread_loop2(fd, x, 512, 2 * 512 + offset);

	/* write it */
	dest = (fsi->psize + 511) / 512;
	dest = (dest * 512) - 3 * 512;
	pwrite_loop2(fd_out, x, 512, dest + offset);

	mem_free(x);

	return;
}

struct file_system hfsp_fs = {
	.id = FSID_HFSP,
	.module = fs_nameP,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_postclone = fs_postclone,
	.fs_debug = fs_debug
};
