
/**
 * Copyright © 1993 Chris Smith
 * Copyright © 1998 - 1999 Mikulas Patocka
 * Copyright © 2012, 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * based on information from linux kernel and this site:
 * http://www.nondot.org/sabre/os/files/FileSystems/HPFS/
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"

/**
 * High Performance File System (HPFS)
 *
 * systems: OS/2, Windows NT, Windows 2000
 * endian:  little
 *
 * tested on:
 * OS/2 Warp 4.0 (qemu img) /TR 2012-04-02
 */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

/* LSN 0 */
struct hpfs_bb {
	u8 jmp[3];
	u8 oem_id[8];		/* OS2 20.0 */
	u16 bytes_per_sector;	/* 512 */
	u8 sectors_per_cluster;
	u8 n_reserved_sectors[2];
	u8 n_fats;
	u8 n_rootdir_entries[2];
	u8 n_sectors_s[2];
	u8 media_byte;
	u16 sectors_per_fat;
	u16 sectors_per_track;
	u16 heads_per_cyl;
	u32 n_hidden_sectors;
	u32 n_sectors_l;	/* size of partition */
	u8 drive_number;
	u8 mbz;
	u8 sig_28h;		/* 28h */
	u8 vol_serno[4];
	u8 vol_label[11];
	const char sig_hpfs[8];	/* "HPFS    " */
	u8 pad[448];		/* padding */
	u16 magic;		/* aa55 */
} __FNG_ATTR_PACKED;

#define SB_MAGIC  0xf995e849
#define SB_MAGIC2 0xfa53e9c5

/* LSN 16 */
struct hpfs_sb {
	u32 magic;		/* 0xf995e849 */
	u32 magic2;		/* 0xfa53e9c5 */
	u8 version;		/* version of a filesystem usually 2 */
	u8 funcversion;		/* functional version */
	u16 zero;		/* 0 */
	u32 root;		/* fnode of root directory */
	u32 n_sectors;		/* size of filesystem */
	u32 n_badblocks;	/* number of bad blocks */
	u32 bitmaps;		/* pointers to free space bit maps */
	u32 zero1;		/* 0 */
	u32 badblocks;		/* bad block list */
	u32 zero3;		/* 0 */
	u32 last_chkdsk;	/* date last checked, 0 if never */
	u32 last_optimize;	/* date last optimized, 0 if never */
	u32 n_dir_band;		/* number of sectors in dir band */
	u32 dir_band_start;	/* first sector in dir band */
	u32 dir_band_end;	/* last sector in dir band */
	u32 dir_band_bitmap;	/* free space map, 1 dnode per bit */
	u8 volume_name[32];	/* not used */
	u32 user_id_table;	/* 8 preallocated sectors - user id */
	u32 zero6[103];		/* 0 */
} __FNG_ATTR_PACKED;

#define SP_MAGIC  0xf9911849
#define SP_MAGIC2 0xfa5229c5

/* LSN 17 */
struct hpfs_sp {
	u32 magic;		/* 0xf9911849 */
	u32 magic2;		/* 0xfa5229c5 */
};

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global hpfs variables */
static struct hpfs_sb sb;
static int fs_getbitmap(int fd);

static char fs_name[] = "hpfs";

static void die_hpfs(char *s)
{
	die_msg(111, s);
}

/**
 * fs_test - check if hpfs filesystem
 *
 * @return 0 - no hpfs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	struct hpfs_bb *bb = mem_talloc(struct hpfs_bb, 1);
	struct hpfs_sp sp;
	ssize_t r;

	/* check boot block */
	r = pread_loop(fd, bb, sizeof(struct hpfs_bb), 0);
	if (r == -1)
		goto no_hpfs;

	if (le16_to_cpu(bb->magic) != 0xaa55)
		goto no_hpfs;

	if (bb->sig_28h != 0x28)
		goto no_hpfs;

	/* LSN 16 - read super block */
	r = pread_loop(fd, &sb, sizeof(struct hpfs_sb), 16 * 512);
	if (r == -1)
		goto no_hpfs;

	if (le32_to_cpu(sb.magic) != SB_MAGIC)
		goto no_hpfs;

	if (le32_to_cpu(sb.magic2) != SB_MAGIC2)
		goto no_hpfs;

	/* LSN 17 - read spare block */
	r = pread_loop(fd, &sp, sizeof(struct hpfs_sp), 17 * 512);
	if (r == -1)
		goto no_hpfs;

	if (le32_to_cpu(sp.magic) != SP_MAGIC)
		goto no_hpfs;

	if (le32_to_cpu(sp.magic2) != SP_MAGIC2)
		goto no_hpfs;

	/* back to bb extra checks */
	if (!str_equal(bb->sig_hpfs, "HPFS    "))
		goto no_hpfs;

	if (le16_to_cpu(bb->bytes_per_sector) != 512)
		die_hpfs("bsize != 512 @ HPFS?");

	/* this should be hpfs */
	mem_free(bb);
	return 1;

 no_hpfs:
	/* no hpfs */
	mem_free(bb);
	return 0;
}

/**
 * fs_info -  gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_getbitmap(fd) == -1) {
		return -1;
	}

	/* not needed @ fs_info() */
	free(fsi->bitmap);
	fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_getbitmap(fd) == -1) {
		return -1;
	}

	free(fsi->bitmap);
	fsi->bitmap = 0;

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s%s\n", "volume name:", sb.volume_name);
	term_print("%20s0x%x\n", "magic:", le32_to_cpu(sb.magic));
	term_print("%20s0x%x\n", "magic2:", le32_to_cpu(sb.magic2));
	term_print("%20s%u\n", "version:", (u32) sb.version);

	term_print("%20s%u\n", "fs block size:", 512);
	term_print("%20s%u\n", "fs # of blocks:", le32_to_cpu(sb.n_sectors));
	term_print("%20s%u\n", "fs # used blocks:", fsi->bused);
	term_print("%20s%u\n\n", "fs # free blocks:", fsi->bfree);

	term_print("%20s%u\n", "last_chkdsk:", le32_to_cpu(sb.last_chkdsk));
	term_print("%20s%u\n", "last_optimize:", le32_to_cpu(sb.last_optimize));
	term_print("%20s%u\n", "n_dir_band:", le32_to_cpu(sb.n_dir_band));
	term_print("%20s%u\n", "dir_band_start:",
		   le32_to_cpu(sb.dir_band_start));
	term_print("%20s%u\n", "dir_band_end:", le32_to_cpu(sb.dir_band_end));
	term_print("%20s%u\n", "dir_band_bitmap:",
		   le32_to_cpu(sb.dir_band_bitmap));
	term_printflush("\n");

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	u32 *bitmaps, nmaps, m;
	u64 bit;
	void *x;

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* set all known infos from super block */
	fsi->name = fs_name;
	fsi->bsize = 512;
	fsi->blocks = le32_to_cpu(sb.n_sectors);
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bused = 0;
	fsi->bfree = 0;

	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + 1);

	/* number of bitmaps */
	nmaps = (le32_to_cpu(sb.n_sectors) + 0x3fff) >> 14;
	x = mem_talloc(u32, nmaps);

	/* read bitmap ptr list */
	pread_loop2(fd, x, nmaps * sizeof(u32),
		    le32_to_cpu(sb.bitmaps) * fsi->bsize);
	bitmaps = x;

	/* bitmap buffer */
	x = mem_alloc(2048);

	/* read each 2K bitmap (16K blocks per map) */
	for (bit = 0, m = 0; m < nmaps; m++) {
		off_t off = bitmaps[m] * fsi->bsize;
		int i;

		/* read bitmap, fill our real one */
		pread_loop2(fd, x, 2048, off);
		for (i = 0; i < 16 * 1024; i++) {
			if (bitmap_get(x, i)) {
				bitmap_set(fsi->bitmap, bit, 0);
				fsi->bfree++;
			} else {
				bitmap_set(fsi->bitmap, bit, 1);
				fsi->bused++;
			}

			bit++;
			if (bit == le32_to_cpu(sb.n_sectors))
				break;
		}
	}
#if 0
	term_print("bused:         %u\n", fsi->bused);
	term_print("bfree:         %u\n", fsi->bfree);
	term_print("bused + bfree: %u\n", fsi->bused + fsi->bfree);
	term_print("gesamt:        %u\n", le32_to_cpu(sb.n_sectors));
	term_flush();
#endif

	mem_free(bitmaps);
	mem_free(x);

	return 0;
};

struct file_system hpfs_fs = {
	.id = FSID_HPFS,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
