
/**
 * Copyright © 2000 - 2004 International Business Machines Corp.
 * Copyright © 2012 - 2013 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "io.h"
#include "byte.h"
#include "uint32.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/bitmap.h"
#include "src/strallocX.h"
#include "src/die.h"
#include "src/mem.h"
#include "src/fs.h"
#include "src/bswap.h"

/**
 * Journaled File System (JFS2)
 *
 * systems: AIX, OS/2, Linux, eComStation
 * endian:  little
 *
 * JFS1 is not supported, I don't have AIX ;)
 * @IBM: send me an cd, I will support it then...
 */
static char fs_name[] = "jfs";

#undef JFS_DEBUG

#define SB_MAGIC_JFS	0x3153464a	/* 'JFS1' => little endian */
#define SB_LOGMAGIC	0x87654321	/* magic of the first log page */

/* important definitions */
#define PSIZE		4096	/* the magic constant @ JFS :) */
#define L2LPERCTL	10	/* l2 num of leaves per dmapctl tree */
#define L2BPERDMAP	13	/* l2 num of blks per dmap */
#define LPERCTL		1024	/* num of leaves per dmapctl tree */
#define BPERDMAP	8192	/* num blocks per dmap @ pmap */

/* important offsets */
#define SUPER1_OFF	 8 * PSIZE	/*  4 Primary Aggregate Superblock */
#define AIMAP_OFF	10 * PSIZE	/*  4 Aggregate Inode Allocation Map */
#define SUPER2_OFF	15 * PSIZE	/*  4 Primary Aggregate Superblock */
#define BMAP_OFF	16 * PSIZE	/* 16 Block Allocation Bitmap */
#define DMAP_OFF	20 * PSIZE	/* xx Allocation Bitmap */

/* maximum l2 number of disk blocks at the various dmapctl levels */
#define L2MAXL0SIZE	(L2BPERDMAP + 1 * L2LPERCTL)
#define L2MAXL1SIZE	(L2BPERDMAP + 2 * L2LPERCTL)
#define L2MAXL2SIZE	(L2BPERDMAP + 3 * L2LPERCTL)

/* maximum number of disk blocks at the various dmapctl levels */
#define MAXL0SIZE	((s64)1 << L2MAXL0SIZE)
#define MAXL1SIZE	((s64)1 << L2MAXL1SIZE)
#define MAXL2SIZE	((s64)1 << L2MAXL2SIZE)
#define MAXMAPSIZE	MAXL2SIZE	/* maximum aggregate map size */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif
struct jfslog_sb {
	u32 magic;		/* 4: log lv identifier */
	s32 version;		/* 4: version number */
	s32 serial;		/* 4: log open/mount counter */
	s32 size;		/* 4: size in number of PSIZE blocks */
	s32 bsize;		/* 4: logical block size in byte */
	s32 l2bsize;		/* 4: log2 of bsize */
	u32 flag;		/* 4: option */
	u32 state;		/* 4: state - see below */
	s32 end;		/* 4: addr of last log record set by logredo */
	u8 uuid[16];		/* 16: 128-bit journal uuid */
	char label[16];		/* 16: journal label */
	/* ... */
} __FNG_ATTR_PACKED;

/* Almost identical to Linux's timespec, but not quite */
struct timestruc_t {
	u32 tv_sec;
	u32 tv_nsec;
};

/* physical xd (pxd) */
typedef struct {
	unsigned int len:24;
	unsigned int addr1:8;
	u32 addr2;
} pxd_t;

/* xd_t field extraction */
#define	lengthPXD(pxd)	le24_to_cpu((pxd)->len)
#define	addressPXD(pxd) ( ((s64)((pxd)->addr1)) << 32 | le32_to_cpu((pxd)->addr2))

/**
 * aggregate superblock
 */
struct jfs_sb {
	u32 s_magic;		/* 4: magic number */
	u32 s_version;		/* 4: version number */

	s64 s_size;		/* 8: aggregate size in hardware/LVM blocks */
	s32 s_bsize;		/* 4: aggregate block size in bytes */
	s16 s_l2bsize;		/* 2: log2 of s_bsize */
	s16 s_l2bfactor;	/* 2: log2(s_bsize/hardware block size) */
	s32 s_pbsize;		/* 4: hardware/LVM block size in bytes */
	s16 s_l2pbsize;		/* 2: log2 of s_pbsize */
	s16 pad;		/* 2: padding necessary for alignment */

	u32 s_agsize;		/* 4: allocation group size in aggr. blocks */
	u32 s_flag;		/* 4: aggregate attributes: see jfs_filsys.h */
	u32 s_state;		/* 4: mount/unmount/recovery state: see jfs_filsys.h */
	s32 s_compress;		/* 4: > 0 if data compression */
	pxd_t s_ait2;		/* 8: first extent of secondary aggregate inode table */
	pxd_t s_aim2;		/* 8: first extent of secondary aggregate inode map */
	u32 s_logdev;		/* 4: device address of log */
	s32 s_logserial;	/* 4: log serial number at aggregate mount */
	pxd_t s_logpxd;		/* 8: inline log extent */
	pxd_t s_fsckpxd;	/* 8: inline fsck work space extent */
	struct timestruc_t s_time;	/* 8: time last updated */

	s32 s_fsckloglen;	/* 4: Number of filesystem blocks reserved for
				 *    the fsck service log.
				 *    N.B. These blocks are divided among the
				 *         versions kept.  This is not a per
				 *         version size.
				 *    N.B. These blocks are included in the
				 *         length field of s_fsckpxd.
				 */

	s8 s_fscklog;		/* 1: which fsck service log is most recent
				 *    0 => no service log data yet
				 *    1 => the first one
				 *    2 => the 2nd one
				 */

	char s_fpack[11];	/* 11: file system volume name
				 *     N.B. This must be 11 bytes to
				 *          conform with the OS/2 BootSector
				 *          requirements
				 *          Only used when s_version is 1
				 */

	/* extendfs() parameter under s_state & FM_EXTENDFS */
	s64 s_xsize;		/* 8: extendfs s_size */
	pxd_t s_xfsckpxd;	/* 8: extendfs fsckpxd */
	pxd_t s_xlogpxd;	/* 8: extendfs logpxd */
	/* - 128 byte boundary - */

	u8 s_uuid[16];		/* 16: 128-bit uuid for volume */
	char s_label[16];	/* 16: volume label */
	u8 s_loguuid[16];	/* 16: 128-bit uuid for log device */
} __FNG_ATTR_PACKED;

/**
 * bmap control page - on-disk aggregate disk allocation map descriptor
 * -> part of Block Allocation Bitmap
 */
#define	MAXAG		128	/* max number of allocation groups */
struct bmapctl {
	s64 dn_mapsize;		/* 8: number of blocks in aggregate  */
	s64 dn_nfree;		/* 8: num free blks in aggregate map */
	s32 dn_l2nbperpage;	/* 4: number of blks per page        */
	s32 dn_numag;		/* 4: total number of ags            */
	s32 dn_maxlevel;	/* 4: number of active ags           */
	s32 dn_maxag;		/* 4: max active alloc group number  */
	s32 dn_agpref;		/* 4: preferred alloc group (hint)   */
	s32 dn_aglevel;		/* 4: dmapctl level holding the AG   */
	s32 dn_agheigth;	/* 4: height in dmapctl of the AG    */
	s32 dn_agwidth;		/* 4: width in dmapctl of the AG     */
	s32 dn_agstart;		/* 4: start tree index at AG height  */
	s32 dn_agl2size;	/* 4: l2 num of blks per alloc group */
	s64 dn_agfree[MAXAG];	/* 8*MAXAG: per AG free count        */
	s64 dn_agsize;		/* 8: num of blks per alloc group    */
	s8 dn_maxfreebud;	/* 1: max free buddy system          */
	u8 pad[3007];		/* 3007: pad to 4096                 */
} __FNG_ATTR_PACKED;		/* - 4096 -                          */

/**
 * dmap control page - disk map control page per level
 * -> part of Block Allocation Bitmap
 */
#define CTLTREESIZE (1024+256+64+16+4+1)	/* size of a dmapctl tree */
struct dmapctl {
	s32 nleafs;		/* 4: number of tree leafs         */
	s32 l2nleafs;		/* 4: l2 number of tree leafs      */
	s32 leafidx;		/* 4: index of the first tree leaf */
	s32 height;		/* 4: height of tree               */
	s8 budmin;		/* 1: minimum l2 tree leaf value   */
	s8 stree[CTLTREESIZE];	/* CTLTREESIZE: dmapctl tree   */
	u8 pad[2714];		/* 2714: pad to 4096               */
} __FNG_ATTR_PACKED;		/* - 4096 -                        */

/**
 * dmap summary tree
 *
 * struct dmaptree must be consistent with struct dmapctl.
 */
#define TREESIZE (256+64+16+4+1)	/* size of a dmap tree (341) */
struct dmaptree {
	s32 nleafs;		/* 4: number of tree leafs      */
	s32 l2nleafs;		/* 4: l2 number of tree leafs   */
	s32 leafidx;		/* 4: index of first tree leaf  */
	s32 height;		/* 4: height of the tree        */
	s8 budmin;		/* 1: min l2 tree leaf value to combine */
	s8 stree[TREESIZE];	/* TREESIZE: tree               */
	u8 pad[2];		/* 2: pad to word boundary      */
} __FNG_ATTR_PACKED;		/* - 360 -                      */

/**
 * dmap page per 8K blocks bitmap
 * -> part of Block Allocation Bitmap
 */
struct dmap {			/* start| len */
	s32 nblocks;		/* 0000 | 4: num blks covered by this dmap */
	s32 nfree;		/* 0004 | 4: num of free blks in this dmap */
	s64 start;		/* 0008 | 8: starting blkno for this dmap  */
	struct dmaptree tree;	/* 0016 | 360: dmap tree                   */
	u8 pad[1672];		/* 0376 | 1672: pad to 2048 bytes          */
	u32 wmap[256];		/* 2048 | 1024: bits of the working map    */
	u32 pmap[256];		/* 3072 | 1024: bits of the persistent map */
	/* stored also in LE32 */
} __FNG_ATTR_PACKED;		/* - 4096 - */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global jfs variables */
struct jfs_sb *sb;
struct bmapctl *bm;
struct jfslog_sb *lp;

/* we will allocate the whole structure one time and leave them in mem */
static int jfs_init_done = 0;

/**
 * @x: value to rotate
 */
static inline u32 rotate32(u32 x)
{
	register u32 n = x;

	n = ((n >> 1) & 0x55555555) | ((n << 1) & 0xaaaaaaaa);
	n = ((n >> 2) & 0x33333333) | ((n << 2) & 0xcccccccc);
	n = ((n >> 4) & 0x0f0f0f0f) | ((n << 4) & 0xf0f0f0f0);
	n = ((n >> 8) & 0x00ff00ff) | ((n << 8) & 0xff00ff00);
	n = ((n >> 16) & 0x0000ffff) | ((n << 16) & 0xffff0000);

	return n;
}

/**
 * jfs_showflags - show some value, or not
 */
static void jfs_showflags(stralloc * sa, u32 flags, u32 val, char *str)
{
	if (is_set(flags, val)) {
		if (!stralloc_cats(sa, str))
			die_nomem();
		if (!stralloc_cats(sa, " "))
			die_nomem();
	}
}

/**
 * fs_test - check if it is a jfs filesysem
 *
 * @return 0 - no jfs fs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	ssize_t r;

	if (jfs_init_done == 0) {
		sb = mem_tzalloc(struct jfs_sb, 1);
		bm = mem_tzalloc(struct bmapctl, 1);
		lp = mem_tzalloc(struct jfslog_sb, 1);

		jfs_init_done = 1;
	}

	r = pread_loop(fd, sb, sizeof(struct jfs_sb), SUPER2_OFF);
	if (r != sizeof(struct jfs_sb))
		goto fail;

	if (le32_to_cpu(sb->s_magic) != SB_MAGIC_JFS)
		goto fail;

	r = pread_loop(fd, sb, sizeof(struct jfs_sb), SUPER1_OFF);
	if (r != sizeof(struct jfs_sb))
		goto fail;

	if (le32_to_cpu(sb->s_magic) != SB_MAGIC_JFS)
		goto fail;

	r = pread_loop(fd, lp, sizeof(struct jfslog_sb),
		       (addressPXD(&sb->s_logpxd) + 1) * PSIZE);
	if (r != sizeof(struct jfslog_sb))
		goto fail;

	if (le32_to_cpu(lp->magic) != SB_LOGMAGIC)
		goto fail;

#if 0
	/* @ linux only 4K is ok, but we could handle any valid value */
	if (le32_to_cpu(sb->s_bsize) != PSIZE)
		goto fail;
#endif

	/* I don't know, if that changes on other systems... */
	if (le32_to_cpu(sb->s_pbsize) != 512)
		goto fail;

	/* last test, the count of total block in fs */
	r = pread_loop(fd, bm, sizeof(struct bmapctl), BMAP_OFF);
	if (r != sizeof(struct bmapctl))
		goto fail;

	/**
	 * validate bmctl / superblock
	 */
	if ((le64_to_cpu(sb->s_size) * 512) / le32_to_cpu(sb->s_bsize)
	    != le64_to_cpu(bm->dn_mapsize)) {
		goto fail;
	}

	/* it seems to be jfs */
	return 1;

 fail:
	jfs_init_done = 0;
	free(sb);
	free(bm);
	free(lp);

	return 0;
}

/**
 * fs_info -  gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* fill in the data */
	fsi->bsize = le32_to_cpu(sb->s_bsize);

	/* out info here is different from statfs() */
	fsi->blocks = fsi->psize / fsi->bsize;
	fsi->bfree = le64_to_cpu(bm->dn_nfree);
	fsi->bused = fsi->blocks - fsi->bfree;
	fsi->fssize = fsi->bsize * fsi->blocks;
	fsi->bitmap = 0;

	/* whole partition / blocksize -> total blocks */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->name = fs_name;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	struct dmapctl dm;
	stralloc sa;
	u32 u;

	s64 npages;		/* total pages */
	s64 ndmaps;		/* dmap's */
	s64 nl0pages;		/* l0 dmap ctl */
	s64 nl1pages;		/* l1 dmap ctl */
	s64 nl2pages;		/* l2 dmap ctl */

	if (fs_info(fd) == -1) {
		return -1;
	}

	stralloc_init(&sa);
	term_print("\n\n");

	term_print("%22s%s\n", "filesystem type:", fs_name);
	term_print("%22s%U\n", "fs psize / bsize:", fsi->psize / fsi->bsize);
	/* should be the same as: fsi->psize / fsi->bsize */
	term_print("%22s%U\n", "dused + log + fschk:",
		   le64_to_cpu(bm->dn_mapsize)
		   + lengthPXD(&sb->s_logpxd) + lengthPXD(&sb->s_fsckpxd));
	term_print("%22s%U\n", "fs block size:", le32_to_cpu(sb->s_bsize));
	term_print("%22s%U\n", "sector size:", le32_to_cpu(sb->s_pbsize));

	term_print("%22s%U\n\n", "bitmap length:", fsi->bitmaplen);

	term_print("%22s%U\n", "fs # of data blocks:",
		   le64_to_cpu(bm->dn_mapsize));
	term_print("%22s%U\n", "fs # used dblocks:",
		   le64_to_cpu(bm->dn_mapsize) - le64_to_cpu(bm->dn_nfree));
	term_print("%22s%U\n\n", "fs # free dblocks:",
		   le64_to_cpu(bm->dn_nfree));

	term_print("%22s0x%x\n", "fs magic:", le32_to_cpu(sb->s_magic));
	term_print("%22s%U\n", "fs version:", le32_to_cpu(sb->s_version));
	term_print("%22s%U\n", "ag size:", le32_to_cpu(sb->s_agsize));
	term_print("%22s%U\n\n", "num ag's:", le32_to_cpu(bm->dn_numag));

	term_print("%22s0x%x\n", "log magic:", le32_to_cpu(lp->magic));
	term_print("%22s%U\n", "log start:", addressPXD(&sb->s_logpxd));
	term_print("%22s%U\n", "log length:", lengthPXD(&sb->s_logpxd));

	term_print("%22s%U\n", "fsck space start:", addressPXD(&sb->s_fsckpxd));
	term_print("%22s%U\n", "fsck space length:", lengthPXD(&sb->s_fsckpxd));

	term_print("%22s%U\n", "ait2 start:", addressPXD(&sb->s_ait2));
	term_print("%22s%U\n", "ait2 length:", lengthPXD(&sb->s_ait2));

	term_print("%22s%U\n", "aim2 start:", addressPXD(&sb->s_aim2));
	term_print("%22s%U\n", "aim2 length:", lengthPXD(&sb->s_aim2));

/* superblock flags */
#define JFS_AIX			0x80000000	/* AIX support */
#define JFS_OS2			0x40000000	/* OS/2 support */
#define JFS_DFS			0x20000000	/* DCE DFS LFS support */
#define JFS_LINUX      		0x10000000	/* Linux support */
#define JFS_UNICODE		0x00000001	/* unicode name */
#define JFS_GROUPCOMMIT		0x00000100	/* group (of 1) commit */
#define JFS_LAZYCOMMIT		0x00000200	/* lazy commit */
#define JFS_TMPFS		0x00000400	/* not log/commit */
#define JFS_INLINELOG		0x00000800	/* inline log within file system */
#define JFS_INLINEMOVE		0x00001000	/* inline log being moved */
#define JFS_BAD_SAIT		0x00010000	/* current secondary ait is bad */
#define JFS_SPARSE		0x00020000	/* sparse regular file */
#define JFS_DASD_ENABLED	0x00040000	/* DASD limits enabled */
#define JFS_DASD_PRIME		0x00080000	/* Prime DASD usage on boot */
#define JFS_DIR_INDEX		0x00200000	/* Persistant index for */

	u = le32_to_cpu(sb->s_flag);
	if (!stralloc_copys(&sa, ""))
		die_nomem();
	jfs_showflags(&sa, u, JFS_AIX, "aix");
	jfs_showflags(&sa, u, JFS_OS2, "os/2");
	jfs_showflags(&sa, u, JFS_DFS, "dfs");
	jfs_showflags(&sa, u, JFS_LINUX, "linux");
	jfs_showflags(&sa, u, JFS_UNICODE, "unicode");
	jfs_showflags(&sa, u, JFS_GROUPCOMMIT, "groupcommit");
	jfs_showflags(&sa, u, JFS_TMPFS, "tmpfs");
	jfs_showflags(&sa, u, JFS_INLINELOG, "inlinelog");
	jfs_showflags(&sa, u, JFS_INLINEMOVE, "inlinemove");
	jfs_showflags(&sa, u, JFS_BAD_SAIT, "bad_sait");
	jfs_showflags(&sa, u, JFS_SPARSE, "sparse");
	jfs_showflags(&sa, u, JFS_DASD_ENABLED, "dasd_enabled");
	jfs_showflags(&sa, u, JFS_DASD_PRIME, "dasd_prime");
	jfs_showflags(&sa, u, JFS_DIR_INDEX, "dir_index");
	term_print("%22s%S\n", "s_flags:", &sa);

/* file system state (superblock state) */
#define FM_CLEAN	0x00000000	/* is unmounted and clean */
#define FM_MOUNT	0x00000001	/* is mounted cleanly */
#define FM_DIRTY	0x00000002	/* was not unmounted and clean */
#define FM_LOGREDO	0x00000004	/* log based recovery failed */
#define FM_EXTENDFS	0x00000008	/* extendfs() in progress */

	u = le32_to_cpu(sb->s_state);
	if (u == FM_CLEAN) {
		if (!stralloc_copys(&sa, "CLEAN"))
			die_nomem();
	} else {
		if (!stralloc_copys(&sa, ""))
			die_nomem();
	}
	jfs_showflags(&sa, u, FM_MOUNT, "mounted");
	jfs_showflags(&sa, u, FM_DIRTY, "dirty");
	jfs_showflags(&sa, u, FM_LOGREDO, "logredo (run fsck!)");
	jfs_showflags(&sa, u, FM_EXTENDFS, "extendfs");
	term_print("%22s%S\n\n", "s_state:", &sa);

	/**
	 * compute the number dmap pages required to cover number_of_blocks;
	 * add one extra dmap page for extendfs(): this is added before
	 * we figure out how many control pages are needed, so we get
	 * the correct number of control pages.
	 */
	npages = ndmaps = ((le64_to_cpu(bm->dn_mapsize) + BPERDMAP - 1)
			   >> L2BPERDMAP) + 1;

	/**
	 * compute number of (logical) control pages at each level of the map
	 */

	/* account for L0 pages to cover dmap pages */
	nl0pages = (ndmaps + LPERCTL - 1) >> L2LPERCTL;
	npages += nl0pages;

	if (nl0pages > 1) {
		/* account for one L2 and actual L1 pages to cover L0 pages */
		nl1pages = (nl0pages + LPERCTL - 1) >> L2LPERCTL;
		nl2pages = 1;	// @ start
	} else {
		/* account for one logical L2 and one logical L1 pages */
		nl1pages = 1;	// @ start
		nl2pages = 1;	// @ start
	}
	npages += nl1pages;
	npages += nl2pages;

	/* account for global control page of map */
	npages++;

	term_print("%22s%D\n", "npages:", npages);
	term_print("%22s%D\n", "ndmaps:", ndmaps);
	term_print("%22s%D\n", "bmapctl:", 1);
	term_print("%22s%D\n", "nl2pages:", nl2pages);
	term_print("%22s%D\n", "nl1pages:", nl1pages);
	term_print("%22s%D\n", "nl0pages:", nl0pages);

	/* L2 Page / XXX: will not work for bsize != PSIZE!!! */
	pread_loop2(fd, &dm, sizeof(struct dmapctl), BMAP_OFF + 1 * PSIZE);
	if (dm.budmin == 33) {
		term_print("\n%s\n", "L2 Page:");
		term_print("%22s%u\n", "l2nleafs:", dm.l2nleafs);
		term_print("%22s%u\n", "leafidx:", dm.leafidx);
		term_print("%22s%u\n", "height:", dm.height);
		term_print("%22s%u\n", "budmin:", dm.budmin);
	}

	/* L1 Page */
	pread_loop2(fd, &dm, sizeof(struct dmapctl), BMAP_OFF + 2 * PSIZE);
	if (dm.budmin == 23) {
		term_print("\n%s\n", "1st L1 Page:");
		term_print("%22s%u\n", "nleafs:", dm.nleafs);
		term_print("%22s%u\n", "l2nleafs:", dm.l2nleafs);
		term_print("%22s%u\n", "leafidx:", dm.leafidx);
		term_print("%22s%u\n", "height:", dm.height);
		term_print("%22s%u\n", "budmin:", dm.budmin);
	}

	/* L0 Page */
	pread_loop2(fd, &dm, sizeof(struct dmapctl), BMAP_OFF + 3 * PSIZE);
	term_print("\n%s\n", "1st L0 Page:");
	term_print("%22s%u\n", "nleafs:", dm.nleafs);
	term_print("%22s%u\n", "l2nleafs:", dm.l2nleafs);
	term_print("%22s%u\n", "leafidx:", dm.leafidx);
	term_print("%22s%u\n", "height:", dm.height);
	term_print("%22s%u\n", "budmin:", dm.budmin);

	term_printflush("\n");
	stralloc_free(&sa);

	return 0;
}

/**
 * +--------+----+------------------------------------------------+
 * |  0..32 | 32 | Reserved                                       |
 * +--------+----+------------------------------------------------+
 * | 32..36 |  4 | Primary Aggregate Superblock                   |
 * +--------+----+------------------------------------------------+
 * | 36..40 |  4 | Control Page                                   |
 * +--------+----+------------------------------------------------+
 * | 40..44 |  4 | Primary Aggregate Inode Allocation Map (IAG)   |
 * +--------+----+------------------------------------------------+
 * | 44..60 | 16 | Primary Aggregate Inode Table                  |
 * |     44 |  4 | dinomap_t                                      |
 * |     48 |  4 |                                                |
 * |     52 |  4 |                                                |
 * |     56 |  4 |                                                |
 * +--------+----+------------------------------------------------+
 * | 60..64 |  4 | Secondary Aggregate Superblock                 |
 * +--------+----+------------------------------------------------+
 * | 64..80 | 16 | Block Allocation Bitmap                        |
 * |     64 |  4 | bmapctl ctl                                    |
 * |     68 |  4 | dmapctl l2                                     |
 * |     72 |  4 | dmapctl l1                                     |
 * |     76 |  4 | dmapctl l0                                     |
 * |     .. | xx | n dmap entries + l0 + l1                       |
 * +--------+----+------------------------------------------------+
 * |        | 16 | Secondary Aggregate Inode Table                |
 * +--------+----+------------------------------------------------+
 * |        |  4 | Control Page                                   |
 * +--------+----+------------------------------------------------+
 * |        |  4 | Secondary Aggregate Inode Allocation Map (IAG) |
 * +--------+----+------------------------------------------------+
 * | DATA   (via dmap bitmaps)                                    |
 * +--------------------------------------------------------------+
 * | space for fsck (start + length is known)                     |
 * +--------------------------------------------------------------+
 * | log (start + length is known)                                |
 * +--------------------------------------------------------------+
 */

/**
 * getbitmap - read u32 values from pmap to our bitmap
 *
 * blocks @ empty 2TiB jfs filesystem:
 * 536870912(total) = 536821710(data) + 16434(fsck) + 32768(log)
 * used total: 114828 = 65626(data) + 32768(log) + 16434(fsck)
 * free total: 536756084(free)
 * dmaps: 65531 = 65531 * 8192 => 536829952 (>536821710) blöcke
 * 536870912 blöcke -> 67.108.864 byte für den bitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	s64 p, dmaps, ctlpages, ndmaps;
	char buf[PSIZE + 1];
	void *map;
	u64 nblocks = 0;

#ifdef JFS_DEBUG
	u64 nfree = 0;
#endif

	if (fs_info(fd) == -1) {
		return -1;
	}

	/* need a bit more, endian stuff */
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + 64);

	/* count the dmaps */
	ndmaps =
	    ((le64_to_cpu(bm->dn_mapsize) + BPERDMAP - 1) >> L2BPERDMAP) + 1;

	/* I don't want to much logic stuff here... we know how many dmaps
	 * must be read and skip all others... until all bits are found ;) */
	dmaps = 0;		/* dmap pages */
	ctlpages = 0;		/* control pages */
	for (p = 0, ctlpages = 0; dmaps < ndmaps - 1; p++) {
		u32 count, i;
		struct dmapctl *dmctl;
		struct dmap *dm;

		pread_loop2(fd, buf, PSIZE, DMAP_OFF + p * PSIZE);
		dmctl = (struct dmapctl *)buf;

		/* skip control pages */
		if (le32_to_cpu(dmctl->budmin) == 13)
			if (le32_to_cpu(dmctl->nleafs) == LPERCTL) {
				ctlpages++;
				continue;
			}

		if (le32_to_cpu(dmctl->budmin) == 23)
			if (le32_to_cpu(dmctl->nleafs) == LPERCTL) {
				ctlpages++;
				continue;
			}

		/* now, we have for sure an dmap page */
		dmaps++;
		dm = (struct dmap *)buf;
		nblocks += le32_to_cpu(dm->nblocks);
#ifdef JFS_DEBUG
		nfree += le32_to_cpu(dm->nfree);
		if (dm->nblocks != 8192)
			term_print("last one: p:%u ctl:%u dmap:%u ndmaps:%u\n",
				   p, ctlpages, dmaps, ndmaps);
#endif

		/* copy and transform the pmap of the dmap pages */
		count = dm->nblocks / 8;	/* 8192 bits -> 1024 bytes */
		count += 4 - count % 4;	/* padding */
		count /= 4;	/* 1024 bytes -> 256 x u32 */

		/* for each 32bit integer, rotate the bits */
		for (i = 0; i < count; i++) {
			dm->pmap[i] = rotate32(le32_to_cpu(dm->pmap[i]));
		}

		/* calculate length for memcpy */
		count = le32_to_cpu(dm->nblocks) / 8;
		map = fsi->bitmap + (le64_to_cpu(dm->start) / 8);
		memcpy(map, dm->pmap, count);
	}

	if (nblocks != (u64) le64_to_cpu(bm->dn_mapsize)) {
		die_msg(111,
			"Calculated data blocks seem to be wrong, mail me!");
	}

	/* mark log and fsck space as used */
	bitmap_set_bits(fsi->bitmap, addressPXD(&sb->s_logpxd),
			lengthPXD(&sb->s_logpxd));
	bitmap_set_bits(fsi->bitmap, addressPXD(&sb->s_fsckpxd),
			lengthPXD(&sb->s_fsckpxd));

#ifdef JFS_DEBUG
	term_print("%22s%U\n", "fs psize / bsize:", fsi->psize / fsi->bsize);
	term_print("%22s%U\n", "dmaps:", p);
	term_print("%22s%U\n", "nblocks:", nblocks);
	term_print("%22s%U\n", "dmaps*8K:", p * 8192);
	term_print("%22s%U\n", "dmaps*8K - mapsize:",
		   p * 8192 - le64_to_cpu(bm->dn_mapsize));
	term_print("%22s%D\n", "nblock - mapsize:",
		   (s64) (nblocks - le64_to_cpu(bm->dn_mapsize)));

	term_print("%22s%U\n", "  mapsize", le64_to_cpu(bm->dn_mapsize));
	term_print("%22s%U\n", "+ log", lengthPXD(&sb->s_logpxd));
	term_print("%22s%U\n", "+ fschk", lengthPXD(&sb->s_fsckpxd));
	term_print("%22s%U\n", "= mapsize + log + fschk:",
		   le64_to_cpu(bm->dn_mapsize)
		   + lengthPXD(&sb->s_logpxd) + lengthPXD(&sb->s_fsckpxd));

	/* debugging: */
	term_print
	    ("dmaps:    free: %u used: %u nblocks: %u [%u = %u = %u + %u]\n",
	     nfree, nblocks - nfree, nblocks, ndmaps, p, ctlpages, dmaps);
	term_print("dmaps:    fsi->blocks: %u log: %u fsck: %u tused:%u\n",
		   (u32) fsi->blocks, (u32) lengthPXD(&sb->s_logpxd),
		   (u32) lengthPXD(&sb->s_fsckpxd),
		   (u32) (lengthPXD(&sb->s_logpxd) + lengthPXD(&sb->s_fsckpxd) +
			  nblocks - nfree));
	term_flush();
#endif

	return 0;
}

struct file_system jfs_fs = {
	.id = FSID_JFS,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
