
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"

/**
 * Minix file system
 *
 * systems: MINIX 1-3, Linux, HelenOS
 * endian:  little (sometimes big endian?)
 */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

/**
 * minix super-block data on disk
 */
struct minix_sb {
	u16 s_ninodes;		/* # usable inodes on the minor device */
	u16 s_nzones;		/* number of data zones @ V2 */
	u16 s_imap_blocks;	/* # of blocks used by inode bit map */
	u16 s_zmap_blocks;	/* # of blocks used by zone bit map */
	u16 s_firstdatazone;	/* number of first data zone (small) */
	u16 s_log_zone_size;	/* size of a data zone (1024 << s_log_zone_size) */
	u32 s_max_size;		/* maximum file size in bytes */
	u16 s_magic;		/* magic number to recognize super-blocks */
	u16 s_state;		/* mount state */
	u32 s_zones;		/* number of zones @ V1 */
} __FNG_ATTR_PACKED;

/**
 * V3 minix super-block data on disk
 */
struct minix3_sb {
	u32 s_ninodes;		/* # usable inodes on the minor device */
	u16 s_pad0;		/* total device size, including bit maps etc */
	u16 s_imap_blocks;	/* # of blocks used by inode bit map */
	u16 s_zmap_blocks;	/* # of blocks used by zone bit map */
	u16 s_firstdatazone;	/* number of first data zone (small) */
	u16 s_log_zone_size;	/* log2 of blocks/zone */
	u16 s_pad1;		/* try to avoid compiler-dependent padding */
	u32 s_max_size;		/* maximum file size on this device */
	u32 s_zones;		/* number of zones (replaces s_nzones in V2) */
	u16 s_magic;		/* magic number to recognize super-blocks */
	u16 s_pad2;		/* try to avoid compiler-dependent padding */
	u16 s_blocksize;	/* block size in bytes. */
	u8 s_disk_version;	/* filesystem format sub-version */
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

#define MINIX1_SUPER_MAGIC      0x137F	/* minix v1 fs, 14 char names */
#define MINIX1_SUPER_MAGIC2     0x138F	/* minix v1 fs, 30 char names */
#define MINIX2_SUPER_MAGIC      0x2468	/* minix v2 fs, 14 char names */
#define MINIX2_SUPER_MAGIC2     0x2478	/* minix v2 fs, 30 char names */
#define MINIX3_SUPER_MAGIC      0x4d5a	/* minix v3 fs, 60 char names */

/* global minixfs variables */
static struct minix_sb sb;
static struct minix3_sb sb3;
static int fs_endian;

#define MINIX_V1  1		/* minix v1, returned by fs_test() */
#define MINIX_V2  2		/* minix v2 */
#define MINIX_V3  3		/* minix v3 */
static int minix_type;

/**
 * fs_test_it - check for minix filesystem
 *
 * @return 0 - no minix
 * @return 1 - yes
 */
static int fs_test_it(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	ssize_t r;

	/* test for minix3 first */
	r = pread_loop(fd, &sb3, sizeof(struct minix3_sb), 1024);
	if (r == -1) {
		return 0;
	}

	if (fs16_to_cpu(fs_endian, sb3.s_magic) == MINIX3_SUPER_MAGIC) {
		fsi->name = "minix v3 (60 char names)";
		return MINIX_V3;
	}

	/* test for minix1 and minix2 */
	r = pread_loop(fd, &sb, sizeof(struct minix_sb), 1024);
	if (r == -1) {
		return 0;
	}

	if (fs16_to_cpu(fs_endian, sb.s_magic) == MINIX1_SUPER_MAGIC) {
		fsi->name = "minix v1 (14 char names)";
		return MINIX_V1;
	} else if (fs16_to_cpu(fs_endian, sb.s_magic) == MINIX1_SUPER_MAGIC2) {
		fsi->name = "minix v1 (30 char names)";
		return MINIX_V1;
	} else if (fs16_to_cpu(fs_endian, sb.s_magic) == MINIX2_SUPER_MAGIC) {
		fsi->name = "minix v2 (14 char names)";
		return MINIX_V2;
	} else if (fs16_to_cpu(fs_endian, sb.s_magic) == MINIX2_SUPER_MAGIC2) {
		fsi->name = "minix v2 (30 char names)";
		return MINIX_V2;
	}

	/* no minix found */
	return 0;
}

/**
 * fs_test - check for minix filesystem
 *
 * @return 0 - no minix
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	ssize_t r;

	/* try it with little endian */
	fs_endian = BYTESEX_LE;
	r = fs_test_it(fd);
	if (r > 0)
		return r;

	/* try it with big endian */
	fs_endian = BYTESEX_BE;
	r = fs_test_it(fd);
	if (r > 0)
		return r;

	/* no minix found */
	return 0;
}

/**
 * minix disk layout for 21 blocks:
 * +---------------------+-------------+------+
 * | 1                   | blank block | 1    | unused
 * +---------------------+-------------+------+
 * | 1                   | superblock  | 1    | system
 * +---------------------+-------------+------+
 * | # inodes / bs       | imap        | 1    | system
 * +---------------------+-------------+------+
 * | # blocks / bs       | zmap (bits) | 1    | system
 * +---------------------+-------------+------+
 * | 32 o. 16 x # inodes | inode table | 1    | system
 * +---------------------+-------------+------+
 * | datasize            | data        | 16   | 1x inode + 15x data
 * +---------------------+-------------+------+
 */
static int fs_getbitmap_v12(int fd, int type)
{
	struct fsinfo *fsi = io_getcookie(fd);
	void *map;
	u32 sysblks;
	u64 c;

	/* 1) blocksize */
	fsi->bsize = (1024 << fs16_to_cpu(fs_endian, sb.s_log_zone_size));

	/* 2) reserved fs blocks (1x empty + 1x sb + inode block + ...) */
	sysblks = fs16_to_cpu(fs_endian, sb.s_firstdatazone);
	if (type == MINIX_V1) {
		/* minix v1 */
		fsi->blocks = fs16_to_cpu(fs_endian, sb.s_nzones);
	} else {
		/* minix v2 */
		fsi->blocks = fs16_to_cpu(fs_endian, sb.s_zones);
	}

	/* 3) total # of blocks */
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen);
	map = mem_zalloc(fsi->bitmaplen);	/* temp */

	/* 4) mark system block used: */
	bitmap_set_bits(fsi->bitmap, 0, sysblks);
	fsi->bused = sysblks;
	fsi->bfree = 0;

	/* 5) read data bitmap */
	pread_loop2(fd, map, fsi->bitmaplen - sysblks / 8,
		    (2 + fs16_to_cpu(fs_endian, sb.s_imap_blocks)) * 1024);
	for (c = 0; c < fsi->blocks - sysblks; c++) {
		if (bitmap_get(map, c)) {
			bitmap_set(fsi->bitmap, c + sysblks, 1);
			fsi->bused++;
		} else {
			fsi->bfree++;
		}
	}

	mem_free(map);

	return 0;
}

/* nearly the same like v1 or v2 - but easier to debug, if extra */
static int fs_getbitmap_v3(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	void *map;
	u32 sysblks;
	u64 c;

	/* 1) blocksize */
	fsi->bsize = (1024 << fs16_to_cpu(fs_endian, sb3.s_log_zone_size));

	/* 2) reserved fs blocks (1x empty + 1x sb + inode block + ...) */
	sysblks = fs16_to_cpu(fs_endian, sb3.s_firstdatazone);
	fsi->blocks = fs32_to_cpu(fs_endian, sb3.s_zones);

	/* 3) total # of blocks */
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen);
	map = mem_zalloc(fsi->bitmaplen);	/* temp */

	/* 4) mark system block used: */
	bitmap_set_bits(fsi->bitmap, 0, sysblks);
	fsi->bused = sysblks;
	fsi->bfree = 0;

	term_print("gesamt:%u system:%u maplen=%u Rmaplen=%u x:%u b/8:%u\n",
		   fsi->blocks, sysblks, fsi->bitmaplen, fs16_to_cpu(fs_endian,
								     sb.
								     s_zmap_blocks
								     * 1024),
		   (2 + fs16_to_cpu(fs_endian, sb3.s_imap_blocks)) * 1024,
		   sysblks / 8);
	term_flush();

	/* 5) read data bitmap */
	pread_loop2(fd, map, fsi->bitmaplen - sysblks / 8,
		    (2 + fs16_to_cpu(fs_endian, sb3.s_imap_blocks)) * 1024);
	for (c = 0; c < fsi->blocks - sysblks; c++) {
		if (bitmap_get(map, c)) {
			bitmap_set(fsi->bitmap, c + sysblks, 1);
			fsi->bused++;
		} else {
			fsi->bfree++;
		}
	}

	mem_free(map);

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	minix_type = fs_test(fd);

	switch (minix_type) {
	case MINIX_V3:
		return fs_getbitmap_v3(fd);
	case MINIX_V2:
	case MINIX_V1:
		return fs_getbitmap_v12(fd, minix_type);
	}

	return -1;		/* no minix */
};

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_getbitmap(fd) == -1) {
		return -1;
	}

	free(fsi->bitmap);
	fsi->bitmaplen = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	if (minix_type != MINIX_V3) {
		term_print("%20s%u\n", "s_ninodes:",
			   (u32) fs16_to_cpu(fs_endian, sb.s_ninodes));
		term_print("%20s%u\n", "s_nzones:",
			   (u32) fs16_to_cpu(fs_endian, sb.s_nzones));
		term_print("%20s%u\n", "s_imap_blocks:",
			   (u32) fs16_to_cpu(fs_endian, sb.s_imap_blocks));
		term_print("%20s%u\n", "s_zmap_blocks:",
			   (u32) fs16_to_cpu(fs_endian, sb.s_zmap_blocks));
		term_print("%20s%u\n", "s_firstdatazone:",
			   (u32) fs16_to_cpu(fs_endian, sb.s_firstdatazone));
		term_print("%20s%u\n", "s_log_zone_size:",
			   (u32) 1024 << fs16_to_cpu(fs_endian,
						     sb.s_log_zone_size));
		term_print("%20s%u\n", "s_max_size:",
			   (u32) fs32_to_cpu(fs_endian, sb.s_max_size));
		term_print("%20s0x%4x\n", "s_magic:",
			   (u32) fs16_to_cpu(fs_endian, sb.s_magic));
		term_print("%20s%u\n", "s_state:",
			   (u32) fs16_to_cpu(fs_endian, sb.s_state));
		term_print("%20s%u\n", "s_zones:",
			   (u32) fs32_to_cpu(fs_endian, sb.s_zones));
	} else {
		term_print("%20s%u\n", "s_ninodes:",
			   (u32) fs32_to_cpu(fs_endian, sb3.s_ninodes));
		term_print("%20s%u\n", "s_pad0:",
			   (u32) fs16_to_cpu(fs_endian, sb3.s_pad0));
		term_print("%20s%u\n", "s_imap_blocks:",
			   (u32) fs16_to_cpu(fs_endian, sb3.s_imap_blocks));
		term_print("%20s%u\n", "s_zmap_blocks:",
			   (u32) fs16_to_cpu(fs_endian, sb3.s_zmap_blocks));
		term_print("%20s%u\n", "s_firstdatazone:",
			   (u32) fs16_to_cpu(fs_endian, sb3.s_firstdatazone));
		term_print("%20s%u\n", "s_log_zone_size:",
			   (u32) 1024 << fs16_to_cpu(fs_endian,
						     sb3.s_log_zone_size));
		term_print("%20s%u\n", "s_pad1:",
			   (u32) fs16_to_cpu(fs_endian, sb3.s_pad1));
		term_print("%20s%u\n", "s_max_size:",
			   (u32) fs32_to_cpu(fs_endian, sb3.s_max_size));
		term_print("%20s%u\n", "s_zones:",
			   (u32) fs32_to_cpu(fs_endian, sb3.s_zones));
		term_print("%20s0x%4x\n", "s_magic:",
			   (u32) fs16_to_cpu(fs_endian, sb3.s_magic));
		term_print("%20s%u\n", "s_pad2:",
			   (u32) fs16_to_cpu(fs_endian, sb3.s_pad2));
		term_print("%20s%u\n", "s_blocksize:",
			   (u32) fs16_to_cpu(fs_endian, sb3.s_blocksize));
		term_print("%20s%u\n", "s_disk_version:",
			   (u32) sb3.s_disk_version);
	}
	term_printflush("\n");

	return 0;
}

struct file_system minix_fs = {
	.id = FSID_MINIX,
	.module = "minix",
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
