
/**
 * Copyright © 2008 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/debug.h"
#include "src/bitmap.h"
#include "src/read_skipfile.h"

/* ntfs-3g */
#include "src/ntfs/volume.h"
#include "src/ntfs/inode.h"
#include "src/ntfs/bitmap.h"
#include "src/ntfs/bootsect.h"
#include "src/ntfs/dir.h"
#include "src/ntfs/unistr.h"

static char fs_name[] = "ntfs";

static ntfs_volume *fs_ntfs_open_device(int fd)
{
	ntfs_volume *vol = ntfs_mount("frisbee", fd);

	if (!vol)
		fatal("Unable to open ntfs volume");

	return vol;
}

static ATTR_RECORD *find_attribute(const ATTR_TYPES type,
				   ntfs_attr_search_ctx * ctx)
{
	if (!ctx)
		return 0;

	if (ntfs_attr_lookup(type, NULL, 0, 0, 0, NULL, 0, ctx) != 0)
		return 0;

	return ctx->attr;
}

/* remove one file from bitmap */
static void fs_ntfs_skipfile_file(ntfs_volume * vol, u8 * bitmap,
				  ntfs_inode * ni)
{
	ntfs_attr_search_ctx *ctx;
	ATTR_RECORD *rec;
	runlist *runs;
	int i;

	/**
	 * virtual cluster number @vcn -> of a file
	 * logical cluster number @lcn -> of a device
	 */
	ctx = ntfs_attr_get_search_ctx(ni, 0);
	while ((rec = find_attribute(AT_UNUSED, ctx))) {
		if (!rec->non_resident)
			continue;

		runs = ntfs_mapping_pairs_decompress(vol, rec, 0);
		if (!runs)
			continue;

		for (i = 0; runs[i].length > 0; i++) {
			if (runs[i].lcn < 0)
				continue;
			bitmap_clear_bits(bitmap, runs[i].lcn, runs[i].length);
		}
		free(runs);
	}
	ntfs_attr_put_search_ctx(ctx);

	return;
}

typedef struct {
	ntfs_volume *vol;
	u8 *bitmap;
} filldir_t;

static int filldir(void *dirent, const ntfschar * name, const int name_len,
		   const int name_type, const s64 pos, const MFT_REF mref,
		   const unsigned dt_type)
{
	filldir_t *de = (filldir_t *) dirent;
	ntfs_inode *ni;
	int i;

	/* will catch posix + win32 + win32/dos combi names */
	if (name_type == FILE_NAME_DOS)
		return 0;

	ni = ntfs_inode_open(de->vol, mref);
	if (!ni)
		return 0;

	/*
	char *filename = mem_alloc(name_len * 4);
	if (ntfs_ucstombs(name, name_len, &filename, name_len * 4) < 0)
		fatal("ntfs_ucstombs()");
	*/

	/* next one, recursiv? */
	if (dt_type == NTFS_DT_DIR) {
		s64 npos = 2;	/* skip . and .. */

		ntfs_readdir(ni, &npos, dirent, filldir);
		goto filldir_out;
	}

	if (dt_type == NTFS_DT_REG)
		fs_ntfs_skipfile_file(de->vol, de->bitmap, ni);

 filldir_out:
	/* mem_free(filename); */
	ntfs_inode_close(ni);

	return 0;
}

/**
 * fs_ntfs_skipfile() - modify bitmap by skipping unwanted files
 * [R-] vol
 * [RW] bitmap
 * [R-] len
 * [R-] filename
 */
static void fs_ntfs_skipfile(ntfs_volume * vol, u8 * bitmap, stralloc * sa)
{
	u32 i, j;

	for (j = 0; j < sa->len; j += str_len(sa->s + j) + 1) {
		ntfs_inode *ni;
		char *fn = sa->s + j;

		debug(12, 5, "fs_ntfs_skipfile(), free for file: %s", fn);
		ni = ntfs_pathname_to_inode(vol, 0, fn);
		if (!ni) {
			debug(12, 1, "not found");
			continue;
		}

		/* if directory, then ... */
		if (ni->mrec->flags & MFT_RECORD_IS_DIRECTORY) {
			s64 pos = 2;	/* skip . and .. */
			filldir_t dirent;

			dirent.vol = vol;
			dirent.bitmap = bitmap;
			ntfs_readdir(ni, &pos, &dirent, filldir);
		} else {
			/* some file */
			fs_ntfs_skipfile_file(vol, bitmap, ni);
		}

		ntfs_inode_close(ni);
	}

	return;
}

/**
 * fs_test - check for ntfs filesystem
 *
 * @return 0 - no ntfs fs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	NTFS_BOOT_SECTOR sb;
	ssize_t r;

	r = pread_loop(fd, &sb, sizeof(NTFS_BOOT_SECTOR), 0);
	if (r == -1) {
		return 0;
	}

	return ntfs_boot_sector_is_ntfs(&sb);
}

/**
 * fs_getbitmap - get allocation bitmap for ntfs filesystem
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	ntfs_volume *vol;
	s64 pos, tmp;

	/* fast test, if there is some ntfs data ... */
	if (fs_test(fd) == 0)
		return -1;

	vol = fs_ntfs_open_device(fd);

	/**
	 * read bitmap => vol->free_clusters is wrong :/
	 */
	fsi->bitmaplen = vol->lcnbmp_na->data_size;
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen);

	tmp = 0;
	for (pos = 0; pos < vol->lcnbmp_na->data_size; pos += tmp) {
		tmp =
		    ntfs_attr_pread(vol->lcnbmp_na, pos,
				    vol->lcnbmp_na->data_size - pos,
				    fsi->bitmap + pos);
		if (tmp < 0) {
			die_msg(2, "ntfs_attr_pread failed");
		}
	}

	/* update bitmap, when skipfile is set */
	if (fng_skipfile.len)
		fs_ntfs_skipfile(vol, fsi->bitmap, &fng_skipfile);

	fsi->bused = 0;
	/* count the clusters @ bitmap */
	for (tmp = 0; tmp < vol->nr_clusters; tmp++) {
		if (bitmap_get(fsi->bitmap, tmp)) {
			fsi->bused++;
		}
	}

	/* fill info structure */
	fsi->blocks = vol->nr_clusters;
	fsi->bsize = vol->cluster_size;
	fsi->bfree = vol->nr_clusters - fsi->bused;
	fsi->name = fs_name;
	fsi->fssize = fsi->bsize * fsi->blocks;

	ntfs_umount(vol);

	return 0;
}

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_getbitmap(fd) == -1) {
		return -1;
	}

	/* not needed @ fs_info() */
	free(fsi->bitmap);
	fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug - give extended information to stdout
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	ntfs_volume *vol;
	s64 pos, tmp;

	/* fast test, if there is some ntfs data ... */
	if (fs_test(fd) == 0)
		return -1;

	vol = fs_ntfs_open_device(fd);
	fsi->bitmaplen = vol->lcnbmp_na->data_size;
	fsi->bitmap = mem_talloc(u8, fsi->bitmaplen + 1);

	tmp = 0;
	for (pos = 0; pos < vol->lcnbmp_na->data_size; pos += tmp) {
		tmp =
		    ntfs_attr_pread(vol->lcnbmp_na, pos,
				    vol->lcnbmp_na->data_size - pos,
				    fsi->bitmap + pos);
		if (tmp < 0) {
			die_msg(2, "ntfs_attr_pread failed");
		}
	}

	/* count the clusters @ bitmap */
	fsi->bused = 0;
	for (tmp = 0; tmp < vol->nr_clusters; tmp++) {
		if (bitmap_get(fsi->bitmap, tmp)) {
			fsi->bused++;
		}
	}

	term_print("%20s%s\n", "filesystem type:", fs_name);
	term_print("%20s%s\n", "volume name:", vol->vol_name);
	term_print("%20s%u\n", "fs block size:", vol->cluster_size);
	term_print("%20s%u\n", "sector size:", vol->sector_size);
	term_print("%20s%u\n", "index record size:", vol->indx_record_size);
	term_print("%20s%u\n", "mft record size:", vol->mft_record_size);
	term_print("%20s%u\n\n", "bitmap length:", vol->lcnbmp_na->data_size);

	term_print("%20s%u\n", "fs # of blocks:", vol->nr_clusters);
	term_print("%20s%u\n", "fs # used blocks:", fsi->bused);
	term_print("%20s%u\n", "fs # free blocks:",
		   vol->nr_clusters - fsi->bused);
	term_printflush("\n");

	free(fsi->bitmap);
	ntfs_umount(vol);

	fsi->name = fs_name;

	return 0;
}

/**
 * fs_postclone
 * - copy 1st sector to the last sector
 * - fd is source fd
 */
static void fs_postclone(int fd, int fd_out, off_t offset)
{
	struct fsinfo *fsi = io_getcookie(fd);
	char *pbr;

	if (!fsi)
		fatal("fs_postclone() *fsi not defined!!!");

	pbr = mem_alloc(512);

	/* read 1st sector and copy to the last one! */
	pread_loop2(fd, pbr, 512, 0 + offset);

	/* write it to the last sector of destination */
	pwrite_loop2(fd_out, pbr, 512, fsi->psize - 512 + offset);
	mem_free(pbr);

	return;
}

struct file_system ntfs_fs = {
	.id = FSID_NTFS,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_postclone = fs_postclone,
	.fs_debug = fs_debug
};
