
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"
#include "src/getblock.h"

/**
 * QNX file system 4
 *
 * systems: QNX4 and newer
 * endian:  little
 */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

static char fs_name[] = "qnx4";

/* qnx4 fs detection */
#define QNX4_SB_MAGIC 0x2f

typedef struct {
	u32 blk;
	u32 size;
} xtnt_t;

struct qnx4_di {
	u8 fname[16];
	u32 size;
	xtnt_t xt0;
	u32 xblk;
	u32 ftime;
	u32 mtime;
	u32 atime;
	u32 ctime;
	u16 num_xtnts;
	u16 mode;
	u16 uid;
	u16 gid;
	u16 nlink;
	u8 zero[4];
	u8 type;
	u8 status;
} __FNG_ATTR_PACKED;

struct qnx4_sb {
	struct qnx4_di root;
	struct qnx4_di inode;
	struct qnx4_di boot;
	struct qnx4_di altboot;
};				/* size: 256 */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global qnx4 variables */
static struct qnx4_sb *sb;

/**
 * fs_test - check for qnx4 filesystem
 *
 * @return 0 - no qnx4
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	void *x;
	u32 *magic;

	x = getblock(fd, 1);
	if (x == 0) {
		return 0;
	}

	/* will be used later */
	sb = x;

	magic = x;
	if (le32_to_cpu(*magic) == QNX4_SB_MAGIC) {
		fsi->name = fs_name;
		return 1;
	}

	/* no qnx4 found */
	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	fsi->bitmaplen = (le32_to_cpu(sb->root.xt0.blk) - 4) * 512;
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen);

	fsi->bsize = 512;
	fsi->blocks = fsi->bitmaplen * 8;
	fsi->bused = 20;
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bfree = fsi->blocks - fsi->bused;

	term_print("passt ;)");
	term_flush();

	return 0;
};

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_getbitmap(fd) == -1) {
		return -1;
	}
	free(fsi->bitmap);

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_getbitmap(fd) == -1) {
		return -1;
	}

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);

	term_print("%20s%u\n", "root.xt0.blk:", le32_to_cpu(sb->root.xt0.blk));
	term_print("%20s%u\n", "root.xt0.size:",
		   le32_to_cpu(sb->root.xt0.size));
	term_print("%20s%u\n", "root.size:", le32_to_cpu(sb->root.size));
	term_print("%20s%u\n", "root.xblk:", le32_to_cpu(sb->root.xblk));

	term_print("%20s%u\n", "inode.xt0.blk:",
		   le32_to_cpu(sb->inode.xt0.blk));
	term_print("%20s%u\n", "inode.xt0.size:",
		   le32_to_cpu(sb->inode.xt0.size));
	term_print("%20s%u\n", "inode.size:", le32_to_cpu(sb->inode.size));
	term_print("%20s%u\n", "inode.xblk:", le32_to_cpu(sb->inode.xblk));

	term_printflush("\n");

	return 0;
}

struct file_system qnx4_fs = {
	.id = FSID_QNX4,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
