
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"
#include "src/getblock.h"

/**
 * QNX file system 6
 *
 * systems: QNX6, Audi MMI 3G
 * endian:  little and big endian
 */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

static char fs_name[] = "qnx6";
static char fs_nameA[] = "qnx6_mmi3g";

/* qnx6 fs detection */
#define QNX6_SB_MAGIC       0x68191122

struct qnx6_root_node {
	u64 size;
	u32 ptr[16];
	u8 levels;
	u8 mode;
	u8 spare[6];
} __FNG_ATTR_PACKED;

/* plain qnx6 superblock layout */
struct qnx6_sb {
	u32 sb_magic;
	u32 sb_checksum;
	u64 sb_serial;
	u32 sb_ctime;		/* time the fs was created */
	u32 sb_atime;		/* last access time */
	u32 sb_flags;
	u16 sb_version1;	/* filesystem version information */
	u16 sb_version2;	/* filesystem version information */
	u8 sb_volumeid[16];
	u32 sb_blocksize;
	u32 sb_num_inodes;
	u32 sb_free_inodes;
	u32 sb_num_blocks;
	u32 sb_free_blocks;
	u32 sb_allocgroup;
	struct qnx6_root_node Inode;
	struct qnx6_root_node Bitmap;
	struct qnx6_root_node Longfile;
	struct qnx6_root_node Unknown;
} __FNG_ATTR_PACKED;

/* Audi MMI 3G superblock layout */
struct qnx6_mmi_sb {
	u32 sb_magic;		/* same like plain qnx6 */
	u32 sb_checksum;	/* same like plain qnx6 */
	u64 sb_serial;		/* same like plain qnx6 */
	u8 sb_spare0[12];
	u8 sb_id[12];
	u32 sb_blocksize;
	u32 sb_num_inodes;
	u32 sb_free_inodes;
	u32 sb_num_blocks;
	u32 sb_free_blocks;
	u8 sb_spare1[4];
	struct qnx6_root_node Inode;
	struct qnx6_root_node Bitmap;
	struct qnx6_root_node Longfile;
	struct qnx6_root_node Unknown;
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global qnx6 variables */
static struct qnx6_sb *sb, *sb2;
static struct qnx6_mmi_sb *sb_mmi;
static int fs_endian;

#define QNX6_PLAIN 1
#define QNX6_MMI3G 2
static int fs_type;

/**
 * fs_test_magic - check for qnx6 magic and endianess
 *
 * @return 0 - no qnx6
 * @return 1 - yes
 */
static int fs_test_magic(int fd, int offset)
{
	void *x;

	x = getblock(fd, offset);
	if (!x) {
		return 0;
	}

	sb = x;
	sb_mmi = x;
	if (le32_to_cpu(sb->sb_magic) == QNX6_SB_MAGIC) {
		fs_endian = BYTESEX_LE;
		return 1;
	} else if (be32_to_cpu(sb->sb_magic) == QNX6_SB_MAGIC) {
		fs_endian = BYTESEX_BE;
		return 1;
	}

	return 0;
}

/**
 * fs_test - check for qnx6 filesystem
 *
 * @return 0 - no qnx6
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	/**
	 * structure plain qnx6:
	 * 8x bsize system
	 * superblock 1
	 * num_blocks data
	 * 4x bsize system
	 * superblock 2
	 *
	 * structure audi mmi 3g:
	 * superblock 1
	 * num_blocks data
	 * 4x bsize system?
	 * superblock 2
	 */

	if (fs_test_magic(fd, 16)) {
		/* plain qnx6 */
		fs_type = QNX6_PLAIN;
		fsi->name = fs_name;
		fsi->bsize = fs32_to_cpu(fs_endian, sb->sb_blocksize);
	} else if (fs_test_magic(fd, 0)) {
		/* audi mmi qnx6 */
		fs_type = QNX6_MMI3G;
		fsi->name = fs_nameA;
		fsi->bsize = fs32_to_cpu(fs_endian, sb_mmi->sb_blocksize);
	} else
		return 0;

	/* check blocksize */
	if (fsi->bsize != 512 && fsi->bsize != 1024 && fsi->bsize != 2048
	    && fsi->bsize != 4096) {
		/* we have no valid blocksize */
		return 0;
	}

	/* qnx6 found */
	return 1;
}

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* set block stuff */
	fsi->bsize = fs32_to_cpu(fs_endian, sb->sb_blocksize);
	fsi->blocks = fs32_to_cpu(fs_endian, sb->sb_num_blocks);
	fsi->bfree = fs32_to_cpu(fs_endian, sb->sb_free_blocks);
	fsi->bused = fsi->blocks - fsi->bfree;

	/* real used space on disk */
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	struct qnx6_root_node *rn;
	off_t offset;
	void *x;
	int i;

	if (fs_info(fd) == -1) {
		return -1;
	}

	/* get offset of second superblock */
	if (fs_type == QNX6_PLAIN) {
		offset = 0x2000 + 0x1000;
	} else {
		offset = 0x1000;	/* 0x1200 */
	}
	offset += fsi->blocks * fsi->bsize;
	x = getblock(fd, offset / 512);
	if (!x)
		return -1;
	sb2 = x;

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s0x%x\n", "magic:",
		   fs32_to_cpu(fs_endian, sb->sb_magic));
	term_print("%20s0x%0x\n", "checksum:",
		   fs32_to_cpu(fs_endian, sb->sb_checksum));
	term_print("%20s%U\n", "serial1:",
		   fs64_to_cpu(fs_endian, sb->sb_serial));
	term_print("%20s%U\n", "serial2:",
		   fs64_to_cpu(fs_endian, sb2->sb_serial));
	term_print("%20s0x%1x\n", "flags:",
		   fs32_to_cpu(fs_endian, sb->sb_flags));
	term_print("%20s%u\n", "version1:",
		   fs16_to_cpu(fs_endian, sb->sb_version1));
	term_print("%20s%u\n", "version2:",
		   fs16_to_cpu(fs_endian, sb->sb_version2));
	term_print("%20s%u\n", "blocksize:",
		   fs32_to_cpu(fs_endian, sb->sb_blocksize));
	term_print("%20s%u\n", "num_inodes:",
		   fs32_to_cpu(fs_endian, sb->sb_num_inodes));
	term_print("%20s%u\n", "free_inodes:",
		   fs32_to_cpu(fs_endian, sb->sb_free_inodes));
	term_print("%20s%u\n", "num_blocks:",
		   fs32_to_cpu(fs_endian, sb->sb_num_blocks));
	term_print("%20s%u\n", "free_blocks:",
		   fs32_to_cpu(fs_endian, sb->sb_free_blocks));
	term_print("%20s%u\n", "allocgroup:",
		   fs32_to_cpu(fs_endian, sb->sb_allocgroup));

	rn = &sb->Inode;
	term_print("%20s%U\n", "inode size:", fs32_to_cpu(fs_endian, rn->size));
	for (i = 0; i < 16; i++)
		term_print("%20s%U\n", "inode ptr:",
			   fs32_to_cpu(fs_endian, rn->ptr[i]));

	rn = &sb->Bitmap;
	term_print("%20s%U\n", "bitmap size:",
		   fs32_to_cpu(fs_endian, rn->size));

	for (i = 0; i < 16; i++)
		term_print("%20s%U\n", "bitmap ptr:",
			   fs32_to_cpu(fs_endian, rn->ptr[i]));

	rn = &sb->Longfile;
	term_print("%20s%U\n", "longfile size:",
		   fs32_to_cpu(fs_endian, rn->size));
	rn = &sb->Unknown;
	term_print("%20s%U\n", "unknown size:",
		   fs32_to_cpu(fs_endian, rn->size));
	term_printflush("\n");

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	struct qnx6_root_node *rn;

	if (fs_info(fd) == -1) {
		return -1;
	}

	rn = &sb->Bitmap;
	if (fsi->bitmaplen != fs32_to_cpu(fs_endian, rn->size)) {
		die_msg(111, "bitmap calculation has failure!");
	}

	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + 1);

	return -1;
};

struct file_system qnx6_fs = {
	.id = FSID_QNX6,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
