
/**
 * Copyright © 2012, 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "io.h"
#include "byte.h"
#include "uint32.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/bitmap.h"
#include "src/strallocX.h"
#include "src/die.h"
#include "src/mem.h"
#include "src/fs.h"
#include "src/bswap.h"

static char fs_name[] = "raw";

/**
 * fs_test - check if it is a unknown filesysem ;)
 *
 * @return 0 - no - it is not unknown ^^
 * @return 1 - yes - it is unknown
 */
static int fs_test(int fd)
{
	char *x = mem_alloc(512);
	ssize_t r;

	r = pread_loop(fd, x, 512, 0);
	mem_free(x);
	if (r != 512) {
		/* we need at least ONE block */
		return 0;
	}

	/* we can read everything */
	return 1;
}

/**
 * fs_info - gathering information about unknown filesystems
 *
 * these are set:
 * - u32 bsize;  -> blocksize of filesystem
 * - u64 fssize; -> really used space in bytes of partition or file
 *
 * - u64 blocks; -> total data blocks in filesystem
 * - u64 bused;  -> used blocks in filesystem
 * - u64 bfree;  -> free blocks in filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* fill in the data */
	fsi->name = fs_name;
	fsi->bsize = 512;	/* hard default to 512 */
	fsi->blocks = fsi->bused = fsi->psize / fsi->bsize;
	fsi->bfree = 0;
	fsi->fssize = fsi->bsize * fsi->blocks;

	/* calculate bitmap length */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fs_name);
	term_print("%20s%u\n", "fs block size:", fsi->bsize);
	term_print("%20s%u\n\n", "bitmap length:", fsi->bitmaplen);

	term_print("%20s%u\n", "fs # of blocks:", fsi->blocks);
	term_print("%20s%u\n", "fs # used blocks:", fsi->bused);
	term_print("%20s%u\n", "fs # free blocks:", fsi->bfree);

	term_printflush("\n");

	return 0;
}

/**
 * fs_getbitmap - get allocation bitmap for unknown filesystems
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = mem_talloc(u8, fsi->bitmaplen);
	memset(fsi->bitmap, 0xff, fsi->bitmaplen);

	return 0;
}

struct file_system raw_fs = {
	.id = FSID_RAW,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
