
/**
 * Copyright © 1996 - 2004 Hans Reiser
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/defs.h"
#include "src/term.h"
#include "src/mem.h"

/**
 * Reiser File System (Version 3)
 *
 * systems: Linux, BSD
 * endian:  little
 */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

/* this is the super from 3.5 + 3.6 + 3.X */
struct reiserfs_sb {
	u32 sb_block_count;	/* 0 number of block on data device */
	u32 sb_free_blocks;	/* 4 free blocks count */
	u32 sb_root_block;	/* 8 root of the tree */
	u32 jp_journal_1st_block;	/* 12 where does journal start from on its device */
	u32 jp_journal_dev;	/* 16 journal device st_rdev */
	u32 jp_journal_size;	/* 20 size of the journal on FS creation. used to make sure they don't overflow it */
	u32 jp_journal_trans_max;	/* 24 max number of blocks in a transaction.  */
	u32 jp_journal_magic;	/* 28 random value made on fs creation (this was sb_journal_block_count) */
	u32 jp_journal_max_batch;	/* 32 max number of blocks to batch into a trans */
	u32 jp_journal_max_commit_age;	/* 36 in seconds, how old can an async commit be */
	u32 jp_journal_max_trans_age;	/* 40 in seconds, how old can a transaction be */
	u16 sb_blocksize;	/* 44 */
	u16 sb_oid_maxsize;	/* 46 max size of object id array, see get_objectid() commentary */
	u16 sb_oid_cursize;	/* 48 current size of object id array */
	u16 sb_umount_state;	/* 50 this is set to 1 when filesystem was umounted, to 2 - when not */
	char s_magic[10];	/* 52 reiserfs magic string indicates that file system is reiserfs: "ReIsErFs" or "ReIsEr2Fs" or "ReIsEr3Fs" */
	u16 sb_fs_state;	/* 62 it is set to used by fsck to mark which phase of rebuilding is done (used for fsck debugging) */
	u32 sb_hash_function_code;	/* 64 code of fuction which was/is/will be used to sort names in a directory. See codes in above */
	u16 sb_tree_height;	/* 68 height of filesytem tree. Tree consisting of only one root block has 2 here */
	u16 sb_bmap_nr;		/* 70 amount of bitmap blocks needed to address each block of file system */
	u16 sb_version;		/* 72 this field is only reliable on filesystem with non-standard journal */
	u16 sb_reserved_for_journal;	/* 74 size in blocks of journal area on main device, we need to keep after non-standard journal relocation */
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global reiserfs variables */
static struct reiserfs_sb sb;
static int reiserfs_offset;

/**
 * fs_reiser_isfs - check if reiser filesystem
 *
 * @return 0 - no reiserfs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	ssize_t r;

	/* try standard offset (64K) */
	reiserfs_offset = 64 * 1024;
	r = pread_loop(fd, &sb, sizeof(struct reiserfs_sb), reiserfs_offset);

	if (r == -1) {
		return 0;
	}

	if (str_start(sb.s_magic, "ReIsEr")) {
		return 1;
	}

	/* try old offset (3.5 - 3.5.11) */
	reiserfs_offset = 8 * 1024;
	r = pread_loop(fd, &sb, sizeof(struct reiserfs_sb), reiserfs_offset);

	if (r == -1) {
		return 0;
	}

	if (str_start(sb.s_magic, "ReIsEr")) {
		return 1;
	}

	return 0;
}

/**
 * fs_info -  gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	/* "ReIsErFs" "ReIsEr2Fs" "ReIsEr3Fs" */
	fsi->name = "reiserfs 3.5";	/* default */
	if (sb.s_magic[7] == '2')
		fsi->name = "reiserfs 3.6";
	if (sb.s_magic[7] == '3')
		fsi->name = "reiserfs JR";

	/* fill in the data */
	fsi->bsize = le16_to_cpu(sb.sb_blocksize);
	fsi->blocks = le32_to_cpu(sb.sb_block_count);
	fsi->bfree = le32_to_cpu(sb.sb_free_blocks);
	fsi->bused = fsi->blocks - fsi->bfree;
	fsi->fssize = fsi->bsize * fsi->blocks;
	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = 0;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	/* XXX: add a bit more info */
	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s%u\n", "fs block size:", fsi->bsize);
	term_print("%20s%u\n\n", "bitmap length:", fsi->bitmaplen);

	term_print("%20s%u\n", "fs # of blocks:", fsi->blocks);
	term_print("%20s%u\n", "fs # used blocks:", fsi->bused);
	term_print("%20s%u\n", "fs # free blocks:", fsi->bfree);
	term_printflush("\n");

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	u16 map;
	u64 bmap;

	if (fs_info(fd) == -1) {
		return -1;
	}

	/**
	 * source: http://homes.cerias.purdue.edu/~florian/reiser/reiserfs.html
	 *
	 * if bs=8K, each map holds 65536 bits
	 * if bs=4K, each map holds 32768 bits
	 * if bs=2K, each map holds 16384 bits
	 * if bs=1K, each map holds 8192 bits
	 * if bs=512, each map holds 4096 bits
	 * the bitmaps start directly after the sb!
	 * the sb is located @ reiserfs_offset, 64K or 8K
	 * the bitmap count is stored in sb via sb_bmap_nr
	 */

	/* fsi->bitmaplen is set in fs_info() */
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + fsi->bsize);

	/* location of superblock # */
	bmap = reiserfs_offset / fsi->bsize;
	bmap += 1;

	/* 1st single map, directly after superblock */
	pread_loop2(fd, fsi->bitmap, fsi->bsize, bmap * fsi->bsize);

	/* read the other maps */
	for (map = 1; map < le16_to_cpu(sb.sb_bmap_nr); map++) {
		//term_print("bmap:%U max:%U map:%U maplen:%U\n", bmap, (u64)le16_to_cpu(sb.sb_bmap_nr), map, fsi->bitmaplen);
		//term_flush();
		bmap = fsi->bsize * 8 * map;
		//term_print("pread(%d, %U, %U, %U)\n",
		//      fd, fsi->bitmap + map * fsi->bsize, fsi->bsize, bmap * fsi->bsize);
		pread_loop2(fd, fsi->bitmap + map * fsi->bsize, fsi->bsize,
			    bmap * fsi->bsize);
	}

	return 0;
}

struct file_system reiser_fs = {
	.id = FSID_REISER,
	.module = "reiser",
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
