
/**
 * Copyright © 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 */

/* system */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "str.h"
#include "io.h"

/* frisbee-ng */
#include "src/wrapper.h"
#include "src/term.h"
#include "src/defs.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/mem.h"
#include "src/bitmap.h"

/**
 * Unix File System (UFS) / Berkeley Fast File System (FFS)
 * - also UFS2 since FreeBSD 5.x
 *
 * systems: FreeBSD, NetBSD, OpenBSD, Solaris, NeXTStep, Linux, Mac OS X,
 *          A/UX, DragonFlyBSD, FreeNAS, HP-UX, SunOS, Tru64 UNIX,
 *          UNIX System V, Minix and others ...
 * endian:  little or big endian (depends on system)
 */

#define FS_UFS1_MAGIC	0x00011954	/* UFS1 fast file system magic number */
#define FS_UFS2_MAGIC	0x19540119	/* UFS2 fast file system magic number */
#define NOCSPTRS	((128 / sizeof(void *)) - 4)

struct csum_total {
	s64 cs_ndir;		/* number of directories */
	s64 cs_nbfree;		/* number of free blocks */
	s64 cs_nifree;		/* number of free inodes */
	s64 cs_nffree;		/* number of free frags */
	s64 cs_numclusters;	/* number of free clusters */
	s64 cs_spare[3];	/* future expansion */
};

struct csum {
	s32 cs_ndir;		/* number of directories */
	s32 cs_nbfree;		/* number of free blocks */
	s32 cs_nifree;		/* number of free inodes */
	s32 cs_nffree;		/* number of free frags */
};

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

/*
 * Super block for an FFS filesystem.
 */
struct ufs_sb {
	s32 fs_firstfield;	/* historic filesystem linked list, */
	s32 fs_unused_1;	/*     used for incore super blocks */
	s32 fs_sblkno;		/* offset of super-block in filesys */
	s32 fs_cblkno;		/* offset of cyl-block in filesys */
	s32 fs_iblkno;		/* offset of inode-blocks in filesys */
	s32 fs_dblkno;		/* offset of first data after cg */
	s32 fs_old_cgoffset;	/* cylinder group offset in cylinder */
	s32 fs_old_cgmask;	/* used to calc mod fs_ntrak */
	s32 fs_old_time;	/* last time written */
	s32 fs_old_size;	/* number of blocks in fs */
	s32 fs_old_dsize;	/* number of data blocks in fs */
	u32 fs_ncg;		/* number of cylinder groups */
	s32 fs_bsize;		/* size of basic blocks in fs */
	s32 fs_fsize;		/* size of frag blocks in fs */
	s32 fs_frag;		/* number of frags in a block in fs */
/* these are configuration parameters */
	s32 fs_minfree;		/* minimum percentage of free blocks */
	s32 fs_old_rotdelay;	/* num of ms for optimal next block */
	s32 fs_old_rps;		/* disk revolutions per second */
/* these fields can be computed from the others */
	s32 fs_bmask;		/* ``blkoff'' calc of blk offsets */
	s32 fs_fmask;		/* ``fragoff'' calc of frag offsets */
	s32 fs_bshift;		/* ``lblkno'' calc of logical blkno */
	s32 fs_fshift;		/* ``numfrags'' calc number of frags */
/* these are configuration parameters */
	s32 fs_maxcontig;	/* max number of contiguous blks */
	s32 fs_maxbpg;		/* max number of blks per cyl group */
/* these fields can be computed from the others */
	s32 fs_fragshift;	/* block to frag shift */
	s32 fs_fsbtodb;		/* fsbtodb and dbtofsb shift constant */
	s32 fs_sbsize;		/* actual size of super block */
	s32 fs_spare1[2];	/* old fs_csmask */
	/* old fs_csshift */
	s32 fs_nindir;		/* value of NINDIR */
	u32 fs_inopb;		/* value of INOPB */
	s32 fs_old_nspf;	/* value of NSPF */
/* yet another configuration parameter */
	s32 fs_optim;		/* optimization preference, see below */
	s32 fs_old_npsect;	/* # sectors/track including spares */
	s32 fs_old_interleave;	/* hardware sector interleave */
	s32 fs_old_trackskew;	/* sector 0 skew, per track */
	s32 fs_id[2];		/* unique filesystem id */
/* sizes determined by number of cylinder groups and their sizes */
	s32 fs_old_csaddr;	/* blk addr of cyl grp summary area */
	s32 fs_cssize;		/* size of cyl grp summary area */
	s32 fs_cgsize;		/* cylinder group size */
	s32 fs_spare2;		/* old fs_ntrak */
	s32 fs_old_nsect;	/* sectors per track */
	s32 fs_old_spc;		/* sectors per cylinder */
	s32 fs_old_ncyl;	/* cylinders in filesystem */
	s32 fs_old_cpg;		/* cylinders per group */
	u32 fs_ipg;		/* inodes per group */
	s32 fs_fpg;		/* blocks per group * fs_frag */
/* this data must be re-computed after crashes */
	struct csum fs_old_cstotal;	/* cylinder summary information */
/* these fields are cleared at mount time */
	s8 fs_fmod;		/* super block modified flag */
	s8 fs_clean;		/* filesystem is clean flag */
	s8 fs_ronly;		/* mounted read-only flag */
	s8 fs_old_flags;	/* old FS_ flags */
	u8 fs_fsmnt[468];	/* name mounted on */
	u8 fs_volname[32];	/* volume name */
	u64 fs_swuid;		/* system-wide uid */
	s32 fs_pad;		/* due to alignment of fs_swuid */
/* these fields retain the current block allocation info */
	s32 fs_cgrotor;		/* last cg searched */
	void *fs_ocsp[NOCSPTRS];	/* padding; was list of fs_cs buffers */
	u8 *fs_contigdirs;	/* (u) # of contig. allocated dirs */
	struct csum *fs_csp;	/* (u) cg summary info buffer */
	s32 *fs_maxcluster;	/* (u) max cluster in each cyl group */
	u32 *fs_active;		/* (u) used by snapshots to track fs */
	s32 fs_old_cpc;		/* cyl per cycle in postbl */
	s32 fs_maxbsize;	/* maximum blocking factor permitted */
	s64 fs_unrefs;		/* number of unreferenced inodes */
	s64 fs_sparecon64[16];	/* old rotation block list head */
	s64 fs_sblockloc;	/* byte offset of standard superblock */
	struct csum_total fs_cstotal;	/* (u) cylinder summary information */
	s64 fs_time;		/* last time written */
	s64 fs_size;		/* number of blocks in fs */
	s64 fs_dsize;		/* number of data blocks in fs */
	s64 fs_csaddr;		/* blk addr of cyl grp summary area */
	s64 fs_pendingblocks;	/* (u) blocks being freed */
	u32 fs_pendinginodes;	/* (u) inodes being freed */
	u32 fs_snapinum[20];	/* list of snapshot inode numbers */
	u32 fs_avgfilesize;	/* expected average file size */
	u32 fs_avgfpdir;	/* expected # of files per directory */
	s32 fs_save_cgsize;	/* save real cg size to use fs_bsize */
	s64 fs_mtime;		/* Last mount or fsck time. */
	s32 fs_sujfree;		/* SUJ free list */
	s32 fs_sparecon32[23];	/* reserved for future constants */
	s32 fs_flags;		/* see FS_ flags below */
	s32 fs_contigsumsize;	/* size of cluster summary array */
	s32 fs_maxsymlinklen;	/* max length of an internal symlink */
	s32 fs_old_inodefmt;	/* format of on-disk inodes */
	u64 fs_maxfilesize;	/* maximum representable file size */
	s64 fs_qbmask;		/* ~fs_bmask for use with 64-bit size */
	s64 fs_qfmask;		/* ~fs_fmask for use with 64-bit size */
	s32 fs_state;		/* validate fs_clean field */
	s32 fs_old_postblformat;	/* format of positional layout tables */
	s32 fs_old_nrpos;	/* number of rotational positions */
	s32 fs_spare5[2];	/* old fs_postbloff */
	/* old fs_rotbloff */
	s32 fs_magic;		/* magic number */
} __FNG_ATTR_PACKED;

/*
 * Cylinder group block for a filesystem.
 */
#define	CG_MAGIC	0x090255
struct cg {
	s32 cg_firstfield;	/* historic cyl groups linked list */
	s32 cg_magic;		/* magic number */
	s32 cg_old_time;	/* time last written */
	u32 cg_cgx;		/* we are the cgx'th cylinder group */
	s16 cg_old_ncyl;	/* number of cyl's this cg */
	s16 cg_old_niblk;	/* number of inode blocks this cg */
	u32 cg_ndblk;		/* number of data blocks this cg */
	struct csum cg_cs;	/* cylinder summary information */
	u32 cg_rotor;		/* position of last used block */
	u32 cg_frotor;		/* position of last used frag */
	u32 cg_irotor;		/* position of last used inode */
	u32 cg_frsum[8];	/* counts of available frags */
	s32 cg_old_btotoff;	/* (int32) block totals per cylinder */
	s32 cg_old_boff;	/* (u_int16) free block positions */
	u32 cg_iusedoff;	/* (u_int8) used inode map */
	u32 cg_freeoff;		/* (u_int8) free block map */
	u32 cg_nextfreeoff;	/* (u_int8) next available space */
	u32 cg_clustersumoff;	/* (u_int32) counts of avail clusters */
	u32 cg_clusteroff;	/* (u_int8) free cluster map */
	u32 cg_nclusterblks;	/* number of clusters this cg */
	u32 cg_niblk;		/* number of inode blocks this cg */
	u32 cg_initediblk;	/* last initialized inode */
	u32 cg_unrefs;		/* number of unreferenced inodes */
	s32 cg_sparecon32[2];	/* reserved for future use */
	s64 cg_time;		/* time last written */
	s64 cg_sparecon64[3];	/* reserved for future use */
	u8 cg_space[1];		/* space for cylinder group maps */
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* global ufs variables */
static struct ufs_sb sb;
static int sblist[] = { 64, 8, 0, 256, -1 };

static int fs_endian;

#define UFS_UFS1 1
#define UFS_UFS2 2
static int fs_type;

/*
*********************************************************************************************************************
[root@freebsd ~]# newfs -O1 /dev/md0
/dev/md0: 5120.0MB (10485760 sectors) block size 32768, fragment size 4096
        using 11 cylinder groups of 508.00MB, 16256 blks, 32512 inodes.
super-block backups (for fsck -b #) at:
 64, 1040448, 2080832, 3121216, 4161600, 5201984, 6242368, 7282752, 8323136, 9363520, 10403904

*********************************************************************************************************************
[root@freebsd ~]# newfs -O2 /dev/md1
/dev/md1: 5120.0MB (10485760 sectors) block size 32768, fragment size 4096
        using 7 cylinder groups of 740.00MB, 23680 blks, 47360 inodes.
super-block backups (for fsck -b #) at:
 192, 1515712, 3031232, 4546752, 6062272, 7577792, 9093312

*********************************************************************************************************************
[root@freebsd ~]# newfs -b 4096 -O2 /dev/md2
/dev/md2: 5120.0MB (10485760 sectors) block size 4096, fragment size 4096
        using 96 cylinder groups of 53.47MB, 13689 blks, 3424 inodes.
super-block backups (for fsck -b #) at:
 144, 109656, 219168, 328680, 438192, 547704, 657216, 766728, 876240, 985752, 1095264, 1204776, 1314288, 1423800, 1533312, 1642824, 1752336, 1861848, 1971360, 2080872, 2190384,
 2299896, 2409408, 2518920, 2628432, 2737944, 2847456, 2956968, 3066480, 3175992, 3285504, 3395016, 3504528, 3614040, 3723552, 3833064, 3942576, 4052088, 4161600, 4271112,
 4380624, 4490136, 4599648, 4709160, 4818672, 4928184, 5037696, 5147208, 5256720, 5366232, 5475744, 5585256, 5694768, 5804280, 5913792, 6023304, 6132816, 6242328, 6351840,
 6461352, 6570864, 6680376, 6789888, 6899400, 7008912, 7118424, 7227936, 7337448, 7446960, 7556472, 7665984, 7775496, 7885008, 7994520, 8104032, 8213544, 8323056, 8432568,
 8542080, 8651592, 8761104, 8870616, 8980128, 9089640, 9199152, 9308664, 9418176, 9527688, 9637200, 9746712, 9856224, 9965736, 10075248, 10184760, 10294272, 10403784

*********************************************************************************************************************
[root@freebsd ~]# newfs -f 1024 -O2 /dev/md2
/dev/md2: 5120.0MB (10485760 sectors) block size 8192, fragment size 1024
        using 114 cylinder groups of 45.13MB, 5777 blks, 11584 inodes.
super-block backups (for fsck -b #) at:
 144, 92576, 185008, 277440, 369872, 462304, 554736, 647168, 739600, 832032, 924464, 1016896, 1109328, 1201760, 1294192, 1386624, 1479056, 1571488, 1663920, 1756352, 1848784,
 1941216, 2033648, 2126080, 2218512, 2310944, 2403376, 2495808, 2588240, 2680672, 2773104, 2865536, 2957968, 3050400, 3142832, 3235264, 3327696, 3420128, 3512560, 3604992,
 3697424, 3789856, 3882288, 3974720, 4067152, 4159584, 4252016, 4344448, 4436880, 4529312, 4621744, 4714176, 4806608, 4899040, 4991472, 5083904, 5176336, 5268768, 5361200,
 5453632, 5546064, 5638496, 5730928, 5823360, 5915792, 6008224, 6100656, 6193088, 6285520, 6377952, 6470384, 6562816, 6655248, 6747680, 6840112, 6932544, 7024976, 7117408,
 7209840, 7302272, 7394704, 7487136, 7579568, 7672000, 7764432, 7856864, 7949296, 8041728, 8134160, 8226592, 8319024, 8411456, 8503888, 8596320, 8688752, 8781184, 8873616,
 8966048, 9058480, 9150912, 9243344, 9335776, 9428208, 9520640, 9613072, 9705504, 9797936, 9890368, 9982800, 10075232, 10167664, 10260096, 10352528, 10444960

*********************************************************************************************************************
- we should just take the used fragments, not the blocks...

fs_bsize = block size (immer >= fs_size)
fs_fsize = frag size
fs_frag = nr of frags per block
fs_fpg = fagment anzahl

freeoff       int32_t    0x0000108e
nextfreeoff   int32_t    0x00005810

ist nicht zu schwer.... aber erstmal fng-*.c fertig machen, dann nochmal die filesysteme checken ;)

*/

/**
 * fs_test - check for ufs filesystem
 *
 * @return 0 - no ufs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	ssize_t r;
	int i;

	for (i = 0; sblist[i] != -1; i++) {
		r = pread_loop(fd, &sb, sizeof(struct ufs_sb), sblist[i]);
		if (r == -1) {
			return 0;
		}

		if (FS_UFS1_MAGIC == le32_to_cpu(sb.fs_magic)) {
			fs_endian = BYTESEX_LE;
			fsi->name = "ufs1 (le)";
			fs_type = UFS_UFS1;
			return 1;
		} else if (FS_UFS1_MAGIC == be32_to_cpu(sb.fs_magic)) {
			fs_endian = BYTESEX_BE;
			fsi->name = "ufs1 (be)";
			fs_type = UFS_UFS1;
			return 1;
		} else if (FS_UFS2_MAGIC == le32_to_cpu(sb.fs_magic)) {
			fs_endian = BYTESEX_LE;
			fsi->name = "ufs2 (le)";
			fs_type = UFS_UFS2;
			return 1;
		} else if (FS_UFS2_MAGIC == be32_to_cpu(sb.fs_magic)) {
			fs_endian = BYTESEX_BE;
			fsi->name = "ufs2 (be)";
			fs_type = UFS_UFS2;
			return 1;
		}
	}

	/* no ufs found */
	return 0;
}

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_test(fd) == 0) {
		return -1;
	}

	fsi->bsize = fs32_to_cpu(fs_endian, sb.fs_bsize);
	fsi->blocks = fs64_to_cpu(fs_endian, sb.fs_dsize);
	fsi->bused = 0;
	fsi->fssize = fsi->blocks * fsi->bsize;
	fsi->bfree = fsi->blocks - fsi->bused;

	/* calculate bitmap for later */
	fsi->bitmaplen = (fsi->blocks + 7) / 8;

	return 0;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
#if 0
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}
	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fsi->name);
	term_print("%20s%s\n", "volume name:", sb.name);
	term_print("%20s0x%x\n", "byte order:",
		   fs32_to_cpu(fs_endian, sb.byte_order));
	term_print("%20s%u\n", "block size:",
		   fs32_to_cpu(fs_endian, sb.block_size));
	term_print("%20s%u\n", "block shift:",
		   fs32_to_cpu(fs_endian, sb.block_shift));
	term_print("%20s%U\n", "# fs blocks:",
		   fs64_to_cpu(fs_endian, sb.num_blocks));
	term_print("%20s%U\n", "# used blocks:",
		   fs64_to_cpu(fs_endian, sb.used_blocks));
	term_print("%20s%u\n", "inode_size:", sb.inode_size);

	term_print("%20s0x%x\n", "magic1:", fs32_to_cpu(fs_endian, sb.magic1));
	term_print("%20s0x%x\n", "magic2:", fs32_to_cpu(fs_endian, sb.magic2));
	term_print("%20s0x%x\n", "magic3:", fs32_to_cpu(fs_endian, sb.magic3));
	term_print("%20s%u\n", "ag shift:",
		   fs32_to_cpu(fs_endian, sb.ag_shift));
	term_print("%20s%u\n", "num ags:", fs32_to_cpu(fs_endian, sb.num_ags));
	term_printflush("\n");
#endif

	unused(fd);

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	if (fs_info(fd) == -1) {
		return -1;
	}

	/* allocate space for the map */
	fsi->bitmap = mem_tzalloc(u8, fsi->bitmaplen + 8);

	/* bitmap is on the first map */
	pread_loop2(fd, fsi->bitmap, fsi->blocks / 8, fsi->bsize);
	/* endian? */

	return 0;
};

struct file_system ufs_fs = {
	.id = FSID_UFS,
	.module = "ufs",
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
