
/**
 * Copyright © 2000 - 2001, 2005 Silicon Graphics, Inc.
 * Copyright © 2012, 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 12
 * based on "XFS Filesystem Structure 2nd Edition Revision 1"
 */

/* system */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "io.h"

/* frisbee-ng */
#include "src/defs.h"
#include "src/wrapper.h"
#include "src/die.h"
#include "src/fs.h"
#include "src/term.h"
#include "src/mem.h"
#include "src/bitmap.h"

/**
 * XFS file system
 *
 * systems: IRIX, Linux, FreeBSD
 * endian:  big
 */

static char fs_name[] = "xfs";

#define XFS_SB_MAGIC    0x58465342	/* 'XFSB' sb magic */
#define XFS_AGF_MAGIC   0x58414746	/* 'XAGF' agf magic */
#define XFS_AGI_MAGIC   0x58414749	/* 'XAGI' agi magic */

#define XFS_ABTB_MAGIC  0x41425442	/* 'ABTB' bno B+tree */
#define XFS_ABTC_MAGIC  0x41425443	/* 'ABTC' cnt B+tree */

#define XFS_MAX_BSIZE    64 * 1024
#define XFS_BTREE_MAXLEVELS      8

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

struct xfs_sb {
	u32 sb_magicnum;	/* magic number == XFS_SB_MAGIC */
	u32 sb_blocksize;	/* logical block size, bytes */
	u64 sb_dblocks;		/* number of data blocks */
	u64 sb_rblocks;		/* number of realtime blocks */
	u64 sb_rextents;	/* number of realtime extents */
	char sb_uuid[16];	/* file system unique id */
	u64 sb_logstart;	/* starting block of log if internal */
	u64 sb_rootino;		/* root inode number */
	u64 sb_rbmino;		/* bitmap inode for realtime extents */
	u64 sb_rsumino;		/* summary inode for rt bitmap */
	u32 sb_rextsize;	/* realtime extent size, blocks */
	u32 sb_agblocks;	/* size of an allocation group */
	u32 sb_agcount;		/* number of allocation groups */
	u32 sb_rbmblocks;	/* number of rt bitmap blocks */
	u32 sb_logblocks;	/* number of log blocks */
	u16 sb_versionnum;	/* header version == XFS_SB_VERSION */
	u16 sb_sectsize;	/* volume sector size, bytes */
	u16 sb_inodesize;	/* inode size, bytes */
	u16 sb_inopblock;	/* inodes per block */
	char sb_fname[12];	/* file system name */
	u8 sb_blocklog;		/* log2 ofsb.sb_blocksize */
	u8 sb_sectlog;		/* log2 ofsb.sb_sectsize */
	u8 sb_inodelog;		/* log2 ofsb.sb_inodesize */
	u8 sb_inopblog;		/* log2 ofsb.sb_inopblock */
	u8 sb_agblklog;		/* log2 ofsb.sb_agblocks (rounded up) */
	u8 sb_rextslog;		/* log2 ofsb.sb_rextents */
	u8 sb_inprogress;	/* mkfs is in progress, don't mount */
	u8 sb_imax_pct;		/* max % of fs for inode space statistics */

	/*
	 * These fields must remain contiguous.  If you really
	 * want to change their layout, make sure you fix the
	 * code in xfs_trans_apply_sb_deltas().
	 */
	u64 sb_icount;		/* allocated inodes */
	u64 sb_ifree;		/* free inodes */
	u64 sb_fdblocks;	/* free data blocks */
	u64 sb_frextents;	/* free realtime extents */

	/*
	 * End contiguous fields.
	 */
	u64 sb_uquotino;	/* user quota inode */
	u64 sb_gquotino;	/* group quota inode */
	u16 sb_qflags;		/* quota flags */
	u8 sb_flags;		/* misc. flags */
	u8 sb_shared_vn;	/* shared version number */
	u32 sb_inoalignmt;	/* inode chunk alignment, fsblocks */
	u32 sb_unit;		/* stripe or raid unit */
	u32 sb_width;		/* stripe or raid width */
	u8 sb_dirblklog;	/* log2 of dir block size (fsbs) */
	u8 sb_logsectlog;	/* log2 of the log sector size */
	u16 sb_logsectsize;	/* sector size for the log, bytes */
	u32 sb_logsunit;	/* stripe unit size for the log */
	u32 sb_features2;	/* additional feature bits */

	/*
	 * bad features2 field as a result of failing to pad the sb
	 * structure to 64 bits. Some machines will be using this field
	 * for features2 bits. Easiest just to mark it bad and not use
	 * it for anything else.
	 */
	u32 sb_bad_features2;	/* must be padded to 64 bit alignment */
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/**
 * sector offset size  name          description
 * 0      0      512   xfs_sb_t      super block
 * 1      512    512   xfs_agf_t     ag free block info
 * 2      1024   512   xfs_agi_t     ag inode B+tree info
 * 3      1536   512   xfs_agfl_t    ag internal free list
 * -> secsize can be 2^9 .. 2^15 (512 .. 32K)
 *
 * block (in fs->bsize) -> eg 4K
 * 1      4096   xfs_inobt_block_t   root of inode B+tree
 * 2      4096   xfs_btree_sblock_t  root of free space B+tree (offset)
 * 3      4096   xfs_btree_sblock_t  root of free space B+tree (count)
 * 4-7    4096   4x free space (free list / agfl)
 * 1      ...    data
 * -> blocksize can be 2^9 .. 2^16 (512 .. 64K)
 */
struct xfs_agf {
	u32 agf_magicnum;
	u32 agf_versionnum;
	u32 agf_seqno;
	u32 agf_length;
	u32 agf_bno_offset;	/* btree+ */
	u32 agf_cnt_offset;	/* btree+ */
	u32 agf_spare0;
	u32 agf_bno_level;	/* btree+ */
	u32 agf_cnt_level;	/* btree+ */
	u32 agf_spare1;
	u32 agf_flfirst;
	u32 agf_fllast;
	u32 agf_flcount;
	u32 agf_freeblks;
	u32 agf_longest;
	u32 agf_btreeblks;
};

struct xfs_agi {
	u32 agi_magicnum;	/* magic number == XFS_AGI_MAGIC */
	u32 agi_versionnum;	/* header version == XFS_AGI_VERSION */
	u32 agi_seqno;		/* sequence # starting from 0 */
	u32 agi_length;		/* size in blocks of a.g. */
	u32 agi_count;		/* count of allocated inodes */
	u32 agi_root;		/* root of inode btree */
	u32 agi_level;		/* levels in inode btree */
	u32 agi_freecount;	/* number of free inodes */
	u32 agi_newino;		/* new inode just allocated */
	u32 agi_dirino;		/* last directory inode chunk */
	u32 agi_unlinked[64];
};

struct xfs_agfl {
	u32 agfl_bno;
};

struct xfs_btree {
	u32 bb_magic;		/*  4 byte */
	u16 bb_level;		/*  6 byte */
	u16 bb_numrecs;		/*  8 byte */
	u32 bb_leftsib;		/* 12 byte */
	u32 bb_rightsib;	/* 16 byte */
};

/* 8 byte per record */
struct xfs_alloc_rec {
	u32 ar_startblock;	/* starting block number */
	u32 ar_blockcount;	/* count of free blocks */
};

struct xfs_alloc_key {
	u32 ak_startblock;	/* starting block number */
	u32 ak_blockcount;	/* count of free blocks */
};

/* global xfs variables */
static struct xfs_sb sb;

/* internal function definitions */
static void read_btree(int fd, struct xfs_agf *agf, off_t block);
static int bno_record_offset(int blocksize);
static void read_agfl(int fd, struct xfs_agf *agf, off_t ag_off);

static void die_xfs(char *s)
{
	die_msg(111, s);
}

/**
 * fs_test - check if it is a xfs filesysem
 *
 * @return 0 - no xfs fs
 * @return 1 - yes
 */
static int fs_test(int fd)
{
	u8 *x = mem_alloc(XFS_MAX_BSIZE);
	struct xfs_sb *sb;
	struct xfs_agf *agf;
	struct xfs_agi *agi;
	ssize_t r;

	r = pread_loop(fd, x, XFS_MAX_BSIZE, 0);
	if (r != XFS_MAX_BSIZE)
		goto no_xfs;

	/* check fs sb magic */
	sb = (struct xfs_sb *)x;
	if (be32_to_cpu(sb->sb_magicnum) != XFS_SB_MAGIC)
		goto no_xfs;

	/* bsize to large */
	if (be32_to_cpu(sb->sb_blocksize) > XFS_MAX_BSIZE)
		goto no_xfs;

	/* bsize to small */
	if (be32_to_cpu(sb->sb_blocksize) < 512)
		goto no_xfs;

	/* sector size to large */
	if (be16_to_cpu(sb->sb_sectsize) > 32 * 1024)
		goto no_xfs;

	/* sector size to small */
	if (be16_to_cpu(sb->sb_sectsize) < 512)
		goto no_xfs;

	/* check fs agf magic */
	agf = (struct xfs_agf *)(x + be16_to_cpu(sb->sb_sectsize));
	if (be32_to_cpu(agf->agf_magicnum) != XFS_AGF_MAGIC)
		goto no_xfs;

	/* check fs agi magic */
	agi = (struct xfs_agi *)(x + (be16_to_cpu(sb->sb_sectsize) * 2));
	if (be32_to_cpu(agi->agi_magicnum) != XFS_AGI_MAGIC)
		goto no_xfs;

	/* YES */
	mem_free(x);
	return 1;

 no_xfs:
	mem_free(x);
	return 0;
}

/**
 * fs_info - gathering information about filesystem
 */
static int fs_info(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);

	/* fast test for xfs filesystem */
	if (fs_test(fd) == 0) {
		return -1;
	}

	/* read sb info */
	pread_loop2(fd, &sb, sizeof(struct xfs_sb), 0);
	fsi->bsize = be32_to_cpu(sb.sb_blocksize);
	fsi->blocks = be64_to_cpu(sb.sb_dblocks);
	fsi->bfree = be64_to_cpu(sb.sb_fdblocks);
	fsi->bused = fsi->blocks - fsi->bfree;

	/* used space on disk */
	fsi->fssize = fsi->blocks * fsi->bsize;

	fsi->bitmaplen = (fsi->blocks + 7) / 8;
	fsi->bitmap = 0;
	fsi->name = fs_name;

	return 0;
}

static void read_agfl(int fd, struct xfs_agf *agf, off_t ag_off)
{
	struct fsinfo *fsi = io_getcookie(fd);
	struct xfs_agfl *agfl;
	off_t off = ag_off + 3 * be16_to_cpu(sb.sb_sectsize);

	u64 agfl_max = be16_to_cpu(sb.sb_sectsize) / 4;
	u64 agfl_start = ag_off / fsi->bsize;
	u32 flfirst = be32_to_cpu(agf->agf_flfirst);
	u32 fllast = be32_to_cpu(agf->agf_fllast);
	u32 flcount = be32_to_cpu(agf->agf_flcount);
	u32 i;

	if (flcount != (fllast - flfirst + 1)) {
		die_xfs("agf flcount not correct!");
	}

	agfl = mem_alloc(agfl_max * sizeof(u32));

#if 0
	term_print(" %17s%U\n", "flfirst:", be32_to_cpu(agf->agf_flfirst));
	term_print(" %17s%U\n", "fllast:", be32_to_cpu(agf->agf_fllast));
	term_print(" %17s%U\n", "flcount:", be32_to_cpu(agf->agf_flcount));
	term_print(" %17s%U\n", "agf_btreeblks:",
		   be32_to_cpu(agf->agf_btreeblks));
#endif

	/* read all the agfl blocks 

	   -> alles klaro, nur 4 blöcke oder so einzulesen, dann ...
	   -> 

	 */
	debug(1, 1, "read_agflP(%d) first:%u last:%u count:%u",
	      be16_to_cpu(sb.sb_sectsize), be32_to_cpu(agf->agf_flfirst),
	      be32_to_cpu(agf->agf_fllast), be32_to_cpu(agf->agf_flcount));
	pread_loop2(fd, agfl, be16_to_cpu(sb.sb_sectsize), off);
	debug(1, 1, "read_agfl1(%d)", be16_to_cpu(sb.sb_sectsize));

	for (i = flfirst; i <= fllast; i++) {
		u32 flb = be32_to_cpu(agfl[i].agfl_bno);

		debug(1, 1, "read_agfl(%u) agfl_start=%u flb=%u beides:%u", i,
		      agfl_start, flb, agfl_start + flb);

		bitmap_set(fsi->bitmap, agfl_start + flb, 0);
		// read_btree(fd, agf, flb);
	}
	debug(1, 1, "read_agfl2(%d)", be16_to_cpu(sb.sb_sectsize));

	mem_free(agfl);

	return;
}

/**
 * bno_record_offset - return offset for ptr entries @ bno/cnt B+trees
 *
 * offset[512]:   [344]   bitize[2752]   recs[41]
 * offset[1024]:  [688]   bitize[5504]   recs[84]
 * offset[2048]:  [1368]  bitize[10944]  recs[169]
 * offset[4096]:  [2736]  bitize[21888]  recs[340]
 * offset[8192]:  [5464]  bitize[43712]  recs[681]
 * offset[16384]: [10928] bitize[87424]  recs[1364]
 * offset[32768]: [21848] bitize[174784] recs[2729]
 * offset[65536]: [43696] bitize[349568] recs[5460]
 */
static int bno_record_offset(int blocksize)
{
	debug(1, 1, "bno_record_offset(%d)", blocksize);

	/* this is a lot easier then all the #define CRAP ;) */
	switch (blocksize) {
	case 512:
		return 344;
	case 1024:
		return 688;
	case 2048:
		return 1368;
	case 4096:
		return 2736;
	case 8192:
		return 5464;
	case 16384:
		return 10928;
	case 32768:
		return 21848;
	case 65536:
		return 43696;
	}

	die_xfs("invalid blocksize!");
	return -1;
}

/**
 * fd fürs bitmap
 * agf für aktuellen agno
 * block ist relativ zum ag
 */
static void read_btree(int fd, struct xfs_agf *agf, off_t block)
{
	struct fsinfo *fsi = io_getcookie(fd);
	struct xfs_btree *bt;
	u8 *x = mem_alloc(XFS_MAX_BSIZE);
	u16 record;
	off_t off, ag_off;

	debug(1, 1, "read_btree(block:%U)", block);

	ag_off = be32_to_cpu(agf->agf_seqno) * be32_to_cpu(sb.sb_agblocks);
	off = (ag_off + block) * fsi->bsize;
	pread_loop2(fd, x, XFS_MAX_BSIZE, off);

	bt = (struct xfs_btree *)x;

#if 0
	term_print(" %17s%x\n", "bb_magic:", be32_to_cpu(bt->bb_magic),
		   XFS_ABTB_MAGIC);
	term_print(" %17s%u\n", "bb_level:", (u32) be16_to_cpu(bt->bb_level));
	term_print(" %17s%u\n", "bb_numrecs:",
		   (u32) be16_to_cpu(bt->bb_numrecs));
	term_print(" %17s%u\n", "bb_leftsib:", be32_to_cpu(bt->bb_leftsib));
	term_print(" %17s%u\n", "bb_rightsib:", be32_to_cpu(bt->bb_rightsib));
	term_flush();

	u32 bb_magic;		/*  4 byte */
	u16 bb_level;		/*  6 byte */
	u16 bb_numrecs;		/*  8 byte */
	u32 bb_leftsib;		/* 12 byte */
	u32 bb_rightsib;	/* 16 byte */
#endif

	/* check bb magic */
	if (be32_to_cpu(bt->bb_magic) != XFS_ABTB_MAGIC) {
		die_xfs("XFS_ABTB_MAGIC");
	}

	/* check for max btree level */
	if (be16_to_cpu(bt->bb_level) > XFS_BTREE_MAXLEVELS) {
		die_xfs("bblevel > XFS_BTREE_MAXLEVELS");
	}

	/* node */
	if (be16_to_cpu(bt->bb_level) > 0) {
		off = bno_record_offset(fsi->bsize);

		for (record = 0; record < be16_to_cpu(bt->bb_numrecs); record++) {
			u32 *b = (u32 *) (x + off);

			read_btree(fd, agf, be32_to_cpu(*b));
			off += sizeof(u32);
		}

		mem_free(x);
		return;
	}

	/* leaf */
	off = 16;
	for (record = 0; record < be16_to_cpu(bt->bb_numrecs); record++) {
		struct xfs_alloc_rec *ar = (struct xfs_alloc_rec *)(x + off);

		bitmap_clear_bits(fsi->bitmap,
				  ag_off + be32_to_cpu(ar->ar_startblock),
				  be32_to_cpu(ar->ar_blockcount));
		off += 8;
	}

	mem_free(x);
	return;
}

/**
 * fs_debug
 */
static int fs_debug(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	u64 agblocks, agcount, ag, sectsize;
	u64 sum_agblocks, sum_agfree;
	u8 *x;

	if (fs_info(fd) == -1)
		return -1;

	x = mem_alloc(XFS_MAX_BSIZE * 12);
	agcount = be32_to_cpu(sb.sb_agcount);
	agblocks = be32_to_cpu(sb.sb_agblocks);
	sectsize = be16_to_cpu(sb.sb_sectsize);

	term_print("\n\n");
	term_print("%20s%s\n", "filesystem type:", fs_name);
	term_print("%20s%s\n", "volume name:", sb.sb_fname);
	term_print("%20s0x%x\n", "fs super magic:",
		   be32_to_cpu(sb.sb_magicnum));
	term_print("%20s%u\n", "fs block size:", fsi->bsize);
	term_print("%20s%u\n", "sector size:", be16_to_cpu(sb.sb_sectsize));
	term_print("%20s%u\n", "sb version:", be16_to_cpu(sb.sb_versionnum));

	term_print("%20s%U\n", "log start:", be64_to_cpu(sb.sb_logstart));
	term_print("%20s%u\n", "log length:", be32_to_cpu(sb.sb_logblocks));

	term_print("%20s%u\n", "sb_rextsize:", be32_to_cpu(sb.sb_rextsize));

	term_print("\n");
	term_print("%20s%U\n", "fs # of blocks:", be64_to_cpu(sb.sb_dblocks));
	term_print("%20s%U\n", "fs # used blocks:",
		   be64_to_cpu(sb.sb_dblocks) - be64_to_cpu(sb.sb_fdblocks));
	term_print("%20s%U\n", "fs # free blocks:",
		   be64_to_cpu(sb.sb_fdblocks));

	term_print("\n");
	term_print("%20s%U\n", "fs # used inodes:", be64_to_cpu(sb.sb_icount));
	term_print("%20s%U\n", "fs # free inodes:", be64_to_cpu(sb.sb_ifree));

	term_print("\n");
	term_print("%20s%U\n", "# of ag's:", agcount);

	/* for each ag ... */
	sum_agblocks = 0;
	sum_agfree = 0;
	for (ag = 0; ag < agcount; ag++) {
		u64 ag_off = agblocks * fsi->bsize * ag;
		struct xfs_sb *agsb;
		struct xfs_agf *agf;

		/* read first 64K of the current ag infos */
		/* term_print("ag[%u]: %U\n", (u32)ag, (u64)ag_off); */
		pread_loop2(fd, x, XFS_MAX_BSIZE, ag_off);

		agsb = (struct xfs_sb *)x;
		agf = (struct xfs_agf *)(x + sectsize);

		/* check fs sb magic */
		if (be32_to_cpu(agsb->sb_magicnum) != XFS_SB_MAGIC) {
			die_xfs("XFS_SB_MAGIC");
		}

		/* check fs agf magic */
		if (be32_to_cpu(agf->agf_magicnum) != XFS_AGF_MAGIC) {
			die_xfs("XFS_AGF_MAGIC");
		}

		/* version check... */
		if (be32_to_cpu(agf->agf_versionnum) != 1) {
			die_xfs("agf version != 1");
		}

		term_print("\nag block %u/%u:\n",
			   be32_to_cpu(agf->agf_seqno) + 1, agcount);

		term_print(" %17s%U\n", "# of blocks:",
			   be32_to_cpu(agf->agf_length));
		term_print(" %17s%U\n", "# used blocks:",
			   be32_to_cpu(agf->agf_length) -
			   be32_to_cpu(agf->agf_freeblks));
		term_print(" %17s%U\n", "# free blocks:",
			   be32_to_cpu(agf->agf_freeblks));

		sum_agblocks += be32_to_cpu(agf->agf_length);
		sum_agfree += be32_to_cpu(agf->agf_freeblks);

		term_print(" %17s%U\n", "flfirst:",
			   be32_to_cpu(agf->agf_flfirst));
		term_print(" %17s%U\n", "fllast:",
			   be32_to_cpu(agf->agf_fllast));
		term_print(" %17s%U\n", "flcount:",
			   be32_to_cpu(agf->agf_flcount));
	}

	term_print("\n");
	term_print("%20s%U\n", "fs # of blocks:", be64_to_cpu(sb.sb_dblocks));
	term_print("%20s%U\n", "fs # free blocks:",
		   be64_to_cpu(sb.sb_fdblocks));
	term_print("%20s%U\n", "fs # used blocks:",
		   be64_to_cpu(sb.sb_dblocks) - be64_to_cpu(sb.sb_fdblocks));

	term_print("\n");
	term_print("%20s%U\n", "sum_agblocks:", sum_agblocks);
	term_print("%20s%U\n", "sum_agfree:", sum_agfree);
	term_print("%20s%U\n", "sum_agused:", sum_agblocks - sum_agfree);

	term_printflush("\n");
	mem_free(x);

	return 0;
}

/**
 * fs_getbitmap
 */
static int fs_getbitmap(int fd)
{
	struct fsinfo *fsi = io_getcookie(fd);
	u64 agblocks, agcount, ag, sectsize;
	u8 *x;

	if (fs_info(fd) == -1)
		return -1;

	x = mem_alloc(XFS_MAX_BSIZE);

	/* fsi->bitmaplen is set in fs_info() */
	fsi->bitmap = mem_talloc(u8, fsi->bitmaplen + 8);

	/* mark all as used */
	memset(fsi->bitmap, 0xff, fsi->bitmaplen);

	agcount = be32_to_cpu(sb.sb_agcount);
	agblocks = be32_to_cpu(sb.sb_agblocks);
	sectsize = be16_to_cpu(sb.sb_sectsize);

	debug(1, 1, "XFS start (%s)", fsi->path);

	/* for each ag we count the FREE bits now ... */
	for (ag = 0; ag < agcount; ag++) {
		u64 ag_off = agblocks * fsi->bsize * ag;
		struct xfs_sb *agsb;
		struct xfs_agf *agf;

		/* read first 64K of the current ag infos */
		pread_loop2(fd, x, XFS_MAX_BSIZE, ag_off);
		debug(1, 1, "read_agg(ag_off:%U)", ag_off);

		agsb = (struct xfs_sb *)x;
		agf = (struct xfs_agf *)(x + sectsize);

		/* check fs sb magic */
		if (be32_to_cpu(agsb->sb_magicnum) != XFS_SB_MAGIC) {
			die_xfs("XFS_SB_MAGIC");
		}

		/* check fs agf magic */
		if (be32_to_cpu(agf->agf_magicnum) != XFS_AGF_MAGIC) {
			die_xfs("XFS_AGF_MAGIC");
		}

		/* version check... */
		if (be32_to_cpu(agf->agf_versionnum) != 1) {
			die_xfs("agf version != 1");
		}

		if (be32_to_cpu(agf->agf_flcount) > 128) {
			die_xfs("agf flcount > 128");
		}

		/* mark the free list blocks free */
		if (be32_to_cpu(agf->agf_flcount) != 0) {
			read_agfl(fd, agf, ag_off);
		}

		/* scan the b+tree for free blocks */
		read_btree(fd, agf, be32_to_cpu(agf->agf_bno_offset));
	}

	mem_free(x);
	return 0;
}

struct file_system xfs_fs = {
	.id = FSID_XFS,
	.module = fs_name,
	.fs_test = fs_test,
	.fs_info = fs_info,
	.fs_getbitmap = fs_getbitmap,
	.fs_debug = fs_debug
};
