
/**
 * Copyright © 2013 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* libowfat */
#include "str.h"
#include "stralloc.h"
#include "byte.h"

/* frisbee-ng */
#include "defs.h"
#include "fs.h"
#include "term.h"
#include "crypt.h"
#include "mem.h"
#include "get_cycles.h"
#include "bswap.h"
#include "compr.h"
#include "wrapper.h"
#include "fmtX.h"
#include "image.h"
#include "main.h"
#include "strallocX.h"
#include "bitmap.h"
#include "openX.h"
#include "workq.h"

/**
 * magic bytes in image
 */
const char *fng_image_fng1 = "FNG1";
const char *fng_image_img1 = "IMG1";
const char *fng_image_img2 = "IMG2";
const char *fng_image_dsk1 = "DSK1";
const char *fng_image_dsk2 = "DSK2";
const char *fng_image_prt1 = "PRT1";
const char *fng_image_prt2 = "PRT2";
const char *fng_image_chk1 = "CHK1";
const char *fng_image_chk2 = "CHK2";
const char *fng_image_idx1 = "IDX1";
const char *fng_image_idx2 = "IDX2";

#define MAGIC_FNG1	0x31474e46
#define MAGIC_IMG1	0x31474d49
#define MAGIC_IMG2	0x32474d49
#define MAGIC_DSK1	0x314b5344
#define MAGIC_DSK2	0x324b5344
#define MAGIC_PRT1	0x31545250
#define MAGIC_PRT2	0x32545250
#define MAGIC_CHK1	0x314b4843
#define MAGIC_CHK2	0x324b4843

/* don't mix the writings to our images */
static pthread_mutex_t image_mutex = PTHREAD_MUTEX_INITIALIZER;
static u64 image_size = 0;

static void img_create_disk_header(struct img_disk *disk, int diskid);
static void img_create_partition_header(struct img_disk *disk, int diskid,
					int partid);
static void img_create_chunks(struct img_disk *disk, int diskid, int partid,
			      int partnum);
static void img_create_index(struct img_disk *disk, int diskid, int partid,
			      int partnum);
static void img_compress(struct img_image *img, stralloc * sa, char *ctype,
			 struct fng_compr *c);
static void image_read_fatal(off_t off);

static void image_read_fatal(off_t off)
{
	term_print("offset: %U\n", off);
	fatal("Wrong bits @ Frisbee-NG Image :/");
}

static void img_compress(struct img_image *img, stralloc * sa, char *ctype,
			 struct fng_compr *c)
{
	static struct compr_method *cm = 0;
	static char *saved_type = 0;
	static u32 methods;
	u32 i;

	/* no option parsing, when the same type is used again and again */
	if (saved_type != ctype) {
		mem_free(cm);
		methods = find_compression(ctype, &cm);
		saved_type = ctype;
	}

	c->clen = 0;
	c->cbuf = 0;
	c->ubuf = sa->s;
	c->ulen = sa->len;

	for (i = 0; i < methods; i++) {

		/* will allocate one new cbuf */
		cm[i].compress(c, cm[i].options);

		/* first method has to set the data id 0x01 */
		if (i == 0)
			compr_set_fnghdr_last(c->cbuf);

		/* free old compressed data, but not the source stralloc */
		if (c->ubuf != sa->s)
			mem_free(c->ubuf);

		/**
		 * next uncompressed buffer is the currently new
		 * compressed one... clen only represents the
		 * compressed data part, not the hdr... */
		c->ulen = c->clen + FNGHDR;
		c->ubuf = c->cbuf;
	}

	if (img->cfg->password.len) {
		c->ubuf = c->cbuf;
		c->ulen = c->clen + FNGHDR;
		compress_encrypt_data(c);
	}

	/* cm is left allocated ... */
	return;
}

/**
 * create frisbee image
 */
void image_create(struct img_image *img)
{
	struct fng_compr *c = img->c;
	stralloc *hdr = mem_talloc(stralloc, 1);
	stralloc *part2 = mem_talloc(stralloc, 1);
	u32 i, partnum, seed;
	char *x;

	/* top status line, show info what is done */
	if (!stralloc_cats(&term_left_status_top, ""))
		die_nomem();
	x = img->cfg->source.s;
	for (i = 0; i < img->cfg->source_count; i++) {
		if (!stralloc_cats(&term_left_status_top, x))
			die_nomem();
		if (!stralloc_cats(&term_left_status_top, " "))
			die_nomem();
		x += str_len(x) + 1;
	}

	if (!stralloc_cats(&term_left_status_top, "-> "))
		die_nomem();

	x = img->cfg->images.s;
	for (i = 0; i < img->cfg->image_count; i++) {
		if (!stralloc_cats(&term_left_status_top, x))
			die_nomem();
		if (!stralloc_cats(&term_left_status_top, " "))
			die_nomem();
		x += str_len(x) + 1;
	}

#if 0
	/**
	 * image hdr, part2 (compressed / encrypted)
	 */
	u32 img_magic2;		/* IMG2 magic */
	u32 disks;		/* number of disks in image */
	u32 partitions;		/* number of partitions in image */
	u32 lshw_len;		/* length of compressed "lshw -short" */
	u32 comment_len;	/* length of compressed short description of image */
	u32 seed;		/* seed used for xxhash */
#endif

	/* 6x 4byte, part2 of frisbee image */
	stralloc_init(part2);
	if (!stralloc_copys(part2, fng_image_img2))
		die_nomem();
	if (!stralloc_cat_le32(part2, img->disks))
		die_nomem();
	if (!stralloc_cat_le32(part2, img->partitions))
		die_nomem();
	if (!stralloc_cat_le32(part2, img->cfg->lshw.len))
		die_nomem();
	if (!stralloc_cat_le32(part2, img->cfg->comment.len))
		die_nomem();
	if (!stralloc_cat_le32(part2, c->seed))
		die_nomem();
	if (!stralloc_cat(part2, &img->cfg->lshw))
		die_nomem();
	if (!stralloc_cat(part2, &img->cfg->comment))
		die_nomem();

	/* frisbee image header has seed=0 */
	seed = c->seed;
	c->seed = 0;
	img_compress(img, part2, img->cfg->compr.s, c);
	c->seed = seed;
	stralloc_free(part2);
	mem_free(part2);

#if 0
	/**
	 * image hdr, part1
	 */
	u32 fng_magic1;		/* FNG1 magic */
	u32 img_magic1;		/* IMG1 magic */
	u32 version;		/* frisbee version used */
	u64 ctime;		/* image timestamp */

	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */
#endif
	stralloc_init(hdr);
	if (!stralloc_cats(hdr, fng_image_fng1))
		die_nomem();
	if (!stralloc_cats(hdr, fng_image_img1))
		die_nomem();
	if (!stralloc_cat_le32(hdr, 1))
		die_nomem();
	if (!stralloc_cat_le64(hdr, time(0)))
		die_nomem();
	if (!stralloc_catb(hdr, c->cbuf, c->clen + FNGHDR))
		die_nomem();
	mem_free(c->cbuf);

	/* will call fatal() if sth. goes bad */
	for (i = 0; i < img->cfg->image_count; i++) {
		write_loop2(img->image_fds[i], hdr->s, hdr->len);
	}
	image_size += hdr->len;

	/* allocate index for the chunk location information */
	img->index = mem_talloc(struct fng_index, img->partitions);

	/* free temporary allocated resources */
	stralloc_free(hdr);

	/* create disk header(img) */
	for (i = 0; i < img->disks; i++) {
		img_create_disk_header(&img->disk[i], i);
	}

	/* create partition header(img) */
	for (i = 0; i < img->disks; i++) {
		u32 p;

		for (p = 0; p < img->disk[i].tlen; p++) {
			img_create_partition_header(&img->disk[i], i, p);
		}
	}

	/* create chunks */
	partnum = 0;
	for (i = 0; i < img->disks; i++) {
		u32 p;

		for (p = 0; p < img->disk[i].tlen; p++) {
			img_create_chunks(&img->disk[i], i, p, partnum);
			partnum++;
		}
	}

	/* index offset, will be the last 8 byte in image */
	stralloc_init(hdr);
	if (!stralloc_cat_le64(hdr, image_size))
		die_nomem();

	/* create chunk index */
	partnum = 0;
	for (i = 0; i < img->disks; i++) {
		u32 p;

		for (p = 0; p < img->disk[i].tlen; p++) {
			img_create_index(&img->disk[i], i, p, partnum);
			partnum++;
		}
	}

	/* will call fatal() if sth. goes bad */
	for (i = 0; i < img->cfg->image_count; i++) {
		write_loop2(img->image_fds[i], hdr->s, hdr->len);
	}
	image_size += hdr->len;

	stralloc_free(hdr);
	mem_free(hdr);

	return;
}

/**
 * create disk header
 */
static void img_create_disk_header(struct img_disk *disk, int diskid)
{
	struct img_image *img = disk->img;
	struct fng_compr *c = img->c;
	stralloc *hdr = mem_talloc(stralloc, 1);
	stralloc *part2 = mem_talloc(stralloc, 1);
	u32 i;

#if 0
	/**
	 * disk hdr, part2 (compressed / encrypted)
	 */
	u32 dsk_magic2;		/* DSK2 magic */
	u32 disk;		/* disk number in image */
	u32 tabletype;		/* dos, gpt, bsd, ... */
	u32 partitions;		/* number of saved partitions in this image */
	u32 bootrec_pairs;	/* bootrecord data pairs in part3 */
	u64 disksize;		/* minimum size of destination harddisk in bytes */
#endif

	/* 6x 4byte, part2 of frisbee image */
	stralloc_init(part2);
	if (!stralloc_copys(part2, fng_image_dsk2))
		die_nomem();
	if (!stralloc_cat_le32(part2, diskid))
		die_nomem();
	if (!stralloc_cat_le32(part2, disk->ptype))
		die_nomem();
	if (!stralloc_cat_le32(part2, disk->tlen))
		die_nomem();
	if (!stralloc_cat_le32(part2, disk->pdlen))
		die_nomem();
	if (!stralloc_cat_le64(part2, disk->devsize))
		die_nomem();

	for (i = 0; i < disk->pdlen; i++) {
		if (!stralloc_cat_le64(part2, disk->pd[i].start))
			die_nomem();
		if (!stralloc_cat_le32(part2, disk->pd[i].data.len))
			die_nomem();
		if (!stralloc_catb
		    (part2, disk->pd[i].data.s, disk->pd[i].data.len))
			die_nomem();
	}

	img_compress(disk->img, part2, img->cfg->compr.s, c);
	stralloc_free(part2);
	mem_free(part2);

#if 0
	/**
	 * disk hdr, part1
	 */
	u32 dsk_magic1;		/* DSK1 magic */

	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */
#endif

	stralloc_init(hdr);
	if (!stralloc_cats(hdr, fng_image_dsk1))
		die_nomem();
	if (!stralloc_catb(hdr, c->cbuf, c->clen + FNGHDR))
		die_nomem();
	mem_free(c->cbuf);

	/* will call fatal() if sth. goes bad */
	for (i = 0; i < img->cfg->image_count; i++) {
		write_loop2(img->image_fds[i], hdr->s, hdr->len);
	}
	image_size += hdr->len;

	/* free temporary allocated resources */
	stralloc_free(hdr);
	mem_free(hdr);
}

/**
 * create partition header
 */
static void img_create_partition_header(struct img_disk *disk, int diskid,
					int partid)
{
	struct img_image *img = disk->img;
	struct fng_compr *c = disk->img->c;
	struct partition *p = &disk->todo[partid];
	stralloc *hdr = mem_talloc(stralloc, 1);
	stralloc *part2 = mem_talloc(stralloc, 1);
	u32 i;

#if 0
	/**
	 * partition hdr, part2 (compressed / encrypted)
	 */
	u32 prt_magic2;		/* PRT2 magic */
	u32 disk;		/* disk number in this image */
	u32 index;		/* index of partition @ disk */
	u32 number;		/* number of this partition to show (1,2,3...) */
	u32 bsize;		/* blocksize -> mostly 4096 */
	u32 blks_chunk;		/* saved blocks per chunk */

	u64 start;		/* starting offset in bytes (0..2^64) */
	u64 partsize;		/* size in bytes (0..2^64) */
	u64 blocks;		/* total number of blocks on this partition */
	u64 bused;		/* stored blocks */
	u64 maplen;		/* bitmap length in bytes */
#endif

	stralloc_init(part2);
	if (!stralloc_copys(part2, fng_image_prt2))
		die_nomem();
	if (!stralloc_cat_le32(part2, diskid))
		die_nomem();
	if (!stralloc_cat_le32(part2, p->index))
		die_nomem();
	if (!stralloc_cat_le32(part2, p->number))
		die_nomem();
	if (!stralloc_cat_le32(part2, p->fsi.bsize))
		die_nomem();
	if (!stralloc_cat_le32(part2, img->cfg->chunksize / p->fsi.bsize))
		die_nomem();
	if (!stralloc_cat_le64(part2, p->start))
		die_nomem();
	if (!stralloc_cat_le64(part2, p->size))
		die_nomem();
	if (!stralloc_cat_le64(part2, p->fsi.blocks))
		die_nomem();
	if (!stralloc_cat_le64(part2, p->fsi.bused))
		die_nomem();
	if (!stralloc_cat_le64(part2, p->fsi.bitmaplen))
		die_nomem();
	if (!stralloc_catb(part2, (char *)p->fsi.bitmap, p->fsi.bitmaplen))
		die_nomem();

	img_compress(disk->img, part2, img->cfg->compr.s, img->c);
	stralloc_free(part2);
	mem_free(part2);

#if 0
	/**
	 * part hdr, part1
	 */
	u32 prt_magic1;		/* PRT1 magic */
	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */
#endif

	stralloc_init(hdr);
	if (!stralloc_cats(hdr, fng_image_prt1))
		die_nomem();
	if (!stralloc_catb(hdr, c->cbuf, c->clen + FNGHDR))
		die_nomem();
	mem_free(c->cbuf);

	/* will call fatal() if sth. goes bad */
	for (i = 0; i < img->cfg->image_count; i++) {
		write_loop2(img->image_fds[i], hdr->s, hdr->len);
	}
	image_size += hdr->len;

	/* free temporary allocated resources */
	stralloc_free(hdr);
	mem_free(hdr);
}

struct todo_data {
	struct img_image *img;
	u32 disk;	/* disk X */
	u32 part;	/* partition X */
	u32 partnum; /* partition number in image */
	u32 blocks; /* blocks in chunk */
	u64 index; /* chunk number */
	u64 first_blk; /* first block in chunk */
	stralloc *data;
};

/* worker function which does the most */
static void worker_imaging(void *arg)
{
	struct todo_data *data = (struct todo_data *)arg;
	struct fng_compr *c = mem_talloc(struct fng_compr, 1);
	struct img_image *img = data->img;
	struct image_config *cfg = img->cfg;
	struct fng_index *idx = &img->index[data->partnum];
	stralloc *chunk = mem_talloc(stralloc, 1);
	stralloc *part2 = mem_talloc(stralloc, 1);
	u32 i;

#if 0
	/**
	 * chunk hdr, part2 (compressed / encrypted)
	 */
	u32 chk_magic2;		/* CHK2 magic */
	u32 disk;		/* disk the chunk belongs to */
	u32 part;		/* partition the chunk belongs to */
	u32 blocks;		/* stored blocks in this chunk */
	u32 size;		/* uncompressed chunk size */

	u64 index;		/* index of chunk (disk-part-index) */
	u64 first_blk;		/* first block (of slice) in this chunk */
#endif

	stralloc_init(part2);
	if (!stralloc_copys(part2, fng_image_chk2))
		die_nomem();
	if (!stralloc_cat_le32(part2, data->disk))
		die_nomem();
	if (!stralloc_cat_le32(part2, data->part))
		die_nomem();
	if (!stralloc_cat_le32(part2, data->blocks))
		die_nomem();
	if (!stralloc_cat_le32(part2, data->data->len))
		die_nomem();

	if (!stralloc_cat_le64(part2, data->index))
		die_nomem();
	if (!stralloc_cat_le64(part2, data->first_blk))
		die_nomem();
	if (!stralloc_cat(part2, data->data))
		die_nomem();

	/* init the compression state */
	c->seed = img->c->seed;
	if (cfg->password.len) {
		c->salt = img->c->salt;
		c->key = img->c->key;
		c->nonce = img->c->nonce;
	}

	img_compress(img, part2, cfg->compr.s, c);
	stralloc_free(part2);
	mem_free(part2);

#if 0
	/**
	 * chunk hdr, part1
	 */
	u32 chk_magic1;		/* CHK1 magic */
	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */
#endif

	stralloc_init(chunk);
	if (!stralloc_cats(chunk, fng_image_chk1))
		die_nomem();
	if (!stralloc_catb(chunk, c->cbuf, c->clen + FNGHDR))
		die_nomem();
	mem_free(c->cbuf);

	/* will call fatal() if sth. goes bad */
	pthread_mutex_lock(&image_mutex);
	for (i = 0; i < cfg->image_count; i++) {
		write_loop2(img->image_fds[i], chunk->s, chunk->len);
	}
	image_size += chunk->len;

	/* old image_size = offset of current chunk */
	idx->chunkmap[data->index - 1] = image_size;
	idx->chunklen[data->index - 1] = chunk->len;
	pthread_mutex_unlock(&image_mutex);

	/* free temporary allocated resources */
	stralloc_free(chunk);
	mem_free(chunk);
	mem_free(c);
	stralloc_free(data->data);
	mem_free(data->data);
	mem_free(data);

	return;
}

/**
 * create chunks of some partition
 */
static void img_create_chunks(struct img_disk *disk, int diskid, int partid,
			      int partnum)
{
	struct img_image *img = disk->img;
	struct partition *p = &disk->todo[partid];
	struct image_config *cfg = img->cfg;
	struct fsinfo *fsi = &p->fsi;
	struct progress *pdp = fsi->progress;	/* progress data cur partition */
	struct progress *pdd = img->progress;	/* progress for all disks */
	workq_t *w;
	struct term_clustermap *tcmap;
	u64 blk, index, first_blk, bytes;
	stralloc *data;
	u32 blocks, blks_chunk = cfg->chunksize / fsi->bsize;
	void *buffer = mem_alloc(fsi->bsize);
	struct fng_index *idx = &img->index[partnum];

	blk = index = first_blk = bytes = 0;
	blocks = 0;

	if (!pdp)
		fatal("transport_doit() pdp not set!");
	if (!pdd)
		fatal("transport_doit() pdd not set!");

	/* allocate chunkmap index */
	idx->chunks = (fsi->bused + blks_chunk - 1) / blks_chunk;
	idx->chunkmap = mem_talloc(u64, idx->chunks);
	idx->chunklen = mem_talloc(u32, idx->chunks);

	/* init terminal screen and show value for number of blocks per unit */
	tcmap = mem_talloc(struct term_clustermap, 1);

	term_clustermap_init(tcmap, fsi->bitmap, fsi->blocks, 2, term_maxx - 1,
			     3, term_maxy - 9);

	term_gotoxy(term_maxx - 16 - 8 - 1, term_maxy - 8);
	term_print("%8u blocks per unit", tcmap->blocks_per_unit);
	term_gotoxy(3, 2);
	term_sa(&term_left_status_top);

	/* clear old stats */
	term_gotoxy(3, term_maxy - 6);
	term_hline(term_maxx - 3, " ");
	term_gotoxy(3, term_maxy - 5);
	term_hline(term_maxx - 3, " ");
	term_gotoxy(3, term_maxy - 4);
	term_hline(term_maxx - 3, " ");
	term_flush();

	/* workq_init(int thr_max, int queue_max, int idle_timeout, fn) */
	w = workq_init(cfg->threads, 2, 2, worker_imaging);

	/* our read block data */
	data = mem_talloc(stralloc, 1);
	stralloc_init(data);
	stralloc_ready(data, cfg->chunksize);

	for (blk = 0; blk < fsi->blocks; blk++) {

		/* alle XY blöcke gibts neuen status... */
		if (blk % tcmap->blocks_per_unit == 0) {

			/* calculate stats */
			progress(pdp);
			progress(pdd);

			/* draw the map */
			term_clustermap(tcmap, blk);

	/**
	 * ├────────────────────────────────────────────────────────────┤ -7
	 * │ running time: 1:23:33, eta: 30:54:15                       │ -6
	 * │ data rate: xxxx.yMiB/s  x778.2kiB / xx28.9GiB              │ -5
	 * │ block: 1.123.123 / 123.123.123                             │ -4
	 * │ xx/xx: [=======================>                    ]  45% │ -3
	 * │ all:   [============>                               ]  33% │ -2
	 * └────────────────────────────────────────────────────────────┘ -1
	 */

			term_gotoxy(3, term_maxy - 6);
			term_print
			    ("running time: %u:%02u:%02u, eta: %u:%02u:%02u, image size: %HiB, ratio: %2u%%   ",
			     min(pdd->ctt_h, 99), pdd->ctt_m, pdd->ctt_s,
			     min(pdd->eta_h, 99), pdd->eta_m, pdd->eta_s,
			     image_size,
			     pdd->current ? image_size * 100 /
			     pdd->current : 0);

			term_gotoxy(3, term_maxy - 5);
			term_print
			    ("data rate: %HiB/s, done: %HiB / %HiB         ",
			     pdd->current / pdd->timediff, pdd->current,
			     pdd->total);

			term_gotoxy(3, term_maxy - 4);
			term_print("block: %10Z / %10Z", blk, fsi->blocks);

			/* status current partition */
			term_gotoxy(3, term_maxy - 3);
			term_print("%d/%d:", partnum + 1, img->partitions);

			term_gotoxy(10, term_maxy - 3);
			term_progress(10, term_maxx - 7, pdp->percent);

			term_gotoxy(term_maxx - 5, term_maxy - 3);
			term_print("%-3u%%", pdp->percent);

			/* overall status */
			term_gotoxy(3, term_maxy - 2);
			term_print("all:");

			term_gotoxy(10, term_maxy - 2);
			term_progress(10, term_maxx - 7, pdd->percent);

			term_gotoxy(term_maxx - 5, term_maxy - 2);
			term_print("%-3u%%", pdd->percent);

			term_flush();

		}

		if (bitmap_get(fsi->bitmap, blk)) {

			/* first block in chunk */
			if (data->len == 0) {
				/* remember first block */
				first_blk = blk;
				/* current blocks in chunk */
				blocks = 0;
				/* first index will be 1 */
				index++;
			}
			pread_loop2(disk->fd, buffer, fsi->bsize,
				    p->start + blk * fsi->bsize);
			if (!stralloc_catb(data, buffer, fsi->bsize))
				die_nomem();

			blocks++;
			progress_add(pdp, fsi->bsize);
			progress_add(pdd, fsi->bsize);
		}

		/* chunk is full, compress and write it */
		if (blocks == blks_chunk) {
			struct todo_data *td = mem_talloc(struct todo_data, 1);

			td->data = data;
			td->disk = diskid;
			td->part = partid;
			td->partnum = partnum;
			td->blocks = blocks;
			td->index = index;
			td->first_blk = first_blk;
			td->img = img;

			workq_add(w, td);

			/* allocate next chunk data */
			data = mem_talloc(stralloc, 1);
			stralloc_init(data);
			stralloc_ready(data, cfg->chunksize);
		}
	}

	{
		struct todo_data *td = mem_talloc(struct todo_data, 1);

		td->data = data;
		td->disk = diskid;
		td->part = partid;
		td->partnum = partnum;
		td->blocks = blocks;
		td->index = index;
		td->first_blk = first_blk;
		td->img = img;

		workq_add(w, td);
	}

	/* wait for worker to finish */
	workq_deinit(w);

	/* done... */
	term_clustermap(tcmap, fsi->blocks);
	term_gotoxy(1, term_maxy);
	term_flush();

	mem_free(tcmap);
	mem_free(buffer);
}

/**
 * create chunks of some partition
 */
static void img_create_index(struct img_disk *disk, int diskid, int partid,
			      int partnum)
{
	struct img_image *img = disk->img;
	struct fng_index *idx = &img->index[partnum];
	struct fng_compr *c = disk->img->c;
	stralloc *hdr = mem_talloc(stralloc, 1);
	stralloc *part2 = mem_talloc(stralloc, 1);
	u64 i;

#if 0
	/**
	 * index hdr, part2 (compressed / encrypted)
	 */
	u32 idx_magic2;		/* IDX2 magic */
	u32 disk;		/* disk the chunk belongs to */
	u32 part;		/* partition the chunk belongs to */
	u64 chunks;		/* chunks in index */

	/**
	 * chunk hdr, part3 (compressed / encrypted)
	 * 8 byte offset + 4 byte length for each chunk in that disk/part
	 */
	u64 *chunkmap;
	u32 *chunklen;
#endif

	stralloc_init(part2);
	if (!stralloc_copys(part2, fng_image_idx2))
		die_nomem();
	if (!stralloc_cat_le32(part2, diskid))
		die_nomem();
	if (!stralloc_cat_le32(part2, partid))
		die_nomem();
	if (!stralloc_cat_le64(part2, idx->chunks))
		die_nomem();

	for (i = 0; i < idx->chunks; i++) {
		if (!stralloc_cat_le64(part2, idx->chunkmap[i]))
			die_nomem();
		if (!stralloc_cat_le32(part2, idx->chunklen[i]))
			die_nomem();
	}

	img_compress(disk->img, part2, img->cfg->compr.s, img->c);
	stralloc_free(part2);
	mem_free(part2);

#if 0
	/**
	 * index hdr, part1
	 */
	u32 idx_magic1;		/* IDX1 magic */

	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */
#endif

	stralloc_init(hdr);
	if (!stralloc_cats(hdr, fng_image_idx1))
		die_nomem();
	if (!stralloc_catb(hdr, c->cbuf, c->clen + FNGHDR))
		die_nomem();
	mem_free(c->cbuf);

	/* will call fatal() if sth. goes bad */
	for (i = 0; i < img->cfg->image_count; i++) {
		write_loop2(img->image_fds[i], hdr->s, hdr->len);
	}
	image_size += hdr->len;

	/* free temporary allocated resources */
	stralloc_free(hdr);
	mem_free(hdr);

}

/**
 * read metadata from some frisbee image into memory
 */
static void image_read_metadata(struct img_restore *ir)
{
	stralloc *sa = mem_talloc(stralloc, 1);
	struct restore_config *cfg = ir->cfg;
	struct fng_compr *c = mem_talloc(struct fng_compr, 1);
	struct fng_image *img = mem_talloc(struct fng_image, 1);
	struct fng_disk *disk = 0;
	struct fng_part *part = 0;
	struct fng_chunk *chunk;
	struct fng_index *idx;
	void *data;
	u64 offset = 0;
	u32 *u, i, j;
	u64 *U;

	/* 1) read image header */
	pread_loop2(cfg->fd, img, FNG_IMAGE_HDR_P1 + FNGHDR, offset);

	/* change endian if needed */
	img->fng_magic1 = le32_to_cpu(img->fng_magic1);
	img->img_magic1 = le32_to_cpu(img->img_magic1);
	img->version = le32_to_cpu(img->version);
	img->ctime = le64_to_cpu(img->ctime);
	img->cflags = le32_to_cpu(img->cflags);
	img->p2clen = le32_to_cpu(img->p2clen);
	img->p2ulen = le32_to_cpu(img->p2ulen);
	img->crc = le32_to_cpu(img->crc);

	if (MAGIC_FNG1 != img->fng_magic1)
		image_read_fatal(offset);
	if (MAGIC_IMG1 != img->img_magic1)
		image_read_fatal(offset + 4);

	/* read the complete compressed part2 */
	offset = FNG_IMAGE_HDR_P1;
	stralloc_init(sa);
	pread_sa(cfg->fd, sa, img->p2clen + FNGHDR, offset);

	compr_init(c);
	if (cfg->password.len)
		compr_set_password(c, &cfg->password);

	c->clen = sa->len;
	c->cbuf = sa->s;

	/* decompress_data() will free sa->s ! */
	decompress_data(c);

	u = c->ubuf;
	img->img_magic2 = le32_to_cpu(*u);

	if (MAGIC_IMG1 != img->img_magic1)
		image_read_fatal(offset);
	offset += img->p2clen + FNGHDR;

	u = c->ubuf + 4;
	img->disks = le32_to_cpu(*u);
	u = c->ubuf + 4 * 2;
	img->partitions = le32_to_cpu(*u);
	u = c->ubuf + 4 * 3;
	img->lshw_len = le32_to_cpu(*u);
	u = c->ubuf + 4 * 4;
	img->comment_len = le32_to_cpu(*u);
	u = c->ubuf + 4 * 5;
	img->seed = le32_to_cpu(*u);

	if (img->disks)
		disk = mem_talloc(struct fng_disk, img->disks);

	if (img->partitions)
		part = mem_talloc(struct fng_part, img->partitions);

	stralloc_init(&ir->lshw);
	data = c->ubuf + c->ulen - img->comment_len - img->lshw_len;
	if (img->lshw_len)
		stralloc_copyb(&ir->lshw, data, img->lshw_len);

	stralloc_init(&ir->comment);
	data = c->ubuf + c->ulen - img->comment_len;
	if (img->comment_len)
		stralloc_copyb(&ir->comment, data, img->comment_len);
	/* free uncompressed buffer */
	mem_free(c->ubuf);

	/* setup the first pointer for the caller */
	ir->c = c;
	ir->img = img;
	ir->disk = disk;
	ir->part = part;
	ir->disks = img->disks;
	ir->partitions = img->partitions;

	/* all other header need this seed */
	c->seed = img->seed;

	for (i = 0; i < img->disks; i++) {
		struct fng_disk *d = &disk[i];

		pread_loop2(cfg->fd, d, 4 + FNGHDR, offset);
		offset += 4;
		d->dsk_magic1 = le32_to_cpu(d->dsk_magic1);
		if (MAGIC_DSK1 != d->dsk_magic1)
			image_read_fatal(offset);
		d->cflags = le32_to_cpu(d->cflags);
		d->p2clen = le32_to_cpu(d->p2clen);
		d->p2ulen = le32_to_cpu(d->p2ulen);
		d->crc = le32_to_cpu(d->crc);

		stralloc_init(sa);
		pread_sa(cfg->fd, sa, d->p2clen + FNGHDR, offset);

		/* read the complete compressed part2 */
		c->clen = sa->len;
		c->cbuf = sa->s;

		/* decompress_data() will free sa->s ! */
		decompress_data(c);

		u = c->ubuf;
		d->dsk_magic2 = le32_to_cpu(*u);
		if (MAGIC_DSK2 != d->dsk_magic2)
			image_read_fatal(offset);
		offset += d->p2clen + FNGHDR;

		u = c->ubuf + 4;
		d->disk = le32_to_cpu(*u);
		u = c->ubuf + 2 * 4;
		d->tabletype = le32_to_cpu(*u);
		u = c->ubuf + 3 * 4;
		d->partitions = le32_to_cpu(*u);
		u = c->ubuf + 4 * 4;
		d->bootrec_pairs = le32_to_cpu(*u);
		U = c->ubuf + 5 * 4;
		d->disksize = le64_to_cpu(*U);

		/*
		   debug(1,1,"read diskid=%u", d->disk);
		   debug(1,1,"read tabletype=%u", d->tabletype);
		   debug(1,1,"read partitions=%u", d->partitions);
		   debug(1,1,"read bootrec_pairs=%u", d->bootrec_pairs);
		   debug(1,1,"read disksize=%U", d->disksize);
		   {
		   int fd=open_device_rw("ubuf");
		   write(fd, c->ubuf, c->ulen);
		   close(fd);
		   }
		 */

		/**
		 * get the bootrec_pairs
		 * u64 start / u32 length / u8 data[length]
		 */
		d->pd = mem_talloc(struct ptable_data, d->bootrec_pairs);

		data = c->ubuf + 5 * 4 + 8;
		for (j = 0; j < d->bootrec_pairs; j++) {
			struct ptable_data *pd = &d->pd[j];
			u32 len;

			U = data;
			pd->start = le64_to_cpu(*U);
			u = data + 8;
			len = le32_to_cpu(*u);
			stralloc_init(&pd->data);
			if (!stralloc_copyb(&pd->data, data + 8 + 4, len))
				die_nomem();
			/* next entry */
			data = data + 8 + 4 + len;
		}

		/* free uncompressed buffer */
		mem_free(c->ubuf);
	}

	for (i = 0; i < img->partitions; i++) {
		struct fng_part *p = &part[i];

		pread_loop2(cfg->fd, p, 4 + FNGHDR, offset);
		p->prt_magic1 = le32_to_cpu(p->prt_magic1);
		if (MAGIC_PRT1 != p->prt_magic1)
			image_read_fatal(offset);
		offset += 4;
		p->cflags = le32_to_cpu(p->cflags);
		p->p2clen = le32_to_cpu(p->p2clen);
		p->p2ulen = le32_to_cpu(p->p2ulen);
		p->crc = le32_to_cpu(p->crc);

		stralloc_init(sa);
		pread_sa(cfg->fd, sa, p->p2clen + FNGHDR, offset);

		/* read the complete compressed part2 */
		c->clen = sa->len;
		c->cbuf = sa->s;

		/* decompress_data() will free sa->s ! */
		decompress_data(c);

		u = c->ubuf;
		p->prt_magic2 = le32_to_cpu(*u);
		if (MAGIC_PRT2 != p->prt_magic2)
			image_read_fatal(offset);
		offset += p->p2clen + FNGHDR;

		u = c->ubuf + 4;
		p->disk = le32_to_cpu(*u);
		u = c->ubuf + 4 * 2;
		p->index = le32_to_cpu(*u);
		u = c->ubuf + 4 * 3;
		p->number = le32_to_cpu(*u);
		u = c->ubuf + 4 * 4;
		p->bsize = le32_to_cpu(*u);
		u = c->ubuf + 4 * 5;
		p->blks_chunk = le32_to_cpu(*u);

		U = c->ubuf + 4 * 6;
		p->start = le64_to_cpu(*U);
		U = c->ubuf + 4 * 6 + 1 * 8;
		p->partsize = le64_to_cpu(*U);
		U = c->ubuf + 4 * 6 + 2 * 8;
		p->blocks = le64_to_cpu(*U);
		U = c->ubuf + 4 * 6 + 3 * 8;
		p->bused = le64_to_cpu(*U);
		U = c->ubuf + 4 * 6 + 4 * 8;
		p->maplen = le64_to_cpu(*U);

		p->bitmap = mem_talloc(u8, p->maplen);
		byte_copy(p->bitmap, p->maplen, c->ubuf + 4 * 6 + 5 * 8);

		/* free uncompressed buffer */
		mem_free(c->ubuf);
	}

	/* calculate number of chunks in image */
	ir->chunks = 0;
	for (i = 0; i < img->partitions; i++) {
		struct fng_part *p = &part[i];

		p->chunks = (p->bused + p->blks_chunk - 1) / p->blks_chunk;
		/* build an index for fast access later */
		p->chunkmap = mem_talloc(u64, p->chunks);
		p->chunklen = mem_talloc(u32, p->chunks);
		ir->chunks += p->chunks;
	}

	/* XXX, if this is not fast enough, we have to create some index */
	#if 0
	chunk = mem_tzalloc(struct fng_chunk, 1);
	idx = mem_talloc(struct fng_index, 1);

	for (i = 0; i < img->partitions; i++) {
		struct fng_part *p = &part[i];
		u32 j;

		for (j = 0; j < p->chunks; j++) {
			/* build a simple offset+len index of alll chunks */
			p->chunkmap[j] = offset;

			pread_loop2(cfg->fd, chunk, 4 + FNGHDR, offset);
			chunk->chk_magic1 = le32_to_cpu(chunk->chk_magic1);
			/* we read each chunk and check it magic */
			if (MAGIC_CHK1 != chunk->chk_magic1)
				image_read_fatal(offset);
			chunk->p2clen = le32_to_cpu(chunk->p2clen);
			p->chunklen[j] = chunk->p2clen;
			/* when everything is fine, goto next */
			offset += 4 + chunk->p2clen + FNGHDR;
		}
	}
	mem_free(chunk);
	#endif

	/* clean up */
	mem_free(sa);
	return;
}

/**
 * read infos
 */
void image_info(struct restore_config *cfg)
{
	struct img_restore *ir = mem_talloc(struct img_restore, 1);
	struct fng_image *img;
	char *ctime = mem_alloc(fmt_datetime(0, 1) + 1);
	u32 i;

	/* read metadata of image */
	ir->cfg = cfg;
	image_read_metadata(ir);
	img = ir->img;

	ctime[fmt_datetime(ctime, img->ctime)] = 0;
	term_print("Version:    %8u\n", img->version);
	term_print("ctime: %s\n", ctime);
	mem_free(ctime);

	term_print("disks:      %8u\n", img->disks);
	term_print("partitions: %8u\n", img->partitions);

	if (ir->lshw.len)
		term_print("lshw data:\n%S\n", &ir->lshw);
	if (ir->comment.len)
		term_print("comment data:\n%S\n", &ir->comment);
	term_print("\n");

	for (i = 0; i < img->disks; i++) {
		struct fng_disk *d = &ir->disk[i];

		term_print("bootrec's:  %8u\n", d->bootrec_pairs);
		term_print("\n");
	}

	for (i = 0; i < img->partitions; i++) {
		struct fng_part *p = &ir->part[i];

		term_print("disk:       %8u\n", p->disk);
		term_print("index:      %8u\n", p->index);
		term_print("number:     %8u\n", p->number);
		term_print("bsize:      %8u\n", p->bsize);
		term_print("blks_chunk: %8u\n", p->blks_chunk);

		term_print("start:      %8Z\n", p->start);
		term_print("partsize:   %8Z\n", p->partsize);
		term_print("blocks:     %8Z\n", p->blocks);
		term_print("bused:      %8Z\n", p->bused);
		term_print("maplen:     %8Z\n", p->maplen);
		term_print("chunks:     %8Z\n", p->chunks);
	}

	term_print("\n");
	term_flush();

	/* XXX, free memory */
	return;
}

/**
 * show a lot info about image file
 */
void image_debug(struct restore_config *cfg)
{
	struct img_restore *ir = mem_talloc(struct img_restore, 1);
	struct fng_image *img;
	u32 i;

	/* read metadata of image */
	ir->cfg = cfg;
	image_read_metadata(ir);
	img = ir->img;

	term_print("FNG1 Magic: %12x\n", img->fng_magic1);
	term_print("IMG1 Magic: %12x\n", img->img_magic1);
	term_print("Version:    %12u\n", img->version);
	term_print("ctime:      %12U\n", img->ctime);
	term_print("cflags:     %12x\n", img->cflags);
	term_print("clen:       %12u\n", img->p2clen);
	term_print("ulen:       %12u\n", img->p2ulen);
	term_print("crc:        %12x\n", img->crc);

	term_print("\n");
	term_print("IMG2 Magic: %12x\n", img->img_magic2);
	term_print("disks:      %12u\n", img->disks);
	term_print("partitions: %12u\n", img->partitions);
	term_print("lshw:       %12u\n", img->lshw_len);
	term_print("comment:    %12u\n", img->comment_len);
	term_print("seed:       %12x\n", img->seed);

	if (ir->lshw.len)
		term_print("lshw data:\n%S\n", &ir->lshw);
	if (ir->comment.len)
		term_print("comment data:\n%S\n", &ir->comment);
	term_print("\n");

	for (i = 0; i < img->disks; i++) {
		struct fng_disk *d = &ir->disk[i];

		term_print("DSK1 Magic: %08x\n", d->dsk_magic1);
		term_print("DSK2 Magic: %08x\n", d->dsk_magic2);
		term_print("cflags:     %08x\n", d->cflags);
		term_print("clen:       %8u\n", d->p2clen);
		term_print("ulen:       %8u\n", d->p2ulen);
		term_print("crc:        %08x\n", d->crc);
		term_print("bootrec's:  %8u\n", d->bootrec_pairs);
		term_print("\n");
	}

	for (i = 0; i < img->partitions; i++) {
		stralloc sa;
		struct fng_part *p = &ir->part[i];
		u32 j;

		term_print("PRT1 Magic: %08x\n", p->prt_magic1);
		term_print("PRT2 Magic: %08x\n", p->prt_magic2);
		term_print("cflags:     %08x\n", p->cflags);
		term_print("clen:       %8u\n", p->p2clen);
		term_print("ulen:       %8u\n", p->p2ulen);
		term_print("crc:        %08x\n", p->crc);

		term_print("disk:       %8u\n", p->disk);
		term_print("index:      %8u\n", p->index);
		term_print("number:     %8u\n", p->number);
		term_print("bsize:      %8u\n", p->bsize);
		term_print("blks_chunk: %8u\n", p->blks_chunk);

		term_print("start:      %8U\n", p->start);
		term_print("partsize:   %8U\n", p->partsize);
		term_print("blocks:     %8U\n", p->blocks);
		term_print("bused:      %8U\n", p->bused);
		term_print("maplen:     %8U\n", p->maplen);
		term_print("chunks:     %8U\n", p->chunks);
		term_print("locations:\n");
		stralloc_init(&sa);
		for (j = 0; j < p->chunks; j++) {
			if (!stralloc_cat_le64(&sa, p->chunkmap[j]))
				die_nomem();
			if (!stralloc_cat_le32(&sa, p->chunklen[j]))
				die_nomem();
			term_print("%U[%U] ", p->chunkmap[j], p->chunklen[j]);
		}
		j = open_append("index");
		write_loop2(j, sa.s, sa.len);
		close(j);
		term_print("\n\n");
	}

	term_print("\n");
	term_flush();

	/* XXX, free memory */
	return;
}

/**
 * restore one partition of image to one or more destinations
 */
void image_restore_partition(struct restore_config *cfg)
{
	struct img_restore *ir = mem_talloc(struct img_restore, 1);
	struct fng_image *img;
	u32 i;

	/* read metadata of image */
	ir->cfg = cfg;
	image_read_metadata(ir);
	img = ir->img;

	term_print("FNG1 Magic: %12x\n", img->fng_magic1);
	term_print("IMG1 Magic: %12x\n", img->img_magic1);
	term_print("Version:    %12u\n", img->version);
	term_print("ctime:      %12U\n", img->ctime);
	term_print("cflags:     %12x\n", img->cflags);
	term_print("clen:       %12u\n", img->p2clen);
	term_print("ulen:       %12u\n", img->p2ulen);
	term_print("crc:        %12x\n", img->crc);

	term_print("parts:      %12x\n", cfg->dest_count);
}

/**
 * restore one disk of image to one or more destinations
 */
void image_restore_disk(struct restore_config *cfg)
{
	struct img_restore *ir = mem_talloc(struct img_restore, 1);
	struct fng_image *img;
	u32 i;

	/* read metadata of image */
	ir->cfg = cfg;
	image_read_metadata(ir);
	img = ir->img;

	term_print("FNG1 Magic: %12x\n", img->fng_magic1);
	term_print("IMG1 Magic: %12x\n", img->img_magic1);
	term_print("Version:    %12u\n", img->version);
	term_print("ctime:      %12U\n", img->ctime);
	term_print("cflags:     %12x\n", img->cflags);
	term_print("clen:       %12u\n", img->p2clen);
	term_print("ulen:       %12u\n", img->p2ulen);
	term_print("crc:        %12x\n", img->crc);

	term_print("disks:      %12x\n", cfg->dest_count);
}
