
/**
 * Copyright © 2012 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef FNG_IMAGE_H
#define FNG_IMAGE_H

/* system */
#include <pthread.h>

/* libowfat */
#include "io.h"
#include "case.h"

/* frisbee-ng */
#include "defs.h"
#include "pt_reader.h"
#include "fs.h"

/**
 * definitions for flags within the structures of the images
 */

extern const char *fng_image_fng1;
extern const char *fng_image_img1;
extern const char *fng_image_dsk1;
extern const char *fng_image_prt1;
extern const char *fng_image_chk1;

extern const char *fng_image_img2;
extern const char *fng_image_dsk2;
extern const char *fng_image_prt2;
extern const char *fng_image_chk2;

/**
 * fng image (struct fng_hdr)
 * - holds one header and mupltiple chunks after it
 * - holds pointer to harddisks (max u32)
 * -> fng_hdr + n fng_chunks
 *
 * fng disk -> are not saved really... only partition data
 * - has type like mbr / gpt / bsd
 * - has partitions
 *
 * fng partition
 * - has some index (0 = first partition of disk)
 * - has some start (byteoffset!) and some size (also in bytes!)
 * - has some fs with block size and count
 *
 * fng chunk
 * - has some blocks/clusters stored
 * - e.g. 512 cluster of 4KiB each => 2MiB uncompressed data per chunk
 * - each chunk can be used, to restore some special parts of a disk
 *
 * fng packet, not used here, it's in packet.c ...
 * - used to split the chunks into packets for network transfer
 *
 * nonce's:
 * - nonce = 0 -> special -> fng_image header
 * - 24 bytes -> 3x 64bit -> u64 disk + u64 part + u64 chunk
 * - fng-disk header nonce: 1+0+0
 * - fng-disk1,part1 header nonce: 1+1+0
 * - fng-disk1,part1,chunk7 header nonce: 1+1+7
 */

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

struct fng_image;		/* IMG1 */
struct fng_disk;		/* DSK1 */
struct fng_part;		/* PRT1 */
struct fng_chunk;		/* CHK1 */

/**
 * superblock of frisbee-ng image
 */
#define FNG_IMAGE_HDR_P1   20	/* without fnghdr */
struct fng_image {

	/**
	 * image hdr, part1
	 */
	u32 fng_magic1;		/* FNG1 magic */
	u32 img_magic1;		/* IMG1 magic */
	u32 version;		/* frisbee version used */
	u64 ctime;		/* image timestamp, seconds since 1970 */

	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */

	/**
	 * image hdr, part2 (compressed / encrypted)
	 */
	u32 img_magic2;		/* IMG2 magic */
	u32 disks;		/* number of disks in image */
	u32 partitions;		/* number of partitions in image */
	u32 lshw_len;		/* length of compressed "lshw -short" */
	u32 comment_len;	/* length of compressed short description of image */
	u32 seed;		/* seed used for xxhash */

	/**
	 * image hdr, part3 (compressed / encrypted)
	 * 1) lshw data (lshw_len)
	 * 2) notes data (notes_len)
	 */

	/**
	 * next:
	 * 1) N x fng_disk
	 * or) N x fng_part, when disks=0
	 */
} __FNG_ATTR_PACKED;

/**
 * disk header
 * - each saved disk gets this header
 */
struct fng_disk {

	/**
	 * disk hdr, part1
	 */
	u32 dsk_magic1;		/* DSK1 magic */

	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */

	/**
	 * disk hdr, part2 (compressed / encrypted)
	 */
	u32 dsk_magic2;		/* DSK2 magic */
	u32 disk;		/* disk number in image */
	u32 tabletype;		/* dos, gpt, bsd, ... */
	u32 partitions;		/* number of saved partitions in this image */
	u32 bootrec_pairs;	/* bootrecord data pairs in part3 */

	u64 disksize;		/* minimum size of destination harddisk in bytes */

	/**
	 * disk hdr, part3 (compressed / encrypted)
	 * 1) bootrecord data (gpt, mbr, grub, ... variable data)
	 * - stored via offset / length pairs
	 * - bootrec_pairs contains number of these pairs:
	 *
	 * u64 start;        -> start @ disk offset
	 * u32 length;       -> length in bytes
	 * u8 data[length];  -> data bytes
	 */

	/* will be set by read_image_metadata */
	struct ptable_data *pd;

	/**
	 * next:
	 * 1) next fng_disk header
	 * 2) or: first fng_part hdr, if no next fng_disk is there
	 */
} __FNG_ATTR_PACKED;

/**
 * partition header
 * - each saved partition gets this header
 */
struct fng_part {

	/**
	 * partition hdr, part1
	 */
	u32 prt_magic1;		/* PRT1 magic */

	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */

	/**
	 * partition hdr, part2 (compressed / encrypted)
	 */
	u32 prt_magic2;		/* PRT2 magic */
	u32 disk;		/* disk number in this image */
	u32 index;		/* index of partition @ disk */
	u32 number;		/* number of this partition to show (1,2,3...) */
	u32 bsize;		/* blocksize -> mostly 4096 */
	u32 blks_chunk;		/* saved blocks per chunk */

	u64 start;		/* starting offset in bytes (0..2^64) */
	u64 partsize;		/* size in bytes (0..2^64) */
	u64 blocks;		/* total number of blocks on this partition */
	u64 bused;		/* stored blocks */
	u64 maplen;		/* bitmap length in bytes */

	/**
	 * partition hdr, part3 (compressed / encrypted)
	 * 1) bitmap data of exact length (maplen)
	 */

	/* will be set by read_image_metadata */
	u64 chunks;
	u8 *bitmap;
	u64 *chunkmap;
	u32 *chunklen;

	/**
	 * next:
	 * 1) next fng_part header
	 * 2) or: first fng_chunk, if no next fng_part is there
	 */
} __FNG_ATTR_PACKED;

/**
 * chunk header
 */
struct fng_chunk {

	/**
	 * chunk hdr, part1
	 */
	u32 chk_magic1;		/* CHK1 magic */

	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */

	/**
	 * chunk hdr, part2 (compressed / encrypted)
	 */
	u32 chk_magic2;		/* CHK2 magic */
	u32 disk;		/* disk the chunk belongs to */
	u32 part;		/* partition the chunk belongs to */
	u32 blocks;		/* stored blocks in this chunk */
	u32 size;		/* uncompressed chunk size */

	u64 index;		/* index of chunk (disk-part-index) */
	u64 first_blk;		/* first block (of slice) in this chunk */

	/**
	 * chunk hdr, part3 (compressed / encrypted)
	 * 1) chunk data (uncompressed length @ hdr->size)
	 */

	/* will be set by read_image_chunk */
	u8 *data;

	/**
	 * next:
	 * -> next fng_chunk or index
	 */
} __FNG_ATTR_PACKED;

/**
 * chunk index
 */
struct fng_index {

	/**
	 * index hdr, part1
	 */
	u32 idx_magic1;		/* IDX1 magic */

	u32 cflags;		/* cflags of this header */
	u32 p2clen;		/* compressed length of part2 + part3 */
	u32 p2ulen;		/* uncompressed length of part2 + part3 */
	u32 crc;		/* crc of full header (crc set to 0) */

	/**
	 * index hdr, part2 (compressed / encrypted)
	 */
	u32 idx_magic2;		/* IDX2 magic */
	u32 disk;		/* disk the chunk belongs to */
	u32 part;		/* partition the chunk belongs to */
	u64 chunks;		/* chunks in index */

	/**
	 * chunk hdr, part3 (compressed / encrypted)
	 * 8 byte offset + 4 byte length for each chunk in that disk/part
	 */
	u64 *chunkmap;
	u32 *chunklen;

	/**
	 * next:
	 * -> next index or end of image
	 */
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/**
 * the big in memory structure
 */
struct img_image;

/**
 * write the frisbee image header/data to fd
 */
void image_create(struct img_image *s);

/**
 * write the frisbee disk header/data to fd
 */
void image_create_disk(int fd, int eflags, int disk, u64 mindevsize, int ptype,
		       u16 partitions, struct ptable_data *pd);

/**
 * write the frisbee partition header/data to fd
 */
void image_create_part(int fd, int eflags, int disk, struct partition *prt);

/**
 * write the frisbee chunk header/data to fd
 */
void image_create_chunk(int fd, int eflags, int cflags, int disk,
			struct partition *prt);

/**
 * configuration of fng-image
 */
struct image_config {

	/**
	 * comment's which should be added to the image
	 */
	stralloc lshw;
	stralloc comment;

	/**
	 * no encryption per default
	 */
	stralloc password;

	/**
	 * Chunk Size in MiB
	 */
	u32 chunksize;

	/**
	 * compression type and options to use
	 */
	stralloc compr;

	/**
	 * source device(s) to read from
	 * /dev/sda
	 * /dev/sda,1,5
	 */
	stralloc source;
	u32 source_count;

	/**
	 * destination image(s) to write to
	 */
	stralloc images;
	u32 image_count;

	/**
	 * number of worker threads to use
	 */
	int threads;
};

/**
 * disk information for imaging
 */
struct img_disk {
	int64 fd;		/* opened source devices in read only mode */
	char *device;		/* path to source devices or files (-p / -d) */
	u64 devsize;		/* disksize or partition size in bytes */

	/* for saving boot manager data */
	struct ptable_data *pd;	/* start + len pairs */
	u32 pdlen;		/* length of *pd */

	/* all slices found on source device */
	struct slice *sl;	/* slice information of device */
	u32 slice_count;	/* number of slices in *sl */
	char *type;		/* name of partition type (gpt, bsd, mbr...) */
	int ptype;		/* id of partition type, see pt_reader.h  */

	/* tlen = partitions to clone/image */
	struct partition *todo;	/* each slice gets it's own thing here */
	u32 tlen;		/* number of todo's */

	struct img_image *img;
};

/**
 * generic handling structure for imaging disks
 */
struct img_image {
	struct image_config *cfg;	/* access to configuration */
	struct fng_compr *c;	/* temporary compression buffers */

	/* fng header infos */
	u32 disks;		/* disks in image */
	struct img_disk *disk;	/* per disk partition info */
	u32 partitions;		/* count of all imaged partitions */

	/* overall progress status */
	progress_t *progress;

	/* destination(s) */
	int64 *image_fds;	/* fd's of the imagefile(s) we will write */
	char **image_files;	/* filename(s) of image(s) */

	/* chunks in image */
	struct fng_index *index;
};

/**
 * configuration of fng-restore
 */
struct restore_config {

	/**
	 * no encryption per default
	 */
	stralloc password;

	/**
	 * _the_ device in image to restore
	 * -d 1  -> means disk 0 in image
	 * -d 2  -> means disk 1 in image
	 * - restoring more than one disk from a multidisk image is not
	 *   wanted currently -> to many failures may happen than
	 */
	u32 disk;

	/**
	 * source partition(s) to restore
	 * -p 0  -> means all saved partition of disk
	 * -p 1  -> means only partition one should be restored
	 */
	u32 partition;

	/**
	 * image(s) to read from
	 */
	stralloc image;
	int64 fd;		/* image fd */

	/**
	 * destination device(s) for writing
	 * /dev/sda /dev/sdb ...
	 */
	stralloc dest;
	u32 dest_count;

	/**
	 * number of worker threads to use
	 */
	int threads;
};

/**
 * generic handling structure for imaging disks
 */
struct img_restore {
	struct restore_config *cfg;	/* access to configuration */
	struct fng_compr *c;	/* temporary compression buffers */

	/**
	 * comment's which should be added to the image
	 */
	struct fng_image *img;
	stralloc lshw;
	stralloc comment;

	u32 disks;		/* disks in image */
	struct fng_disk *disk;	/* not allocated, when disks=0 */

	u32 partitions;		/* partitions in image */
	struct fng_part *part;

	u64 chunks;		/* chunks in image */
	struct fng_chunk *chunk;

	/* chunks in image */
	u64 index_offset;	/* saved in le64 @ end of image */
	struct fng_index *index;

	/* overall progress status */
	progress_t *progress;

	/* destination(s) */
	int64 *out_fds;		/* fd's of the destination(s) */
	char **out_files;	/* filename(s) of destination(s) */
};

/**
 * image_info() - read frisbee-ng image and show info
 */
void image_info(struct restore_config *cfg);
void image_debug(struct restore_config *cfg);
void image_restore_partition(struct restore_config *cfg);
void image_restore_disk(struct restore_config *cfg);

#endif
