
/**
 * Copyright © 2005-2008 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/**
 * logging / debugging functions
 */

#include <sys/types.h>
#include <pthread.h>
#include <string.h>
#include <unistd.h>

/* libowfat */
#include "buffer.h"
#include "byte.h"
#include "io.h"

/* frisbee-ng */
#include "main.h"
#include "timestamp.h"
#include "fmtX.h"
#include "log.h"

/* mutex for pid and fd_log */
static pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER;

/* mutex for logbuf */
static pthread_mutex_t logbuf_mutex = PTHREAD_MUTEX_INITIALIZER;

static const char *logfile = "/var/tmp/fng.log";
static const char *devnull = "/dev/null";
static char log_BUF[1024 * 8];
static buffer logbuf;
static int64 fd_log;
static pid_t pid;

/**
 * log_init - initialize logfile and it's buffering
 *
 * @param r       the current thread request structure
 * @param reinit  should the logfile re-opened (1=yes)
 */
void log_init(void)
{
	if (!logfile) {
		logfile = devnull;
		return;		/* XXX logging disabled!!! */
	}

	pthread_mutex_lock(&log_mutex);
	byte_zero(log_BUF, sizeof log_BUF);
	if (!io_appendfile(&fd_log, logfile)) {
		if (!io_appendfile(&fd_log, devnull)) {
			pthread_mutex_unlock(&log_mutex);
			fng_exit(111);
		}
	}
	pid = getpid();
	buffer_init(&logbuf, log_write, fd_log, log_BUF, sizeof log_BUF);
	pthread_mutex_unlock(&log_mutex);

	return;
}

void log_s(char *s)
{
	pthread_mutex_lock(&logbuf_mutex);
	buffer_puts(&logbuf, s);
	pthread_mutex_unlock(&logbuf_mutex);

	return;
}

void log_sa(stralloc * sa)
{
	pthread_mutex_lock(&logbuf_mutex);
	buffer_putsa(&logbuf, sa);
	pthread_mutex_unlock(&logbuf_mutex);

	return;
}

void log_flush(void)
{
	pthread_mutex_lock(&logbuf_mutex);
	buffer_flush(&logbuf);
	pthread_mutex_unlock(&logbuf_mutex);

	return;
}

void log_sa_flush_nl(stralloc * sa)
{
	pthread_mutex_lock(&logbuf_mutex);
	buffer_putsa(&logbuf, sa);
	buffer_puts(&logbuf, "\n");
	buffer_flush(&logbuf);
	pthread_mutex_unlock(&logbuf_mutex);

	return;
}

void log_headline(char *headline)
{
	char s[FMT_TIMESTAMP];

	/* @400000004826b41f0e79602c reading[21749]: ufilter_md5: 1 */

	timestamp(s);
	pthread_mutex_lock(&logbuf_mutex);
	buffer_put(&logbuf, s, FMT_TIMESTAMP);
	buffer_puts(&logbuf, " ");
	buffer_puts(&logbuf, headline);
	pthread_mutex_unlock(&logbuf_mutex);

	return;
}

ssize_t log_write(int fd, const void *buf, size_t count)
{
	int offset, block;

	if (!logfile)
		return count;	/* logging disabled!!! */

	pthread_mutex_lock(&log_mutex);
	offset = 0;
	while (count > 0) {
		block = write(fd, buf + offset, count);
		if (block < 0)
			return block;
		if (block == 0)
			return offset;

		offset += block;
		count -= block;
	}
	pthread_mutex_unlock(&log_mutex);

	return offset;
}
