
/**
 * Copyright © 2008 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 1
 */

/* system */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

/* libowfat */
#include "errmsg.h"
#include "str.h"

/* fribee-ng */
#include "main.h"
#include "crypt.h"
#include "tty.h"
#include "fs.h"
#include "compr.h"
#include "version.h"

#include "sig.h"
#include "log.h"
#include "term.h"

/* these variables get exported */
const char *fng_version = FNG_VERSION;
const char *fng_name = "";

int fng_main;
int fng_cpucount;

static const char *me0 = "fng-image";
static const char *me1 = "fng-restore";
static const char *me2 = "fng-info";
static const char *me3 = "fng-shredder";
static const char *me4 = "fng-clone";
static const char *me5 = "fng-client";
static const char *me6 = "fng-server";
static const char *me7 = "fng-adhoc";
static const char *me8 = "fng-encode";
static const char *me9 = "fng-decode";
static const char *me10 = "fng-overview";

/* should be the only call to exit() related functions */
int fng_exit(int status)
{
	log_flush();
	term_quit();
	tty_restore();

	exit(status);
	return 0;
}

/* create symlinks, when needed */
void check_symlinks(void)
{
#ifdef __linux__
	stralloc prefix, linkpath;
	char binfile[PATH_MAX];
	size_t filelen = readlink("/proc/self/exe", binfile, PATH_MAX - 2);
	int i;

	binfile[filelen] = 0;
	stralloc_init(&prefix);
	stralloc_init(&linkpath);

	/**
	 * copy prefix from binfile:
	 * "/usr/bin/fng-image" -> remove last 9 chars -> "/usr/bin/"
	 */
	if (!stralloc_copyb(&prefix, binfile, str_len(binfile) - 9))
		die_nomem();

	static const char *me[] = {
		"fng-image", "fng-restore", "fng-info", "fng-shredder",
		"fng-clone", "fng-client", "fng-server", "fng-adhoc",
		"fng-encode", "fng-decode", "fng-overview"
	};

	for (i = 1; i < 11; i++) {
		const char *x = me[i];

		if (!stralloc_copy(&linkpath, &prefix))
			die_nomem();
		if (!stralloc_cats(&linkpath, x))
			die_nomem();
		if (!stralloc_0(&linkpath))
			die_nomem();
		if (access(linkpath.s, F_OK) != 0)
			symlink(me[0], linkpath.s);
	}

	stralloc_free(&prefix);
	stralloc_free(&linkpath);
#endif

	return;
}

void version(void)
{
	term_print(fng_version);
	term_printflush("\n");

	fng_exit(0);
}

static int usage(void)
{
	term_print("%s multi-call binary.\n", fng_version);
	term_print
	    ("Copyright © 2008 - 2014 Tino Reichardt and others. Licensed under GPLv2.\n\n");
	term_print("Available commands  [status] description\n");
	term_print(" %18s %s", me0, "[123456] image creation\n");
	term_print(" %18s %s", me1, "[123   ] restoring data of some image\n");
	term_print(" %18s %s", me2,
		   "[123456] information about disks, partitions or images\n");
	term_print(" %18s %s", me3,
		   "[123456] shredder data on disk or files\n");
	term_print(" %18s %s", me4,
		   "[123456] fast local cloning of disks or partitions\n");
	term_print(" %18s %s", me5, "[1     ] client of image server\n");
	term_print(" %18s %s", me6, "[1     ] image server\n");
	term_print(" %18s %s", me7,
		   "[123   ] adhoc mode for fast disk cloning over local network\n");
	term_print(" %18s %s", me8, "[123456] encoding tests\n");
	term_print(" %18s %s", me9, "[123456] decoding tests\n");
	term_print(" %18s %s", me10, "[123456] show this summary ;)\n");
	term_quit();

	return fng_exit(0);
}

int main(int argc, char **argv)
{
	char *me;

	me = strrchr(argv[0], '/');
	if (me)
		++me;
	else
		me = argv[0];

	/* we remove setuid/setgid bit + execute for owner */
	umask(06177);

	fng_cpucount = sysconf(_SC_NPROCESSORS_ONLN);
	if (fng_cpucount == -1)
		die(1, "_SC_NPROCESSORS_ONLN not set!");

	sig_init();
	term_init();
	log_init();

	/* tty_init(); -> nur dahin, wo es gebraucht wird! */

	/* call only, if needed... */
	register_filesystems();
	register_compression_modes();

	if (!strcmp(me, me0)) {
		/* image storing */
		fng_main = FNG_MAIN_IMAGE;
		fng_name = me0;
		errmsg_iam(fng_name);

		/* fng-image is the main binary,
		 * check if all symlinks are there
		 */
		check_symlinks();

		/* now the real program */
		return fng_image(argc, argv);
	} else if (!strcmp(me, me1)) {
		/* image restoring / direct io */
		fng_main = FNG_MAIN_RESTORE;
		fng_name = me1;
		errmsg_iam(fng_name);
		return fng_restore(argc, argv);
	} else if (!strcmp(me, me2)) {
		/* disc / image info */
		fng_main = FNG_MAIN_INFO;
		fng_name = me2;
		errmsg_iam(fng_name);
		return fng_info(argc, argv);
	} else if (!strcmp(me, me3)) {
		/* disc shredder / direct io */
		fng_main = FNG_MAIN_SHREDDER;
		fng_name = me3;
		errmsg_iam(fng_name);
		return fng_shredder(argc, argv);
	} else if (!strcmp(me, me4)) {
		/* disc cloning / direct io */
		fng_main = FNG_MAIN_CLONE;
		fng_name = me4;
		errmsg_iam(fng_name);
		return fng_clone(argc, argv);
	} else if (!strcmp(me, me5)) {
		/* client of server */
		fng_main = FNG_MAIN_CLIENT;
		fng_name = me5;
		errmsg_iam(fng_name);
		return fng_client(argc, argv);
	} else if (!strcmp(me, me6)) {
		/* image distribution */
		fng_main = FNG_MAIN_SERVER;
		fng_name = me6;
		errmsg_iam(fng_name);
		return fng_server(argc, argv);
	} else if (!strcmp(me, me7)) {
		/* frisbee adhoc mode */
		fng_main = FNG_MAIN_ADHOC;
		fng_name = me7;
		errmsg_iam(fng_name);
		return fng_adhoc(argc, argv);
	} else if (!strcmp(me, me8)) {
		/* frisbee encode tests */
		fng_main = FNG_MAIN_ENCODE;
		fng_name = me8;
		errmsg_iam(fng_name);
		return fng_encode(argc, argv, 1);
	} else if (!strcmp(me, me9)) {
		/* frisbee decode tests */
		fng_main = FNG_MAIN_DECODE;
		fng_name = me9;
		errmsg_iam(fng_name);
		return fng_encode(argc, argv, 2);
	}

	return usage();
}
