
/**
 * Copyright © 2010 - 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef FNG_MEM_H
#define FNG_MEM_H

/* system */
#include <stdlib.h>
#include <string.h>

/* frisbee */
#include "debug.h"
#include "defs.h"
#include "die.h"

/* tests */
#include "haveinline.h"

/* warning, will produce a lot messages... only for debugging strange things... */
//#define MEM_DEBUG

#ifdef MEM_DEBUG
#include "term.h"
#define mem_alloc(s)     mem_alloc_r(s, __func__, __FILE__, __LINE__)
#define mem_zalloc(s)    mem_zalloc_r(s, __func__, __FILE__, __LINE__)
#define mem_realloc(p,s) mem_realloc_r(p, s, __func__, __FILE__, __LINE__)
#define mem_calloc(n,m)  mem_calloc_r(n, m, __func__, __FILE__, __LINE__)
#define mem_free(s)      mem_free_r(s, __func__, __FILE__, __LINE__)
#else
#define mem_alloc(s)     mem_alloc_r(s)
#define mem_zalloc(s)    mem_zalloc_r(s)
#define mem_realloc(p,s) mem_realloc_r(p,s)
#define mem_calloc(n,m)  mem_calloc_r(n,m)
#define mem_free(s)      mem_free_r(s)
#endif

/**
 * wrapper around malloc()...
 * - calls die_nomem(), if no space could be allocated
 */
#ifdef MEM_DEBUG
static inline void *mem_alloc_r(size_t size, const char *func, const char *file,
				int line)
{
	void *x = malloc(size);

	debug(1,1,"%s(%s): %s,%d = %08X (len=%z)", __func__, file, func,
			line, x, size);
#else
static inline void *mem_alloc_r(size_t size)
{
	void *x = malloc(size);
#endif
	if (unlikely(!x))
		die_nomem();

	return x;
}

/**
 * mem_talloc - allocating n members of type t
 * - talloc means "type alloc"
 */
#define mem_talloc(t,n)  (t*)mem_alloc(sizeof(t)*n)

/**
 * wrapper around malloc()...
 * - calls die_nomem(), if no space could be allocated
 * - initialize new allocated memory with zero
 */

#ifdef MEM_DEBUG
static inline void *mem_zalloc_r(size_t size, const char *func,
				 const char *file, int line)
{
	void *x = malloc(size);

	debug(1,1,"%s(%s): %s,%d = %08X (len=%z)", __func__, file, func,
			line, x, size);
#else
static inline void *mem_zalloc_r(size_t size)
{
	void *x = malloc(size);
#endif

	if (unlikely(!x))
		die_nomem();
	memset(x, 0, size);

	return x;
}

/**
 * mem_tzalloc - allocating n members of type t
 * - tzalloc means "type zalloc"
 */
#define mem_tzalloc(t,n)  (t*)mem_zalloc(sizeof(t)*n)

/**
 * wrapper around realloc()...
 * - calls die_nomem(), if no space could be allocated
 */

#ifdef MEM_DEBUG
static inline void *mem_realloc_r(void *p, size_t size, const char *func,
				  const char *file, int line)
{
	void *x = realloc(p, size);

	debug(1,1,"%s(%s): %s,%d = %08X (len=%z)", __func__, file, func,
			line, x, size);
#else
static inline void *mem_realloc_r(void *p, size_t size)
{
	void *x = realloc(p, size);
#endif

	if (unlikely(!x))
		die_nomem();

	return x;
}

/**
 * wrapper around calloc()...
 * - calls die_nomem(), if no space could be allocated
 */

#ifdef MEM_DEBUG
static inline void *mem_calloc_r(size_t nmemb, size_t size, const char *func,
				 const char *file, int line)
{
	void *x = calloc(nmemb, size);

	debug(1,1,"%s(%s): %s,%d = %08X (len=%z * %z)", __func__, file,
			func, line, x, nmemb, size);
#else
static inline void *mem_calloc_r(size_t nmemb, size_t size)
{
	void *x = calloc(nmemb, size);
#endif

	if (unlikely(!x))
		die_nomem();

	return x;
}

/**
 * wrapper around free()...
 */

#ifdef MEM_DEBUG
static inline void mem_free_r(void *p, const char *func, const char *file,
			      int line)
{
	debug(1,1,"%s(%s): %s(%08X) on line %d", __func__, file, func,
			p, line);
#else
static inline void mem_free_r(void *p)
{
#endif
	if (likely(p))
		free(p);
	p = 0;

	return;
}

#endif				/* FNG_MEM_H */
