
/**
 * Copyright © 2012 - 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <string.h>
#include <arpa/inet.h>
#include <stdio.h>

/* libowfat */
#include "stralloc.h"
#include "socket.h"
#include "io.h"

/* frisbee-ng */
#include "packet.h"
#include "die.h"
#include "sc.h"

#include "fng-adhoc.h"

#define MAXUDP 64*1024

/* the magic for frisbee-ng ad-hoc packets */
static char *fng_adhoc_server = "FNGA";

/* incomplete chunk queue */
struct icq {
	u64 chunk_idx;
	u16 partition;
};

/*

----------------------------------------------------------------------------------------------------------------------------------------------------

- chunks are defined by:
  1) disk
  2) partition
  3) chunk number

- chunk packets are defined by:
  - std header
  - disk / partiton / chunk number / chunk fragment

- Reassemble chunk fragments, so wie @ linux lkernel

@linux:
- ip_local_deliver() @ ip_input.c wird durch ip_rcv_finish() aufgerufen
- sie fügt neue fragmente zur warteschlange, und wenn alle da sind, wird ip_local_deliver_finish() aufgerufen
- ip_local_deliver_finish() schiebt dann das fertige packet auf die platte

----------------------------------------------------------------------------------------------------------------------------------------------------

packet details client:
1) client hello (broadcast check for servers)
2) client get-configuration (via udp, maxmtu, chunksize, min_devsize)
3) client get-bootrecord (via udp, mbr bzw gpt... packet 1/10 2/10 ... je nach maxmtu, je partition die werte für belegt... und max chunkcount)
4) client get-chunk 0..chunkcount
5) client get-chunkpart (0..chunkcount) part 0..N

packet details server:
typen:
1) antwort zum client hello
2) antwort zum client get-configuration
3) antwort zum client get-bootrecord
4) antwort zum client get-chunk und get-chunkpart

wie tcp fragmente teilen:

offset


*/

static void SendBroadcast(stralloc * packet)
{
	int s = socket_udp();
	char ip[4] = { 255, 255, 255, 255 };

	if (s == -1)
		die_msg(111, "socket_udp6()");
	if (socket_broadcast(s) == -1)
		die_msg(111, "socket_broadcast()");
	socket_send4(s, packet->s, packet->len, ip, 7983);
}

void ClientHello(int sock)
{
	struct sstate *s = io_getcookie(sock);
	stralloc sa;
	int i;

	unused(s);

	stralloc_init(&sa);
	if (!stralloc_copys(&sa, ""))
		die_nomem();

	for (i = 0; i < 10000; i++)
		if (!stralloc_cats(&sa, "x"))
			die_nomem();

	SendBroadcast(&sa);
	stralloc_free(&sa);
}

void handle_adhoc_server_init(int64 sock)
{
	struct sstate *s = io_getcookie(sock);
	struct adhoc_config *cfg = s->cfg;

	u16 b16;
	u32 b32;
	u64 b64;

	void *p16 = &b16;
	void *p32 = &b32;
	void *p64 = &b64;

	/* HELLO reply, id 0x01 - static */
	stralloc_init(&s->p_hello);
	if (!stralloc_copys(&s->p_hello, fng_adhoc_server))
		die_nomem();
	b32 = cpu_to_le32(1);
	if (!stralloc_catb(&s->p_hello, p32, 4))
		die_nomem();
	if (!stralloc_catb(&s->p_hello, cfg->fng_mcip, 16))
		die_nomem();
	b32 = cpu_to_le16(str_len(s->group));
	if (!stralloc_catb(&s->p_hello, p32, 4))
		die_nomem();
	if (!stralloc_cats(&s->p_hello, s->group))
		die_nomem();

	/* GETCONFIG reply, id 0x02 - static */
	stralloc_init(&s->p_config);
	if (!stralloc_copys(&s->p_config, fng_adhoc_server))
		die_nomem();
	b32 = cpu_to_le32(2);
	if (!stralloc_catb(&s->p_config, p32, 4))
		die_nomem();
	b32 = cpu_to_le32(cfg->blks_chunk);
	if (!stralloc_catb(&s->p_config, p32, 4))
		die_nomem();
	b16 = cpu_to_le16(cfg->mtu);
	if (!stralloc_catb(&s->p_config, p16, 2))
		die_nomem();
	b16 = cpu_to_le16(s->tlen);
	if (!stralloc_catb(&s->p_config, p16, 2))
		die_nomem();
	b64 = cpu_to_le64(s->min_devsize);
	if (!stralloc_catb(&s->p_config, p64, 8))
		die_nomem();

	/* GETPTABLE reply, id 0x03 - static */
	stralloc_init(&s->p_ptable);
	if (!stralloc_copys(&s->p_ptable, fng_adhoc_server))
		die_nomem();
	b32 = cpu_to_le32(2);
	if (!stralloc_catb(&s->p_ptable, p32, 4))
		die_nomem();
	b32 = cpu_to_le32(cfg->blks_chunk);
	if (!stralloc_catb(&s->p_ptable, p32, 4))
		die_nomem();
	b16 = cpu_to_le16(cfg->mtu);
	if (!stralloc_catb(&s->p_ptable, p16, 2))
		die_nomem();
	b16 = cpu_to_le16(s->tlen);
	if (!stralloc_catb(&s->p_ptable, p16, 2))
		die_nomem();
	b64 = cpu_to_le64(s->min_devsize);
	if (!stralloc_catb(&s->p_ptable, p64, 8))
		die_nomem();

	return;
}

void handle_adhoc_server_read(int64 sock)
{
	static int initdone = 0;
	char remoteip[16];
	u16 remoteport;
	size_t byte_count;
	stralloc sa;
	char *x = "huhu haha";

	if (!initdone) {
		initdone = 1;
	}

	stralloc_init(&sa);
	if (!stralloc_ready(&sa, MAXUDP))
		die_nomem();

	byte_count = sc_recv(sock, sa.s, MAXUDP, remoteip, &remoteport);

	term_print("socket=%d got %d bytes on port %d\n", sock, byte_count,
		   remoteport);

	sc_send(sock, x, strlen(x), remoteip, remoteport);

	return;
}

void handle_adhoc_client_init(int64 sock)
{
	struct cstate *s = io_getcookie(sock);
	struct adhoc_config *cfg = s->cfg;

	u16 b16;

	void *p16 = &b16;

	/* request id 01 */
	stralloc_init(&s->p_hello);
	if (!stralloc_copys(&s->p_hello, "FNGa"))
		die_nomem();
	b16 = cpu_to_le16(str_len(cfg->group.s));
	if (!stralloc_catb(&s->p_hello, p16, 2))
		die_nomem();
	if (!stralloc_cats(&s->p_hello, cfg->group.s))
		die_nomem();

	return;
}

void handle_adhoc_client_read(int64 sock)
{
	static int initdone = 0;
	char remoteip[16];
	u16 remoteport;
	size_t byte_count;
	stralloc sa;
	char *x = "huhu haha";

	if (!initdone) {
		initdone = 1;
	}

	stralloc_init(&sa);
	if (!stralloc_ready(&sa, MAXUDP))
		die_nomem();

	byte_count = sc_recv(sock, sa.s, MAXUDP, remoteip, &remoteport);

	/*

	   1) einlsen
	   2) checken ob id=1 oder id=2, dann bearbeiten und return
	   3) checken ob id>2, dann defragmenierung mit einbeziehen, sie linux kernel

	 */

	term_print("socket=%d got %d bytes on port %d\n", sock, byte_count,
		   remoteport);

	sc_send(sock, x, strlen(x), remoteip, remoteport);

	return;
}

#if 0
extern int64 sc_udp(char ip[16], u16 port);
extern int64 sc_multicast(char ip[16], u16 port, char ttl, char loop);
extern int64 sc_broadcast(void);
extern ssize_t sc_send(int64 sock, char *buf, size_t len, const char ip[16],
		       u16 port);
extern ssize_t sc_recv(int64 s, char *buf, size_t len, char ip[16], u16 * port);
#endif
