
/**
 * Copyright © 2012 - 2014 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* libowfat */
#include "byte.h"
#include "io.h"
#include "uint16.h"
#include "uint32.h"
#include "uint64.h"

/* frisbee-ng */
#include "pt_reader.h"
#include "term.h"
#include "mem.h"
#include "die.h"
#include "wrapper.h"
#include "bswap.h"

/* must be 512 ... */
#define BLKSZ 512

/**
 * we take care for grub and other boot managers
 * - read/write also past the mbr @ dos partition type
 * - same for gpt...
 * - but not more then around 10 MiB
 */

#define EXTRALEN 10

static void pt_clone_gpt(int fd, int len, int64 * fds, u64 * devsize,
			 int first_part)
{
	void *mbr = mem_alloc(BLKSZ);	/* protective MBR (LBA0) */
	void *hdr = mem_alloc(BLKSZ);	/* GPT header (LBA1) */
	void *gpe = mem_alloc(BLKSZ * 32);	/* Partition entries LBA2..LBA33 */
	void *x;
	int i;

	/* lba0 */
	pread_loop2(fd, mbr, BLKSZ, 0);

	/* lba1 */
	pread_loop2(fd, hdr, BLKSZ, BLKSZ);

	/* lba2..34 */
	pread_loop2(fd, gpe, BLKSZ * 32, BLKSZ * 2);

	for (i = 0; i < len; i++) {
		u32 crc;
		u64 dsize = devsize[i] / BLKSZ;

		/* protective MBR */
		pwrite_loop2(fds[i], mbr, BLKSZ, 0);

		/* gpt header1 */
		uint32_pack(hdr + 16, 0);
		uint64_pack(hdr + 24, 1);
		uint64_pack(hdr + 32, dsize - 1);
		crc = efi_crc32(hdr, 92);
		uint32_pack(hdr + 16, crc);
		pwrite_loop2(fds[i], hdr, BLKSZ, BLKSZ);

		/* gpt header2 */
		uint32_pack(hdr + 16, 0);
		uint64_pack(hdr + 24, dsize - 1);
		uint64_pack(hdr + 32, 1);
		crc = efi_crc32(hdr, 92);
		uint32_pack(hdr + 16, crc);
		pwrite_loop2(fds[i], hdr, BLKSZ, dsize * BLKSZ - BLKSZ);

		/* gpt entries */
		pwrite_loop2(fds[i], gpe, BLKSZ * 32, BLKSZ * 2);
		pwrite_loop2(fds[i], gpe, BLKSZ * 32,
			     dsize * BLKSZ - BLKSZ * 33);
	}

	/* read also past mbr bytes... if they contain only zeros... we are
	 * finished, if not, we raw copy the data, it is mostly grub or some
	 * other boot manager stuff...
	 * XXX
	 */
	x = hdr;
	if (!x)
		die_nomem();
	byte_zero(x, BLKSZ);
	for (i = 35; i < first_part; i++) {
		int j;

		/* only zeros here... nothing to do then */
		pread_loop2(fd, mbr, BLKSZ, i * BLKSZ);
		//if (byte_diff(x, BLKSZ, mbr) == 0)
		//      break;

		/* copy the boot manager stuff */
		for (j = 0; j < len; j++) {
			pwrite_loop2(fds[j], mbr, BLKSZ, i * BLKSZ);
		}
	}

	mem_free(mbr);
	mem_free(hdr);
	mem_free(gpe);

	return;
}

static void pt_clone_dos(int fd, int len, int64 * fds, int first_part)
{
	void *mbr = mem_alloc(BLKSZ);
	struct mbrpart *p;
	void *x;
	int i;
	u32 lba, start;

	if (!mbr)
		die_nomem();

	/* 1) copy mbr - this is easy ;) */
	pread_loop2(fd, mbr, BLKSZ, 0);
	for (i = 0; i < len; i++)
		pwrite_loop2(fds[i], mbr, BLKSZ, 0);

	/* 2) search for ebr's and copy them... */
	lba = 0;
	for (i = 0; i < 4; i++) {
		x = mbr + 446 + 16 * i;
		p = x;

		if (is_extended(p->sys_type)) {
			lba = p->start_sect;
			break;
		}
	}

	/* read also past mbr bytes... if they contain only zeros... we are
	 * finished, if not, we raw copy the data, it is mostly grub or some
	 * other boot manager stuff... */
	x = mem_alloc(BLKSZ);
	byte_zero(x, BLKSZ);
	for (i = 1; i < first_part; i++) {
		int j;

		/* only zeros here... nothing to do then */
		pread_loop2(fd, mbr, BLKSZ, i * BLKSZ);
		//if (byte_diff(x, BLKSZ, mbr) == 0)
		//      break;

		/* copy the boot manager stuff */
		for (j = 0; j < len; j++) {
			pwrite_loop2(fds[j], mbr, BLKSZ, i * BLKSZ);
		}
	}

	/* no extended partition found */
	if (lba == 0)
		goto exit;

	for (start = 0, i = 0; i < 64; i++) {
		u16 sig;
		int j;

		if (lba == 0)
			break;

		pread_loop2(fd, mbr, BLKSZ, (start + lba) * BLKSZ);
		for (j = 0; j < len; j++)
			pwrite_loop2(fds[j], mbr, BLKSZ, (start + lba) * BLKSZ);

		/* have we found our way out? */
		uint16_unpack(mbr + 510, &sig);
		if (sig != 0xaa55)
			break;

		/* reset to go away value */
		if (start == 0)
			start = lba;
		lba = 0;

		x = mbr + 446;
		p = x;
		if (is_extended(p->sys_type))
			lba = p->start_sect;

		x = mbr + 446 + 16;
		p = x;
		if (is_extended(p->sys_type))
			lba = p->start_sect;
	}

 exit:
	mem_free(x);
	mem_free(mbr);

	return;
}

/**
 * pt_clone() - clone one partitiontable to one or more destinations
 *
 * @fd: harddisk to read from
 * @type: type of ptable
 * @sl: data of source partitions
 * @slice_count: number of source partitions
 * @len: number of destination devices
 * @fds: destination fd's
 * @devsize: destination device sizes
 *
 * @return: nothing
 */
void pt_clone(int fd, int ptype, struct slice *sl, u32 slice_count, u32 len,
	      int64 * fds, u64 * devsize)
{
	u32 i;
	u64 first_part = 0;

	/* find the first real starting partition... */
	for (i = 0; i < slice_count; i++) {
		if (first_part == 0)
			first_part = sl[i].start;
		first_part = min(sl[i].start, first_part);
	}

	/* no real boot manager is over 10 MiB ;) */
	first_part = min(first_part, EXTRALEN * 1024 * 1024);
	debug(1, 1, "first_part=%d ... extra copy up to ", first_part);

	if (ptype == PTYPE_GPT)
		pt_clone_gpt(fd, len, fds, devsize, first_part);
	else if (ptype == PTYPE_DOS)
		pt_clone_dos(fd, len, fds, first_part);
	else
		fatal("Only mbr/gpt are properly cloned currently!");

	/**
	 * I don't use BSD or other systems...
	 * -> if you need it, send me a patch or give me some hint...
	 */

	return;
}
