
/**
 * Copyright © 2000 - 2001 Matt Domsch, Dell Computer Corporation
 * Copyright © 2011 - 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Initially taken from multipath-tools, which derived from util-linux :)
 */

/* system */
#include <sys/stat.h>
#include <inttypes.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>

#include "byte.h"

#include "zlib/zlib.h"
#include "pt_reader.h"
#include "term.h"
#include "bswap.h"
#include "mem.h"
#include "dev.h"

int force_gpt = 0;

/* shut up fprintf()! /TR */
#define fprintf(dst,fmt,args...) do { } while (0)

typedef struct {
	u8 b[16];
} efi_guid_t;

#define EFI_GUID(a,b,c,d0,d1,d2,d3,d4,d5,d6,d7) \
((efi_guid_t) \
{{ (a) & 0xff, ((a) >> 8) & 0xff, ((a) >> 16) & 0xff, ((a) >> 24) & 0xff, \
   (b) & 0xff, ((b) >> 8) & 0xff, \
   (c) & 0xff, ((c) >> 8) & 0xff, \
  (d0), (d1), (d2), (d3), (d4), (d5), (d6), (d7) }})

/******************************************************
 * GUIDs
 ******************************************************/
#define NULL_GUID \
EFI_GUID( 0x00000000, 0x0000, 0x0000, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00)

static int efi_guidcmp(efi_guid_t left, efi_guid_t right)
{
	return memcmp(&left, &right, sizeof(efi_guid_t));
}

typedef u16 efi_char16_t;	/* UNICODE character */

#define EFI_PMBR_OSTYPE_EFI       0xEF
#define EFI_PMBR_OSTYPE_EFI_GPT   0xEE
#define MSDOS_MBR_SIGNATURE       0xaa55
#define GPT_BLOCK_SIZE            512
#define GPT_HEADER_SIGNATURE      0x5452415020494645ULL

#define GPT_PRIMARY_PARTITION_TABLE_LBA 1
#define EFI_GPT_PRIMARY_PARTITION_TABLE_LBA 1

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

typedef struct _gpt_header {
	u64 signature;
	u32 revision;
	u32 header_size;
	u32 header_crc32;
	u32 reserved1;
	u64 my_lba;
	u64 alternate_lba;
	u64 first_usable_lba;
	u64 last_usable_lba;
	efi_guid_t disk_guid;
	u64 partition_entry_lba;
	u32 num_partition_entries;
	u32 sizeof_partition_entry;
	u32 partition_entry_array_crc32;
	u8 reserved2[GPT_BLOCK_SIZE - 92];
} __FNG_ATTR_PACKED gpt_header;

typedef struct _gpt_entry___attribute__s {
	u64 required_to_function:1;
	u64 reserved:47;
	u64 type_guid_specific:16;
} __FNG_ATTR_PACKED gpt_entry___attribute__s;

typedef struct _gpt_entry {
	efi_guid_t partition_type_guid;
	efi_guid_t unique_partition_guid;
	u64 starting_lba;
	u64 ending_lba;
	gpt_entry___attribute__s __attribute__s;
	efi_char16_t partition_name[72 / sizeof(efi_char16_t)];
} __FNG_ATTR_PACKED gpt_entry;

/* Protected Master Boot Record  & Legacy MBR share same structure */
/* Needs to be packed because the u16 types force misalignment. */
typedef struct _legacy_mbr {
	u8 bootcode[440];
	u32 unique_mbr_signature;
	u16 unknown;
	struct mbrpart partition[4];
	u16 signature;
} __FNG_ATTR_PACKED legacy_mbr;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/**
 * efi_crc32() - EFI version of crc32 function
 * @buf: buffer to calculate crc32 of
 * @len - length of buf
 *
 * Description: Returns EFI-style CRC32 value for @buf
 * 
 * This function uses the little endian Ethernet polynomial
 * but seeds the function with ~0, and xor's with ~0 at the end.
 * Note, the EFI Specification, v1.02, has a reference to
 * Dr. Dobbs Journal, May 1994 (actually it's in May 1992).
 */
u32 efi_crc32(const void *buf, unsigned long len)
{
	return (crc32(~0L, buf, len) ^ ~0L);
}

/**
 * is_pmbr_valid(): test Protective MBR for validity
 * @mbr: pointer to a legacy mbr structure
 *
 * Description: Returns 1 if PMBR is valid, 0 otherwise.
 * Validity depends on two things:
 *  1) MSDOS signature is in the last two bytes of the MBR
 *  2) One partition of type 0xEE is found
 */
static int is_pmbr_valid(legacy_mbr * mbr)
{
	int i, found = 0, signature = 0;

	if (!mbr)
		return 0;
	signature = (le16_to_cpu(mbr->signature) == MSDOS_MBR_SIGNATURE);
	for (i = 0; signature && i < 4; i++) {
		if (mbr->partition[i].sys_type == EFI_PMBR_OSTYPE_EFI_GPT) {
			found = 1;
			break;
		}
	}

	return (signature && found);
}

/************************************************************
 * _get_num_sectors
 * Requires:
 *  - filedes is an open file descriptor, suitable for reading
 * Modifies: nothing
 * Returns:
 *  Last LBA value on success 
 *  0 on error
 *
 * Try getting BLKGETSIZE64 and BLKSSZGET first,
 * then BLKGETSIZE if necessary.
 *  Kernels 2.4.15-2.4.18 and 2.5.0-2.5.3 have a broken BLKGETSIZE64
 *  which returns the number of 512-byte sectors, not the size of
 *  the disk in bytes. Fixed in kernels 2.4.18-pre8 and 2.5.4-pre3.
 ************************************************************/
static u64 _get_num_sectors(int fd)
{
	u64 bytes = dev_getdisksize(fd);

	return bytes / dev_getsectorsize(fd);
}

/************************************************************
 * last_lba(): return number of last logical block of device
 * 
 * @fd
 * 
 * Description: returns Last LBA value on success, 0 on error.
 * Notes: The value st_blocks gives the size of the file
 *        in 512-byte blocks, which is OK if
 *        EFI_BLOCK_SIZE_SHIFT == 9.
 ************************************************************/

static u64 last_lba(int filedes)
{
	u64 sectors = _get_num_sectors(filedes);

	return sectors ? sectors - 1 : 0;
}

static ssize_t read_lba(int fd, u64 lba, void *buffer, size_t bytes)
{
	int sector_size = dev_getsectorsize(fd);
	off_t offset = lba * sector_size;
	u64 lastlba;
	ssize_t bytesread;

	lseek(fd, offset, SEEK_SET);
	bytesread = read(fd, buffer, bytes);

	lastlba = last_lba(fd);
	if (!lastlba)
		return bytesread;

	return bytesread;
}

/**
 * alloc_read_gpt_entries(): reads partition entries from disk
 *
 * @fd  is an open file descriptor to the whole disk
 * @gpt is a buffer into which the GPT will be put  
 *
 * Description: Returns ptes on success,  NULL on error.
 * Allocates space for PTEs based on information found in @gpt.
 * Notes: remember to free pte when you're done!
 */
static gpt_entry *alloc_read_gpt_entries(int fd, gpt_header * gpt)
{
	gpt_entry *pte;
	size_t count =
	    le32_to_cpu(gpt->num_partition_entries) *
	    le32_to_cpu(gpt->sizeof_partition_entry);

	if (!count)
		return NULL;

	pte = (gpt_entry *) mem_alloc(count);
	byte_zero(pte, count);

	if (!read_lba(fd, le64_to_cpu(gpt->partition_entry_lba), pte, count)) {
		mem_free(pte);
		return NULL;
	}

	return pte;
}

/**
 * alloc_read_gpt_header(): Allocates GPT header, reads into it from disk
 * @fd  is an open file descriptor to the whole disk
 * @lba is the Logical Block Address of the partition table
 * 
 * Description: returns GPT header on success, NULL on error.   Allocates
 * and fills a GPT header starting at @ from @bdev.
 * Note: remember to free gpt when finished with it.
 */
static gpt_header *alloc_read_gpt_header(int fd, u64 lba)
{
	gpt_header *gpt = mem_tzalloc(gpt_header, 1);

	if (!read_lba(fd, lba, gpt, sizeof(gpt_header))) {
		mem_free(gpt);
		return NULL;
	}

	return gpt;
}

/**
 * is_gpt_valid() - tests one GPT header and PTEs for validity
 * @fd  is an open file descriptor to the whole disk
 * @lba is the logical block address of the GPT header to test
 * @gpt is a GPT header ptr, filled on return.
 * @ptes is a PTEs ptr, filled on return.
 *
 * Description: returns 1 if valid,  0 on error.
 * If valid, returns pointers to newly allocated GPT header and PTEs.
 */
static int is_gpt_valid(int fd, u64 lba, gpt_header ** gpt, gpt_entry ** ptes)
{
	int rc = 0;		/* default to not valid */
	u32 crc, origcrc;

	if (!gpt || !ptes)
		return 0;
	if (!(*gpt = alloc_read_gpt_header(fd, lba)))
		return 0;

	/* Check the GUID Partition Table signature */
	if (le64_to_cpu((*gpt)->signature) != GPT_HEADER_SIGNATURE) {
		mem_free(*gpt);
		*gpt = NULL;
		return rc;
	}

	/* Check the GUID Partition Table Header CRC */
	origcrc = le32_to_cpu((*gpt)->header_crc32);
	(*gpt)->header_crc32 = 0;
	crc = efi_crc32(*gpt, le32_to_cpu((*gpt)->header_size));
	if (crc != origcrc) {
		term_printflush("GPTH CRC check failed, %x != %x.\n", origcrc,
				crc);
		(*gpt)->header_crc32 = cpu_to_le32(origcrc);
		mem_free(*gpt);
		*gpt = NULL;
		return 0;
	}
	(*gpt)->header_crc32 = cpu_to_le32(origcrc);

	/* Check that the my_lba entry points to the LBA
	 * that contains the GPT we read */
	if (le64_to_cpu((*gpt)->my_lba) != lba) {
		mem_free(*gpt);
		*gpt = NULL;
		return 0;
	}

	/* Check that sizeof_partition_entry has the correct value */
	if (le32_to_cpu((*gpt)->sizeof_partition_entry) != sizeof(gpt_entry)) {
		// printf("GUID partition entry size check failed.\n");
		mem_free(*gpt);
		*gpt = NULL;
		return 0;
	}

	if (!(*ptes = alloc_read_gpt_entries(fd, *gpt))) {
		mem_free(*gpt);
		*gpt = NULL;
		return 0;
	}

	/* Check the GUID Partition Entry Array CRC */
	crc =
	    efi_crc32(*ptes,
		      le32_to_cpu((*gpt)->num_partition_entries) *
		      le32_to_cpu((*gpt)->sizeof_partition_entry));
	if (crc != le32_to_cpu((*gpt)->partition_entry_array_crc32)) {
		// printf("GUID Partitition Entry Array CRC check failed.\n");
		mem_free(*gpt);
		*gpt = NULL;
		mem_free(*ptes);
		*ptes = NULL;
		return 0;
	}

	/* We're done, all's well */
	return 1;
}

/**
 * compare_gpts() - Search disk for valid GPT headers and PTEs
 * @pgpt is the primary GPT header
 * @agpt is the alternate GPT header
 * @lastlba is the last LBA number
 * Description: Returns nothing.  Sanity checks pgpt and agpt fields
 * and prints warnings on discrepancies.
 */
static void compare_gpts(gpt_header * pgpt, gpt_header * agpt, u64 lastlba)
{
	int error_found = 0;

	if (!pgpt || !agpt)
		return;
	if (le64_to_cpu(pgpt->my_lba) != le64_to_cpu(agpt->alternate_lba)) {
		error_found++;
		fprintf(stderr,
			"GPT:Primary header LBA != Alt. header alternate_lba\n");
	}
	if (le64_to_cpu(pgpt->alternate_lba) != le64_to_cpu(agpt->my_lba)) {
		error_found++;
		fprintf(stderr,
			"GPT:Primary header alternate_lba != Alt. header my_lba\n");
	}
	if (le64_to_cpu(pgpt->first_usable_lba) !=
	    le64_to_cpu(agpt->first_usable_lba)) {
		error_found++;
		fprintf(stderr, "GPT:first_usable_lbas don't match.\n");
	}
	if (le64_to_cpu(pgpt->last_usable_lba) !=
	    le64_to_cpu(agpt->last_usable_lba)) {
		error_found++;
		fprintf(stderr, "GPT:last_usable_lbas don't match.\n");
	}
	if (efi_guidcmp(pgpt->disk_guid, agpt->disk_guid)) {
		error_found++;
		fprintf(stderr, "GPT:disk_guids don't match.\n");
	}
	if (le32_to_cpu(pgpt->num_partition_entries) !=
	    le32_to_cpu(agpt->num_partition_entries)) {
		error_found++;
		fprintf(stderr,
			"GPT:num_partition_entries don't match: "
			"0x%x != 0x%x\n",
			le32_to_cpu(pgpt->num_partition_entries),
			le32_to_cpu(agpt->num_partition_entries));
	}
	if (le32_to_cpu(pgpt->sizeof_partition_entry) !=
	    le32_to_cpu(agpt->sizeof_partition_entry)) {
		error_found++;
		fprintf(stderr,
			"GPT:sizeof_partition_entry values don't match: "
			"0x%x != 0x%x\n",
			le32_to_cpu(pgpt->sizeof_partition_entry),
			le32_to_cpu(agpt->sizeof_partition_entry));
	}
	if (le32_to_cpu(pgpt->partition_entry_array_crc32) !=
	    le32_to_cpu(agpt->partition_entry_array_crc32)) {
		error_found++;
		fprintf(stderr,
			"GPT:partition_entry_array_crc32 values don't match: "
			"0x%x != 0x%x\n",
			le32_to_cpu(pgpt->partition_entry_array_crc32),
			le32_to_cpu(agpt->partition_entry_array_crc32));
	}
	if (le64_to_cpu(pgpt->alternate_lba) != lastlba) {
		error_found++;
		fprintf(stderr,
			"GPT:Primary header thinks Alt. header is not at the end of the disk.\n");
	}

	if (le64_to_cpu(agpt->my_lba) != lastlba) {
		error_found++;
		fprintf(stderr,
			"GPT:Alternate GPT header not at the end of the disk.\n");
	}

	if (error_found)
		fprintf(stderr, "GPT: Use GNU Parted to correct GPT errors.\n");

	/* XXX: error_found is ignored currently!! */
	return;
}

/**
 * find_valid_gpt() - Search disk for valid GPT headers and PTEs
 * @fd  is an open file descriptor to the whole disk
 * @gpt is a GPT header ptr, filled on return.
 * @ptes is a PTEs ptr, filled on return.
 * Description: Returns 1 if valid, 0 on error.
 * If valid, returns pointers to newly allocated GPT header and PTEs.
 * Validity depends on finding either the Primary GPT header and PTEs valid,
 * or the Alternate GPT header and PTEs valid, and the PMBR valid.
 */
static int find_valid_gpt(int fd, gpt_header ** gpt, gpt_entry ** ptes)
{
	int good_pgpt = 0, good_agpt = 0, good_pmbr = 0;
	gpt_header *pgpt = NULL, *agpt = NULL;
	gpt_entry *pptes = NULL, *aptes = NULL;
	legacy_mbr *legacymbr = NULL;
	u64 lastlba;

	if (!gpt || !ptes)
		return -1;

	if (!(lastlba = last_lba(fd)))
		return -1;

	good_pgpt =
	    is_gpt_valid(fd, GPT_PRIMARY_PARTITION_TABLE_LBA, &pgpt, &pptes);
	if (good_pgpt) {
		good_agpt =
		    is_gpt_valid(fd, le64_to_cpu(pgpt->alternate_lba), &agpt,
				 &aptes);
		if (!good_agpt) {
			good_agpt = is_gpt_valid(fd, lastlba, &agpt, &aptes);
		}
	} else {
		good_agpt = is_gpt_valid(fd, lastlba, &agpt, &aptes);
	}

	/* The obviously unsuccessful case */
	if (!good_pgpt && !good_agpt) {
		goto fail;
	}

	/* This will be added to the EFI Spec. per Intel after v1.02. */
	legacymbr = mem_talloc(legacy_mbr, 1);
	if (legacymbr) {
		byte_zero(legacymbr, sizeof(*legacymbr));
		read_lba(fd, 0, (u8 *) legacymbr, sizeof(*legacymbr));
		good_pmbr = is_pmbr_valid(legacymbr);
		mem_free(legacymbr);
		legacymbr = NULL;
	}

	/* Failure due to bad PMBR */
	if ((good_pgpt || good_agpt) && !good_pmbr && !force_gpt) {
		fprintf(stderr,
			"  Warning: Disk has a valid GPT signature "
			"but invalid PMBR.\n"
			"  Assuming this disk is *not* a GPT disk anymore.\n"
			"  Use gpt kernel option to override.  "
			"Use GNU Parted to correct disk.\n");
		goto fail;
	}

	/* Would fail due to bad PMBR, but force GPT anyhow */
	if ((good_pgpt || good_agpt) && !good_pmbr && force_gpt) {
		fprintf(stderr,
			"  Warning: Disk has a valid GPT signature but "
			"invalid PMBR.\n" "  Use GNU Parted to correct disk.\n"
			"  gpt option taken, disk treated as GPT.\n");
	}

	compare_gpts(pgpt, agpt, lastlba);

	/* The good cases */
	if (good_pgpt && (good_pmbr || force_gpt)) {
		*gpt = pgpt;
		*ptes = pptes;
		if (agpt) {
			mem_free(agpt);
			agpt = NULL;
		}
		if (aptes) {
			mem_free(aptes);
			aptes = NULL;
		}
		if (!good_agpt) {
			fprintf(stderr,
				"Alternate GPT is invalid, "
				"using primary GPT.\n");
		}
		return 1;
	} else if (good_agpt && (good_pmbr || force_gpt)) {
		*gpt = agpt;
		*ptes = aptes;
		if (pgpt) {
			mem_free(pgpt);
			pgpt = NULL;
		}
		if (pptes) {
			mem_free(pptes);
			pptes = NULL;
		}
		fprintf(stderr,
			"Primary GPT is invalid, using alternate GPT.\n");
		return 1;
	}

 fail:
	if (pgpt) {
		mem_free(pgpt);
		pgpt = NULL;
	}
	if (agpt) {
		mem_free(agpt);
		agpt = NULL;
	}
	if (pptes) {
		mem_free(pptes);
		pptes = NULL;
	}
	if (aptes) {
		mem_free(aptes);
		aptes = NULL;
	}
	*gpt = NULL;
	*ptes = NULL;

	return 0;
}

/**
 * read_gpt_pt()
 * @fd
 * @all - slice with start/size of whole disk
 *
 *  0 if this isn't our partition table
 *  number of partitions if successful
 */
int read_gpt_pt(int fd, struct slice *sp, int ns)
{
	gpt_header *gpt = NULL;
	gpt_entry *ptes = NULL;
	int i;
	int n = 0;
	int last_used_index = -1;

	if (!find_valid_gpt(fd, &gpt, &ptes) || !gpt || !ptes) {
		if (gpt)
			mem_free(gpt);
		if (ptes)
			mem_free(ptes);
		return 0;
	}

	for (i = 0; i < (int)le32_to_cpu(gpt->num_partition_entries) && i < ns;
	     i++) {
		if (!efi_guidcmp(NULL_GUID, ptes[i].partition_type_guid)) {
			sp[n].start = 0;
			sp[n].size = 0;
			n++;
		} else {
			sp[n].start = le64_to_cpu(ptes[i].starting_lba);
			sp[n].size =
			    le64_to_cpu(ptes[i].ending_lba) -
			    le64_to_cpu(ptes[i].starting_lba) + 1;
			last_used_index = n;
			n++;
		}
	}
	mem_free(ptes);
	mem_free(gpt);

	return last_used_index + 1;
}
