
/**
 * Copyright © 2011 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Source: multipath-tools, Copyrights of the original file applies
 */

#ifndef FNG_PT_READER_H
#define FNG_PT_READER_H

/**
 * For each partition type there is a routine that takes
 * a block device and a range, and returns the list of
 * slices found there in the supplied array SP that can
 * hold NS entries. The return value is the number of
 * entries stored, or -1 if the appropriate type is not
 * present.
 */

/* libowfat */
#include "io.h"
#include "stralloc.h"

/* frisbee-ng */
#include "defs.h"
#include "getblock.h"

/**
 * unit: lba (512 byte / sector)
 */
struct slice {
	u32 number;
	u64 start;
	u64 size;
	u32 container;

	/* marker */
	u32 todo;
};

struct mbrpart {
	u8 boot_ind;		/* 0x80 - active */
	u8 bh, bs, bc;
	u8 sys_type;
	u8 eh, es, ec;
	u32 start_sect;
	u32 nr_sects;
};

/**
 * ptreader() - read partitiontable of some specific format
 *
 * @fd: harddisk to read from
 * @sp: ptr to be filled
 * @ns: max number of slices, which sp can hold
 *
 * @return: number of found partitions
 */
typedef int (ptreader) (int fd, struct slice * sp, int ns);

#define KNOWN_PT_READER 7
extern ptreader read_gpt_pt;
extern ptreader read_dos_pt;
extern ptreader read_mac_pt;
extern ptreader read_bsd_pt;
extern ptreader read_sun_pt;
extern ptreader read_solaris_pt;
extern ptreader read_unixware_pt;

#define PTYPE_UNKNOWN   0
#define PTYPE_GPT       1
#define PTYPE_DOS       2
#define PTYPE_BSD       3
#define PTYPE_SOLARIS   4
#define PTYPE_MAC       5
#define PTYPE_SUN       6
#define PTYPE_UNIXWARE  7

/**
 * ptreader() - read partitiontable of some specific format
 *
 * @fd: harddisk to read from
 * @sl: ptr to be filled
 * @len: number of slices found
 * @type: short name of partitiontable type
 *
 * @return: type of found partitions (PTYPE_GPT, ...)
 */
extern int pt_getlist(int fd, struct slice **sl, u32 * slen, char **type);

/**
 * efi_crc32() - EFI version of crc32 function
 * @buf: buffer to calculate crc32 of
 * @len - length of buf
 *
 * Description: Returns EFI-style CRC32 value for @buf
 */
extern u32 efi_crc32(const void *buf, unsigned long len);

/* check, if partition type is extended */
extern int is_extended(int type);

/**
 * pt_clone() - clone one partitiontable to one/some destination(s)
 *
 * @fd: harddisk to read from
 * @ptype: type of ptable
 * @len: number of destination devices
 * @fds: destination fd's
 * @devsize: destination device sizes
 */
extern void pt_clone(int fd, int ptype, struct slice *sl, u32 slice_count,
		     u32 len, int64 * fds, u64 * devsize);

/**
 * used for restoring mbr/gpt or whatever data
 */
struct ptable_data {
	u64 start;		/* in bytes */
	stralloc data;		/* data pair */
};

/**
 * pt_image() - image partiton table for later restoration
 *
 * @fd: source disk to read from
 * @ptype: type of partition table (gpt, dos, bsd, ...)
 * @len: number of data entries written to pd
 * @pd: the data for imaging
 */
extern void pt_image(int fd, int ptype, u64 devsize, u32 * len,
		     struct ptable_data **pd);

/* FNG_PT_READER_H */
#endif
