
/**
 * Copyright © 2007 Hannes Reinecke
 * Copyright © 2011 - 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Initially taken from multipath-tools, which derived from util-linux :)
 */

/* system */
#include <sys/types.h>

/* frisbee-ng */
#include "bswap.h"
#include "pt_reader.h"

#define SUN_DISK_MAGIC		0xDABE	/* Disk magic number */
#define SUN_DISK_MAXPARTITIONS	8

#ifdef HAVE_PRAGMA_PACK
#pragma pack(push,1)
#endif

struct sun_raw_part {
	u32 start_cylinder;	/* where the part starts... */
	u32 num_sectors;	/* ...and it's length */
};

struct sun_part_info {
	u8 spare1;
	u8 id;			/* Partition type */
	u8 spare2;
	u8 flags;		/* Partition flags */
};

struct sun_disk_label {
	char info[128];		/* Informative text string */
	u8 spare0[14];
	struct sun_part_info infos[SUN_DISK_MAXPARTITIONS];
	u8 spare1[246];		/* Boot information etc. */
	u16 rspeed;		/* Disk rotational speed */
	u16 pcylcount;		/* Physical cylinder count */
	u16 sparecyl;		/* extra sects per cylinder */
	u8 spare2[4];		/* More magic... */
	u16 ilfact;		/* Interleave factor */
	u16 ncyl;		/* Data cylinder count */
	u16 nacyl;		/* Alt. cylinder count */
	u16 ntrks;		/* Tracks per cylinder */
	u16 nsect;		/* Sectors per track */
	u8 spare3[4];		/* Even more magic... */
	struct sun_raw_part partitions[SUN_DISK_MAXPARTITIONS];
	u16 magic;		/* Magic number */
	u16 csum;		/* Label xor'd checksum */
} __FNG_ATTR_PACKED;

#ifdef HAVE_PRAGMA_PACK
#pragma pack(pop)
#endif

/* Checksum Verification */
static int sun_verify_checksum(struct sun_disk_label *label)
{
	u16 *ush = ((u16 *) (label + 1)) - 1;
	u16 csum = 0;

	while (ush >= (u16 *) label)
		csum ^= *ush--;

	return !csum;
}

int read_sun_pt(int fd, struct slice *sp, int ns)
{
	struct sun_disk_label *l;
	struct sun_raw_part *s;
	unsigned int end;
	int i, j, n;
	u8 *bp;

	bp = getblock(fd, 0);
	if (bp == 0)
		return -1;

	l = (struct sun_disk_label *)bp;
	if (be16_to_cpu(l->magic) != SUN_DISK_MAGIC)
		return -1;

	if (!sun_verify_checksum(l)) {
		/* fprintf(stderr, "Corrupted Sun disk label\n"); */
		return -1;
	}

	for (i = 0, n = 0; i < SUN_DISK_MAXPARTITIONS; i++) {
		s = &l->partitions[i];

		if (s->num_sectors == 0)
			continue;
		if (n < ns) {
			sp[n].start =
			    be32_to_cpu(s->start_cylinder) *
			    be16_to_cpu(l->nsect) * be16_to_cpu(l->ntrks);
			sp[n].size = be32_to_cpu(s->num_sectors);
			n++;
		} else {
			/* fprintf(stderr, "sun_disklabel: too many slices\n"); */
			break;
		}
	}
	/*
	 * Convention has it that the SUN disklabel will always have
	 * the 'c' partition spanning the entire disk.
	 * So we have to check for contained slices.
	 */
	for (i = 0; i < SUN_DISK_MAXPARTITIONS; i++) {
		if (sp[i].size == 0)
			continue;

		end = sp[i].start + sp[i].size;
		for (j = 0; j < SUN_DISK_MAXPARTITIONS; j++) {
			if (i == j)
				continue;
			if (sp[j].size == 0)
				continue;

			if (sp[i].start < sp[j].start) {
				if (end > sp[j].start
				    && end < sp[j].start + sp[j].size) {
					/* Invalid slice */
					/* fprintf(stderr,
					   "sun_disklabel: slice %d overlaps with %d\n", i , j);
					 */
					sp[i].size = 0;
				}
			} else {
				if (end <= sp[j].start + sp[j].size) {
					sp[i].container = j + 1;
				}
			}
		}
	}

	return n;
}
