
/**
 * Copyright © 2005 Felix von Leitner
 * Copyright © 2013 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <sys/uio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

/* libowfat */
#include "byte.h"

/* frisbee-ng */
#include "mem.h"
#include "ringbuf.h"

void ringbuf_init(struct ringbuf *r, char *buf, size_t size)
{
	r->buf = buf;
	r->size = size;
	r->first = r->bytes = 0;
}

int ringbuf_init_malloc(struct ringbuf *r, size_t size)
{
	ringbuf_init(r, mem_alloc(size), size);

	if (r->buf == 0)
		return -1;

	return 0;
}

size_t ringbuf_spaceleft(struct ringbuf * r)
{
	return r->size - r->bytes;
}

ssize_t ringbuf_appendbin(struct ringbuf * r, const char *s, size_t len)
{
	size_t len2 = len;
	size_t next;
	size_t firstchunk;

	if (ringbuf_spaceleft(r) < len)
		return -1;

	next = r->first + r->bytes;
	if (next > r->size)
		next -= r->size;

	if ((firstchunk = r->size - next) < len) {
		byte_copy(r->buf + next, firstchunk, s);
		s += firstchunk;
		len -= firstchunk;
		next = 0;
	}
	byte_copy(r->buf + next, len, s);

	return (r->bytes += len2) == len2 ? 1 : 0;
}

ssize_t ringbuf_append(struct ringbuf * r, const char *s)
{
	return ringbuf_appendbin(r, s, strlen(s));
}

ssize_t ringbuf_write(struct ringbuf * r, int s)
{
	ssize_t x;

	if (r->first + r->bytes < r->size)
		x = write(s, r->buf + r->first, r->bytes);
	else {
		struct iovec v[2];

		v[0].iov_base = r->buf + r->first;
		v[0].iov_len = r->size - r->first;
		v[1].iov_base = r->buf;
		v[1].iov_len = r->bytes - (r->size - r->first);
		x = writev(s, v, 2);
	}

	if (x > 0) {
		r->bytes -= x;
		if ((r->first += x) > r->size)
			r->first -= r->size;
	}

	if (x == -1 && errno != EAGAIN)
		x = -3;

	return x;
}

ssize_t ringbuf_read(struct ringbuf * r, int s)
{
	struct iovec v[2];
	ssize_t x;

	if (r->bytes == 0)
		r->first = 0;

	if (r->size - r->bytes == 0)
		return 0;

	x = r->first + r->bytes;

	if (x > (ssize_t) r->size)
		x -= r->size;
	v[0].iov_base = r->buf + x;
	v[0].iov_len = r->size - x;
	v[1].iov_base = r->buf;
	v[1].iov_len = r->first;

	if (!v[0].iov_len)
		x = read(s, r->buf, v[1].iov_len);
	else if (!v[1].iov_len)
		x = read(s, v[0].iov_base, v[0].iov_len);
	else
		x = readv(s, v, 2);

	if (x > 0)
		r->bytes += x;

	if (x == -1 && errno != EAGAIN)
		x = -3;

	return x;
}

void ringbuf_free(struct ringbuf *r)
{
	mem_free(r->buf);
}

int ringbuf_empty(struct ringbuf *r)
{
	return r->bytes == 0;
}

size_t fmt_ringbuf(char *s, struct ringbuf * r)
{
	if (!s)
		return r->bytes;

	if (r->first + r->bytes > r->size) {
		byte_copy(s, r->size - r->first, r->buf + r->first);
		byte_copy(s + r->size - r->first,
			  r->bytes - (r->size - r->first), r->buf);
	} else
		byte_copy(s, r->bytes, r->buf + r->first);

	return r->bytes;
}
