
/**
 * Copyright © 2011 - 2013 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Transparent handling of IPv4 and IPv6 UDP packets ;)
 */

/* system */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>

/* libowfat */
#include "stralloc.h"
#include "socket.h"
#include "byte.h"
#include "scan.h"
#include "str.h"
#include "ip4.h"
#include "ip6.h"
#include "fmt.h"
#include "ndelay.h"

/* frisbee-ng */
#include "io.h"
#include "sc.h"
#include "die.h"
#include "debug.h"
#include "wrapper.h"

/**
 * sc_scope_id is a global variable used by all ipv6 functions
 */
u32 sc_scope_id = 0;

/**
 * ...
 */
int64 sc_udp(char ip[16], u16 port)
{
	int64 sock;

	if (ip6_isv4mapped(ip)) {

		sock = socket_udp4();
		if (sock == -1)
			fatal("socket_udp4()");

		if (socket_bind4_reuse(sock, ip + 12, port) == -1)
			fatal("socket_bind4_reuse()");

	} else {

		sock = socket_udp6();
		if (sock == -1)
			fatal("socket_udp6()");

		if (socket_bind6_reuse(sock, ip, port, sc_scope_id) == -1)
			fatal("socket_bind6_reuse()");
	}

	io_fd2(sock);

	return sock;
}

/**
 * sc_multicast - create multicast socket and bind it
 *
 * @ip - ipv4 or ipv6 addr, which should be opened
 * @port - port, which sould be bind to
 * @ttl - time to live, important for ipv4 addresses
 * @loop - should the sender get also the packets
 */
int64 sc_multicast(char ip[16], u16 port, char ttl, char loop)
{
	int64 sock = sc_udp(ip, port);
	int r;

	if (ip6_isv4mapped(ip)) {
		r = socket_mcttl4(sock, ttl);
		if (r == -1)
			fatal("socket_mcttl4()");
		r = socket_mcloop4(sock, loop);
		if (r == -1)
			fatal("socket_mcloop4()");
	} else {
		r = socket_mchopcount6(sock, ttl);
		if (r == -1)
			fatal("socket_mchopcount6()");
		r = socket_mcloop6(sock, loop);
		if (r == -1)
			fatal("socket_mcloop6()");
	}

	return sock;
}

/**
 * sc_broadcast - create broadcast socket
 */
int64 sc_broadcast(void)
{
	int64 sock = socket_udp6();

	if (sock == -1)
		fatal("socket_udp6()");

	if (socket_broadcast(sock) == -1)
		fatal("socket_broadcast()");

	io_fd2(sock);

	return sock;
}

/**
 * sc_send - send a UDP datagram over ipv4 or ipv6
 *
 * @buf - buffer to send
 * @len - length of buffer
 * @ip - destination ipv4 or ipv6 address
 * @port - destination UDP port
 */
ssize_t sc_send(int64 s, char *buf, size_t len, const char ip[16], u16 port)
{
	ssize_t r;

	if (ip6_isv4mapped(ip)) {
		r = socket_send4(s, buf, len, ip + 12, port);
		if (r == -1)
			fatal("socket_send4()");
	} else {
		r = socket_send6(s, buf, len, ip, port, sc_scope_id);
		if (r == -1)
			fatal("socket_send6()");
	}

	return r;
}

/**
 * sc_recv - receive a UDP datagram over ipv4 or ipv6
 *
 * @buf - buffer to fill
 * @len - length of buffer
 * @ip - destination ipv4 or ipv6 address
 * @port - destination UDP port
 */
ssize_t sc_recv(int64 s, char *buf, size_t len, char ip[16], u16 * port)
{
	ssize_t r;
	u32 scope;

	if (ip6_isv4mapped(ip)) {
		r = socket_recv4(s, buf, len, ip, port);
		if (r == -1)
			fatal("socket_recv4()");
	} else {
		r = socket_recv6(s, buf, len, ip, port, &scope);
		if (r == -1)
			fatal("socket_recv6()");
	}

	return r;
}
