
/**
 * Copyright © 2010 - 2012 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* system */
#include <pthread.h>
#include <stdarg.h>
#include <string.h>

/* libowfat */
#include "stralloc.h"
#include "strallocX.h"
#include "str.h"
#include "scan.h"

#include "defs.h"
#include "mem.h"
#include "die.h"

/**
 * debugging with some formatting ;)
 * - there only some minimum of formatting
 * - adding new features should be simple, if needed
 *
 * @sa  stralloc to be filled, must be initialized
 */
int stralloc_fmt(stralloc * sa, const char *fmt, va_list ap)
{
	while (*fmt) {
		int d, len, pad = 0;
		int padright;
		char *s, pw;
		u32 u;
		u64 U;
		s64 S;

		if (*fmt != '%') {
			if (!stralloc_catb(sa, fmt++, 1))
				die_nomem();
			continue;
		}

		fmt++;		/* skip '%' */
		if (*fmt == 0)
			break;

		/* look for %% */
		if (*fmt == '%') {
			if (!stralloc_cats(sa, "%"))
				die_nomem();
			fmt++;
			continue;	/* next loop */
		}

		/* look for stralloc's */
		if (*fmt == 'S') {	/* stralloc */
			stralloc *tmp;

			tmp = va_arg(ap, stralloc *);
			if (!stralloc_cat(sa, tmp))
				die_nomem();
			fmt++;	/* skip 'S' */
			continue;	/* next loop */
		}

		/* all others can be padded */
		if (*fmt == '-') {	/* left / right padding */
			padright = 1;
			fmt++;	/* skip '-' */
		} else {
			padright = 0;
		}

		if (*fmt == '0') {	/* zero / space padding */
			pw = '0';
			fmt++;	/* skip '0' */
		} else {
			pw = ' ';
		}

		len = scan_int(fmt, &d);	/* pad length */
		if (len) {
			fmt += len;	/* skip numbers */
			pad = d;
		} else {
			pad = 0;
		}

		switch (*fmt) {

		case 'c':	/* character */
			s = va_arg(ap, char *);

			if (!stralloc_catb(sa, (char *)&s, 1))
				die_nomem();
			break;

		case 's':	/* string */
			s = va_arg(ap, char *);

			if (!s)
				s = "";

			len = str_len(s);
			if (len < pad) {
				if (padright) {
					if (!stralloc_cat_padright
					    (sa, s, len, pad, pad - 1, pw))
						die_nomem();
				} else {
					if (!stralloc_cats(sa, s))
						die_nomem();
					if (!stralloc_cat_padleft
					    (sa, len, pad, pad - 1, pw))
						die_nomem();
				}
			} else {
				if (!stralloc_cats(sa, s))
					die_nomem();
			}
			break;

		case 'b':	/* bitmap u32 */
			{
				int i, len = pad ? pad : 32;
				char *mask = mem_alloc(len);

				u = va_arg(ap, u32);
				memset(mask, '0', len);
				for (i = 0; i < len; i++) {
					if ((u32) 1 << i > u)
						break;
					if (is_set(u, (u32) 1 << i))
						mask[i] = '1';
				}
				if (!stralloc_catb(sa, mask, len))
					die_nomem();
				mem_free(mask);
			}
			break;

		case 'B':	/* bitmap u64 */
			{
				int i, len = pad ? pad : 64;
				char *mask = mem_alloc(len);

				U = va_arg(ap, u64);
				memset(mask, '0', len);
				for (i = 0; i < len; i++) {
					if ((u64) 1 << i > U)
						break;
					if (is_set(U, (u64) 1 << i))
						mask[i] = '1';
				}
				if (!stralloc_catb(sa, mask, len))
					die_nomem();
				mem_free(mask);
			}
			break;

		case 'd':	/* s32 */
			d = va_arg(ap, s32);
			if (!stralloc_cat_s64pw(sa, d, pad ? pad : 0, pw))
				die_nomem();
			break;

		case 'D':	/* s64 */
			S = va_arg(ap, s64);
			if (!stralloc_cat_s64pw(sa, S, pad ? pad : 0, pw))
				die_nomem();
			break;

		case 'u':	/* u32 */
			u = va_arg(ap, u32);
			if (!stralloc_cat_u64pw(sa, u, pad ? pad : 0, pw))
				die_nomem();
			break;

		case 'U':	/* u64 */
			U = va_arg(ap, u64);
			if (!stralloc_cat_u64pw(sa, U, pad ? pad : 0, pw))
				die_nomem();
			break;

		case 'x':	/* u32 - 0xXX.. */
			u = va_arg(ap, u32);
			if (!stralloc_cat_x64pw(sa, u, pad ? pad : 8, pw))
				die_nomem();
			break;

		case 'X':	/* u64 - 0xXX.. */
			U = va_arg(ap, u64);
			if (!stralloc_cat_x64pw(sa, U, pad ? pad : 16, pw))
				die_nomem();
			break;

		case 'h':	/* u64 - KB/MB/GB/TB (/1000) */
			U = va_arg(ap, u64);
			if (!stralloc_cat_human(sa, U))
				die_nomem();
			break;

		case 'H':	/* u64 - KiB/MiB/GiB/TiB (/1024) */
			U = va_arg(ap, u64);
			if (!stralloc_cat_humank(sa, U))
				die_nomem();
			break;

		case 'z':	/* u32 - 0xXX.. (for benchmarks) */
			U = va_arg(ap, u32);
			if (!stralloc_cat_dotted(sa, U, '.'))
				die_nomem();
			break;

		case 'Z':	/* u64 - 0xXX.. (for benchmarks) */
			U = va_arg(ap, u64);
			if (!stralloc_cat_dotted(sa, U, '.'))
				die_nomem();
			break;

		}
		fmt++;
	}

	return 1;
}

int stralloc_print(stralloc * sa, const char *fmt, ...)
{
	stralloc buf;
	va_list ap;

	stralloc_init(&buf);

	va_start(ap, fmt);
	if (!stralloc_fmt(&buf, fmt, ap))
		die_nomem();
	va_end(ap);

	if (!stralloc_cat(sa, &buf))
		die_nomem();
	stralloc_free(&buf);

	return 1;
}
