
/**
 * Copyright © 2010 - 2016 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * debug section: 10
 */

/* system */
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* libowfat */
#include "byte.h"
#include "io.h"
#include "case.h"

/* fribee-ng */
#include "bitmap.h"
#include "main.h"
#include "mem.h"
#include "fs.h"
#include "die.h"
#include "dev.h"
#include "bitmap.h"
#include "term.h"
#include "workq.h"
#include "debug.h"
#include "progress.h"
#include "splice.h"
#include "get_cycles.h"
#include "wrapper.h"
#include "list.h"
#include "transport.h"
#include "term.h"
#include "workq.h"

struct transport_data {
	int64 fd;
	int dst_count;
	int64 *fds;
	u64 start;
	u64 bytes;
	tp_clone *fn;
	int readtest;
};

/* worker function which does the whole work... */
static void worker_cloning(void *arg)
{
	struct transport_data *td = (struct transport_data *)arg;

	td->fn(td->fd, td->dst_count, td->fds, td->start, td->bytes);

	/* XXX: threading needed */
	if (unlikely(td->readtest))
		tp_readtest(td->fd, td->dst_count, td->fds, td->start,
			    td->bytes);
	mem_free(td);

	return;
}

int find_transportmode(const char *name)
{
	const char *x = name;

	if (!x)
		return FNG_TMODE_DEFAULT;

	if (case_equals("readwrite", x))
		return FNG_TMODE_READWRITE;
	if (case_equals("rdwr", x))
		return FNG_TMODE_READWRITE;
#ifdef HAVE_SPLICE
	if (case_equals("splice-tee", x))
		return FNG_TMODE_SPLICE_TEE;
	if (case_equals("splice", x))
		return FNG_TMODE_SPLICE;
#endif
	if (case_equals("sendfile", x))
		return FNG_TMODE_SENDFILE;
	if (case_equals("dummy", x))
		return FNG_TMODE_DUMMY;

	die_msg(111, "invalid transport mode!\n");
	return FNG_TMODE_DEFAULT;
}

/**
 * transport_doit() - copy/clone data of a partition
 *
 * @fd:			source fd
 * @p:			information of source partition
 * @dst_count:		number of destinations
 * @fds:		destination fd's (opened in rw mode)
 * @transportmode:	type of transport (splice, tee, rdwr...)
 * @readtest:		if set, a readtest will be done
 * @pdd:		progress data for disk
 * @part1:		first partition
 * @partN:		first partition
 * @paths:		dst pathnames
 */
#define START_USED  1
#define START_FREE  2
void transport_doit(int64 fd, struct partition *p, int dst_count, int64 * fds,
		    int transportmode, int readtest, struct progress *pdd,
		    int part1, int partN, int threads)
{
	struct fsinfo *fsi = &p->fsi;
	u64 blk, start, bytes;
	int state, i;
	workq_t *w;
	struct progress *pdp = fsi->progress;
	static struct term_clustermap *tcmap = 0;

	tp_init *do_init;
	tp_clone *do_clone;
	tp_done *do_done;

	blk = start = bytes = 0;
	state = START_FREE;

	if (!pdp)
		fatal("transport_doit() pdp not set!");
	if (!pdd)
		fatal("transport_doit() pdd not set!");

	/* default mode is dummy */
	do_init = tp_dummy_init;
	do_clone = tp_dummy_clone;
	do_done = tp_dummy_done;

	switch (transportmode) {
	case FNG_TMODE_READWRITE:
		do_clone = tp_rdwr_clone;
		break;
#ifdef HAVE_SPLICE
	case FNG_TMODE_SPLICE:
		do_clone = tp_splice_clone;
		break;
		/* seems to be the fastest on my box */
	case FNG_TMODE_SPLICE_TEE:
		do_clone = tp_splice_tee_clone;
		break;
#endif
	case FNG_TMODE_SENDFILE:
		do_clone = tp_sendfile_clone;
		break;
	case FNG_TMODE_DUMMY:
		do_clone = tp_dummy_clone;
		break;
	default:
		fatal("strange transportmode!");
	}

	/* init terminal screen and show value for number of blocks per unit */
	if (!tcmap)
		tcmap = mem_talloc(struct term_clustermap, 1);

	term_clustermap_init(tcmap, fsi->bitmap, fsi->blocks, 2, term_maxx - 1,
			     3, term_maxy - 9);

	term_gotoxy(term_maxx - 16 - 8 - 1, term_maxy - 8);
	term_print("%8u blocks per unit", tcmap->blocks_per_unit);
	term_gotoxy(3, 2);
	term_sa(&term_left_status_top);

	/* clear old stats */
	term_gotoxy(3, term_maxy - 6);
	term_hline(term_maxx - 3, " ");
	term_gotoxy(3, term_maxy - 5);
	term_hline(term_maxx - 3, " ");
	term_gotoxy(3, term_maxy - 4);
	term_hline(term_maxx - 3, " ");
	term_flush();

	/* init resources used by this mode */
	do_init(fd, dst_count, fds);

	/* workq_init(int thr_max, int queue_max, int idle_timeout, fn) */
	w = workq_init(threads, 2, 2, worker_cloning);

	for (blk = 0; blk < fsi->blocks; blk++) {

		/* alle XY blöcke gibts neuen status... (XXX, test and adjust) */
		if (blk % tcmap->blocks_per_unit == 0) {

			/* calculate stats */
			progress(pdp);
			progress(pdd);

			/* draw the map */
			term_clustermap(tcmap, blk);

	/**
	 * ├────────────────────────────────────────────────────────────┤ -7
	 * │ running time: 1:23:33, eta: 30:54:15                       │ -6
	 * │ data rate: xxxx.yMiB/s  x778.2kiB / xx28.9GiB              │ -5
	 * │ block: 1.123.123 / 123.123.123                             │ -4
	 * │ xx/xx: [=======================>                    ]  45% │ -3
	 * │ disk:  [============>                               ]  33% │ -2
	 * └────────────────────────────────────────────────────────────┘ -1
	 */

			term_gotoxy(3, term_maxy - 6);
			term_print
			    ("running time: %u:%02u:%02u,  eta: %u:%02u:%02u",
			     min(pdd->ctt_h, 99), pdd->ctt_m, pdd->ctt_s,
			     min(pdd->eta_h, 99), pdd->eta_m, pdd->eta_s);

			term_gotoxy(3, term_maxy - 5);
			term_print
			    ("data rate: %HiB/s  done: %HiB / %HiB        ",
			     pdd->current / pdd->timediff, pdd->current,
			     pdd->total);

			term_gotoxy(3, term_maxy - 4);
			term_print("block: %10Z / %10Z", blk, fsi->blocks);

			/* status current partition */
			term_gotoxy(3, term_maxy - 3);
			term_print("%d/%d:", part1, partN);

			term_gotoxy(10, term_maxy - 3);
			term_progress(10, term_maxx - 7, pdp->percent);

			term_gotoxy(term_maxx - 5, term_maxy - 3);
			term_print("%-3u%%", pdp->percent);

			/* overall status */
			term_gotoxy(3, term_maxy - 2);
			term_print("disk:");

			term_gotoxy(10, term_maxy - 2);
			term_progress(10, term_maxx - 7, pdd->percent);

			term_gotoxy(term_maxx - 5, term_maxy - 2);
			term_print("%-3u%%", pdd->percent);

			term_flush();

		}

		if (bitmap_get(fsi->bitmap, blk)) {
			/* block is used */
			if (state == START_FREE) {
				/* setup new start point */
				start = blk * fsi->bsize;
				state = START_USED;
			}
			bytes += fsi->bsize;

			/* if 100 MiB reached, clone it ... */
			if (bytes >= 100 * 1024 * 1024) {
				struct transport_data *td =
				    mem_talloc(struct transport_data, 1);

				td->fd = fd;
				td->dst_count = dst_count;
				td->fds = fds;
				td->start = start + p->start;
				td->bytes = bytes;
				td->fn = do_clone;
				td->readtest = readtest;
				progress_add(pdp, bytes);
				progress_add(pdd, bytes);
				state = START_FREE;
				bytes = 0;

				workq_add(w, td);
			}

		} else {
			/* block is NOT used... so ignore this empty block
			 * when not in used state */
			if (state == START_FREE)
				continue;

			struct transport_data *td =
			    mem_talloc(struct transport_data, 1);

			td->fd = fd;
			td->dst_count = dst_count;
			td->fds = fds;
			td->start = start + p->start;
			td->bytes = bytes;
			td->fn = do_clone;
			td->readtest = readtest;
			progress_add(pdp, bytes);
			progress_add(pdd, bytes);
			state = START_FREE;
			bytes = 0;

			workq_add(w, td);
		}
	}

	/* copy the last used blocks now */
	if (state == START_USED) {
		struct transport_data *td =
		    mem_talloc(struct transport_data, 1);

		td->fd = fd;
		td->dst_count = dst_count;
		td->fds = fds;
		td->start = start + p->start;
		td->bytes = bytes;
		td->fn = do_clone;
		td->readtest = readtest;
		progress_add(pdp, bytes);
		progress_add(pdd, bytes);
		bytes = 0;

		workq_add(w, td);
	}

	/* nearly done... */
	term_clustermap(tcmap, fsi->blocks);
	term_gotoxy(1, term_maxy);
	term_flush();

	/* wait for worker to finish */
	workq_deinit(w);

	/* clean up resources used by this mode */
	do_done(fd, dst_count, fds);

	/* run fs_postclone() function, if defined */
	for (i = 0; i < dst_count; i++)
		if (fsi->fs->fs_postclone)
			fsi->fs->fs_postclone(fd, fds[i], p->start);

	return;
}
